#ifndef GLWIDGET_H
#define GLWIDGET_H

/**
 * glfractal.h
 *
 * Brief Definition of GLFractal
 *
 * This is a simple QGLWidget displaying an openGL Sierpinski Gasket
 */

#include <qgl.h>

struct MyPoint {
  float x;
  float y;
}; // end Point

/**
 * GLFractal inherits from QGLWidget (http://doc.trolltech.com/2.3/qglwidget.html) <BR>
 * QGLWidget basically allows you to control OpenGL through method calls
 */
class GLWidget : public QGLWidget
{
  Q_OBJECT
public:
  GLWidget(QWidget* parent, const std::string &name);
  virtual ~GLWidget();

 public slots:
    void setXRotation(int degrees);
    void setYRotation(int degrees);
    void setZRotation(int degrees);
    // void setScale(int newscale);

protected:
    void initializeGL();
    void paintGL();
    void resizeGL(int w, int h );
    virtual GLuint makeObject();
    void drawTriangle(MyPoint a, MyPoint b, MyPoint c);
    void drawSierpinski(MyPoint a, MyPoint b, MyPoint c, int level);

private:
    GLuint object;
    GLfloat xRot, yRot, zRot, scale;
};

#endif

