#ifndef GLFUNCTIONS_H
#define GLFUNCTIONS_H

#include <GL/gl.h>
#include <GL/glut.h>
#include <math.h>

#include "Vector3D.h"
#include "Color.h"

#include "Defaults.h"


void enterSelectionMode(int cursorX,int cursorY,double viewDistance,unsigned int bufferSize,GLuint buffer[]);
unsigned int exitSelectionMode();
void setOrthographicProjection();
void resetPerspectiveProjection();



inline void drawCircle(Vector3D location,double radius,int numTriangles,Color color){
	double angle,vectorY1,vectorX1,vectorX,vectorY;	

	vectorX1=location.x;
	vectorY1=location.y;
 
	glBegin(GL_TRIANGLES);	
		glColor4f(color.red,color.green,color.blue,color.alpha);
		for(int i=0;i<=360;i+=(360/numTriangles)){
			angle=(double)(((double)i)/57.29577957795135); // number is (360 / 2*PI)
			vectorX=location.x+(radius*(double)sin((double)angle));
			vectorY=location.y+(radius*(double)cos((double)angle));		
			glVertex3d(location.x,location.y,location.z);
			glVertex3d(vectorX1,vectorY1,location.z);
			glVertex3d(vectorX,vectorY,location.z);
			vectorY1=vectorY;
			vectorX1=vectorX;
		}
	glEnd();
}

inline void drawSphere(Vector3D location,double radius,Color color){
	
	glPushMatrix();
	
	glColor4f(color.red,color.green,color.blue,color.alpha);
	glTranslatef(location.x,location.y,location.z);
	
	glutSolidSphere(radius,25,25);
	glPopMatrix();

}
	
// this method just draw a 2D square..
inline void drawSquare(Vector3D location,double size,Color color){
	glBegin(GL_QUADS);
		glColor4f(color.red,color.green,color.blue,color.alpha);
		glVertex3d(location.x-(size/2.0),location.y-(size/2.0),location.z);
		glVertex3d(location.x+(size/2.0),location.y-(size/2.0),location.z);
		glVertex3d(location.x+(size/2.0),location.y+(size/2.0),location.z);
		glVertex3d(location.x-(size/2.0),location.y+(size/2.0),location.z);
	glEnd();
}

inline void drawLine(Vector3D start,Vector3D end,Color color){
	glBegin(GL_LINE_STRIP);
	glColor4f(color.red,color.green,color.blue,color.alpha);

	glVertex3d(start.x,start.y,start.z);
	glVertex3d(end.x,end.y,end.z);
	
	glEnd();	
}

inline void drawArrow(Vector3D start,Vector3D end,double arrowLength,Color color){
	// first, draw the main line...
	drawLine(start,end,color);
	
	// now draw a cone at the end of this line...
	glPushMatrix();
	glColor4f(color.red,color.green,color.blue,color.alpha);
	
	Vector3D lineVector = end-start;  // get the vector of the line fro end to start
	
	Vector3D shortenedVector = start + (lineVector | (lineVector.length() - arrowLength)); // shorten it by arrow length, so tip of arrow is at the right point
	
	glTranslatef(shortenedVector.x,shortenedVector.y,start.z);
	
	// now, we have to rotate the cone around a line... in this case, the line is a vector perpendicular to the line itself
	
	glRotatef(-90.0,lineVector.y,-lineVector.x,start.z);
	
	glutSolidCone(1.0,arrowLength,15,15); // finally, draw the cone.
	
	glPopMatrix();
	
}
inline void drawRectangle(Vector3D start,Vector3D end, Color color){
	//fprintf(stderr,"drawing: %lf %lf %lf %lf\n",start.x,start.y,end.x,end.y);
	glBegin(GL_LINE_STRIP);
	glColor4f(color.red,color.green,color.blue,color.alpha);
	glVertex3d(start.x,start.y,start.z);
	glVertex3d(end.x,start.y,start.z);
	glVertex3d(end.x,end.y,start.z);
	glVertex3d(start.x,end.y,start.z);
	glVertex3d(start.x,start.y,start.z);
	glEnd();
}

inline void draw3DString(double x,double y,std::string string,Color color,double scale){
	
	glColor4f(color.red,color.green,color.blue,color.alpha);
	
	glPushMatrix();
	glTranslatef(x,y,-1.0);	
	
	glScalef(scale,scale,scale);
	
	for (unsigned int i=0;i<string.length();i++){
		glutStrokeCharacter(GLUT_STROKE_ROMAN,(GLubyte) string.at(i));	
	}
	
	glPopMatrix();
}


inline void draw2DString(double x,double y,std::string string,Color color,bool smallFont){
	
	glColor4f(color.red,color.green,color.blue,color.alpha);
	
	//glPushMatrix();
	//glLoadIdentity();
	
	glRasterPos3f(x,y,-1.0); // -1 to make sure text is always above everything else
	
	for (unsigned int i=0;i<string.length();i++){
		if(smallFont){
			glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10,(GLubyte) string.at(i));
		}else{
			glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18,(GLubyte) string.at(i));
		}
	}

	//glPopMatrix();

}
inline void drawOrthoString(double x,double y,std::string string,Color color,bool smallFont){
	setOrthographicProjection();
	
	glColor4f(color.red,color.green,color.blue,color.alpha);
	
	glPushMatrix();
	glLoadIdentity();
	
	glRasterPos3f(x,y,-1.0); // -1 to make sure text is always above everything else
	
	for (unsigned int i=0;i<string.length();i++){
		if(smallFont){
			glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10,(GLubyte) string.at(i));
		}else{
			glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18,(GLubyte) string.at(i));
		}
	}
	
	glPopMatrix();
	
	resetPerspectiveProjection();
}


inline void enterSelectionMode(int cursorX,int cursorY,double viewDistance,unsigned int bufferSize,GLuint selectionBuffer[]){
	
	GLint viewport[4];
	glSelectBuffer(bufferSize,selectionBuffer);
	glRenderMode(GL_SELECT);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	
	glGetIntegerv(GL_VIEWPORT,viewport);
	gluPickMatrix(cursorX,viewport[3]-cursorY,5,5,viewport);
	
	gluPerspective(45.0,(double)glutGet(GLUT_WINDOW_WIDTH) / (double)glutGet(GLUT_WINDOW_HEIGHT),1.0,viewDistance);
	glMatrixMode(GL_MODELVIEW);
	glInitNames();
}
inline unsigned int exitSelectionMode(){
	
	unsigned int numHits = 0;
	
	// restoring the original projection matrix
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glFlush();
	
	// returning to normal rendering mode
	numHits = glRenderMode(GL_RENDER);
	
	return numHits;
}


inline void setOrthographicProjection() {
	
	// switch to projection mode
	glMatrixMode(GL_PROJECTION);
	// save previous matrix which contains the 
	//settings for the perspective projection
	glPushMatrix();
	// reset matrix
	glLoadIdentity();
	// set a 2D orthographic projection
	gluOrtho2D(0, glutGet(GLUT_WINDOW_WIDTH),0,glutGet(GLUT_WINDOW_HEIGHT));
	// invert the y axis, down is positive
	glScalef(1, -1, 1);
	// mover the origin from the bottom left corner
	// to the upper left corner
	glTranslatef(0, -glutGet(GLUT_WINDOW_HEIGHT), 0);
	glMatrixMode(GL_MODELVIEW);
}

inline void resetPerspectiveProjection() {
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

inline void draw2DText(){
	if(!renderText) return;
	
	//setOrthographicProjection();
	
	// now draw the strings for the status buffer
	if(smallFont){
		for(unsigned int i=0;i<statusBuffer.size();i++){
			drawOrthoString((int)statusMessagePos.x,(int)(statusMessagePos.y+(15*i)),statusBuffer[i],red,smallFont);
		}
	}else{
		for(unsigned int i=0;i<statusBuffer.size();i++){
			drawOrthoString((int)statusMessagePos.x,(int)(statusMessagePos.y+(22*i)),statusBuffer[i],red,smallFont);
		}
	}
	
	drawOrthoString(glutGet(GLUT_WINDOW_WIDTH)-80,20,std::string(fpsString),red,smallFont);
	
	if(paused){
		drawOrthoString(glutGet(GLUT_WINDOW_WIDTH)/2.0-20,glutGet(GLUT_WINDOW_HEIGHT)/2.0,"PAUSED",red,false);
	}
	//resetPerspectiveProjection();
}

#endif
