#ifndef GROUP_H
#define GROUP_H

#include "SDL_thread.h"

#include "GLfunctions.h"
#include "MapObject.h"


#define MIN_GROUP_SIZE 0.5

// shows how many triangles will make up the circle 
#define CIRCLE_DETAIL 24

class Unit;
class Order;

typedef std::map<const std::string,std::vector<Unit*>* > UnitMapVector;
typedef std::map<const std::string,double> UnitRecharges;

class Group : public MapObject {
	
public:
	Group(const std::string& name,unsigned int playerID);
	~Group();
	Group(const Group& b);
	Group* clone() const;
	
	
	bool withinRange(Group* otherGroup,const std::string& unitType);
	bool withinRange(Group* otherGroup);
	bool takeDamage(const double& damage,const std::string& unitType);
	bool eliminateDead();
	
	std::string getMostDamagingUnit();
	std::vector<Unit*>* getUnitsOfType(const std::string& unitType);
	double getAttackRangeForType(const std::string& unitType);
	
	void recalculateRadius();
	void recalculateMaxValues();
	void getChargedUnits(std::vector<std::string>& unitTypeList,double currentTime);
	double getNextShotTime(std::vector<std::string>& unitsInRange,double currentTime);
	
	void draw();
	void drawInSelectionMode();
	
	void addUnit(const std::string& name,Unit* newUnit);
	void printUnits();
	
	double getExtents() const;
	
	std::string printUnitComposition();
	std::string toString();
	
	unsigned int getNumUnits();
	unsigned int getNumUnitsOfType(const std::string& unitType);
	
	void hasShot(const std::string& unitType,double currentTime);
	
	void setOrder(Order* newOrder,SDL_cond* conditional,unsigned int* numIncompleteTasks);
	Order* getOrder() const;
	void executeOrders(double timeElapsed);
	void executeOrderAction(const std::string& actionType);
	void cleanupOrder();
	
	void join(Group* groupToJoin);
	bool enemyInRange();
	void enemyInRangeOfUnitType(std::vector<std::string>& unitTypes);
	
	double getMaxSpeed() const;
	double getMaxAttackRange() const;
	double getLastDistanceMoved() const;
	
	UnitRecharges getLastShotTimes() const;
	
	
private:
	Group& operator=(const Group& b) {}; // no = operator for now
	
	Order* currentOrder;
	double maxSpeed;
	double maxAttackRange; // used just for processing speed reasons, can be calculated from objects in group if needed
	double lastDistanceMoved;
	
	UnitMapVector units;
	UnitRecharges lastShotTimes;
	
	double radius;
	
	SDL_cond* conditional;
	unsigned int* numIncompleteTasks;

};

#endif


