#ifndef MAPOBJECT_H
#define MAPOBJECT_H

#include <string>
#include <sstream>

#include <GL/gl.h>

// how large is each unit represented
#define UNIT_SIZE 5.0

// shows how many triangles will make up the circle 
#define CIRCLE_DETAIL 24

#define GROUP_TYPE_ID 0
#define BASE_TYPE_ID 1

#include "Vector3D.h"
#include "Color.h"

class Group;

class MapObject{

public:
	MapObject(const std::string& name,unsigned int playerID);
	//MapObject(const MapObject& obj);
	virtual ~MapObject();
	// FIXME need to add copy constructors and assignment operators maybe for the subclasses as well?
	//MapObject& operator=(const MapObject& b);
	
	virtual MapObject* clone() const = 0;
	
	virtual bool withinRange(Group* otherObject,const std::string& unitType)=0;
	virtual bool withinRange(Group* otherObject)=0;
	virtual bool takeDamage(const double& damage,const std::string& unitType)=0;
	virtual bool eliminateDead()=0;
	virtual double getExtents() const=0;
	virtual void draw()=0;
	virtual void drawInSelectionMode();
	
	virtual std::string toString();
	
	unsigned int getID() const;

	std::string getName() const;
	Vector3D getPosition() const;
	void setPosition(const Vector3D& pos);
	Color getColor() const;

protected:
		
	std::string name;
	Vector3D position;
	Color color;
	
	unsigned int objectID;
	unsigned int playerID;
	bool wasHit;

private:
	static unsigned int numObjects;
	
};

#endif
