
#include "Player.h"

#include "Group.h"
#include "Order.h"
#include "MapObject.h"
#include "Base.h"

Player::Player(unsigned int playerID){
	this->playerID = playerID;
}

Player::~Player(){
	PlayerObjectMap::iterator it;
	for(it = objects.begin();it != objects.end(); ++it){
		delete  (it->second);
	}
	
	objects.clear();
}
Player::Player(const Player& p){
	playerID = p.playerID;

	PlayerObjectMap objectMap = p.objects;

	PlayerObjectMap::iterator it;
	for(it = objectMap.begin();it != objectMap.end(); ++it){
		MapObject* currObject = it->second;

		addObject(currObject->clone());
		
	}
}

Player* Player::clone() const{
	return (new Player(*this));
}

PlayerObjectMap* Player::getObjects(){
	return &objects;
}

unsigned int Player::getID() const{
	return playerID;
}

void Player::addObject(MapObject* newObject){
	objects[newObject->getID()] = newObject;
}

void Player::drawObjects(){
	PlayerObjectMap::iterator it;
	for(it = objects.begin();it != objects.end(); ++it){
		(it->second)->draw();
	}
}
void Player::drawObjectsInSelectionMode(){
	PlayerObjectMap::iterator it;
	for(it = objects.begin();it != objects.end(); ++it){
		(it->second)->drawInSelectionMode();
	}
}

void Player::executeOrders(double timeElapsed){
	PlayerObjectMap::iterator it;
	for(it = objects.begin();it != objects.end(); ++it){
		Group* group = dynamic_cast<Group*> (it->second);
		if(group != NULL && group->getNumUnits() > 0){
			group->executeOrders(timeElapsed);
		}
	}
}


void Player::addContainedObjectsToList(const Vector3D& selectionRectangleStart,const Vector3D& selectionRectangleEnd,MapObjectSet& selectedObjects){
	
	PlayerObjectMap::iterator it;
	for(it = objects.begin();it != objects.end(); ++it){

		Vector3D pos = (it->second)->getPosition();
		
		// create a new rectangle to make the checking a lot easier
		double minX = std::min(selectionRectangleStart.x,selectionRectangleEnd.x);
		double minY = std::min(selectionRectangleStart.y,selectionRectangleEnd.y);
		double maxX = std::max(selectionRectangleStart.x,selectionRectangleEnd.x);
		double maxY = std::max(selectionRectangleStart.y,selectionRectangleEnd.y);
		
		if(pos.x >= minX && pos.x <= maxX && pos.y >= minY && pos.y <= maxY){
			selectedObjects.insert(it->second);
		}
	}
}

void Player::printObjects(std::ofstream& stream){
	PlayerObjectMap::iterator it;
	for(it = objects.begin();it != objects.end(); ++it){
		stream << (it->second)->toString();
	}
}

std::string Player::printObjects(){
	std::ostringstream stream;
	PlayerObjectMap::iterator it;
	for(it = this->objects.begin();it != this->objects.end(); ++it){
		stream << (it->second)->toString();
	}	
	return stream.str();
}

MapObject* Player::getObject(unsigned int objectID){
	PlayerObjectMap::iterator it;
	it = objects.find(objectID);
	if(it != objects.end()){
		return objects[objectID];
	}else{
		return NULL;
	}
}


std::string Player::toString(){
	std::ostringstream stream;
		
	stream << "player: "  <<  playerID << "\n";
	stream << printObjects();
	
	return stream.str();
}

void Player::checkForDestroyedObjects(std::vector<MapObject*>& deletedObjects){
	//std::vector<MapObject*> deletedObjects;
	
	PlayerObjectMap::iterator it;
	for(it = objects.begin();it != objects.end(); ++it){
		MapObject* currentObject = it->second;
		if(currentObject->eliminateDead()){ // if unit is dead
			deletedObjects.push_back(currentObject);
		}
	}
	// clear some guys out of the map....
	for(unsigned int i=0;i<deletedObjects.size();++i){
		objects.erase(deletedObjects[i]->getID());
	}
}

void Player::getGroupsAtLocation(const Group* otherGroup,std::vector<Group*>& groupsAtLocation){
	PlayerObjectMap::iterator it;
	for(it = objects.begin();it != objects.end(); ++it){
		Group* group = dynamic_cast<Group*> (it->second);
		// make sure it is a unit, and not the same unit as is doing the searching, also make sure the unit is a valid one(not dead, or JUST joined)
		if(group != NULL && group != otherGroup && group->getNumUnits() > 0){
			if((group->getPosition() - otherGroup->getPosition()).length() < positionCompareRange){
				groupsAtLocation.push_back(group);
			}
		}
	}
}


Base* Player::getNearestEnemyBase(const Vector3D& position){
	double nearest = 10000000000.0;
	Base* nearestBase = NULL;
	PlayerMap::iterator it;
	for(it = playerMap.begin();it != playerMap.end();++it){
		if(this == (it->second)) continue; // don't search yourself
		PlayerObjectMap* otherPlayerObjects = (it->second)->getObjects();
		PlayerObjectMap::iterator it2;
		for(it2 = otherPlayerObjects->begin();it2 != otherPlayerObjects->end(); ++it2){
			Base* currentBase = dynamic_cast<Base*> (it2->second);
			if(currentBase != NULL){
				double dist = (position - currentBase->getPosition()).length();
				if(dist < nearest){
					nearestBase = currentBase;
					nearest = dist;
				}
			}
		}
	}
	return nearestBase;
}
