
#include "SimFunctions.h"

#include "UtilFunctions.h"
#include "GLfunctions.h"

#include "Player.h"

#include "Group.h"
#include "Base.h"
#include "Unit.h"
#include "Order.h"

double lastSimulationTime = 0;

void moveObjects(PlayerMap& players,double timeElapsed){
	if(unitSemaphore || paused) return; // don't mess with these if in middle of deleting or loading
	
	PlayerMap::iterator it;
	for(it = players.begin();it != players.end(); ++it){
		(it->second)->executeOrders(timeElapsed);
	}
}

void blah(PlayerMap& players,double currentTime){
		
	PlayerMap::iterator currentPlayerIterator;
	PlayerMap::iterator otherPlayerIterator;
	
	for(currentPlayerIterator = players.begin(); currentPlayerIterator != players.end(); ++currentPlayerIterator){
		PlayerObjectMap* currPlayerObjects =  (currentPlayerIterator->second)->getObjects();
		
		PlayerObjectMap::iterator it;
		for(it = currPlayerObjects->begin();it != currPlayerObjects->end(); ++it){
			Group* currentObject = dynamic_cast<Group*> (it->second);
			if(currentObject == NULL) continue;
			//if(currentObject->getOrder() == NULL) continue;
			
			for(otherPlayerIterator = players.begin(); otherPlayerIterator != players.end(); ++otherPlayerIterator){
				if(currentPlayerIterator == otherPlayerIterator) continue; // don't check your own objects...
				PlayerObjectMap* otherPlayerObjects =  (otherPlayerIterator->second)->getObjects();
				PlayerObjectMap::iterator it2;
				for(it2 = otherPlayerObjects->begin();it2 != otherPlayerObjects->end(); ++it2){
					Group* otherObject = dynamic_cast<Group*> (it2->second);
					if(otherObject == NULL) continue;
					if(otherObject->withinRange(currentObject)){
						
						print("time at collision is: " + convertToString(currentTime));
						
					}
				}
			}
		}
	}
}

void calculateCollisions(PlayerMap& players,double currentTime){

	PlayerMap::iterator currentPlayerIterator;
	PlayerMap::iterator secondPlayerIterator;
	
	// yes,, it is n^2, and could be made a bit better by doing both evaluations for collisions, but total computing time is around the same...(i think)
	//for(currentPlayerIterator = players->begin(); currentPlayerIterator != players->end(); ++currentPlayerIterator){
	for(currentPlayerIterator = players.begin(); currentPlayerIterator != players.end(); ++currentPlayerIterator){
		PlayerObjectMap* currPlayerObjects =  (currentPlayerIterator->second)->getObjects();
		PlayerObjectMap::iterator it;
		for(it = currPlayerObjects->begin();it != currPlayerObjects->end(); ++it){
			Group* currentGroup = dynamic_cast<Group*> (it->second);
			if(currentGroup == NULL) continue; // if the current object is not a group, ignore it, since it can't attack
			
			std::vector<std::string> chargedUnits;
			// check if the group has the ability to fire(not recharging) and which units in the group are in range...
			currentGroup->getChargedUnits(chargedUnits,currentTime);
			if(chargedUnits.size() == 0) continue; // no need to do anything if this unit has no way to fire...
			if(currentGroup->getLastDistanceMoved() > 0.0) continue;// can only attack if we are not moving...
			
			std::map<std::string,std::vector<MapObject*> > objectsInRangeMap;
			
			for(secondPlayerIterator= players.begin();secondPlayerIterator != players.end(); ++secondPlayerIterator){
				if(currentPlayerIterator == secondPlayerIterator) continue; // don't check your own objects...
				
				PlayerObjectMap* secondPlayerObjects =  (secondPlayerIterator->second)->getObjects();
				PlayerObjectMap::iterator it2;
				for(it2 = secondPlayerObjects->begin();it2 != secondPlayerObjects->end(); ++it2){
					// if one object is within the maximum range of another, then do something....
					// check if the second object is within the range of the current object...
					for(unsigned int k=0;k<chargedUnits.size();k++){
						if((it2->second)->withinRange(currentGroup,chargedUnits[k])){
							objectsInRangeMap[chargedUnits[k]].push_back(it2->second);
							//print("OBJECT HIT!");
						}	
					}
					
				}
			}
			// now that we have all the objects that this current one hits, we have to decide what to do with the available targets, and which ones to attack
			
			
			// TODO right now, just attack the first one found by default
			std::map<std::string,std::vector<MapObject*> >::iterator validTargetsIterator;
			for(validTargetsIterator= objectsInRangeMap.begin();validTargetsIterator != objectsInRangeMap.end(); ++validTargetsIterator){
				std::vector<Unit*>* currentUnits = currentGroup->getUnitsOfType(validTargetsIterator->first);
				bool shotAtAll = false;
				for(unsigned int a=0;a<currentUnits->size();a++){
					for(unsigned int b=0;b<(validTargetsIterator->second).size();b++){
						MapObject* currentEnemyObject = (validTargetsIterator->second)[b];
						Group* enemyGroup = dynamic_cast<Group*> (currentEnemyObject);
						if(enemyGroup != NULL){
							std::string bestEnemyUnit = enemyGroup->getMostDamagingUnit();
							if(bestEnemyUnit == "NONE"){
								continue; // go to the next enemy object in range, if any...since all units of this one have been destroyed
							}else{
								shotAtAll = true;
								enemyGroup->takeDamage((*currentUnits)[a]->getAttackPower(),bestEnemyUnit);
								break;
							}
						}else{
							Base* enemyBase = dynamic_cast<Base*> (currentEnemyObject);
							if(enemyBase == NULL){
								print("WARNING: should never get here in damage calculations");
							}else{
								if(enemyBase->getHitPoints() > 0){
									enemyBase->takeDamage((*currentUnits)[a]->getAttackPower(),"");
									shotAtAll = true;
									break;
								}else{
									continue;
								}
							}
						}
					}
				}
				// now if even ONE of these units has shot, we consider all of them to have shot, not perfect, but alternative is more costly
				if(shotAtAll) currentGroup->hasShot(validTargetsIterator->first,currentTime);
						
			}
		}
	}
	
}

// function originally from Collision2D.C from ORTS

/** Solves \f$ at^2 + bt = c; a > 0 \f$
@return true iff solution exists

robust solution (numerical recipies)
*/

bool solveQuadratic(double a,double b,double c,double& t1,double& t2){
	
	if (a <= 0.0){
		//print("ERROR: Trying to solve quadratic with a <=0");
		return false;
	}
	// no need to do this, pass a negative c instead
	//c = -c;
	double squareRootPart = b*b-4.0*a*c;
	
	// can't have a negative square root, so return false if this is negative
	if (squareRootPart < 0.0){
		//print("error, squareRootPart is negative: "+convertToString(squareRootPart));
		return false;
	}
  	// robust root computation (Numerical Recipies p.184)
	
	double q = -0.5L * (b + sgn(b) * sqrt(squareRootPart));
	if(q == 0.0){
		//print("error in quadratic, q was 0");
		return false;
	}
	t1 = q / a;
	t2 = c / q;
	
	if (t1 > t2){
		double temp = t2;
		t2 = t1;
		t1 = temp;
	}
	//assert(!(t1 > t2));
	
	double e1 = fabs(a*t1*t1+b*t1+c);
	double e2 = fabs(a*t2*t2+b*t2+c);
	if (e1 > 1.5e-5) print("error in quadratic: " + convertToString(e1));
	if (e2 > 1.5e-5)  print("error in quadratic: " + convertToString(e2));
	
	return true;
}

// originally from Collision2D.H from ORTS

/** Compute collision time of a (moving) circle and a (moving) point.
    Used to check for collision with the endpoint of a line.
@param circlePos     circle center
@param radius    square of the circle's radius
@param pointPos     point's location
@param relativeHeading     circle's heading relative to the point
@param startTime  start simulation at this time
*/

double getCirclePointCollisionTime(const Vector3D& circlePos,double radius,const Vector3D& pointPos,const Vector3D& relativeHeading,double startTime){
	
	Vector3D positionDifference = circlePos - pointPos;
	
	double quadraticValueC = (radius*radius) - (positionDifference.x*positionDifference.x) - (positionDifference.y*positionDifference.y);
	
	//int c_h = c_p * h; // c dot h
	double differenceHeading = positionDifference.dot(relativeHeading);
	
	// if they are heading away from each other and not touching, then no need to do any more work...
	if (differenceHeading > 0.0) {   // moving away from each other
		if (quadraticValueC >= 0.0) { // they don't intersect
			//print("moving from each other");
			return -1.0;
		}
	}
	
	double t1= 0.0;
	double t2 = 0.0;
	
	double quadraticValueA = (relativeHeading.x*relativeHeading.x) + (relativeHeading.y*relativeHeading.y);
	double quadraticValueB = 2.0 * differenceHeading;
	
	bool isCollision = solveQuadratic(quadraticValueA,quadraticValueB,-quadraticValueC, t1, t2); // (a)t^2 + (b)t - c = 0
	
	t1*=1000.0;
	t2*=1000.0;	
	//print("t1= " + convertToString(t1));
	//print("t2= " + convertToString(t2));
	
	if (!isCollision){
		//print("no collision");
		return -1.0;
	}

	t1+=startTime;
	t2+=startTime;
	if (t1 == t2){
		//print("t1 and t2 are equal!");
		return -1.0; // point intersects tangentially
	}
	
	if (t2 <= startTime){
		//print("t2 <= starttime");
		return -1.0;
	}
	if (t1 == startTime){
		print("WARNING: time is equal to startTime");
		return -1.0;
		//return startTime; // intersecting at time
	}
	if(t1 < startTime || t2 < startTime){
		print("WARNING: time of collision is less than start time, this is wrong!");
		//return startTime;
		return -1.0;
	}
	
	return t1;
}

double getNextCollisionTime(PlayerMap& players,double currentTime){
	double minimumTime = DOUBLE_MAX;
	
	PlayerMap::iterator currentPlayerIterator;
	PlayerMap::iterator otherPlayerIterator;

	for(currentPlayerIterator = players.begin(); currentPlayerIterator != players.end(); ++currentPlayerIterator){
		PlayerObjectMap* currPlayerObjects =  (currentPlayerIterator->second)->getObjects();

		PlayerObjectMap::iterator it;
		
		for(it = currPlayerObjects->begin();it != currPlayerObjects->end(); ++it){
			// FIXME later make this work for bases as well
			Group* currentObject = dynamic_cast<Group*> (it->second);
			if(currentObject == NULL) continue;
			Vector3D currentObjectHeading;
			if(currentObject->getOrder() == NULL){
				currentObjectHeading = Vector3D(0.0,0.0,0.0);
			}else{
				currentObjectHeading = (currentObject->getOrder()->getDestination()-currentObject->getPosition())|currentObject->getOrder()->getSpeed();
			}
			
			for(otherPlayerIterator = players.begin(); otherPlayerIterator != players.end(); ++otherPlayerIterator){
				if(currentPlayerIterator == otherPlayerIterator) continue; // don't check your own objects...
				PlayerObjectMap* otherPlayerObjects =  (otherPlayerIterator->second)->getObjects();
				PlayerObjectMap::iterator it2;
				for(it2 = otherPlayerObjects->begin();it2 != otherPlayerObjects->end(); ++it2){
					Group* otherObject = dynamic_cast<Group*> (it2->second);
					if(otherObject == NULL) continue;
					
					// if objects are already within each other's range, don't do them anymore.
					
					if((currentObject->getPosition() - otherObject->getPosition()).length() <= currentObject->getExtents()+currentObject->getMaxAttackRange()+otherObject->getExtents()){
						print("inside object!!!");
						continue;
					}
					
					Vector3D otherObjectHeading;
					if(otherObject->getOrder() == NULL){
						otherObjectHeading = Vector3D(0.0,0.0,0.0);
					}else{
						otherObjectHeading = (otherObject->getOrder()->getDestination()-otherObject->getPosition())|otherObject->getOrder()->getSpeed();
					}
				
					Vector3D relativeHeading = currentObjectHeading - otherObjectHeading;
					
					double collisionTime = getCirclePointCollisionTime(currentObject->getPosition(),currentObject->getExtents()+currentObject->getMaxAttackRange()+otherObject->getExtents(),otherObject->getPosition(),relativeHeading,currentTime);
					
					print("found time: "+convertToString(collisionTime));
					if(collisionTime == -1.0){
						print("skipping this one");
						continue;
					}
					
					if(collisionTime < minimumTime) minimumTime = collisionTime;
					
				}
			}
		}
	}
	return minimumTime;
}

double getNextShootingTime(PlayerMap& players,double currentTime){
	double minShotTime = DOUBLE_MAX;
	PlayerMap::iterator playerIterator;
	for(playerIterator = players.begin(); playerIterator != players.end(); ++playerIterator){
		PlayerObjectMap* currPlayerObjects =  (playerIterator->second)->getObjects();
		PlayerObjectMap::iterator it;
		for(it = currPlayerObjects->begin();it != currPlayerObjects->end(); ++it){
			Group* currentGroup = dynamic_cast<Group*> (it->second);
			if(currentGroup == NULL) continue; // if the current object is not a group, ignore it, since it can't attack
			if(currentGroup->getLastDistanceMoved() > 0.0) continue; // if it's moved, no need to check when it can shoot next
			
			std::vector<std::string> unitsWithEnemyInRange;
			currentGroup->enemyInRangeOfUnitType(unitsWithEnemyInRange);
				
			/*
			if(!currentGroup->enemyInRange()){
				print("enemy not in range");
				continue;
			}
			*/
			double currentShotTime = currentGroup->getNextShotTime(unitsWithEnemyInRange,currentTime);
			if(currentShotTime < minShotTime){
				minShotTime = currentShotTime;
			}
		}
	}
	return minShotTime;
}

double getNextOrderCompletionTime(PlayerMap& players,double currentTime){
	double minShotTime = DOUBLE_MAX;
	PlayerMap::iterator playerIterator;
	for(playerIterator = players.begin(); playerIterator != players.end(); ++playerIterator){
		PlayerObjectMap* currPlayerObjects =  (playerIterator->second)->getObjects();
		PlayerObjectMap::iterator it;
		for(it = currPlayerObjects->begin();it != currPlayerObjects->end(); ++it){
			Group* currentGroup = dynamic_cast<Group*> (it->second);
			if(currentGroup == NULL) continue; // if the current object is not a group, ignore it, since it can't attack
			if(currentGroup->getOrder() == NULL) continue;
			if(currentGroup->getPosition() == currentGroup->getOrder()->getDestination()) continue;
			
			print("unit " + currentGroup->getName() + " order start time: "+convertToString(currentTime));
					
			double estimatedTime = currentTime + (currentGroup->getOrder()->getDestination() - currentGroup->getPosition()).length() / ((double)currentGroup->getOrder()->getSpeed()/1000.0);
			
			if(estimatedTime == currentTime) continue;
			
			print("unit " + currentGroup->getName() + " order estimated end time: "+convertToString(estimatedTime));
			
			if(estimatedTime < minShotTime){
				minShotTime = estimatedTime;
			}
		}
	}
	return minShotTime/*+1.0*/;
}

double getNextTime(PlayerMap& players,double currentTime){
	double minTime = DOUBLE_MAX;
	
	// FIXME  collisions do not work right now... gets stuck after colliding
	
	double collisionTime = getNextCollisionTime(players,currentTime);
	print("min collision time: " + convertToString(collisionTime));
	if(collisionTime < minTime) minTime = collisionTime;
	
	double orderCompletionTime = getNextOrderCompletionTime(players,currentTime)+0.01;
	print("min order completion time: "+ convertToString(orderCompletionTime));
	if(orderCompletionTime < minTime) minTime = orderCompletionTime;
	
	double shootingTime = getNextShootingTime(players,currentTime);
	print("min shooting time: " + convertToString(shootingTime));
	if(shootingTime < minTime) minTime = shootingTime;
	
	if(minTime == DOUBLE_MAX){
		print("ERROR: no time found for simulation, this should never happen!");
		return currentTime;
	}
	
	if(minTime < currentTime){
		print("error, mintime of " + convertToString(minTime) + " is less than the currentTime of: " + convertToString(currentTime));
		return currentTime;
	}
	
	print("new minTime = " + convertToString(minTime));
	
	// IMPORTANT: the +0.01 is just to prevent some rounding off problems which can be inefficient, since some units don't quite reach their goal sometimes
	// it's equivalent to 1/100 milliseconds so not significant at all
	return minTime;
}

void advanceTime(PlayerMap& players){
	double lastMoveUpdateMillis = lastSimulationTime;
	lastSimulationTime = getNextTime(players,lastSimulationTime);
	moveObjects(players,lastSimulationTime - lastMoveUpdateMillis);
	//blah(playerMap,lastSimulationTime);
	calculateCollisions(playerMap,lastSimulationTime);
	
	// now we delete any objects that have < 0 hitpoints
	PlayerMap::iterator it;
	for(it = playerMap.begin();it != playerMap.end(); ++it){
		std::vector<MapObject*> deletedObjects; // get the addresses of all the deleted objects...
		(it->second)->checkForDestroyedObjects(deletedObjects);
		
		for(unsigned int i=0;i<deletedObjects.size();i++){
			MapObjectSet::iterator it2 = selectedObjects.find(deletedObjects[i]);
			if(it2 != selectedObjects.end()){
				selectedObjects.erase(it2);
			}
			delete deletedObjects[i];
		}
	}	
	//glutPostRedisplay();
}

