
#include "Strategies.h"

#include "Player.h"
#include "Group.h"
#include "Order.h"
#include "Base.h"
#include "UtilFunctions.h"


int massAttack(void* player){
	Player* currentPlayer = (Player*)player;
	PlayerObjectMap* playerObjects = currentPlayer->getObjects();

	SDL_cond* conditional = SDL_CreateCond();
	SDL_mutex* mutex = SDL_CreateMutex();

	TaskStruct* newTask = new TaskStruct;
	newTask->player = currentPlayer;
	newTask->conditional = conditional;

	PlayerObjectMap::iterator it;
	for(it = playerObjects->begin();it != playerObjects->end(); ++it){
		Group* group = dynamic_cast<Group*> (it->second);
		if(group != NULL){
			newTask->groups.push_back(group);
		}
	}

	SDL_CreateThread(join,(void*)newTask);

	SDL_mutexP(mutex);
	SDL_CondWait(conditional,mutex);
	SDL_mutexV(mutex);
	
	while(currentPlayer->getNearestEnemyBase(Vector3D(0.0,0.0,0.0)) != NULL){
		// do some attacking
		PlayerObjectMap::iterator it;
		for(it = playerObjects->begin();it != playerObjects->end(); ++it){
			Group* group = dynamic_cast<Group*> (it->second);
			if(group != NULL){
				if(group->getOrder() == NULL){
					Base* nearestBase =  currentPlayer->getNearestEnemyBase(group->getPosition());
					if(nearestBase != NULL){
						TaskStruct* newTask = new TaskStruct;
						newTask->player = currentPlayer;
						newTask->conditional = conditional;
						newTask->groups.push_back(group);
						newTask->location = nearestBase->getPosition();
							
						SDL_CreateThread(attackNearestBase,(void*)newTask);
					}
				}
			}
		}
		
		SDL_mutexP(mutex);
		SDL_CondWait(conditional,mutex);
		SDL_mutexV(mutex);
		
	}
	
	SDL_DestroyCond(conditional);
	SDL_DestroyMutex(mutex);
	conditional = NULL;
}

int spreadAttack(void* player){
	Player* currentPlayer = (Player*)player;
	PlayerObjectMap* playerObjects = currentPlayer->getObjects();
	
	SDL_cond* conditional = SDL_CreateCond();
	SDL_mutex* mutex = SDL_CreateMutex();
	
	while(currentPlayer->getNearestEnemyBase(Vector3D(0.0,0.0,0.0)) != NULL){
		// do some attacking
		PlayerObjectMap::iterator it;
		for(it = playerObjects->begin();it != playerObjects->end(); ++it){
			Group* group = dynamic_cast<Group*> (it->second);
			if(group != NULL){
				if(group->getOrder() == NULL){
					Base* nearestBase =  currentPlayer->getNearestEnemyBase(group->getPosition());
					if(nearestBase != NULL){
						TaskStruct* newTask = new TaskStruct;
						newTask->player = currentPlayer;
						newTask->conditional = conditional;
						newTask->groups.push_back(group);
						newTask->location = nearestBase->getPosition();
						
						SDL_CreateThread(attackNearestBase,(void*)newTask);
					}
				}
			}
		}
		
		SDL_mutexP(mutex);
		SDL_CondWait(conditional,mutex);
		SDL_mutexV(mutex);
		
	}
	
	SDL_DestroyCond(conditional);
	SDL_DestroyMutex(mutex);
	conditional = NULL;
}

// all units will attack the ONE base that is closest to any one of the units
int attackNearestBase(void* taskPtr){
	TaskStruct* task = (TaskStruct*)taskPtr;
	
	SDL_cond* conditional = SDL_CreateCond();
	SDL_mutex* mutex = SDL_CreateMutex();
	
	TaskStruct* newTask = new TaskStruct;
	newTask->groups = task->groups;
	newTask->taskType = "ATTACK";
	newTask->conditional = conditional;
	
	double minDistance = 1000000.0;
	Base* nearestBase = NULL;
	// get the nearest base of this group of units...
	for(unsigned int i=0; i<task->groups.size();++i){
		Base* currNearBase = task->player->getNearestEnemyBase((task->groups)[i]->getPosition());
		if(currNearBase != NULL){
			double currDistance = (currNearBase->getPosition() - (task->groups)[i]->getPosition()).length();
			if(currDistance < minDistance){
				minDistance = currDistance;
				nearestBase = currNearBase;
			}
		}
	}
	
	if(nearestBase != NULL){
		newTask->location = nearestBase->getPosition();
		SDL_CreateThread(executeTask,(void*)newTask);
		
		SDL_mutexP(mutex);
		SDL_CondWait(conditional,mutex);
		SDL_mutexV(mutex);
	}
	
	SDL_mutexP(mutex);
	SDL_CondBroadcast(task->conditional);
	SDL_mutexV(mutex);
	
	SDL_DestroyCond(conditional);
	SDL_DestroyMutex(mutex);
	conditional = NULL;
	delete task;
}

int join(void* taskPtr){
	TaskStruct* task = (TaskStruct*)taskPtr;
	
	SDL_cond* conditional = SDL_CreateCond();
	SDL_mutex* mutex = SDL_CreateMutex();
	
	TaskStruct* newTask = new TaskStruct;
	newTask->groups = task->groups;
	newTask->taskType = "JOIN";
	newTask->conditional = conditional;
	
	newTask->location = getCenterOfGroups(newTask->groups);
	
	SDL_CreateThread(executeTask,(void*)newTask);
	
	SDL_mutexP(mutex);
	SDL_CondWait(conditional,mutex);
	SDL_mutexV(mutex);
	
	SDL_mutexP(mutex);
	SDL_CondBroadcast(task->conditional);
	SDL_mutexV(mutex);
	
	SDL_DestroyCond(conditional);
	SDL_DestroyMutex(mutex);
	conditional = NULL;
	
	delete task;
}

int move(void* taskPtr){
	TaskStruct* task = (TaskStruct*)taskPtr;
	
	SDL_cond* conditional = SDL_CreateCond();
	SDL_mutex* mutex = SDL_CreateMutex();

	TaskStruct* newTask = new TaskStruct;
	newTask->groups = task->groups;
	newTask->location = task->location;
	newTask->taskType = "MOVE";
	newTask->conditional = conditional;

	SDL_CreateThread(executeTask,(void*)newTask);

	SDL_mutexP(mutex);
	SDL_CondWait(conditional,mutex);
	SDL_mutexV(mutex);
	
	SDL_mutexP(mutex);
	SDL_CondBroadcast(task->conditional);
	SDL_mutexV(mutex);

	SDL_DestroyCond(conditional);
	SDL_DestroyMutex(mutex);
	conditional = NULL;
	delete task;
}


int executeTask(void* taskPtr){
	TaskStruct* task = (TaskStruct*)taskPtr;

	SDL_cond* conditional = SDL_CreateCond();
	SDL_mutex* mutex = SDL_CreateMutex();
	
	unsigned int numIncompleteTasks=0;
	for(unsigned int i=0;i<task->groups.size();++i){
		Group* currGroup = (task->groups)[i];
		currGroup->setOrder(new Order(task->taskType,task->location,currGroup->getMaxSpeed()),conditional,&numIncompleteTasks);
		//print("number of not done tasks: " + convertToString(numIncompleteTasks));
	}
	
	while(numIncompleteTasks > 0){
		SDL_mutexP(mutex);
		SDL_CondWait(conditional,mutex);
		SDL_mutexV(mutex);
	}
	
	SDL_mutexP(mutex);
	SDL_CondBroadcast(task->conditional);
	SDL_mutexV(mutex);
	
	SDL_DestroyCond(conditional);
	SDL_DestroyMutex(mutex);
	conditional = NULL;
	delete task;
}

Vector3D getCenterOfGroups(std::vector<Group*> groups){
	double maxX = -100000.0;
	double maxY = -100000.0;
	double minX = 1000000.0;
	double minY = 1000000.0;
	
	for(unsigned int i=0;i<groups.size();++i){
		Vector3D position = groups[i]->getPosition();
		if(position.x > maxX) maxX = position.x;
		if(position.y > maxY) maxY = position.y;
		if(position.x < minX) minX = position.x;
		if(position.y < minY) minY = position.y;
	}
	
	return Vector3D((maxX + minX)/2.0,(maxY+minY)/2.0,0.0);
}


