// $Id: dcdtpath_main.C,v 1.7 2006/06/01 20:24:08 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "GfxGlobal.H"
#include "Options.H"
#include "GfxModule.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "PWidget.H"
#include "ClientState.H"
#include "MyEventHandler.H"
#include "Game.H"
#include "SegmentMerge.H"

using namespace std;

const char *ORTSG_VERSION = "GfxClient version 0.8 (2004-Oct-5)";

#define USE_EFENCE 0

extern int EF_PROTECT_FREE, EF_FILL, EF_PROTECT_BELOW, EF_ALLOW_MALLOC_0;

struct GfxEventHandler : public EventHandler
{
  ClientState &state;
  
  GfxEventHandler(ClientState &state_) : state(state_) { }

  bool handle_event(const Event& e)
  {
    if (e.get_who() == GameStateModule::FROM) {

      //cout << "Event: " << e.who() << " " << e.what() << endl;
      
      if (e.get_what() == GameStateModule::STOP_MSG) exit(0);

      if (e.get_what() == GameStateModule::READ_ERROR_MSG) exit(10);      
      
      if (e.get_what() == GameStateModule::VIEW_MSG) {

#if 0	
	if (state.gui) {
	  state.gui->event();
	  state.gui->display();
	  if (state.gui->quit) exit(0);
	}
#endif
	
	state.gfxm->process_changes();
	
	////////////////
	Vector<string> temp;
	Game &game = state.gsm->get_game();
	FORS (i, game.get_player_num()) temp.push_back(""); // dummy list of commands
	game.exec_actions(temp); // evaluate client actions and send actions to the server
	////////////////
	
	state.gsm->send_actions();
	
	sint4 vf = game.get_view_frame();
	sint4 ai = game.get_action_frame();
	sint4 sk = game.get_skipped_actions();
	sint4 md = game.get_merged_actions();
	
	if (vf != ai || sk || md > 1) {
	    
          cout << "frame " << vf;
          if (ai < 0) cout << " [no action]";
          else if (vf != ai) cout << " [behind by " << (vf - ai) << " frame(s)]";
          if (sk) cout << " [skipped " << sk << "]";
          if (md > 1) cout << " [merged " << md << "]";
          cout << endl;
	}
	
	// don't draw if we are far behind
	if (abs(vf-ai) < 10){
          state.gfxm->draw();
          state.just_drew = true;
	}
	
	return 1;
      }
      
    }
    
    return 0;
  }
};


class PolypathLooper : public LoopFunctor
{
public :
  PolypathLooper(ClientState* state_) : state(state_) {}
  
  void loop() const {

    if (state->gfxm->get_quit()) exit(0);

    bool refresh;
    Options::get("-refresh", refresh);

    if (refresh) {

      if (!state->just_drew) // allow some waiting time after a message
        state->gfxm->draw();
      else
        state->just_drew = false;
    }

    // server events
    if (!state->gsm->recv_view()) SDL_Delay(1);
  }

private:
  ClientState* state;
};


void set_label(SegmentMerge &sm)
{
  static int n = 1;
  
  FORS (i, 4) {
    FORALL (sm.add[i], it) {
      (**it).label = n++; // old
    }
  }
}
  
void write_delta(const SegmentMerge &sm)
{
  cout  << "DELETE" << endl;
  
  FORS (i, 4) {
    cout << "kind " << i << endl;
    FORALL (sm.del[i], it) {
      cout << *it << endl;
    }
  }

  cout << "ADD" << endl;
  
  FORS (i, 4) {
    cout << "kind " << i << endl;
    FORALL (sm.add[i], it) {
      (**it).write(); cout << endl;
    }
  }
}


int main(int argc, char **argv)
{
  glutInit(&argc, argv);

#if 0

  // test merging

  { // horizontal 
    SegmentMerge sm;
    SortSegment ss;
    Vector<SortSegment> segs;

    segs.push_back(SortSegment(0, 0, 1, 0));
    segs.push_back(SortSegment(2, 0, 3, 0));  

    FORALL (segs, it) sm.new_segment(*it);
    write_delta(sm);
    set_label(sm);

    cout << endl;
    cout << "------------------------" << endl;
    cout << endl;

    sm.clear_delta();
    sm.new_segment(SortSegment(1, 0, 2, 0));
    write_delta(sm);  
    set_label(sm);
    // must be [0,0]-[3,0]        
  }
  
  { // vertical 
    SegmentMerge sm;
    SortSegment ss;
    Vector<SortSegment> segs;

    segs.push_back(SortSegment(0, 0, 0, 1));
    segs.push_back(SortSegment(0, 2, 0, 3));  

    FORALL (segs, it) sm.new_segment(*it);
    write_delta(sm);
    set_label(sm);

    cout << endl;
    cout << "------------------------" << endl;
    cout << endl;

    sm.clear_delta();
    sm.new_segment(SortSegment(0, 1, 0, 2));
    write_delta(sm);  
    set_label(sm);
    // must be [0,0]-[0,3]    
  }
  { //   dm
    SegmentMerge sm;
    SortSegment ss;
    Vector<SortSegment> segs;

    segs.push_back(SortSegment(0, 0, 1, 1));
    segs.push_back(SortSegment(2, 2, 3, 3));  

    FORALL (segs, it) sm.new_segment(*it);
    write_delta(sm);
    set_label(sm);

    cout << endl;
    cout << "------------------------" << endl;
    cout << endl;

    sm.clear_delta();
    sm.new_segment(SortSegment(1, 1, 2, 2));
    write_delta(sm);  
    set_label(sm);
    // must be [0,0]-[3,3]    
  }
  { //   dp
    SegmentMerge sm;
    SortSegment ss;
    Vector<SortSegment> segs;

    segs.push_back(SortSegment(0, 0, 1, -1));
    segs.push_back(SortSegment(2, -2, 3, -3));  

    FORALL (segs, it) sm.new_segment(*it);
    write_delta(sm);
    set_label(sm);

    cout << endl;
    cout << "------------------------" << endl;
    cout << endl;

    sm.clear_delta();
    sm.new_segment(SortSegment(1, -1, 2, -2));
    write_delta(sm);  
    set_label(sm);

    // must be [0,0]-[3,-3]
  }
  exit(0);

#endif

  
#if 0
  
  if (argc == 2 && argv[1][0] == 't'){

    //cout<<argv[1]<<endl;

    SectorDecomp* decomposition = new SectorDecomp(500, 500, 500);

    //insert some rectangles
    Decomp::Rect r;

    r.set(100,200,100,200);
    decomposition->insert(r, 1, 1);
	 
    r.set(150,250,150,250);
    decomposition->insert(r, 1, 1);

    cout<<"\n/****START****/\n";
    
    FORALL (decomposition->sectors, i) {
		
      FORALL(i->nodes,j){
	Decomp::Node *n = *j;
		    
	if(n->type!=Decomp::Node::R_NODE) continue;
		    
	n->write();
      }
		
    }
    cout<<"/****END****/\n";

    PolyGraph* myGraph = new PolyGraph;

    myGraph->build_graph(&(decomposition->sectors.front()));

    myGraph->print_graph();
	  
    exit(0);
  }
#endif
  
#if USE_EFENCE
  // aggressive efence checks
  EF_PROTECT_FREE = 1;
  EF_FILL = 0xee;
  EF_PROTECT_BELOW = 1;
  EF_ALLOW_MALLOC_0 = 1;
#endif

  signals_shut_down(true);

  Options opt("DCDT Path");
  opt.put("-redundant", "redundant graph mode");
  opt.put("-dynamic",  "handling dynamic obstacles");
  
  // populate opt with options from various modules
  
  GameStateModule::Options::add();
  GfxModule::Options::add();
  // path_add_options();

  // process command line arguments
  if (Options::process(argc, argv, cerr, ORTSG_VERSION)) exit(20);

  SDLinit::video_init();
  SDLinit::network_init();
  
  // populate option objects with command line values
  // and create modules
  
  GameStateModule::Options gsmo;
  GameStateModule gsm(gsmo);
 
  GfxModule::Options gfxmo;
  GfxModule gfxm;

  // gives event handler access to the state
  ClientState state(gsm, gfxm);

  // ortsg event handler
  GfxEventHandler gev(state);
  MyEventHandler mev(state);

  //bool value=false;
  //Options::get("-redundant",value);
  //mev.set_graph_mode(value);
  
  //value = false;  
  //Options::get("-dynamic", value);
  // mev.copying_available(value);
  //cout<<"good"<<endl;
  
#if 0  
  // create pathfinding module with command line options
  PathFinder astar_pf(state, opt);

  // register pathfinder as gfx and gsm event handler
  gfxm.add_handler(&astar_pf);
  gsm.add_handler(&astar_pf);
#endif
  
  // gsm events also handled in gev and mev
  gsm.add_handler(&gev);
  gsm.add_handler(&mev);
  gfxm.add_handler(&mev);
  
  // connect game state module to server
  if (!gsm.connect()) ERR("connection problems");
  gfxm.init(gsm, gfxmo);

  // gfx loop (should become a thread)
  PolypathLooper l(&state);
  gfxm.start_loop(&l);

#if 0
  // old
  Client client(opt);
  Gfxmlient gfxm;
  ClientState state(client, gfxm);
  GfxEventHandler gev(state);
  MyEventHandler mev(state);

  bool value=false;
  Options::get("-redundant",value);
  mev.set_graph_mode(value);
  
  value = false;  
  Options::get("-dynamic", value);
  mev.copying_available(value);
  //cout<<"good"<<endl;
  
  client.add_handler(&gev);
  client.add_handler(&mev);
  client.connect();
  gfxm.init(client);
 
  PolypathLooper l(&state);
  gfxm.start_loop(&l);
#endif  

  /*
    FOREVER {

    // gui events
    state.gfxm->handle_events();
    if (state.gfxm->get_quit()) break;

    bool refresh;
    Options::get("-refresh", refresh);

    if (refresh) {

    if (!state.just_drew) // allow some waiting time after a message
    state.gfxm->draw();
    else
    state.just_drew = false;
    }

    // server events
    if (!state.gsm->server_event()) SDL_Delay(1);
    }
  */
  return 0;
}
