//-----------------------------------------------------------------------------
/** @file example.cpp
    Example program using the pathfinding library.

    $Id: hpastar_main.cpp,v 1.2 2005/01/14 20:04:47 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/apps/hpastar/src/hpastar_main.cpp,v $
*/
//-----------------------------------------------------------------------------

#include <memory>
#include "pathfind.h"
#include "abswizard.h"
#include "smoothwizard.h"
#include "experiment.h"

using namespace std;
using namespace PathFind;

//-----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  for (int i = 0; i < argc; i++) cerr << argv[i] << " ";
  cerr << "\n";

  if (argc < 6) { cerr << "<6 parameters" << endl; exit(10); }
  
  int readFromFile = atoi(argv[1]);
  int nrRuns = atoi(argv[2]);
  int clSize = atoi(argv[3]);
  int maxLevel = atoi(argv[4]);
  int llSearch = atoi(argv[5]);

  if (readFromFile == 0) {

    if (argc != 9) { cerr << "!=9 parameters" << endl; exit(10); }
    int rows = atoi(argv[6]);
    int cols = atoi(argv[7]);
    float obstacle = atof(argv[8]);
    try
      {
	//srand(time(0));
	Experiment experiment(nrRuns, 10000000L, rows, cols,
			      obstacle, (bool)llSearch, true, true,
			      clSize, maxLevel, AbsWizard::END_ENTRANCE,
			      A_STAR, Tiling::OCTILE);
	experiment.setupExperiment();
	experiment.runExperiment();
      }
    catch (const exception& e)
      {
	cerr << "Error: " << e.what() << '\n';
	return -1;
      }
  }
  else if (readFromFile == 1) {

    if (argc != 9) { cerr << "!=9 parameters" << endl; exit(10); }
    
    int start = atoi(argv[6]);
    int target = atoi(argv[7]);
    try
      {
	//srand(time(0));
	Experiment experiment(nrRuns, 10000000L, 0, 0,
			      0, (bool)llSearch, true, true,
			      clSize, maxLevel, AbsWizard::END_ENTRANCE,
			      A_STAR, Tiling::OCTILE);
	experiment.setupExperiment();
	experiment.runExperiment(start, target, argv[8]);
      }
    catch (const exception& e)
      {
	cerr << "Error: " << e.what() << '\n';
	return -1;
      }
  }
  else if (readFromFile == 2) {
    if (argc != 7) { cerr << "!=7 parameters" << endl; exit(10); }      
    try
      {
	//srand(time(0));
	Experiment experiment(nrRuns, 10000000L, 0, 0,
			      0, (bool)llSearch, true, true,
			      clSize, maxLevel, AbsWizard::END_ENTRANCE,
			      A_STAR, Tiling::OCTILE);
	experiment.setupExperiment();
	experiment.runExperiment(argv[6]);
      }
    catch (const exception& e)
      {
	cerr << "Error: " << e.what() << '\n';
	return -1;
      }
  }
  else if (readFromFile == 3) {
    if (argc != 7) { cerr << "!=7 parameters" << endl; exit(10); }    
    try
      {
	//srand(time(0));
	Experiment experiment(nrRuns, 10000000L, 0, 0,
			      0, (bool)llSearch, true, true,
			      clSize, maxLevel, AbsWizard::END_ENTRANCE,
			      A_STAR, Tiling::OCTILE);
	experiment.setupExperiment();
	experiment.runStorageExperiment(argv[6]);
      }
    catch (const exception& e)
      {
	cerr << "Error: " << e.what() << '\n';
	return -1;
      }
  }
  return 0;
}

//-----------------------------------------------------------------------------
