#include <iostream>

#ifdef OS_MAC
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#include <GL/glut.h>
#endif

//#include "main.h"
#include "ma_trackball.h"
#include "ma_common.h"
#include "a_star.h"
#include "pra_star.h"
#include "pra_starUnit.h"
#include "humanUnit.h"
#include "searchUnit.h"
#include "LRTS.h"
#include "PR_LRTS.h"
#include "hLRTS.h"
#include "sarsa.h"
#include "LRTSUnit.h"
#include "PR_LRTSUnit.h"
#include "hLRTSUnit.h"
#include "sarsaUnit.h"
#include "sharedAMapGroup.h"
#include "LRTSGroup.h"
#include "PR_LRTSGroup.h"
#include "hLRTSGroup.h"
#include "sarsaGroup.h"
#include "unitRaceSimulation.h"
#include "unitRewardSimulation.h"
#include "rewardUnit.h"
#include "constants.h"

// For printing debug info
static bool const verbose = false;

char gDefaultMap[1024] = "";
recVec gOrigin = { 0.0, 0.0, 0.0 };

using namespace std;

// intializes context conditions
void initialConditions(pRecContext pContextInfo)
{
#ifdef OS_MAC
	BlockZero (pContextInfo, sizeof (recContext));
#endif
	resetCamera (&pContextInfo->camera);
	pContextInfo->fVel[0] = 0.3;
	pContextInfo->fVel[1] = 0.1;
	pContextInfo->fVel[2] = 0.2; 
	pContextInfo->fAccel[0] =  0.003;
	pContextInfo->fAccel[1] = -0.005;
	pContextInfo->fAccel[2] =  0.004;
	pContextInfo->info = kInfoState;
	pContextInfo->animate = kAnimateState;
	pContextInfo->drawCaps = 0;
	pContextInfo->drawHelp = 1;
	pContextInfo->polygons = 1;
	pContextInfo->lines = 0;
	pContextInfo->points = 0;
	pContextInfo->showCredits = 1;
	pContextInfo->showLand = 1;
	pContextInfo->lighting = 1;
	pContextInfo->paused = false;
	pContextInfo->drawing = true;
	
	pContextInfo->worldRotation[0] = 0;
	pContextInfo->worldRotation[1] = 0;
	pContextInfo->worldRotation[2] = 0;
	pContextInfo->worldRotation[3] = 0;
	pContextInfo->objectRotation[0] = 180;
	pContextInfo->objectRotation[1] = 0;
	pContextInfo->objectRotation[2] = -0.5;
	pContextInfo->objectRotation[3] = 0.5;
	
#ifdef OS_MAC
	pContextInfo->timer = NULL;
#endif
	
	//	gTrackBallRotation[0] = 180;
	//	gTrackBallRotation[1] = 0;
	//	gTrackBallRotation[2] = -0.5;
	//	gTrackBallRotation[3] = 0.5;
	//	
	//	addToRotationTrackball(gTrackBallRotation, pContextInfo->worldRotation);	
	gTrackBallRotation [0] = gTrackBallRotation [1] = gTrackBallRotation [2] = gTrackBallRotation [3] = 0.0f;
	
	pContextInfo->surface = 0; 
	pContextInfo->colorScheme = 4;
	pContextInfo->subdivisions = 64;
	pContextInfo->xyRatio = 1;
	
	pContextInfo->modeFSAA = 0;
	
	if (gDefaultMap[0] == 0)
		pContextInfo->map = new Map(64, 64);
	else
		pContextInfo->map = new Map(gDefaultMap);
	pContextInfo->abstrMap = 0;
	//pContextInfo->abstrMap = new abstractMap(pContextInfo->map);
	pContextInfo->unitLayer = new unitSimulation(pContextInfo->map);
	pContextInfo->human = 0;
}

void SetLighting(unsigned int mode)
{
	GLfloat mat_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat mat_shininess[] = {90.0};
	
	GLfloat position[4] = {7.0,-7.0,12.0,0.0};
	GLfloat ambient[4]  = {0.2,0.2,0.2,1.0};
	GLfloat diffuse[4]  = {1.0,1.0,1.0,1.0};
	GLfloat specular[4] = {1.0,1.0,1.0,1.0};
	
	glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
	glMaterialfv (GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
	
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	
	switch (mode) {
		case 0:
			break;
		case 1:
			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
			glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
			break;
		case 2:
			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
			glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
			break;
		case 3:
			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
			glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
			break;
		case 4:
			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
			glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
			break;
	}
	
	glLightfv(GL_LIGHT0,GL_POSITION,position);
	glLightfv(GL_LIGHT0,GL_AMBIENT,ambient);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuse);
	glLightfv(GL_LIGHT0,GL_SPECULAR,specular);
	glEnable(GL_LIGHT0);
}

bool doKeyboardCommand(pRecContext pContextInfo, char keyHit, bool shift)
{
	bool rebuild = false;
	
	switch (tolower(keyHit))
	{
		case 'e':
		{
			int x1, y1;
			delete pContextInfo->unitLayer;
			pContextInfo->unitLayer = new unitRewardSimulation(pContextInfo->map);
			pContextInfo->unitLayer->setUseBlocking(false);
			pContextInfo->unitLayer->getRandomLocation(x1, y1);
			rewardSeekingUnit *rsu = new rewardSeekingUnit(x1, y1);
			
			for (int x = 0; x < 5; x++)
			{
				pContextInfo->unitLayer->getRandomLocation(x1, y1);
				rewardUnit *ru = new rewardUnit(x1, y1);
				rsu->addRewardLocation(ru);
				pContextInfo->unitLayer->addUnit(ru);
			}
			rsu->setSpeed(0.01);
			pContextInfo->unitLayer->addUnit(rsu);
		}
		case '\\':
			pContextInfo->unitLayer->setRealTime(!pContextInfo->unitLayer->getRealTime());
			break;
		case '\t':
			if (shift && 
				pContextInfo->unitLayer->getUnitGroup(pContextInfo->unitLayer->getDisplayMapNumber()-1) != NULL)
				pContextInfo->unitLayer->getUnitGroup(pContextInfo->unitLayer->getDisplayMapNumber()-1)->cycleDisplayMode();
			else {
				pContextInfo->unitLayer->cycleAbstractMapDisplay();
				// ... here comes code to display the group number in Open GL
			}
			break;			
		case '/':
			// runRace(gDefaultMap,-1,-1,-1,-1,1,1.0,999999.0,2,false);
			// reset view to front view. Doesn't fix camera position
			// or zoom
			//			if (pContextInfo->worldRotation[0] != 0)
			//			{
			//				for (int x = 0; x < 4; x++)
			//					gTrackBallRotation[x] = -pContextInfo->worldRotation[x];
			//				gTrackBallRotation[0] = -gTrackBallRotation[0];
			//				addToRotationTrackball(gTrackBallRotation, pContextInfo->worldRotation);
			//				gTrackBallRotation [0] = gTrackBallRotation [1] = gTrackBallRotation [2] = gTrackBallRotation [3] = 0.0f;
			//			}
			break;
		case '[':
			if (pContextInfo->unitLayer->getUnitGroup(pContextInfo->unitLayer->getDisplayMapNumber()-1) != NULL)
				pContextInfo->unitLayer->getUnitGroup(pContextInfo->unitLayer->getDisplayMapNumber()-1)->decreaseDisplayALevel();
			break;
			
		case ']':
			if (pContextInfo->unitLayer->getUnitGroup(pContextInfo->unitLayer->getDisplayMapNumber()-1) != NULL)
				pContextInfo->unitLayer->getUnitGroup(pContextInfo->unitLayer->getDisplayMapNumber()-1)->increaseDisplayALevel();
			break;			
		case 'p':
			if (shift)
				pContextInfo->drawing = !pContextInfo->drawing;
			else
				pContextInfo->paused = !pContextInfo->paused;
			break;
		case 'q':
		{
			for (int x = 0; x < (shift?100:1); x++)
			{
				graph *g = pContextInfo->unitLayer->getAbstractMap()->getAbstractGraph(0);
				node *n = g->getRandomNode();
				if (n)
				{
					pContextInfo->unitLayer->getAbstractMap()->removeNode(n);
					pContextInfo->unitLayer->getAbstractMap()->getMap()->setTerrainType(n->getLabelL(kFirstData),
																						n->getLabelL(kFirstData+1), kOutOfBounds);
					//					pContextInfo->unitLayer->getAbstractMap()->clearMarkedNodes();
				}
				//				if (!g->verifyGraph())
				//					break;
			}
			pContextInfo->unitLayer->getAbstractMap()->repairAbstraction();
			pContextInfo->unitLayer->getAbstractMap()->verifyHierarchy();
		}
			break;
		case 'v':
			pContextInfo->unitLayer->getAbstractMap()->verifyHierarchy();
			if (pContextInfo->unitLayer->getAbstractMap()->getAbstractGraph(0)->verifyGraph())
				printf("graph (0) verifies\n");
			else
				printf("graph (0) doesn't verify\n");
			break;
		case '.':
		{
			pContextInfo->unitLayer->addUnit(new teleportUnit(0, 0, 10)); // move every 10 steps
		}
			break;
		case ',':
		{
			pContextInfo->unitLayer->addUnit(new rhrUnit(0, 0));
		}
			break;
			// Clear all units
		case '|':
			pContextInfo->unitLayer->clearAllUnits();
			pContextInfo->unitLayer->getAbstractMap()->clearMarkedNodes();
			pContextInfo->human = 0;
			break;
		case 's':
		{
			delete pContextInfo->unitLayer;
			pContextInfo->unitLayer = new unitRaceSimulation(pContextInfo->map);
			pContextInfo->unitLayer->setLogFile(stdout);
			//			praStar *pra = new praStar(); pra->setPartialPathLimit(4);
			//			unit *u;
			//			pContextInfo->unitLayer->addUnit(u = new teleportUnit(0, 0, 0));
			//			pContextInfo->unitLayer->addUnit(new searchUnit(30, 30, u, pra));
		}	
			break;
		case 'd':
			for (int x = 0; x < (shift?(50):(1)); x++)
			{
				if (pContextInfo->unitLayer->getUnitGroup(0) == 0)
				{
					pContextInfo->unitLayer->addUnitGroup(new sharedAMapGroup());
					pContextInfo->unitLayer->setAbstractMapDisplay(1);
				}
				int xx1, yy1, xx2, yy2;
				pContextInfo->unitLayer->getRandomLocation(xx1, yy1);
				pContextInfo->unitLayer->getRandomLocation(xx2, yy2);
				
				unit *u, *u2 = new unit(xx2, yy2, 0);
				
				praStar *pra = new praStar(); pra->setPartialPathLimit(4);
				
				pContextInfo->unitLayer->addUnit(u2);
				u = new searchUnit(xx1, yy1, u2, pra);
				// just set the group of the unit, and it will share a map with those
				// units.
				pContextInfo->unitLayer->getUnitGroup(0)->addUnit(u);
				pContextInfo->unitLayer->addUnit(u);
				u->setSpeed(0); // time to go 1 distance						
			}
			break;
			// Add 1 (when 'a' is pressed) or 100 (when 'A' is pressed) PRA* units chasing random goal states
		case 'a':
			pContextInfo->unitLayer->setAsynchronous();
			for (int x = 0; x < (shift?100:1); x++)
			{
				int xx1, yy1, xx2, yy2;
				pContextInfo->unitLayer->getRandomLocation(xx1, yy1);
				pContextInfo->unitLayer->getRandomLocation(xx2, yy2);

				// xx1 = 31;
				// yy1 = 62;
				// xx2 = 91;
				// yy2 = 120;
				
				unit *u, *u2 = new unit(xx2, yy2, 0);
						
				//aStar *pra = new aStar();
				praStar *pra = new praStar(); pra->setPartialPathLimit(4);
				
				pContextInfo->unitLayer->addUnit(u2);
				pContextInfo->unitLayer->addUnit(u = new searchUnit(xx1, yy1, u2, pra));
				u->setSpeed(0); // time to go 1 distance
				
				//unit *u2 = new unit(xx2, yy2, 0);
				//pContextInfo->unitLayer->addUnit(u2);
//			for (int y = 0; y < 10; y++)
//				{
//					pra = new praStar(); pra->setPartialPathLimit(3);
//					pra->setSimulationEnvironment(pContextInfo->unitLayer);
//					
//					pContextInfo->unitLayer->addUnit(u2 = new searchUnit(xx1+y/2, yy1+y%2, u, pra));
//					u2->setSpeed(0);
//				}
			}
				break;
			
			// Add 1 (when 'z' is pressed) or 4 (when 'Z' is pressed) LRTS units chasing random goal states
			// Each unit will have their own copy of the LRTS algorithm (and, therefore, heuristic)
		case 'z':
			pContextInfo->unitLayer->setSynchronous();
			for (int x = 0; x < (shift?4:1); x++)
			{
				// We will try to create start and goal states until we succeed 
				// and then we will break out of the loop
				while (1) {
					// Compute coordinates of the goal state
					int xx1 = random()%pContextInfo->map->getMapWidth();
					int yy1 = random()%pContextInfo->map->getMapHeight();
					
					// Compute coordinates of the start state
					int xx2 = random()%pContextInfo->map->getMapWidth();
					int yy2 = random()%pContextInfo->map->getMapHeight();
					
					// If both are on the ground (as opposed to water?) then
					if ((pContextInfo->map->getTerrainType(xx1, yy1) == kGround) && (pContextInfo->map->getTerrainType(xx2, yy2) == kGround))
					{
						// Create the static goal state
						unit *u, *u2 = new unit(xx2, yy2, 0);
						pContextInfo->unitLayer->addUnit(u2);
						
						// Create a copy of the LRTS algorithm LRTS(gamma,d,T)
						LRTS *LRTSAlgNew = new LRTS(1.0f, 1, 0);
						printf("Create a new instance of LRTS algorithm\n");
						
						
						// Create the LRTS unit
						pContextInfo->unitLayer->addUnit(u = new LRTSUnit(xx1, yy1, u2, LRTSAlgNew));
						u->setSpeed(0.1);
						
						break;
					}
				}
			}
				break;
			
		// Creates a racing environment between PR_LRTS and PRA* and LRTS
		case 'x': 
		{
		    // Set up the control parameters
		    double penalty = 1.0;			// thinking penalty multiplier
		    double speed = 0.0;				// speed 
		    double minDifficulty = 10;		// minimum allowed difficulty of the path
		    double maxDifficulty = 20;		// maximum allowed difficulty of the path
		    int minLength = 0; //pContextInfo->map->getMapHeight(); // min allowed path length
		    unsigned numFails = 1000;		// number of attempts before relaxing constraints
			const static int visRadius = 10;	// vis radius
		    
		    // switch over to racing simulation
		    delete pContextInfo->unitLayer;
		    pContextInfo->unitLayer = new unitRaceSimulation(pContextInfo->map);
			
			
			// Open the trial log file
			char tLName[1034];
			for (int i=0; i<1034; tLName[i++]='\0');
			if (gDefaultMap[0] != '\0')
				for (int i=0; (gDefaultMap[i] != '.' && 
							   gDefaultMap[i] != '\0' && i<1033); i++) 
					tLName[i] = gDefaultMap[i];
			else
				strcpy(tLName,"default");
			strcat(tLName,".tlog.txt");
			FILE *trialLog = fopen(tLName,"w+");
			pContextInfo->unitLayer->setLogFile(trialLog);
			
			// Maps and unit simulation
			unitSimulation *unitSim = pContextInfo->unitLayer;
			Map *map = pContextInfo->map;
			abstractMap* aMap = unitSim->getAbstractMap();
			sharedAMapGroup* praGroup = NULL;
			
			// Creating a group for PRA* unit
			if (unitSim->getUnitGroup(0) == NULL)
			{
				praGroup = new sharedAMapGroup();
				unitSim->addUnitGroup(praGroup);
				praGroup->setVisibilityRadius(visRadius);
			}
			
			
			// Set parameters
			unitSim->setAsynchronous();
			unitSim->setPenalty(penalty);
			unitSim->setUseBlocking(false);
			
			// A hack to fix HOG's problem with lighting on Mac
			pContextInfo->lighting = 1;
			SetLighting(pContextInfo->lighting);
			
			
			// Instantiate the A* algorithm to make sure the start and goal states are
			// at a reasonable distance from each other
			aStar* myAStar = new aStar();
			
			// Get the base-level graph for this map
			graph *g = aMap->getAbstractGraph(0);
			
			
			// Create one goal, one PR_LRTS and one PRA* unit. Let them race
			unsigned failures = 0;
			while (true) {
				int xx1 = random()%map->getMapWidth();
				int yy1 = random()%map->getMapHeight();
				int xx2 = random()%map->getMapWidth();
				int yy2 = random()%map->getMapHeight();
				
//				xx2 = 31; yy2 = 62;			// start
//				xx1 = 91; yy1 = 120;		// goal
				
				// Compute the optimal path between the two
				int startNodeNum = map->getNodeNum(xx2,yy2);
				int goalNodeNum = map->getNodeNum(xx1,yy1);
				node* start = g->getNode(startNodeNum);
				node* goal = g->getNode(goalNodeNum);
				
				// Compute the length of the optimal path between them
				unsigned expanded = 0;			// number of expanded states
				path* pathBetween = myAStar->getPath(aMap, start, goal);
				expanded = myAStar->getNodesTouched();
				// Check if the path exists
				if (pathBetween == 0)
					continue;           // nope, continue generating the coordinates
				
				// Check of the path between them is reasonable
				int pathLength = pathBetween->length();
				double noDist = aMap->h(start,goal);
				double spDist = pathBetween->distance(aMap);
				double difficulty = spDist/noDist;
				// If A* had h* then A* would have expanded this many states
				unsigned minAStar = pathBetween->degree();
				// This allows us to compute "A* difficulty"
				double aStarDifficulty = (double)expanded / (double)minAStar;
				
				
				// Don't allow paths that are too short
				if (pathLength < minLength) {
					// Once again, we failed to find an appropriate path
					// Once we exceed numFails attempts we will relax the constraints
					if (++failures > numFails) {
						minLength /= 2;
						failures = 0;
					}
					continue;
				}
				
				// Check if the path difficulty is within allowed limits
				if (difficulty < minDifficulty || difficulty > maxDifficulty) {
					// Once again, we failed to find an appropriate path
					// Once we exceed numFails attempts we will relax the constraints
					if (++failures > numFails) {
						minDifficulty *= 0.9;
						maxDifficulty *= 1.1;
						failures = 0;
					}
					continue;
				}
				
				// Print statistics
				printf("\nPenalty:\t\t\t\t\t%1.3f\n",penalty);
				printf("Start state:\t\t\t\t(%d,%d)\nGoal state:\t\t\t\t(%d,%d)\n",xx2,yy2,xx1,yy1);
				printf("Path length:\t\t\t\t%d\n",pathLength);
				printf("Eucleadian distance:\t\t%1.2f\n",
					   std::sqrt((double)(xx2-xx1)*(xx2-xx1)+(yy2-yy1)*(yy2-yy1)));
				
				printf("No-obstacle distance:\t%1.2f\n",noDist);
				printf("Shortest path distance:\t%1.2f\n",spDist);
				printf("Distance difficulty:\t\t%1.2f\n",difficulty);
				printf("A* difficulty:\t\t\t%1.2f\n\n",aStarDifficulty);
				fflush(stdout);
				
				// Ok, all checks clear. We got a good start and goal state pair
				// Create the static goal state (of white color)
				unit *u = new unit(xx1, yy1, 255, 255, 255, 0);
				unitSim->addUnit(u);
				
				// Create A*
				// praStar *pra = new praStar(); //pra->setPartialPathLimit(7);
				aStar *pra = new aStar();
				
				
				searchUnit* aSUnit = new searchUnit(xx2, yy2, 255, 0, 0, u, pra);
				praGroup->addUnit(aSUnit);
				aSUnit->setSpeed(speed);
				unitSim->addUnit(aSUnit);
				
				// Ok, we got the units in the simulation, break out of the creation loop
				break;
			}
		}
		break;
			
			// Creates a racing environment between aLRTS and PRA* or hLRTS and PRA* 
			// when Shift-U is pressed
		case 'u': 
			if (!shift)
			{
				// Set up the control parameters
				const double penalty = 1.0;			// thinking penalty multiplier
				const double speed = 0.0;			// speed 
				double minDifficulty = 10;			// minimum allowed difficulty of the path
				double maxDifficulty = 20;		// maximum allowed difficulty of the path
				int minLength = 0; // pContextInfo->map->getMapHeight(); // min allowed path length
				const unsigned numFails = 1000;			// number of attempts before relaxing constraints
				const int visRadius = 10;				// visibility radius
				
				// switch over to racing simulation
				delete pContextInfo->unitLayer;
				pContextInfo->unitLayer = new unitRaceSimulation(pContextInfo->map);
				
				// Open the trial log file
				char tLName[1034];
				for (int i=0; i<1034; tLName[i++]='\0');
				if (gDefaultMap[0] != '\0')
					for (int i=0; (gDefaultMap[i] != '.' && 
								   gDefaultMap[i] != '\0' && i<1033); i++) 
						tLName[i] = gDefaultMap[i];
				else
					strcpy(tLName,"default");
				strcat(tLName,".tlog.txt");
				FILE *trialLog = fopen(tLName,"w+");
				pContextInfo->unitLayer->setLogFile(trialLog);
				
				// Maps and unit simulation
				unitSimulation *unitSim = pContextInfo->unitLayer;
				Map *map = pContextInfo->map;
				abstractMap* aMap = unitSim->getAbstractMap();
				
				// Creating a group for PRA* unit
				if (unitSim->getUnitGroup(0) == NULL)
				{
					sharedAMapGroup* praGroup = new sharedAMapGroup();
					unitSim->addUnitGroup(praGroup);
					praGroup->setVisibilityRadius(visRadius);
				}
				
				// Creating a group for the LRTS unit
				if (unitSim->getUnitGroup(1) == NULL)
				{
					unitSim->addUnitGroup(new LRTSGroup());
				}
				
				// Set parameters
				unitSim->setAsynchronous();
				unitSim->setPenalty(penalty);
				unitSim->setUseBlocking(true);
				
				// A hack to fix HOG's problem with lighting on Mac
				pContextInfo->lighting = 1;
				SetLighting(pContextInfo->lighting);
				
				
				// Instantiate the A* algorithm to make sure the start and goal states are
				// at a reasonable distance from each other
				aStar* myAStar = new aStar();
				
				// Get the base-level graph for this map
				graph *g = aMap->getAbstractGraph(0);
				
				// Create one goal, one LRTS and one PRA* unit. Let them race
				unsigned failures = 0;
				while (true) {
					int xx1 = random()%map->getMapWidth();
					int yy1 = random()%map->getMapHeight();
					int xx2 = random()%map->getMapWidth();
					int yy2 = random()%map->getMapHeight();

					xx2 = 39; yy2 = 10;			// start (for CSC 2F)
					xx1 = 14; yy1 = 45;		// goal
										
					// Compute the optimal path between the two
					int startNodeNum = map->getNodeNum(xx2,yy2);
					int goalNodeNum = map->getNodeNum(xx1,yy1);
					node* start = g->getNode(startNodeNum);
					node* goal = g->getNode(goalNodeNum);
					
					// Compute the length of the optimal path between them
					unsigned expanded = 0;			// number of expanded states
					path* pathBetween = myAStar->getPath(aMap, start, goal);
					expanded = myAStar->getNodesTouched();
					
					// Check if the path exists
					if (pathBetween == 0)
						continue;           // nope, continue generating the coordinates
					
					// Check of the path between them is reasonable
					int pathLength = pathBetween->length();
					double noDist = aMap->h(start,goal);
					double spDist = pathBetween->distance(aMap);
					double difficulty = spDist/noDist;
					// If A* had h* then A* would have expanded this many states
					unsigned minAStar = pathBetween->degree();
					delete pathBetween;
					// This allows us to compute "A* difficulty"
					double aStarDifficulty = (double)expanded / (double)minAStar;
					
					
					// Don't allow paths that are too short
					if (pathLength < minLength) {
						// Once again, we failed to find an appropriate path
						// Once we exceed numFails attempts we will relax the constraints
						if (++failures > numFails) {
							minLength /= 2;
							failures = 0;
						}
						continue;
					}
					
					// Check if the path difficulty is within allowed limits
					if (aStarDifficulty < minDifficulty || aStarDifficulty > maxDifficulty) {
						// Once again, we failed to find an appropriate path
						// Once we exceed numFails attempts we will relax the constraints
						if (++failures > numFails) {
							minDifficulty *= 0.9;
							maxDifficulty *= 1.1;
							failures = 0;
						}
						continue;
					}
					
					// Print statistics
					printf("\nPenalty:\t\t\t\t\t%1.3f\n",penalty);
					printf("Speed:\t\t\t\t\t%1.3f\n",speed);
					printf("Vis rad:\t\t\t\t\t%2d\n",visRadius);
					printf("Start state:\t\t\t\t(%d,%d)\nGoal state:\t\t\t\t(%d,%d)\n",xx2,yy2,xx1,yy1);
					printf("Path length:\t\t\t\t%d\n",pathLength);
					printf("Eucleadian distance:\t\t%1.2f\n",
						   std::sqrt((double)(xx2-xx1)*(xx2-xx1)+(yy2-yy1)*(yy2-yy1)));
					
					printf("No-obstacle distance:\t%1.2f\n",noDist);
					printf("Shortest path distance:\t%1.2f\n",spDist);
					printf("Distance difficulty:\t\t%1.2f\n",difficulty);
					printf("A* difficulty:\t\t\t%1.2f\n\n",aStarDifficulty);
					fflush(stdout);
					
					// Ok, all checks clear. We got a good start and goal state pair
					// Create the static goal state (of white color)
					searchUnit *u1, *u3;
					unit *u = new unit(xx1, yy1, 255, 255, 255, NULL);
					unitSim->addUnit(u);
					
					
					// Create PRA*
					//praStar *pra = new praStar(); 
					// aStar *pra = new aStar(); 
					// pra->setPartialPathLimit(4);
					
#pragma mark GUI LRTS
					// Create LRTS
					const static int lookaheadDepth = 2;
					assert(lookaheadDepth <= visRadius);
					LRTS *lrts = new LRTS(1.0, lookaheadDepth, 10000000000.0);
					
					// Create the units and add them to the simulation
//					u1 = new searchUnit(xx2, yy2, 255, 0, 0, u, pra);
//					unitSim->getUnitGroup(0)->addUnit(u1);
//					u1->setSpeed(speed);
//					unitSim->addUnit(u1,false);
					
					u3 = new LRTSUnit(xx2, yy2, 0, 255, 0, u, lrts);
					u3->setSpeed(speed);
					unitSim->getUnitGroup(1)->addUnit(u3);
					unitSim->addUnit(u3, false);
					
					// Create a number of random units
					const static int numRandom = 0;
					for (int i=0; i<numRandom; i++) {
						int xx, yy;
						do {
							xx = random()%map->getMapWidth();
							yy = random()%map->getMapHeight();
						}
						while ((map->getTerrainType(xx, yy) != kGround) &&
							   !(xx==xx1 && yy==yy1) && !(xx==xx2 && yy==yy2));

						randomUnit* ru = new randomUnit(xx, yy, 10, 128, 10);
						ru->setObjectType(kIncidentalUnit);
						unitSim->addUnit(ru,true);
						ru->setSpeed(0.02);
					}
					
					// Ok, we got the units in the simulation, break out of the creation loop
					break;
				}
				
			}
			else
			{
				// Set up the control parameters
				double penalty = 1.0;			// thinking penalty multiplier
				double speed = 0.0;				// speed 
				double minDifficulty = 1.5;		// minimum allowed difficulty of the path
				double maxDifficulty = 5.0;		// maximum allowed difficulty of the path
				int minLength = pContextInfo->map->getMapHeight(); // min allowed path length
				unsigned numFails = 1000;		// number of attempts before relaxing constraints
				
				// switch over to racing simulation
				delete pContextInfo->unitLayer;
				pContextInfo->unitLayer = new unitRaceSimulation(pContextInfo->map);
				
				
				// Open the trial log file
				char tLName[1034];
				for (int i=0; i<1034; tLName[i++]='\0');
				if (gDefaultMap[0] != '\0')
					for (int i=0; (gDefaultMap[i] != '.' && 
								   gDefaultMap[i] != '\0' && i<1033); i++) 
						tLName[i] = gDefaultMap[i];
				else
					strcpy(tLName,"default");
				strcat(tLName,".tlog.txt");
				FILE *trialLog = fopen(tLName,"w+");
				
				pContextInfo->unitLayer->setLogFile(trialLog);
				
				// Maps and unit simulation
				unitSimulation *unitSim = pContextInfo->unitLayer;
				Map *map = pContextInfo->map;
				abstractMap* aMap = unitSim->getAbstractMap();
				
				// Creating a group for PRA* unit
				if (unitSim->getUnitGroup(0) == NULL)
					unitSim->addUnitGroup(new sharedAMapGroup());
				
				// Creating a group for the hLRTS unit
				if (unitSim->getUnitGroup(1) == NULL)
					unitSim->addUnitGroup(new hLRTSGroup());
				
				// Set parameters
				unitSim->setAsynchronous();
				unitSim->setPenalty(penalty);
				unitSim->setUseBlocking(true);
				
				// A hack to fix HOG's problem with lighting on Mac
				pContextInfo->lighting = 1;
				SetLighting(pContextInfo->lighting);
				
				// Instantiate the A* algorithm to make sure the start and goal states are
				// at a reasonable distance from each other
				aStar* myAStar = new aStar();
				
				// Get the base-level graph for this map
				graph *g = aMap->getAbstractGraph(0);
				
				// Create one goal, one hLRTS and one PRA* unit. Let them race
				unsigned failures = 0;
				while (true) {
					int xx1 = random()%map->getMapWidth();
					int yy1 = random()%map->getMapHeight();
					int xx2 = random()%map->getMapWidth();
					int yy2 = random()%map->getMapHeight();
					
					// Compute the optimal path between the two
					int startNodeNum = map->getNodeNum(xx2,yy2);
					int goalNodeNum = map->getNodeNum(xx1,yy1);
					node* start = g->getNode(startNodeNum);
					node* goal = g->getNode(goalNodeNum);
					
					// Compute the length of the optimal path between them
					unsigned expanded = 0;			// number of expanded states
					path* pathBetween = myAStar->getPath(aMap, start, goal);
					expanded = myAStar->getNodesTouched();
					
					// Check if the path exists
					if (pathBetween == 0)
						continue;           // nope, continue generating the coordinates
					
					// Check of the path between them is reasonable
					int pathLength = pathBetween->length();
					double noDist = aMap->h(start,goal);
					double spDist = pathBetween->distance(aMap);
					double difficulty = spDist/noDist;
					// If A* had h* then A* would have expanded this many states
					unsigned minAStar = pathBetween->degree();
					delete pathBetween;
					// This allows us to compute "A* difficulty"
					double aStarDifficulty = (double)expanded / (double)minAStar;
					
					
					// Don't allow paths that are too short
					if (pathLength < minLength) {
						// Once again, we failed to find an appropriate path
						// Once we exceed numFails attempts we will relax the constraints
						if (++failures > numFails) {
							minLength /= 2;
							failures = 0;
						}
						continue;
					}
					
					// Check if the path difficulty is within allowed limits
					if (difficulty < minDifficulty || difficulty > maxDifficulty) {
						// Once again, we failed to find an appropriate path
						// Once we exceed numFails attempts we will relax the constraints
						if (++failures > numFails) {
							minDifficulty *= 0.9;
							maxDifficulty *= 1.1;
							failures = 0;
						}
						continue;
					}
					
					// Print statistics
					printf("\nPenalty:\t\t\t\t\t%1.3f\n",penalty);
					printf("Start state:\t\t\t\t(%d,%d)\nGoal state:\t\t\t\t(%d,%d)\n",xx2,yy2,xx1,yy1);
					printf("Path length:\t\t\t\t%d\n",pathLength);
					printf("Eucleadian distance:\t\t%1.2f\n",
						   std::sqrt((double)(xx2-xx1)*(xx2-xx1)+(yy2-yy1)*(yy2-yy1)));
					
					printf("No-obstacle distance:\t%1.2f\n",noDist);
					printf("Shortest path distance:\t%1.2f\n",spDist);
					printf("Distance difficulty:\t\t%1.2f\n",difficulty);
					printf("A* difficulty:\t\t\t%1.2f\n\n",aStarDifficulty);
					fflush(stdout);
					
					// Ok, all checks clear. We got a good start and goal state pair
					// Create the static goal state (of white color)
					searchUnit *u1, *u3;
					unit *u = new unit(xx1, yy1, 255, 255, 255, 0);
					unitSim->addUnit(u);
					
					// Create PRA*
					praStar *pra = new praStar(); //pra->setPartialPathLimit(4);
					
#pragma mark GUI hLRTS
					// Create hLRTS
					hLRTS *hlrts = new hLRTS(1.0, 3);    // gamma, aLevel
					
					// Create the units and add them to the simulation
					u1 = new searchUnit(xx2, yy2, 255, 0, 0, u, pra);
					pContextInfo->unitLayer->getUnitGroup(0)->addUnit(u1);
					u1->setSpeed(speed);
					unitSim->addUnit(u1,false);
					
					u3 = new hLRTSUnit(xx2, yy2, 0, 0, 255, u, hlrts);
					u3->setSpeed(speed);
					unitSim->addUnit(u3,false);
					pContextInfo->unitLayer->getUnitGroup(1)->addUnit(u3);
					
					
					// Create a number of random units
					const static int numRandom = 0;
					for (int i=0; i<numRandom; i++) {
						int xx, yy;
						do {
							xx = random()%map->getMapWidth();
							yy = random()%map->getMapHeight();
						}
						while ((map->getTerrainType(xx, yy) != kGround) &&
							   !(xx==xx1 && yy==yy1) && !(xx==xx2 && yy==yy2));
						
						randomUnit* ru = new randomUnit(xx,yy,10,128,10);
						ru->setObjectType(kIncidentalUnit);
						unitSim->addUnit(ru,true);
						ru->setSpeed(0.02);
					}
					
					// Ok, we got the units in the simulation, break out of the creation loop
					break;
				}
				
			}
			break;
			
				
		// Create a racing environment between sarsa and PRA*
		case ';':
		{
				// Set up the control parameters
				double penalty = 1.0;			// thinking penalty multiplier
				double speed = 0.0;				// speed 
				double minDifficulty = 1.0;		// minimum allowed difficulty of the path
				double maxDifficulty = 5.0;		// maximum allowed difficulty of the path
				int minLength = pContextInfo->map->getMapHeight(); // min allowed path length
				unsigned numFails = 1000;		// number of attempts before relaxing constraints
				
				// Set up the parameters for sarsa:
				double alpha = 0.1;
				double gamma = 1.0;
				double lambda = 0.3;
				
				// switch over to racing simulation
				delete pContextInfo->unitLayer;
				pContextInfo->unitLayer = new unitRaceSimulation(pContextInfo->map);
				
				
				// Open the trial log file
				char tLName[1034];
				for (int i=0; i<1034; tLName[i++]='\0');
				if (gDefaultMap[0] != '\0')
					for (int i=0; (gDefaultMap[i] != '.' && 
								   gDefaultMap[i] != '\0' && i<1033); i++) 
						tLName[i] = gDefaultMap[i];
				else
					strcpy(tLName,"default");
				strcat(tLName,".tlog.txt");
				FILE *trialLog = fopen(tLName,"w+");
				pContextInfo->unitLayer->setLogFile(trialLog);
				
				// Maps and unit simulation
				unitSimulation *unitSim = pContextInfo->unitLayer;
				Map *map = pContextInfo->map;
				abstractMap* aMap = unitSim->getAbstractMap();
				
				// Creating a group for PRA* unit
				if (unitSim->getUnitGroup(0) == NULL)
					// unitSim->addUnitGroup(new sharedAMapGroup());
					unitSim->addUnitGroup(new LRTSGroup());
				
				// Creating a group for the sarsa unit
				if (unitSim->getUnitGroup(1) == NULL)
					unitSim->addUnitGroup(new sarsaGroup());
				
				// Set parameters
				unitSim->setAsynchronous();
				unitSim->setPenalty(penalty);
				unitSim->setUseBlocking(true);
				
				// A hack to fix HOG's problem with lighting on Mac
				pContextInfo->lighting = 1;
				SetLighting(pContextInfo->lighting);
				
				// Instantiate the A* algorithm to make sure the start and goal states are
				// at a reasonable distance from each other
				aStar* myAStar = new aStar();
				
				// Get the base-level graph for this map
				graph *g = aMap->getAbstractGraph(0);
				
				// Create one goal, one sarsa and one PRA* unit. Let them race
				unsigned failures = 0;
				while (true) {
					int xx1 = random()%map->getMapWidth();
					int yy1 = random()%map->getMapHeight();
					int xx2 = random()%map->getMapWidth();
					int yy2 = random()%map->getMapHeight();
					
					// Compute the optimal path between the two
					int startNodeNum = map->getNodeNum(xx2,yy2);
					int goalNodeNum = map->getNodeNum(xx1,yy1);
					node* start = g->getNode(startNodeNum);
					node* goal = g->getNode(goalNodeNum);
					
					// Compute the length of the optimal path between them
					unsigned expanded = 0;			// number of expanded states
					path* pathBetween = myAStar->getPath(aMap, start, goal);
					expanded = myAStar->getNodesTouched();
					
					// Check if the path exists
					if (pathBetween == NULL)
						continue;           // nope, continue generating the coordinates
					
					// Check of the path between them is reasonable
					int pathLength = pathBetween->length();
					double noDist = aMap->h(start,goal);
					double spDist = pathBetween->distance(aMap);
					double difficulty = spDist/noDist;
					// If A* had h* then A* would have expanded this many states
					unsigned minAStar = pathBetween->degree();
					delete pathBetween;
					// This allows us to compute "A* difficulty"
					double aStarDifficulty = (double)expanded / (double)minAStar;
					
					
					// Don't allow paths that are too short
					if (pathLength < minLength) {
						// Once again, we failed to find an appropriate path
						// Once we exceed numFails attempts we will relax the constraints
						if (++failures > numFails) {
							minLength /= 2;
							failures = 0;
						}
						continue;
					}
					
					// Check if the path difficulty is within allowed limits
					if (difficulty < minDifficulty || difficulty > maxDifficulty) {
						// Once again, we failed to find an appropriate path
						// Once we exceed numFails attempts we will relax the constraints
						if (++failures > numFails) {
							minDifficulty *= 0.9;
							maxDifficulty *= 1.1;
							failures = 0;
						}
						continue;
					}
					
					// Print statistics
					printf("\nPenalty:\t\t\t\t\t%1.3f\n",penalty);
					printf("Start state:\t\t\t\t(%d,%d)\nGoal state:\t\t\t\t(%d,%d)\n",xx2,yy2,xx1,yy1);
					printf("Path length:\t\t\t\t%d\n",pathLength);
					printf("Eucleadian distance:\t\t%1.2f\n",
						   std::sqrt((double)(xx2-xx1)*(xx2-xx1)+(yy2-yy1)*(yy2-yy1)));
					
					printf("No-obstacle distance:\t%1.2f\n",noDist);
					printf("Shortest path distance:\t%1.2f\n",spDist);
					printf("Distance difficulty:\t\t%1.2f\n",difficulty);
					printf("A* difficulty:\t\t\t%1.2f\n\n",aStarDifficulty);
					fflush(stdout);
					
					// Ok, all checks clear. We got a good start and goal state pair
					// Create the static goal state (of white color)
					searchUnit *u1, *u3;
					unit *u = new unit(xx1, yy1, 255, 255, 255, 0);
					unitSim->addUnit(u);
					
					// Create PRA*
					// praStar *pra = new praStar(); //pra->setPartialPathLimit(4);
					const static int lookaheadDepth = 1;
					LRTS *lrts = new LRTS(1.0, lookaheadDepth, 999999999.0);
					
#pragma mark GUI SARSA
					// Create sarsa
					sarsa *sarsarl = new sarsa();
					
					// Create the units and add them to the simulation
					u1 = new LRTSUnit(xx2, yy2, 255, 0, 0, u, lrts);
					pContextInfo->unitLayer->getUnitGroup(0)->addUnit(u1);
					u1->setSpeed(speed);
					unitSim->addUnit(u1,false);
					
					u3 = new sarsaUnit(xx2, yy2, 0, 0, 255, u, sarsarl, alpha, gamma, lambda);
					u3->setSpeed(speed);
					unitSim->addUnit(u3,false);
					pContextInfo->unitLayer->getUnitGroup(1)->addUnit(u3);
					
					
					// Create a number of random units
					const static int numRandom = 0;
					for (int i=0; i<numRandom; i++) {
						int xx, yy;
						do {
							xx = random()%map->getMapWidth();
							yy = random()%map->getMapHeight();
						}
						while ((map->getTerrainType(xx, yy) != kGround) &&
							   !(xx==xx1 && yy==yy1) && !(xx==xx2 && yy==yy2));
						
						randomUnit* ru = new randomUnit(xx,yy,10,128,10);
						ru->setObjectType(kIncidentalUnit);
						unitSim->addUnit(ru,true);
						ru->setSpeed(0.02);
					}
					
					// Ok, we got the units in the simulation, break out of the creation loop
					break;
				}
				break;
			}

			
		// Creates a racing environment between aLRTS and aLRTS dynamic lookahead
		case 'o': 
		{
			// Set up the control parameters
			double penalty = 1.0;			// thinking penalty multiplier
			double speed = 0.0;				// speed 
			double minDifficulty = 10;		// minimum allowed difficulty of the path
			double maxDifficulty = 20;		// maximum allowed difficulty of the path
			int minLength = pContextInfo->map->getMapHeight(); // min allowed path length
			unsigned numFails = 1000;		// number of attempts before relaxing constraints
			
			// switch over to racing simulation
			delete pContextInfo->unitLayer;
			pContextInfo->unitLayer = new unitRaceSimulation(pContextInfo->map);
			
			// Open the trial log file
			char tLName[1034];
			for (int i=0; i<1034; tLName[i++]='\0');
			if (gDefaultMap[0] != '\0')
				for (int i=0; (gDefaultMap[i] != '.' && 
							   gDefaultMap[i] != '\0' && i<1033); i++) 
					tLName[i] = gDefaultMap[i];
			else
				strcpy(tLName,"default");
			strcat(tLName,".tlog.txt");
			FILE *trialLog = fopen(tLName,"w+");
			pContextInfo->unitLayer->setLogFile(trialLog);
			
			// Maps and unit simulation
			unitSimulation *unitSim = pContextInfo->unitLayer;
			Map *map = pContextInfo->map;
			abstractMap* aMap = unitSim->getAbstractMap();
			
			// Creating a group for LRTS unit
			if (unitSim->getUnitGroup(0) == NULL)
			{
				unitSim->addUnitGroup(new LRTSGroup());
			}
				
				// Creating a group for the LRTS unit
				if (unitSim->getUnitGroup(1) == NULL)
				{
					unitSim->addUnitGroup(new LRTSGroup());
				}
				
				// Set parameters
				unitSim->setAsynchronous();
			unitSim->setPenalty(penalty);
			unitSim->setUseBlocking(false);
			
			// A hack to fix HOG's problem with lighting on Mac
			pContextInfo->lighting = 1;
			SetLighting(pContextInfo->lighting);
			
			
			// Instantiate the A* algorithm to make sure the start and goal states are
			// at a reasonable distance from each other
			aStar* myAStar = new aStar();
			
			// Get the base-level graph for this map
			graph *g = aMap->getAbstractGraph(0);
			
			
			// Create one goal, one LRTS and one PRA* unit. Let them race
			unsigned failures = 0;
			while (true) {
				int xx1 = random()%map->getMapWidth();
				int yy1 = random()%map->getMapHeight();
				int xx2 = random()%map->getMapWidth();
				int yy2 = random()%map->getMapHeight();
				
				// Compute the optimal path between the two
				int startNodeNum = map->getNodeNum(xx2,yy2);
				int goalNodeNum = map->getNodeNum(xx1,yy1);
				node* start = g->getNode(startNodeNum);
				node* goal = g->getNode(goalNodeNum);
				
				// Compute the length of the optimal path between them
				unsigned expanded = 0;			// number of expanded states
				path* pathBetween = myAStar->getPath(aMap, start, goal);
				expanded = myAStar->getNodesTouched();
				// Check if the path exists
				if (pathBetween == 0)
					continue;           // nope, continue generating the coordinates
				
				// Check of the path between them is reasonable
				int pathLength = pathBetween->length();
				double noDist = aMap->h(start,goal);
				double spDist = pathBetween->distance(aMap);
				double difficulty = spDist/noDist;
				// If A* had h* then A* would have expanded this many states
				unsigned minAStar = pathBetween->degree();
				delete pathBetween;
				// This allows us to compute "A* difficulty"
				double aStarDifficulty = (double)expanded / (double)minAStar;
				
				
				// Don't allow paths that are too short
				if (pathLength < minLength) {
					// Once again, we failed to find an appropriate path
					// Once we exceed numFails attempts we will relax the constraints
					if (++failures > numFails) {
						minLength /= 2;
						failures = 0;
					}
					continue;
				}
				
				// Check if the path difficulty is within allowed limits
				if (difficulty < minDifficulty || difficulty > maxDifficulty) {
					// Once again, we failed to find an appropriate path
					// Once we exceed numFails attempts we will relax the constraints
					if (++failures > numFails) {
						minDifficulty *= 0.9;
						maxDifficulty *= 1.1;
						failures = 0;
					}
					continue;
				}
				
				// Print statistics
				printf("\nPenalty:\t\t\t\t\t%1.3f\n",penalty);
				printf("Start state:\t\t\t\t(%d,%d)\nGoal state:\t\t\t\t(%d,%d)\n",xx2,yy2,xx1,yy1);
				printf("Path length:\t\t\t\t%d\n",pathLength);
				printf("Eucleadian distance:\t\t%1.2f\n",
					   std::sqrt((double)(xx2-xx1)*(xx2-xx1)+(yy2-yy1)*(yy2-yy1)));
				
				printf("No-obstacle distance:\t%1.2f\n",noDist);
				printf("Shortest path distance:\t%1.2f\n",spDist);
				printf("Distance difficulty:\t\t%1.2f\n",difficulty);
				printf("A* difficulty:\t\t\t%1.2f\n\n",aStarDifficulty);
				fflush(stdout);
				
				// Ok, all checks clear. We got a good start and goal state pair
				// Create the static goal state (of white color)
				searchUnit *u1, *u3;
				unit *u = new unit(xx1, yy1, 255, 255, 255, 0);
				unitSim->addUnit(u);
				
#pragma mark GUI dLRTS
				// Create LRTS
				LRTS *lrts1 = new LRTS(0.5f, 6, 99999999.0);    // gamma, d, T
				LRTS *lrts2 = new LRTS(1.0f, 6, 99999999.0);    // gamma, d, T
				
				// Create the units and add them to the simulation
				u1 = new LRTSUnit(xx2, yy2, 255, 0, 0, u, lrts1);
				pContextInfo->unitLayer->getUnitGroup(0)->addUnit(u1);
				u1->setSpeed(speed);
				unitSim->addUnit(u1);
				
				u3 = new LRTSUnit(xx2, yy2, 0, 0, 255, u, lrts2);
				u3->setSpeed(speed);
				unitSim->addUnit(u3);
				pContextInfo->unitLayer->getUnitGroup(1)->addUnit(u3);
				
				// Ok, we got the units in the simulation, break out of the creation loop
				break;
			}
				break;
			
		}
		// Creates a racing environment between PR_LRTS and PRA* and LRTS
		case 'f': 
		{
		    // Set up the control parameters
		    double penalty = 1.0;			// thinking penalty multiplier
		    double speed = 0.0;				// speed 
		    double minDifficulty = 0;		// minimum allowed difficulty of the path
		    double maxDifficulty = 1000;		// maximum allowed difficulty of the path
		    int minLength = 0; // pContextInfo->map->getMapHeight(); // min allowed path length
		    unsigned numFails = 1000;		// number of attempts before relaxing constraints
			const static int visRadius = 10;	// vis radius
		    
		    // switch over to racing simulation
		    delete pContextInfo->unitLayer;
		    pContextInfo->unitLayer = new unitRaceSimulation(pContextInfo->map);
			
			
			// Open the trial log file
			char tLName[1034];
			for (int i=0; i<1034; tLName[i++]='\0');
			if (gDefaultMap[0] != '\0')
				for (int i=0; (gDefaultMap[i] != '.' && 
							   gDefaultMap[i] != '\0' && i<1033); i++) 
					tLName[i] = gDefaultMap[i];
			else
				strcpy(tLName,"default");
			strcat(tLName,".tlog.txt");
			FILE *trialLog = fopen(tLName,"w+");
			pContextInfo->unitLayer->setLogFile(trialLog);
			
			// Maps and unit simulation
			unitSimulation *unitSim = pContextInfo->unitLayer;
			Map *map = pContextInfo->map;
			abstractMap* aMap = unitSim->getAbstractMap();

			// Creating a group for the PR_LRTS unit
			if (unitSim->getUnitGroup(0) == NULL)
			  {
				PR_LRTSGroup* prlrtsGroup = new PR_LRTSGroup();
				unitSim->addUnitGroup(prlrtsGroup);
				prlrtsGroup->setVisibilityRadius(visRadius);
			  }
			
			// Creating a group for PRA* unit
			if (unitSim->getUnitGroup(1) == NULL)
			  {
				sharedAMapGroup* praGroup = new sharedAMapGroup();
				unitSim->addUnitGroup(praGroup);
				praGroup->setVisibilityRadius(visRadius);
			  }

//			// Creating a group for the LRTS unit
//			if (unitSim->getUnitGroup(1) == NULL)
//			  {
//				unitSim->addUnitGroup(new LRTSGroup());
//			  }
//			
						
			// Set parameters
			unitSim->setAsynchronous();
			unitSim->setPenalty(penalty);
			unitSim->setUseBlocking(false);
			
			// A hack to fix HOG's problem with lighting on Mac
			pContextInfo->lighting = 1;
			SetLighting(pContextInfo->lighting);
			
			
			// Instantiate the A* algorithm to make sure the start and goal states are
			// at a reasonable distance from each other
			aStar* myAStar = new aStar();
			
			// Get the base-level graph for this map
			graph *g = aMap->getAbstractGraph(0);
			
			
			// Create one goal, one PR_LRTS and one PRA* unit. Let them race
			unsigned failures = 0;
			while (true) {
				int xx1 = random()%map->getMapWidth();
				int yy1 = random()%map->getMapHeight();
				int xx2 = random()%map->getMapWidth();
				int yy2 = random()%map->getMapHeight();
				
				xx2 = 39; yy2 = 10;			// start (for CSC 2F)
				xx1 = 14; yy1 = 45;		// goal
				
				// Compute the optimal path between the two
				int startNodeNum = map->getNodeNum(xx2,yy2);
				int goalNodeNum = map->getNodeNum(xx1,yy1);
				node* start = g->getNode(startNodeNum);
				node* goal = g->getNode(goalNodeNum);
				
				// Compute the length of the optimal path between them
				unsigned expanded = 0;			// number of expanded states
				path* pathBetween = myAStar->getPath(aMap, start, goal);
				expanded = myAStar->getNodesTouched();
				// Check if the path exists
				if (pathBetween == 0)
					continue;           // nope, continue generating the coordinates
				
				// Check of the path between them is reasonable
				int pathLength = pathBetween->length();
				double noDist = aMap->h(start,goal);
				double spDist = pathBetween->distance(aMap);
				double difficulty = spDist/noDist;
				// If A* had h* then A* would have expanded this many states
				unsigned minAStar = pathBetween->degree();
				delete pathBetween;
				// This allows us to compute "A* difficulty"
				double aStarDifficulty = (double)expanded / (double)minAStar;
				
				
				// Don't allow paths that are too short
				if (pathLength < minLength) {
					// Once again, we failed to find an appropriate path
					// Once we exceed numFails attempts we will relax the constraints
					if (++failures > numFails) {
						minLength /= 2;
						failures = 0;
					}
					continue;
				}
				
				// Check if the path difficulty is within allowed limits
				if (difficulty < minDifficulty || difficulty > maxDifficulty) {
					// Once again, we failed to find an appropriate path
					// Once we exceed numFails attempts we will relax the constraints
					if (++failures > numFails) {
						minDifficulty *= 0.9;
						maxDifficulty *= 1.1;
						failures = 0;
					}
					continue;
				}
				
				// Print statistics
				printf("\nPenalty:\t\t\t\t\t%1.3f\n",penalty);
				printf("Start state:\t\t\t\t(%d,%d)\nGoal state:\t\t\t\t(%d,%d)\n",xx2,yy2,xx1,yy1);
				printf("Path length:\t\t\t\t%d\n",pathLength);
				printf("Eucleadian distance:\t\t%1.2f\n",
					   std::sqrt((double)(xx2-xx1)*(xx2-xx1)+(yy2-yy1)*(yy2-yy1)));
				
				printf("No-obstacle distance:\t%1.2f\n",noDist);
				printf("Shortest path distance:\t%1.2f\n",spDist);
				printf("Distance difficulty:\t\t%1.2f\n",difficulty);
				printf("A* difficulty:\t\t\t%1.2f\n\n",aStarDifficulty);
				fflush(stdout);
				
				// Ok, all checks clear. We got a good start and goal state pair
				// Create the static goal state (of white color)
				searchUnit *u1; // *u2; // *u3;
				unit *u = new unit(xx1, yy1, 255, 255, 255, 0);
				unitSim->addUnit(u);
				
				// Create A*
				// praStar *pra = new praStar(); //pra->setPartialPathLimit(7);
				// aStar *pra = new aStar();
				
#pragma mark GUI PR_LRTS
				// Create PR_LRTS
				PR_LRTS *pr_lrts = new PR_LRTS(1, 2, 10000000000.0);    // gamma, d, learningQuota

				// Create LRTS
				// LRTS *lrts = new LRTS(0.5, 7, 0.0);    // gamma, d, learningQuota
				
				// Create the units and add them to the simulation
				u1 = new PR_LRTSUnit(xx2, yy2, 0, 255, 0, u, pr_lrts);
				u1->setSpeed(speed);
				pContextInfo->unitLayer->getUnitGroup(0)->addUnit(u1);
				unitSim->addUnit(u1);
				
//				u2 = new searchUnit(xx2, yy2, 255, 0, 0, u, pra);
//				pContextInfo->unitLayer->getUnitGroup(1)->addUnit(u2);
//				u2->setSpeed(speed);
//				unitSim->addUnit(u2);
				
//				u3 = new LRTSUnit(xx2, yy2, 0, 0, 255, u, lrts);
//				u3->setSpeed(speed);
//				pContextInfo->unitLayer->getUnitGroup(1)->addUnit(u3);
//				unitSim->addUnit(u3);
				
				// Ok, we got the units in the simulation, break out of the creation loop
				break;
			}
		}
			break;
			
		case '`':
			pContextInfo->showLand =  1 - pContextInfo->showLand;
			pContextInfo->map->setDrawLand(pContextInfo->showLand);
			break;
		case '0': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?10:0); break;
		case '1': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?11:1); break;
		case '2': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?12:2); break;
		case '3': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?13:3); break;
		case '4': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?14:4); break;
		case '5': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?15:5); break;
		case '6': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?16:6); break;
		case '7': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?17:7); break;
		case '8': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?18:8); break;
		case '9': if (pContextInfo->unitLayer->getAbstractMap())
			pContextInfo->unitLayer->getAbstractMap()->toggleDrawAbstraction((shift)?19:9); break;
		case ' ':
		{
			if (shift)
			{
				pContextInfo->map->scale(512, 512);
			}
			else {
				int width = pContextInfo->map->getMapWidth();
				int height = pContextInfo->map->getMapHeight();
				for (int x = 0; x < (width+height)/4; x++)
				{
					if (random()%2)
					{
						int x1 = (random()%(width-2))+1;
						int y1 = (random()%(height-2))+1;
						int x2 = random()%(width/5);
						int y2 = random()%(height/5);
						pContextInfo->map->smoothSetRectHeight(x1, y1,
																 ((x1+x2)>=width-1)?(x1-x2):(x1+x2),
																 ((y1+y2)>=height-1)?(y1-y2):(y1+y2),
																 0, kWater);
						//0, (random()%5)?kWater:kSwamp);
					}
					else
						pContextInfo->map->smoothSetRectHeight((random()%(width-2))+1, (random()%(height-2))+1,
																 (random()%(width-2))+1, (random()%(height-2))+1,
																 random()%5+1, kGround);
				}
			}
			rebuild = true;
		}
			break;
			
		case 'r':
		{
			delete pContextInfo->unitLayer;
			pContextInfo->unitLayer = new unitRaceSimulation(pContextInfo->map);
			pContextInfo->unitLayer->setAsynchronous();
			pContextInfo->unitLayer->setPenalty(1.0);
			pContextInfo->unitLayer->setUseBlocking(false);
			((unitRaceSimulation*)pContextInfo->unitLayer)->setUseSameStart(false);
			if (shift)
				pContextInfo->unitLayer->setMoveStochasticity(0); // 33% of moves fail

			int xx1, yy1, xx2, yy2;
			do {
				pContextInfo->unitLayer->getRandomLocation(xx1, yy1);
				pContextInfo->unitLayer->getRandomLocation(xx2, yy2);
			} while (!pContextInfo->unitLayer->getAbstractMap()->
							 pathable(pContextInfo->map->getNodeNum(xx1, yy1),
												pContextInfo->map->getNodeNum(xx2, yy2)));
			unit *u, *u1, *u2, *u3 = new randomUnit(xx2, yy2);
			
			praStar *pra = new praStar();
			pra->setPartialPathLimit(4);
			//pra->setExpandSearchRadius(false);
			aStar *a = new aStar();
			
			pContextInfo->unitLayer->addUnit(u3, false);
			pContextInfo->unitLayer->addUnit(u = new searchUnit(xx1, yy1, 255, 0, 0, u3, pra), false);
			pContextInfo->unitLayer->addUnit(u1 = new searchUnit(xx1, yy1, 0, 255, 0, u3, a), false);
			pContextInfo->unitLayer->addUnit(u2 = new praStarUnit(xx1, yy1, 0, 0, 255, u3, pra), false);
			u->setSpeed(0); // time to go 1 distance
			u1->setSpeed(0); // time to go 1 distance
			u2->setSpeed(0); // time to go 1 distance
			
			for (int i=0; i<700; i++) {
				int xx, yy;
				pContextInfo->unitLayer->getRandomLocation(xx, yy);
				randomUnit* ru = new randomUnit(xx,yy,10,128,10);
				ru->setObjectType(kIncidentalUnit);
				ru->setSpeed(0.02);
				pContextInfo->unitLayer->addUnit(ru,true);
			}
			
			break;
		}
			
			//for (int x = 0; x < 10000; x++)
			//			praStar pra;
			//			pra.setPartialPathLimit(4);
			//			doRandomPath(pContextInfo->unitLayer->getAbstractMap(), &pra, shift);
			//pContextInfo->abstrMap->doRandomPath(shift);
		case 't':
		{
			//aStar2 astar2;
			praStar astar;
			doRandomPath(pContextInfo->unitLayer->getAbstractMap(), &astar, true);
		}
			break;
		case 'y':
		{
			praStar pra;
			aStar astar;
			pra.setPlanFromMiddle(true);
			pra.setExpandSearchRadius(true);
			//pra.setPartialPathLimit(4);
			doRandomPath(pContextInfo->unitLayer->getAbstractMap(), &pra, shift);
			doRandomPath(pContextInfo->unitLayer->getAbstractMap(), &astar, true);
			printf("pra*\t%d\ta*\t%d\n", pra.getNodesExpanded(), astar.getNodesExpanded());
		}
			break;
		case 'c':
			// toggle caps
			pContextInfo->drawCaps = 1 - pContextInfo->drawCaps;
			//		GetWindowPortBounds (window, &rectPort);
			//		InvalWindowRect (window, &rectPort);
			//		result = noErr;
			break;
		case '?': // credits
			pContextInfo->showCredits =  1 - pContextInfo->showCredits;
			//		GetWindowPortBounds (window, &rectPort);
			//		InvalWindowRect (window, &rectPort);
			break;
		case 'h':
			pContextInfo->human = new humanUnit(0, 0, 0);
			pContextInfo->unitLayer->addUnit(pContextInfo->human);
			// toggle help
			//		pContextInfo->drawHelp = 1 - pContextInfo->drawHelp;
			//		GetWindowPortBounds(window, &rectPort);
			//		InvalWindowRect(window, &rectPort);
			//		result = noErr;
			break;
		case 'j':
			if (pContextInfo->human)
				pContextInfo->human->setNextDirection(kW);
			break;
		case 'k':
			if (pContextInfo->human)
				pContextInfo->human->setNextDirection(kS);
			break;
		case 'i':
			if (pContextInfo->human)
				pContextInfo->human->setNextDirection(kN);
			break;
		case 'l': // lighting
			if (pContextInfo->human)
				pContextInfo->human->setNextDirection(kE);
			
			pContextInfo->lighting++;
			if (pContextInfo->lighting > 4)
				pContextInfo->lighting = 0;
				//		aglSetCurrentContext (pContextInfo->aglContext);
				SetLighting(pContextInfo->lighting);
			//		GetWindowPortBounds (window, &rectPort);
			//		InvalWindowRect (window, &rectPort);
			break;
		case 'm':
			makeMaze(pContextInfo->map, 3);
			rebuild = true;
			break;
		case '+':
			pContextInfo->unitLayer->setPenalty(pContextInfo->unitLayer->getPenalty() * 2.0);
			//if(verbose)
			printf("Penalty set to: %f\n", pContextInfo->unitLayer->getPenalty());
			break;
		case '-':
			pContextInfo->unitLayer->setPenalty(pContextInfo->unitLayer->getPenalty() / 2.0);
			//if(verbose)
			printf("Penalty set to: %f\n", pContextInfo->unitLayer->getPenalty());
			break;
	}
	return rebuild;
}


// sets the camera data to initial conditions
void resetCamera(recCamera * pCamera)
{
	pCamera->aperture = 5.0;
	pCamera->rotPoint = gOrigin;
	
	pCamera->viewPos.x = 0.0;
	pCamera->viewPos.y = 0.0;
	pCamera->viewPos.z = -12.0;
	pCamera->viewDir.x = -pCamera->viewPos.x; 
	pCamera->viewDir.y = -pCamera->viewPos.y; 
	pCamera->viewDir.z = -pCamera->viewPos.z;
	
	pCamera->viewUp.x = 0;  
	pCamera->viewUp.y = 1; 
	pCamera->viewUp.z = 0;
	
	// will be set in resize once the target view size and position is known
	pCamera->viewOriginY = 0;
	pCamera->viewOriginX = 0;
	pCamera->viewHeight = 0;
	pCamera->viewWidth = 0;
}

// Process command line arguments
// NOTE: "-batch" has to be the last switch on the command line
void processCommandLineArgs(int argc, char *argv[])
{
	// Initialize the parameters
	gDefaultMap[0] = 0;
	int batch = -1;

	// Process all command line arguments
	for (int x = 1; x < argc; x++)
	{
		
		// Process the map switch
		if (strcmp(argv[x], "-map") == 0)
		{
			if (x+1 < argc)
				strncpy(gDefaultMap, argv[x+1], 1024);
			else {
				cerr << "ERROR: -map flag used, but no map specified!" << endl;
				exit(10);
			}
		}
		
		// Process the generation switch
		if (strcmp(argv[x], "-generate") == 0)
		{
			scenarioGenerator(gDefaultMap);
			exit(0);
		}

		// Process the scenario switch
		if (strcmp(argv[x], "-scenario") == 0)
		{
			char scenFileName[1024];
			char algName[128];
			if (x+2 < argc) {
				strncpy(scenFileName, argv[x+1], 1024);
				strncpy(algName, argv[x+2], 128);
				if (x+5 >= argc) {
					runScenario(scenFileName, algName);
				}
				else {
					int _d; sscanf(argv[x+3],"%d",&_d);
					double _gamma; sscanf(argv[x+4],"%lf",&_gamma);
					double _LQ; sscanf(argv[x+5],"%lf",&_LQ);
					runScenario(scenFileName, algName, _d, _gamma, _LQ);
				}
				exit(0);
			}
			else {
				cerr << "ERROR: -scenario flag used, but no scenario file and alg name are specified!" << endl;
				exit(10);
			}
		}
		
		// Process the discrepancy switch
		if (strcmp(argv[x], "-discrepancy") == 0)
		{
			char scenFileName[1024];
			char algName[128];
			if (x+2 < argc) {
				strncpy(scenFileName, argv[x+1], 1024);
				strncpy(algName, argv[x+2], 128);
				if (x+5 >= argc) {
					discrepancy(scenFileName, algName);
				}
				else {
					int _d; sscanf(argv[x+3],"%d",&_d);
					double _gamma; sscanf(argv[x+4],"%lf",&_gamma);
					double _LQ; sscanf(argv[x+5],"%lf",&_LQ);
					discrepancy(scenFileName, algName, _d, _gamma, _LQ);
				}
				exit(0);
			}
			else {
				cerr << "ERROR: -discrepancy flag used, but no scenario file and alg name are specified!" << endl;
				exit(10);
			}
		}
		
		if (strcmp(argv[x], "-batch") == 0)
		{
			// set the index of batch mode, so we can find it quickly later
			batch = x+1;
			if (x == argc-1)
			{
				cerr << "ERROR: no batch mode specified." << endl;
				exit(10);
			}
		}
	}
	if (batch == -1)
		return;
	
	if (strstr(argv[batch], "RTS"))
	{
		handleBatchRTS(argc, argv, batch, gDefaultMap);
	}
	else {
		handleBatchPRA(argc, argv, batch, gDefaultMap);
	}
	
	exit(0);
}

void handleBatchPRA(int argc, char *argv[], int mode, char *_map)
{
	double stochastic = 0.0;
	int coordFile = -1;
	int scale = -1;
	for (int x = 1; x < argc; x++)
	{
		if (strcmp(argv[x], "-stochastic") == 0)
		{
			if (x+1 < argc) {
				stochastic = atof(argv[x+1]);
			}
		}
		if (strcmp(argv[x], "-scale") == 0)
		{
			if (x+1 < argc) {
				scale = atoi(argv[x+1]);
			}
		}
		if (strcmp(argv[x], "-coordfile") == 0)
		{
			if (x+1 < argc) {
				coordFile = x+1;
			}
		}
	}

	if (strcmp(argv[mode], "time") == 0)
	{
		calculateAverageTimePerStep(_map, (coordFile!=-1)?argv[coordFile]:0, scale);
	}
	else if (strcmp(argv[mode], "optimality") == 0)
	{
		calculateSubOptimality(_map, (coordFile!=-1)?argv[coordFile]:0, scale);
	}
}

void handleBatchRTS(int argc, char *argv[], int mode, char *_map)
{
	int xxs, yys, xxg, yyg;
	xxs = yys = xxg = yyg = -1;
	
#pragma mark Batch LRTS
	int d = 1;
	double gamma = 1.0;
	double LQ = 999999999.0;
	unsigned numRuns = 1;
	int dist = 50;
	bool drawMap = false;

	for (int x = 1; x < argc; x++)
	{
		// Process the display switch
		if (strcmp(argv[x], "-display") == 0)
			drawMap = true;
		if (strcmp(argv[x], "-coord") == 0)
		{
			if (x+4 < argc) {
				sscanf(argv[x+1],"%i",&xxs);
				sscanf(argv[x+2],"%i",&yys);
				sscanf(argv[x+3],"%i",&xxg);
				sscanf(argv[x+4],"%i",&yyg);
			}
			else {
				cerr << "ERROR: -coord flag used, but no xxs yys xxg yyg specified!" 
				<< endl;
				exit(10);
			}
		}
		// Process the numRuns switch
		if (strcmp(argv[x], "-runs") == 0)
		{
			if (x+1 < argc) {
				sscanf(argv[x+1],"%u",&numRuns);
			}
			else {
				cerr << "ERROR: -runs flag used, but no numRuns specified!" 
				<< endl;
				exit(10);
			}
		}
		// Process the 'dist' switch
		if (strcmp(argv[x], "-dist") == 0)
		{
			if (x+1 < argc) {
				sscanf(argv[x+1],"%d",&dist);
			}
			else {
				cerr << "ERROR: -dist flag used, but no dist specified!" 
				<< endl;
				exit(10);
			}
		}
		// Process the lrts switch
		if (strcmp(argv[x], "-lrts") == 0)
		{
			if (x+3 < argc) {
				sscanf(argv[x+1],"%i",&d);
				sscanf(argv[x+2],"%lf",&gamma);
				sscanf(argv[x+3],"%lf",&LQ);
			}
			else {
				cerr << "ERROR: -lrts flag used, but no d, gamma, T specified!" 
				<< endl;
				exit(10);
			}
		}
	}			

	if (strcmp(argv[mode], "LRTS") == 0)
		runLRTS(_map,xxs,yys,xxg,yyg, d, gamma, LQ, numRuns, drawMap, dist);
	else if (strcmp(argv[mode], "PRLRTS") == 0)
		runPRLRTS(_map,xxs,yys,xxg,yyg, d, gamma, LQ, numRuns, drawMap, dist);
}

// Run different scenarios (tests)
void runBatchMode(char *mode, char *_map, int xxs, int yys, int xxg, int yyg, 
				  int _d, double _gamma, double _LQ, unsigned _numRuns, 
				  bool drawMap, double _stochastic, int _dist)
{
	// longer tests should be put in functions
	if (strcmp(mode, "test") == 0)
		runSimpleUnitTest(_map);
	else if (strcmp(mode, "path") == 0)
	{
		randomPathTesting(_map);
	}
	else {
		cerr << "Error; no applicable batch mode " << mode << endl;
	}
}

// Nathan's simple test
void runSimpleUnitTest(char *_map)
{
	// Control parameters
	double stepSize = 0.0625, currTime = 0;
	double maxTime = 2000.0;
	double penalty = 1.0;
    
    // Map and unit simulations
	Map *map;
	unitSimulation *unitSim;
	
    // Set up the map (either given or a new map of 100x100)
	if (_map[0] == 0)
		map = new Map(100, 100);
	else
		map = new Map(_map);
	
	// Create the unit simulation 
	unitSim = new unitSimulation(map);
	unitSim->setPenalty(penalty);
	unitSim->setAsynchronous();
	unitSim->setRealTime(false);
	
	// Create the units
	for (int x = 0; x < 1000; x++)
	{
		while (1) {
            // Get random coordinates
			int xx1 = random()%map->getMapWidth();
			int yy1 = random()%map->getMapHeight();
			int xx2 = random()%map->getMapWidth();
			int yy2 = random()%map->getMapHeight();
			if ((map->getTerrainType(xx1, yy1) == kGround) && 
                (map->getTerrainType(xx2, yy2) == kGround))
			{
				unit *u, *u2 = new unit(xx2, yy2, 0);
				
				praStar *pra = new praStar();
				pra->setPartialPathLimit(4);
				
				unitSim->addUnit(u2);
				unitSim->addUnit(u = new searchUnit(xx1, yy1, u2, pra));
				u->setSpeed(0); // time to go 1 distance
				break;
			}
		}
	}
	while (currTime < maxTime)
	{
		currTime+=stepSize;
		unitSim->advanceTime(stepSize);
	}
}

//void calculateSubOptimality(char *_map, float stochastic)
//{
//	Map map(_map);
//	map.scale(256, 256);
//	unitRaceSimulation *unitLayer;
//	unitLayer = new unitRaceSimulation(&map);
//	unitLayer->setAsynchronous();
//	unitLayer->setPenalty(0.0);
//	unitLayer->setUseBlocking(false);
//	unitLayer->setUseSameStart(false);
//	unitLayer->setRealTime(false);
//	unitLayer->setMoveStochasticity(stochastic); // % of moves fail
//	
//	int xx1, yy1, xx2, yy2;
//	do {
//				unitLayer->getRandomLocation(xx1, yy1);
//				unitLayer->getRandomLocation(xx2, yy2);
//	} while (!unitLayer->getAbstractMap()->
//					 pathable(map.getNodeNum(xx1, yy1),
//										map.getNodeNum(xx2, yy2)));
//
//	unit *u, *u3;	
//	praStar praINF, pra2, pra4, pra8, pra16;
//	//pra.setExpandSearchRadius(false);
//	pra2.setPartialPathLimit(2);
//	pra4.setPartialPathLimit(4);
//	pra8.setPartialPathLimit(8);
//	pra16.setPartialPathLimit(16);
//	aStar a;
//	
//	unitLayer->addUnit(u3 = new unit(&map, xx2, yy2, 0), false);
//	unitLayer->addUnit(u = new searchUnit(&map, xx1, yy1, u3, &a), false);
//	u->setSpeed(0); // time to go 1 distance
//	unitLayer->addUnit(u = new searchUnit(&map, xx1, yy1, u3, &praINF), false);
//	u->setSpeed(0); // time to go 1 distance
//	unitLayer->addUnit(u = new searchUnit(&map, xx1, yy1, u3, &pra2), false);
//	u->setSpeed(0); // time to go 1 distance
//	unitLayer->addUnit(u = new searchUnit(&map, xx1, yy1, u3, &pra4), false);
//	u->setSpeed(0); // time to go 1 distance
//	unitLayer->addUnit(u = new searchUnit(&map, xx1, yy1, u3, &pra8), false);
//	u->setSpeed(0); // time to go 1 distance
//	unitLayer->addUnit(u = new praStarUnit(&map, xx1, yy1, u3, &pra2), false);
//	u->setSpeed(0); // time to go 1 distance
//	unitLayer->addUnit(u = new praStarUnit(&map, xx1, yy1, u3, &pra4), false);
//	u->setSpeed(0); // time to go 1 distance
//	unitLayer->addUnit(u = new praStarUnit(&map, xx1, yy1, u3, &pra8), false);
//	u->setSpeed(0); // time to go 1 distance
//	
//	char name[256];
//	sprintf(name, "optimality%2d.dat", (int)(stochastic*100));
////	FILE *log = fopen(name,"a+");
////	if (log)
////	{
////		unitLayer->setLogFile(log);
//		while (unitLayer->getCurrRound() < 3000)
//		{
//			unitLayer->advanceTime(.01);
//		}
////	}
////	fclose(log);
//}

// outputs algorithm/path length/time per step
void calculateAverageTimePerStep(char *_map, char *coordFile, int MAP_SIZE)
{
	//const int MAP_SIZE = 512;
	int BUCKET = (int)(MAP_SIZE/4.0);
	const int pathsPerBucket = 10;
	
	int pathsLeft = BUCKET*pathsPerBucket;
	int iterations = 0;
	Map map(_map);
	map.scale(MAP_SIZE, MAP_SIZE);

	abstractMap aMap(&map);
  graph *g = aMap.getAbstractGraph(0);
	FILE *f=0;
	if (coordFile != 0)
	{
		f = fopen(coordFile, "r");
		if (f == 0)
		{
			cerr << "Error opening " << coordFile << endl;
			exit(10);
		}
	}
	
	node *r1, *r2;

	printf("bucket\tfromx\tfromy\ttox\ttoy\tA*len\tA*time\tA*nodes\tPRAlen\tPRA*time\tPRA*nodes\tPRA4time\tPRA4nodes\n");
	int buckets[BUCKET];
	for (int x = 0; x < BUCKET; x++)
		buckets[x] = 0;
//	for (int x = 0; x < 3; x++)
//	{
//		buckets[x] = pathsPerBucket;
//		pathsLeft -= pathsPerBucket;
//	}
	while ((pathsLeft > 0) && (iterations < 5*BUCKET*pathsPerBucket))
	{
		if (coordFile == 0)
		{
			iterations++;
			do {
				r1 = g->getRandomNode();
				r2 = g->getRandomNode();
			} while (!aMap.pathable(r1, r2) || (r1 == r2) ||
							 ((iterations > BUCKET*pathsPerBucket) && (aMap.h(r1, r2) < MAP_SIZE/2)));
		}
		else {
			if (feof(f))
				break;
			int x1, y1, x2, y2;
			fscanf(f, "%d\t%d\t%d\t%d\n", &x1, &y1, &x2, &y2);
			r1 = aMap.getNodeFromMap(x1, y1);
			r2 = aMap.getNodeFromMap(x2, y2);
		}

		aStar a;
		praStar pra;
#ifdef OS_MAC
		AbsoluteTime startTime = UpTime();
#endif
		path *p;
		p = a.getPath(&aMap, r1, r2);
#ifdef OS_MAC
		AbsoluteTime stopTime = UpTime();
		Nanoseconds diff = AbsoluteDeltaToNanoseconds(stopTime, startTime);
		unsigned long long nanosecs = UnsignedWideToUInt64(diff);
#else
		unsigned long long nanosecs = 1;
#endif

		int cnt = 0;
		double length = 0;
		for (path *q = p; q; q = q->next) {
			if (q && q->next) {
				double t1, t2;
				t1 = q->n->getLabelL(kFirstData)-q->next->n->getLabelL(kFirstData);
				t2 = q->n->getLabelL(kFirstData+1)-q->next->n->getLabelL(kFirstData+1);
				length += sqrt(t1*t1+t2*t2);
			}
			cnt++;
		}
		if ((length/4 >= BUCKET) || ((iterations > BUCKET*pathsPerBucket/2) && (buckets[(int)length/4] == pathsPerBucket)))
		{
			while (p && p->next)
			{
				double t1, t2;
				t1 = p->n->getLabelL(kFirstData)-p->next->n->getLabelL(kFirstData);
				t2 = p->n->getLabelL(kFirstData+1)-p->next->n->getLabelL(kFirstData+1);
				length -= sqrt(t1*t1+t2*t2);
				path *t = p;
				p = p->next;
				t->next = 0; delete t;
				if (p->next == 0)
				{
					delete p;
					p = 0;
				}
				if (p == 0)
					break;
				if (buckets[(int)length/4] < pathsPerBucket)
					break;
			}
			if (p && p->next)
			{
				r1 = p->n;
				for (path *q = p; q; q = q->next) {
					r2 = q->n;
				}
				delete p;
#ifdef OS_MAC
				startTime = UpTime();
#endif
				p = a.getPath(&aMap, r1, r2);
#ifdef OS_MAC
				stopTime = UpTime();
				diff = AbsoluteDeltaToNanoseconds(stopTime, startTime);
				nanosecs = UnsignedWideToUInt64(diff);
#else
				nanosecs = 1;
#endif
				length = 0; cnt = 0;
				for (path *q = p; q; q = q->next) {
					if (q && q->next) {
						double t1, t2;
						t1 = q->n->getLabelL(kFirstData)-q->next->n->getLabelL(kFirstData);
						t2 = q->n->getLabelL(kFirstData+1)-q->next->n->getLabelL(kFirstData+1);
						length += sqrt(t1*t1+t2*t2);
					}
					cnt++;
				}
			}
		}
		if (p == 0)
			continue;
		if ((length/4 < BUCKET) && (buckets[(int)length/4] < pathsPerBucket))
		{
			buckets[(int)length/4]++;
			pathsLeft--;
		}
		else {
			continue;
		}
//		printf("aStar\t%5.4f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%d\n",
//					 length/aMap.h(r1, r2), nanosecs/(1000.0*length), nanosecs/(1000.0*cnt),
//					 nanosecs/1000000.0, length, cnt);
		int x1, x2, y1, y2;
		x1 = r1->getLabelL(kFirstData); y1 = r1->getLabelL(kFirstData+1);
		x2 = r2->getLabelL(kFirstData); y2 = r2->getLabelL(kFirstData+1);
		printf("%d\t%d\t%d\t%d\t%d\t%1.2f\t%1.2f\t%d\t",
					 (int)(length/4), x1, y1, x2, y2, length, nanosecs/1000.0,
					 a.getNodesExpanded());
		delete p;
#ifdef OS_MAC
		startTime = UpTime();
#endif
		p = pra.getPath(&aMap, r1, r2);
#ifdef OS_MAC
		stopTime = UpTime();
		diff = AbsoluteDeltaToNanoseconds(stopTime, startTime);
		nanosecs = UnsignedWideToUInt64(diff);
#else
		nanosecs = 1;
#endif
		
		cnt = 0;
		length = 0;
		for (path *q = p; q; q = q->next) {
			if (q && q->next) {
				double t1, t2;
				t1 = q->n->getLabelL(kFirstData)-q->next->n->getLabelL(kFirstData);
				t2 = q->n->getLabelL(kFirstData+1)-q->next->n->getLabelL(kFirstData+1);
				length += sqrt(t1*t1+t2*t2);
			}
			cnt++;
		}
		printf("%1.2f\t%1.2f\t%d\t", length, nanosecs/1000.0, pra.getNodesExpanded());

		pra.setPartialPathLimit(4);
#ifdef OS_MAC
		startTime = UpTime();
#endif
		p = pra.getPath(&aMap, r1, r2);
#ifdef OS_MAC
		stopTime = UpTime();
		diff = AbsoluteDeltaToNanoseconds(stopTime, startTime);
		nanosecs = UnsignedWideToUInt64(diff);
#else
		nanosecs = 1;
#endif
		
		cnt = 0;
		length = 0;
		for (path *q = p; q; q = q->next) {
			if (q && q->next) {
				double t1, t2;
				t1 = q->n->getLabelL(kFirstData)-q->next->n->getLabelL(kFirstData);
				t2 = q->n->getLabelL(kFirstData+1)-q->next->n->getLabelL(kFirstData+1);
				length += sqrt(t1*t1+t2*t2);
			}
			cnt++;
		}
		printf("%1.2f\t%d\t%1.2f\n", nanosecs/1000.0, pra.getNodesExpanded(), length);
	}	
}

// outputs algorithm/path length/time per step
void calculateSubOptimality(char *_map, char *coordFile, int MAP_SIZE)
{
	Map map(_map);
	map.scale(MAP_SIZE, MAP_SIZE);
	
	abstractMap aMap(&map);
	FILE *f=0;
	if (coordFile == 0)
	{
		cerr << "Error; no coordinate file provided" << endl;
		exit(10);
	}

	f = fopen(coordFile, "r");
	if (f == 0)
	{
		cerr << "Error opening " << coordFile << endl;
		exit(10);
	}
	
	node *r1, *r2;
	
	while (!feof(f))
	{
		int x1, y1, x2, y2;
		fscanf(f, "%d\t%d\t%d\t%d\n", &x1, &y1, &x2, &y2);
		r1 = aMap.getNodeFromMap(x1, y1);
		r2 = aMap.getNodeFromMap(x2, y2);
		
		praStar pra;
#ifdef OS_MAC
		AbsoluteTime startTime;
		AbsoluteTime stopTime;
		Nanoseconds diff;
#endif
		unsigned long long nanosecs;
		path *p;
		
		int cnt = 0;
		double length = 0;
		printf("%s\t%d\t%d\t%d\t%d\n", _map, x1, y1, x2, y2);
		path *cache = 0;
		for (int rLen = 2; rLen <= 16; rLen*= 2)
		{
			for (int expansion = 0; expansion <= 0; expansion++)
			{
				r1 = aMap.getNodeFromMap(x1, y1);
				r2 = aMap.getNodeFromMap(x2, y2);

				pra.setExpandSearchRadius(expansion);
				pra.setPartialPathLimit(rLen);
				printf("cPRA(%d:%d)", rLen, expansion);

				while (r1 != r2)
				{
					if (cache != 0)
					{
						node *l = aMap.getNthParent(r1, cache->n->getLabelL(kAbstractionLevel));
						while (cache && (l != cache->n))
						{
							path *tmp = cache;
							cache = cache->next;
							tmp->next = 0;
							delete tmp;
						}
						if (cache && (cache->next == 0))
						{
							delete cache;
							cache = 0;
						}						
					}
					pra.setCache(&cache);
#ifdef OS_MAC
					startTime = UpTime();
#endif
					p = pra.getPath(&aMap, r1, r2);
#ifdef OS_MAC
					stopTime = UpTime();
					diff = AbsoluteDeltaToNanoseconds(stopTime, startTime);
					nanosecs = UnsignedWideToUInt64(diff);
#endif
					
					cnt = 0;
					length = 0;
					for (path *q = p; q; q = q->next) {
						if (q && q->next) {
							double t1, t2;
							t1 = q->n->getLabelL(kFirstData)-q->next->n->getLabelL(kFirstData);
							t2 = q->n->getLabelL(kFirstData+1)-q->next->n->getLabelL(kFirstData+1);
							length += sqrt(t1*t1+t2*t2);
							r1 = q->next->n;
						}
						cnt++;
					}
					printf("\t%1.2f\t%1.2f\t%d", length, nanosecs/1000.0, pra.getNodesExpanded());
				}
				printf("\n");
			}
		}	
	}
}

void randomPathTesting(char *_map)
{
	Map map(_map);
	unitRaceSimulation *unitLayer = new unitRaceSimulation(&map);
	unitLayer->setAsynchronous();
	unitLayer->setPenalty(0);
	unitLayer->setUseBlocking(false);
	unitLayer->setUseSameStart(false);
	unitLayer->setRealTime(false);
	while (1) {
		int xx1 = random()%map.getMapWidth();
		int yy1 = random()%map.getMapHeight();
		int xx2 = random()%map.getMapWidth();
		int yy2 = random()%map.getMapHeight();
		if ((map.getTerrainType(xx1, yy1) == kGround) &&
			(map.getTerrainType(xx2, yy2) == kGround) &&
			(unitLayer->getAbstractMap()->pathable(map.getNodeNum(xx1, yy1),
												   map.getNodeNum(xx2, yy2))))
		{
			unit *u, *u1, *u2 = new unit(xx2, yy2, 0);
			
			praStar *pra = new praStar();
			pra->setPartialPathLimit(4);
			aStar *a = new aStar();
			
			unitLayer->addUnit(u2);
			unitLayer->addUnit(u = new searchUnit(xx1, yy1, u2, pra));
			unitLayer->addUnit(u1 = new searchUnit(xx1, yy1, u2, a));
			u->setSpeed(0); // time to go 1 distance
			u1->setSpeed(0); // time to go 1 distance
			break;
		}
	}
	char logFile[1024];
	sprintf(logFile, "%s_tests.xls", _map);
	FILE *log = fopen(logFile,"w+");
	if (log)
	{
		unitLayer->setLogFile(log);
		
		while (unitLayer->getCurrRound() < 1000)
		{
			unitLayer->advanceTime(.01);
		}
		fclose(log);
	}
}


// Race between PRA* and LRTS
void runLRTS(char *_map, int _xxs, int _yys, int _xxg, int _yyg, 
			 int _d, double _gamma, double _LQ, unsigned numRuns, 
			 bool drawMap, int _dist)
{
	// Set up the control parameters
	const double stepSize = 0.0001;			// time step
	double currTime = 0.0;					// initial time
	const double penalty = 1;				// thinking penalty multiplier
	const double maxTime = 10000.0*penalty;	// simulation time limit
	const double speed = 0.0;				// speed 
	const int visRadius = 10;			// visibility radius for the simulation
	
	printf("Penalty\t\t\t%1.5f\n",penalty);
	printf("Speed\t\t\t%1.5f\n",speed);
	printf("Step size\t\t%1.5f\n",stepSize);
	printf("Max simulation time\t%1.0f\n\n",maxTime);
	
	// Map and simulation
	Map *map = NULL;
	unitSimulation *unitSim = NULL;
	
	// If the map is given then use it. Otherwise create a default 100x100 map
	if (_map[0] == 0)
		map = new Map(100, 100);
	else
		map = new Map(_map);
	
	// Algorithms
	// praStar* pra = new praStar(); // pra->setPartialPathLimit(4);
	aStar* pra = new aStar();
	assert(_d <= visRadius);		// check for cheating
	LRTS* lrts = new LRTS(_gamma, _d, _LQ);   // gamma, d, T
	
	// Run the converge run numRuns times for different start and finish
	for (unsigned run = 0; run < numRuns; run++) {
		
		// Create the unit simulation
		unitSim = new unitRaceSimulation(map);
		abstractMap* aMap = unitSim->getAbstractMap();
		unitSim->setAsynchronous();
		unitSim->setPenalty(penalty);
		unitSim->setUseBlocking(true);
		unitSim->setRealTime(false);
		
		// Creating a group for the PRA* unit
		sharedAMapGroup* praGroup = NULL;
		praGroup = new sharedAMapGroup();
		unitSim->addUnitGroup(praGroup);
		praGroup->setVisibilityRadius(visRadius);
		
		// Creating a group for the LRTS unit
		LRTSGroup* lrtsGroup = NULL;
		lrtsGroup = new LRTSGroup(); 
		unitSim->addUnitGroup(lrtsGroup);
		
		// Get the base-level graph for this map
		graph *g = aMap->getAbstractGraph(0);
		
		// Create one goal and one start
		double spDist = 0.0;
		double aStarDifficulty = 0.0;
		int xxs = _xxs; 
		int yys = _yys;
		int xxg = _xxg;
		int yyg = _yyg;
		pickStartEnd(xxs, yys, xxg, yyg, map, g, aMap, spDist, aStarDifficulty, _dist);
		
		// Create the static goal state (of white color)
		unit* targetUnit = new unit(xxg, yyg, 255, 255, 255, 0);
		unitSim->addUnit(targetUnit);
		
		// Create the units and add them to the simulation
		searchUnit* praUnit = new searchUnit(xxs, yys, 255, 0, 0, targetUnit, pra);
		praGroup->addUnit(praUnit);
		unitSim->addUnit(praUnit,false);
		praUnit->setSpeed(speed);
		
		LRTSUnit* lrtsUnit = new LRTSUnit(xxs, yys, 0, 0, 255, targetUnit, lrts);
		lrtsGroup->addUnit(lrtsUnit);
		unitSim->addUnit(lrtsUnit,false);
		lrtsUnit->setSpeed(speed);
		
		// Produces the prefix name: e.g., 16.LRTS(1,1.0,inf)
		char prefixName[1034];
		for (int i=0; i<1034; prefixName[i++]='\0');
		if (_map[0] != '\0')
			for (int i=0; (_map[i] != '.' && 
						   _map[i] != '\0' && i<1033); i++) 
				prefixName[i] = _map[i];
		else
			strcpy(prefixName,"default");
		strcat(prefixName,".");
		strcat(prefixName,lrtsUnit->getName());
		
		// Produces the two file name
		// e.g., 16.LRTS(1,1.0,inf).003.txt
		// e.g., 16.LRTS(1,1.0,inf).003.dat
		char logFileName[1200];
		char datFileName[1200];
		sprintf(logFileName,"%s.%03d.txt",prefixName,run);
		sprintf(datFileName,"%s.%03d.dat",prefixName,run);
		
		// Open the main log file
		FILE *trialLog = fopen(logFileName,"w");
		unitSim->setLogFile(trialLog);
		
		// Now dump some statistics into a separate file
		FILE *datLog = fopen(datFileName,"w");
		fprintf(datLog,"%1.6f %1.6f\n",spDist,aStarDifficulty);
		fprintf(datLog,"%d %d\n",g->getNumNodes(), 
				map->getMapWidth()*map->getMapHeight());
		fprintf(datLog,"%d %d %d %d\n",xxs,yys,xxg,yyg);
		fclose(datLog);
		
		
		if (drawMap)
			printf("%c[2J", 27);
		
		// Now simulate until convergence (or time out)
		do 
		{
			currTime+=stepSize;
			unitSim->advanceTime(stepSize);
			
			if (drawMap)
				unitSim->print(false);
		}
		while (currTime < maxTime && !unitSim->done());
		
		// Clean things up
		if (unitSim != NULL) delete unitSim;
	}
	if (map)
		delete map;
}


// Scenario generator
void scenarioGenerator(char *_map) {
	// Check if the map name is valid
	assert(_map[0] != 0);
	
	srand(7);
	
	// Some control parameters
	const int numBucket = 200;		// problems to produce per bucket (per map)
	const int numFailedMAX = 10000;	// at most ten attempts per bucket
	
	// Map and simulation
	Map *map = new Map(_map);
	unitRaceSimulation* unitSim = new unitRaceSimulation(map);
	abstractMap* aMap = unitSim->getAbstractMap();
	graph *g = aMap->getAbstractGraph(0);

	aStar* myAStar = new aStar();
	
	// Set up the output file
	FILE* scenFile = fopen("scen1000.txt","a");
	
	// Go through the buckets and fill them in
	for (int cBucket = distBucketFirst; cBucket <= distBucketLast; 
		 cBucket += distBucketDelta) {

		// Initialize the counters
		int numFailed = 0;
		int numGenerated = 0;
		printf("\n");

		// Create numBucket problems
		while (true) {
			
			// Generate random coordinates
			int xs = random()%map->getMapWidth();
			int ys = random()%map->getMapHeight();
			int xg = random()%map->getMapWidth();
			int yg = random()%map->getMapHeight();
			
			// Compute the optimal path between the two
			int startNodeNum = map->getNodeNum(xs,ys);
			int goalNodeNum = map->getNodeNum(xg,yg);
			node* start = g->getNode(startNodeNum);
			node* goal = g->getNode(goalNodeNum);
			
			// Compute the length of the optimal path between them
			path* pathBetween = myAStar->getPath(aMap, start, goal);
			
			// Check if the path exists
			if (pathBetween == NULL)
				continue;
			
			// Check the distance
			double spDist = pathBetween->distance(aMap);
			delete pathBetween;
			
			if (spDist <= cBucket-distBucketDelta || spDist > cBucket) {
				// We failed to find an appropriate path
				// Once we exceed numFailMAX attempts we will give up
				if (++numFailed > numFailedMAX) {
					printf("We failed %d times to generate a problem in (%d,%d]\n",
						   numFailed, cBucket-distBucketDelta, cBucket);
					fclose(scenFile);
					exit(10);
				}
			}
			else {
				// Successfully generated a path
				numFailed = 0;
				numGenerated++;
				fprintf(scenFile,"%3d %s %3d %3d %3d %3d %1.10f\n",
						cBucket,_map,xs,ys,xg,yg,spDist);
				printf("%3d %s %3d %3d %3d %3d %1.10f\n",
						cBucket,_map,xs,ys,xg,yg,spDist);
				
				if (numGenerated >= numBucket)
					break;
			}
			
		}
	}
	
	// Close the file and clean up the memory
	fclose(scenFile);
	delete map;
	delete unitSim;		// deletes aMap and g as well
	delete myAStar;
}


// Run a scenario
void runScenario(char* scenName, char* algName, int _d, double _gamma, double _LQ) {

	// printf("%s, %s(%d,%f,%f)\n",scenName,algName,_d,_gamma,_LQ); fflush(stdout);
	
	char* fileName = algName;
	
	// Open the file
	FILE* scenFile = fopen(scenName,"r");
	assert(scenFile != NULL);
	
	// Simulation data structures
	char mapLoaded[1024];
	mapLoaded[0] = 0;
	Map* map = NULL;
	unitRaceSimulation* unitSim = NULL;
	abstractMap* aMap = NULL;
	graph *g = NULL;
	
	// Measurement data structures
	// convergence execution cost
	double convEC[distNumBuckets];						// sample mean
	double convECvar[distNumBuckets];					// sample variance
	double convECsample[distNumBuckets][maxSamples];	// samples
	
	// convergence planning cost
	double convPC[distNumBuckets];						// sample mean
	double convPCvar[distNumBuckets];					// sample variance
	double convPCsample[distNumBuckets][maxSamples];	// samples
	
	// first-move lag
	double lag[distNumBuckets];							// sample mean
	double lagvar[distNumBuckets];						// sample variance
	double lagsample[distNumBuckets][maxSamples];		// samples

	// suboptimality of the final solution
	double subOpt[distNumBuckets];						// sample mean
	double subOptvar[distNumBuckets];					// sample variance
	double subOptsample[distNumBuckets][maxSamples];	// samples

	// convergence memory cost
	double memory[distNumBuckets];						// sample mean
	double memoryvar[distNumBuckets];					// sample variance
	double memorysample[distNumBuckets][maxSamples];	// samples

	int samples[distNumBuckets];						// samples per bucket

	// Reset the stats
	for (int k=0; k<distNumBuckets; k++) {
		convEC[k] = convPC[k] = memory[k] = lag[k] = subOpt[k] = 0.0;
		for (int j=0; j<maxSamples; j++) {
			convECsample[k][j] = convPCsample[k][j] = 0.0;
			memorysample[k][j] = lagsample[k][j] = subOptsample[k][j] = 0.0;
		}
		samples[k] = 0;
	}
	
	
	// Read through the file
	while (!feof(scenFile)) {
		
		// Read in the problem description
		int bucket;
		char mapName[1024]; 
		int xs, ys, xg, yg;
		double dist;
		fscanf(scenFile,"%d %s %d %d %d %d %lf\n",&bucket,mapName,&xs,&ys,&xg,&yg,&dist);
		int bNum = (bucket - distBucketFirst) / distBucketDelta;
		
		// Check if we have this map already loaded
		if (strcmp(mapName,mapLoaded) != 0) {
			// No, need to load the new map
			if (map != NULL) 
				delete map;
			map = new Map(mapName);
			strcpy(mapLoaded,mapName);
			printf("\nLoaded map %s\n",mapLoaded);
		}
		
		// Control constants
		const double stepSize = 0.0001;			// time step
		double currTime = 0.0;					// initial time
		const double penalty = 1;				// thinking penalty multiplier
		const double maxTime = 10000.0*penalty;	// simulation time limit
		const double speed = 0.0;				// speed 
		const int visRadius = 10;			// visibility radius for the simulation
		
		// Create all the necessary structures
		if (unitSim != NULL)
			delete unitSim;
		unitSim = new unitRaceSimulation(map);
		aMap = unitSim->getAbstractMap();
		g = aMap->getAbstractGraph(0);
		unitSim->setAsynchronous();
		unitSim->setPenalty(penalty);
		unitSim->setUseBlocking(true);
		unitSim->setRealTime(false);
		
		// Do a convergence run on this problem
		printf("\n------- %20s : (%3d,%3d) -> (%3d,%3d); %8.2f; bucket[%2d]\n",
			   mapLoaded,xs,ys,xg,yg,dist,bNum);

		// Create the static goal state/unit (of white color)
		unit* targetUnit = new unit(xg, yg, 255, 255, 255, 0);
		unitSim->addUnit(targetUnit);
		
		// Create groups and units
		if (strcmp(algName,"aStar") == 0) {
			// Creating a group for the incremental A* unit
			sharedAMapGroup* asGroup = new sharedAMapGroup();
			unitSim->addUnitGroup(asGroup);
			asGroup->setVisibilityRadius(visRadius);
			
			// Create an A* unit and add it to the simulation
			aStar* as = new aStar();
			searchUnit* asUnit = new searchUnit(xs, ys, targetUnit, as);
			asGroup->addUnit(asUnit);
			unitSim->addUnit(asUnit,false);
			asUnit->setSpeed(speed);
		}
		else if (strcmp(algName,"waStar") == 0) {
			// Creating a group for the incremental weighted-A* unit
			sharedAMapGroup* asGroup = new sharedAMapGroup();
			unitSim->addUnitGroup(asGroup);
			asGroup->setVisibilityRadius(visRadius);
			
			// Create a weighted A* unit and add it to the simulation
			aStar* as = new aStar(1.0/_gamma);
			searchUnit* asUnit = new searchUnit(xs, ys, targetUnit, as);
			asGroup->addUnit(asUnit);
			unitSim->addUnit(asUnit,false);
			asUnit->setSpeed(speed);
			
			fileName = (char*) asUnit->getName();
		}
		else if (strcmp(algName,"praStar") == 0) {
			// Creating a group for the PRA* unit
			sharedAMapGroup* praGroup = new sharedAMapGroup();
			unitSim->addUnitGroup(praGroup);
			praGroup->setVisibilityRadius(visRadius);
			
			// PRA* without partial refinement
			praStar* pra = new praStar(); // pra->setPartialPathLimit(4);
			searchUnit* praUnit = new searchUnit(xs, ys, targetUnit, pra);
			praGroup->addUnit(praUnit);
			unitSim->addUnit(praUnit,false);
			praUnit->setSpeed(speed);
		}
		else if (strcmp(algName,"LRTA") == 0) {
			// Creating a group for the LRTA unit
			LRTSGroup* lrtsGroup = new LRTSGroup(); 
			unitSim->addUnitGroup(lrtsGroup);
			
			// LRTS Unit
			LRTS* lrts = new LRTS(1.0, 1, 9999999999999999999.0);   // gamma, d, T
			LRTSUnit* lrtsUnit = new LRTSUnit(xs, ys, targetUnit, lrts);
			lrtsGroup->addUnit(lrtsUnit);
			unitSim->addUnit(lrtsUnit,false);
			lrtsUnit->setSpeed(speed);
		}
		else if (strcmp(algName,"LRTS") == 0) {
			// Creating a group for the LRTS unit
			LRTSGroup* lrtsGroup = new LRTSGroup(); 
			unitSim->addUnitGroup(lrtsGroup);
			
			// LRTS Unit
			LRTS* lrts = new LRTS(_gamma, _d, _LQ);   // gamma, d, T
			LRTSUnit* lrtsUnit = new LRTSUnit(xs, ys, targetUnit, lrts);
			lrtsGroup->addUnit(lrtsUnit);
			unitSim->addUnit(lrtsUnit,false);
			lrtsUnit->setSpeed(speed);
			
			fileName = (char*) lrtsUnit->getName();
		}
		else if (strcmp(algName,"PRLRTS") == 0) {
			// Creating a group for the PR_LRTS unit
			PR_LRTSGroup* pr_lrtsGroup = new PR_LRTSGroup(); 
			unitSim->addUnitGroup(pr_lrtsGroup);
			pr_lrtsGroup->setVisibilityRadius(visRadius);
			
			// PR LRTS unit
			PR_LRTS* pr_lrts = new PR_LRTS(_gamma, _d, _LQ);
			PR_LRTSUnit* pr_lrtsUnit = new PR_LRTSUnit(xs, ys, targetUnit, pr_lrts);
			pr_lrtsGroup->addUnit(pr_lrtsUnit);
			unitSim->addUnit(pr_lrtsUnit,false);
			pr_lrtsUnit->setSpeed(speed);
			
			fileName = (char*) pr_lrtsUnit->getName();
		}
		else {
			printf("\nUnknown unit %s \n",algName);
			exit(10);
		}
		
		// Set the statistics variables
		double _convTravel = 0.0;
		double _convThink = 0.0;
		double _fmThink = 0.0;
		double _fTravel = 0.0;
		double _memory = 0.0;
		unitSim->setVars(&_convTravel, &_convThink, &_fmThink, &_fTravel, &_memory);
		
		// Now simulate until convergence (or time out)
		do 
		{
			currTime+=stepSize;
			unitSim->advanceTime(stepSize);
		}
		while (currTime < maxTime && !unitSim->done());

		// Update the statistics

		lag[bNum] += _fmThink;
		lagsample[bNum][samples[bNum]] = _fmThink;
		
		convEC[bNum] += _convTravel;
		convECsample[bNum][samples[bNum]] = _convTravel;
		
		convPC[bNum] += _convThink / _convTravel;
		convPCsample[bNum][samples[bNum]] = _convThink / _convTravel;
		
		memory[bNum] += _memory;
		memorysample[bNum][samples[bNum]] = _memory;

		double _subOpt = (_fTravel-dist) / dist;
		_subOpt = fequal(_subOpt,0.0) ? 0.0 : _subOpt;
		subOpt[bNum] += _subOpt;
		subOptsample[bNum][samples[bNum]] = _subOpt;
		
		samples[bNum]++;
		
		// Clean things up
		if (unitSim != NULL) { 
			delete unitSim;
			unitSim = NULL;
		}
	}
	
	// Compute the mean and standard deviations ===================================
	for (int k=0; k<distNumBuckets; k++) {

		// Compute sample means
		lag[k] /= (double) samples[k];
		convEC[k] /= (double) samples[k];
		convPC[k] /= (double) samples[k];
		memory[k] /= (double) samples[k];
		subOpt[k] /= (double) samples[k];
		
		// Compute sample variances
		lagvar[k] = 0.0;
		convECvar[k] = 0.0;
		convPCvar[k] = 0.0;
		memoryvar[k] = 0.0;
		subOptvar[k] = 0.0;
		
		for (int j=0; j<samples[k]; j++) {
			lagvar[k] += (lagsample[k][j]-lag[k])*(lagsample[k][j]-lag[k]);
			convECvar[k] += (convECsample[k][j]-convEC[k])*(convECsample[k][j]-convEC[k]);
			convPCvar[k] += (convPCsample[k][j]-convPC[k])*(convPCsample[k][j]-convPC[k]);
			memoryvar[k] += (memorysample[k][j]-memory[k])*(memorysample[k][j]-memory[k]);
			subOptvar[k] += (subOptsample[k][j]-subOpt[k])*(subOptsample[k][j]-subOpt[k]);
		}
		
		lagvar[k] /= (double) (samples[k]-1);
		convECvar[k] /= (double) (samples[k]-1);
		convPCvar[k] /= (double) (samples[k]-1);
		memoryvar[k] /= (double) (samples[k]-1);
		subOptvar[k] /= (double) (samples[k]-1);
	}
	
	// Print out the results ======================================================
	
	char ofN[1024];
	FILE* outFile = fopen("numSamples.txt","w");
	for (int k=0; k<distNumBuckets; k++) {
		fprintf(outFile,"%d %d\n",
			   distBucketFirst+k*distBucketDelta-distBucketDelta/2,samples[k]);
	}
	fclose(outFile);

	sprintf(ofN,"%s.lag.txt",fileName);
	outFile = fopen(ofN,"w");
	for (int k=0; k<distNumBuckets; k++) {
		fprintf(outFile,"%d %1.10f %1.10f\n",
			   distBucketFirst+k*distBucketDelta-distBucketDelta/2,
				lag[k], sqrt(lagvar[k]) / sqrt((double) samples[k]));
	}
	fclose(outFile);
	
	sprintf(ofN,"%s.convEC.txt",fileName);
	outFile = fopen(ofN,"w");
	for (int k=0; k<distNumBuckets; k++) {
		fprintf(outFile,"%d %1.10f %1.10f\n",
			   distBucketFirst+k*distBucketDelta-distBucketDelta/2,
				convEC[k], sqrt(convECvar[k]) / sqrt((double) samples[k]));
	}
	fclose(outFile);
	
	sprintf(ofN,"%s.convPC.txt",fileName);
	outFile = fopen(ofN,"w");
	for (int k=0; k<distNumBuckets; k++) {
		fprintf(outFile,"%d %1.10f %1.10f\n",
				distBucketFirst+k*distBucketDelta-distBucketDelta/2,
				convPC[k], sqrt(convPCvar[k]) / sqrt((double) samples[k]));
	}
	fclose(outFile);
	
	sprintf(ofN,"%s.memory.txt",fileName);
	outFile = fopen(ofN,"w");
	for (int k=0; k<distNumBuckets; k++) {
		fprintf(outFile,"%d %1.10f %1.10f\n",
				distBucketFirst+k*distBucketDelta-distBucketDelta/2,
				memory[k], sqrt(memoryvar[k]) / sqrt((double) samples[k]));
//		fprintf(outFile,"[");
//		for (int j=0; j<samples[k]; j++)
//			fprintf(outFile,"%1.10f ",memorysample[k][j]);
//		fprintf(outFile,"]\n");
	}
	fclose(outFile);
	
	sprintf(ofN,"%s.subOpt.txt",fileName);
	outFile = fopen(ofN,"w");
	for (int k=0; k<distNumBuckets; k++) {
		fprintf(outFile,"%d %1.10f %1.10f\n",
			   distBucketFirst+k*distBucketDelta-distBucketDelta/2,
				subOpt[k], sqrt(subOptvar[k]) / sqrt((double) samples[k]));
	}
	fclose(outFile);
	
	// Close the scenario file and clean-up
	fclose(scenFile);
	delete map;
	delete unitSim;		// deletes aMap and g as well
}

// Discrepancy computations
void discrepancy(char* scenName, char* algName, int _d, double _gamma, double _LQ) {
	
	// Open the file
	FILE* scenFile = fopen(scenName,"r");
	assert(scenFile != NULL);
	
	// Data structures
	char mapLoaded[1024];
	mapLoaded[0] = 0;
	Map* map = NULL;
	abstractMap* aMap = NULL;
	unitRaceSimulation* unitSim = NULL;
	graph *g = NULL;
	aStar* myAStar = new aStar();
	
	const int NUMD = 5;				// number of gamma values
	double gamma[5] = {0.3, 0.5, 0.7, 0.9, 1.0};
	double discrepancy[NUMD];
	for (int ek=0; ek < NUMD; ek++)
		discrepancy[ek] = 0.0;
	long numStates = 0;

	
	// Read through the file
	while (!feof(scenFile)) {
		
		// Read in the problem description
		int bucket;
		char mapName[1024];
		int xs, ys, xg, yg;
		double dist;
		fscanf(scenFile,"%d %s %d %d %d %d %lf\n",&bucket,mapName,&xs,&ys,&xg,&yg,&dist);
		
		// Check if we have this map already loaded
		if (strcmp(mapName,mapLoaded) != 0) {
			// No, need to load the new map
			if (map != NULL) {
				delete map;
			}
			map = new Map(mapName);
			strcpy(mapLoaded,mapName);
			printf("\nLoaded map %s\n",mapLoaded);

			// Create all the necessary structures
			unitSim = new unitRaceSimulation(map);
			aMap = unitSim->getAbstractMap();
			g = aMap->getAbstractGraph(0);
		}
		
		
		printf("\nGoal state: (%3d,%3d) ",xg,yg); fflush(stdout);
		double discrepancyGS[NUMD];
		for (int ek=0; ek < NUMD; ek++)
			discrepancyGS[ek] = 0.0;
		long numStatesGS = 0;
		
		// Compute the discrepancy
		// Go through all possible start states
		int xSize = map->getMapWidth();
		int ySize = map->getMapHeight();
		for (int x = 0; x < xSize; x++) {
			for (int y = 0; y < ySize; y++) {
				
				// Check if this is a passable state
				if (map->getTerrainType(x, y) != kGround)
					continue;
				
				// Compute the optimal path between the current state and the goal
				int startNodeNum = map->getNodeNum(x,y);
				int goalNodeNum = map->getNodeNum(xg,yg);
				node* start = g->getNode(startNodeNum);
				node* goal = g->getNode(goalNodeNum);
				
				// Compute the length of the optimal path between them
				path* pathBetween = myAStar->getPath(aMap, start, goal);
				
				// Check if the path exists
				if (pathBetween == NULL)
					continue;
				
				// Calculate the discrepancy
				double spDist = pathBetween->distance(aMap);
				double heDist = aMap->h(start,goal);
				delete pathBetween;
				
				numStatesGS++;
				for (int ek=0; ek<NUMD; ek++) {
					double scaledHeDist = heDist / gamma[ek];
					double diff = spDist - scaledHeDist;
					discrepancyGS[ek] += 100.0 * (fgreater(diff,0.0) ? diff : 0.0) / spDist;
//					discrepancyGS[ek] += 100.0 * fabs(diff) / spDist;
				}
			}
		}
		
		// Done with that goal state
		printf("...num states: %ld\n",numStatesGS);
		for (int ek=0; ek<NUMD; ek++) {
			printf("avg. discrepancyGS[%3.2f] = %1.3f\n", 1.0/gamma[ek], 
				   discrepancyGS[ek]/(double)numStatesGS);
			discrepancy[ek] += discrepancyGS[ek];
		}
		numStates += numStatesGS;
	}

	// Print stats for all maps
	printf("\n\n ============== All maps\n");
	for (int ek=0; ek<NUMD; ek++)
		printf("%1.9f\n",discrepancy[ek]/(double)numStates);
	
	// Close the scenario file and clean-up
	fclose(scenFile);
	delete map;
}



// Pick reasonable coordinates
void pickStartEnd(int& xs, int& ys, int& xg, int& yg, 
				  Map *map, graph *g, abstractMap* aMap,
				  double& spDist, double& aStarDifficulty, 
				  int _dist) {
	
	aStar* myAStar = new aStar();
//	double minDifficulty = 0.0;
//	double maxDifficulty = 10000.0;
	bool randCoord = (xs < 0 || ys < 0 || xg < 0 || yg < 0);

	unsigned failures = 0;
	int tolerance = 5;
	while (true) {
			
		if (randCoord) {
			xs = random()%map->getMapWidth();
			ys = random()%map->getMapHeight();
			xg = random()%map->getMapWidth();
			yg = random()%map->getMapHeight();
		}
		
		// Compute the optimal path between the two
		int startNodeNum = map->getNodeNum(xs,ys);
		int goalNodeNum = map->getNodeNum(xg,yg);
		node* start = g->getNode(startNodeNum);
		node* goal = g->getNode(goalNodeNum);
		
		// Compute the length of the optimal path between them
		path* pathBetween = myAStar->getPath(aMap, start, goal);
		unsigned expanded = myAStar->getNodesTouched();
		// Check if the path exists
		if (pathBetween == NULL)
			if (randCoord)
				continue;           // no path & random coordinates --> try again
			else {
				printf("Coordinates are given but there is no path between them!\n");
				exit(10);
			}
				
		// Check of the path between them is reasonable
		double noDist = aMap->h(start,goal);
		spDist = pathBetween->distance(aMap);
		double difficulty = spDist/noDist;
		// If A* had h* it would have expanded this many states
		unsigned minAStar = pathBetween->degree();
		delete pathBetween;
		// This allows us to compute "A* difficulty"
		aStarDifficulty = (double)expanded / (double)minAStar;
		
		// Check if the path length is around the one requested
		if (abs(spDist-_dist) > tolerance) {
			// Once again, we failed to find an appropriate path
			// Once we exceed numFails attempts we will relax the constraints
			if (++failures > 10) {
				tolerance++;
				failures = 0;
			}
			continue;
		}
		
//		// Check if the path difficulty is within allowed limits
//		if (aStarDifficulty < minDifficulty || aStarDifficulty > maxDifficulty) {
//			// Once again, we failed to find an appropriate path
//			// Once we exceed numFails attempts we will relax the constraints
//			if (++failures > 100) {
//				minDifficulty *= 0.8;
//				maxDifficulty *= 1.2;
//				failures = 0;
//			}
//			continue;
//		}
		
		// Print statistics
		printf("Start state:\t\t\t(%d,%d)\nGoal state:\t\t\t(%d,%d)\n",xs,ys,xg,yg);
		printf("No-obstacle distance:\t\t%1.2f\n",noDist);
		printf("Shortest path distance:\t\t%1.2f\n",spDist);
		printf("Distance difficulty:\t\t%1.2f\n",difficulty);
		printf("A* difficulty:\t\t\t%1.2f\n\n",aStarDifficulty);
		fflush(stdout);
		break;
	}
	
	delete myAStar;
}


// Race between PRLRTS and LRTS
void runPRLRTS(char *_map, int _xxs, int _yys, int _xxg, int _yyg, 
			 int _d, double _gamma, double _LQ, unsigned numRuns, 
			 bool drawMap, int _dist)
{
	// Set up the control parameters
	const double stepSize = 0.0001;			// time step
	double currTime = 0.0;					// initial time
	const double penalty = 1;				// thinking penalty multiplier
	const double maxTime = 10000.0*penalty;	// simulation time limit
	const double speed = 0.0;				// speed 
	const int visRadius = 10;			// visibility radius for the simulation
	
	printf("Penalty\t\t\t%1.5f\n",penalty);
	printf("Speed\t\t\t%1.5f\n",speed);
	printf("Step size\t\t%1.5f\n",stepSize);
	printf("Max simulation time\t%1.0f\n\n",maxTime);
	
	// Map and simulation
	Map *map = NULL;
	unitSimulation *unitSim = NULL;
	
	// If the map is given then use it. Otherwise create a default 100x100 map
	if (_map[0] == 0)
		map = new Map(100, 100);
	else
		map = new Map(_map);
	
	// Algorithms
//	praStar* pra = new praStar(); // pra->setPartialPathLimit(4);
//	praStar* pra1 = new praStar(); pra1->setPartialPathLimit(4);
//	aStar* as = new aStar();
	assert(_d <= visRadius);		// check for cheating
//	LRTS* lrts = new LRTS(_gamma, _d, _LQ);   // gamma, d, T
	LRTS* lrts = new LRTS(0.6, 10, 0.0);   // gamma, d, T
	PR_LRTS* pr_lrts = new PR_LRTS(_gamma, _d, _LQ);

	// Run the converge run numRuns times for different start and finish
	for (unsigned run = 0; run < numRuns; run++) {
		
		// Create the unit simulation
		unitSim = new unitRaceSimulation(map);
		abstractMap* aMap = unitSim->getAbstractMap();
		unitSim->setAsynchronous();
		unitSim->setPenalty(penalty);
		unitSim->setUseBlocking(true);
		unitSim->setRealTime(false);
		
		// Creating a group for the incremental A* unit
//		sharedAMapGroup* asGroup = new sharedAMapGroup();
//		unitSim->addUnitGroup(asGroup);
//		asGroup->setVisibilityRadius(visRadius);
		
		// Creating a group for the PRA* unit
//		sharedAMapGroup* praGroup = new sharedAMapGroup();
//		unitSim->addUnitGroup(praGroup);
//		praGroup->setVisibilityRadius(visRadius);
		
		// Creating a group for the PRA* unit with partial refinement
//		sharedAMapGroup* praGroup1 = new sharedAMapGroup();
//		unitSim->addUnitGroup(praGroup1);
//		praGroup1->setVisibilityRadius(visRadius);
		
		// Creating a group for the LRTS unit
		LRTSGroup* lrtsGroup = new LRTSGroup(); 
		unitSim->addUnitGroup(lrtsGroup);
		
//		// Creating a group for the PR_LRTS unit
		PR_LRTSGroup* pr_lrtsGroup = new PR_LRTSGroup(); 
		unitSim->addUnitGroup(pr_lrtsGroup);
		pr_lrtsGroup->setVisibilityRadius(visRadius);
		
		// Get the base-level graph for this map
		graph *g = aMap->getAbstractGraph(0);
		
		// Create one goal and one start
		double spDist = 0.0;
		double aStarDifficulty = 0.0;
		int xxs = _xxs; 
		int yys = _yys;
		int xxg = _xxg;
		int yyg = _yyg;
		pickStartEnd(xxs, yys, xxg, yyg, map, g, aMap, spDist, aStarDifficulty, _dist);
		
		// Create the static goal state (of white color)
		unit* targetUnit = new unit(xxg, yyg, 255, 255, 255, 0);
		unitSim->addUnit(targetUnit);
		
		// Create the units and add them to the simulation
		
		// A*
//		searchUnit* asUnit = new searchUnit(xxs, yys, targetUnit, as);
//		asGroup->addUnit(asUnit);
//		unitSim->addUnit(asUnit,false);
//		asUnit->setSpeed(speed);
		
		// PRA* without partial refinement
//		searchUnit* praUnit = new searchUnit(xxs, yys, targetUnit, pra);
//		praGroup->addUnit(praUnit);
//		unitSim->addUnit(praUnit,false);
//		praUnit->setSpeed(speed);
	
		// PRA* with partial refinement
//		searchUnit* praUnit1 = new searchUnit(xxs, yys, targetUnit, pra1);
//		praGroup1->addUnit(praUnit1);
//		unitSim->addUnit(praUnit1,false);
//		praUnit1->setSpeed(speed);
		
		// LRTS Unit
		LRTSUnit* lrtsUnit = new LRTSUnit(xxs, yys, targetUnit, lrts);
		lrtsGroup->addUnit(lrtsUnit);
		unitSim->addUnit(lrtsUnit,false);
		lrtsUnit->setSpeed(speed);
		
//		// PR LRTS unit
		PR_LRTSUnit* pr_lrtsUnit = new PR_LRTSUnit(xxs, yys, targetUnit, pr_lrts);
		pr_lrtsGroup->addUnit(pr_lrtsUnit);
		unitSim->addUnit(pr_lrtsUnit,false);
		pr_lrtsUnit->setSpeed(speed);
		
		// Produces the prefix name: e.g., 16.LRTS(1,1.0,inf)
		char prefixName[1034];
		for (int i=0; i<1034; prefixName[i++]='\0');
		if (_map[0] != '\0')
			for (int i=0; (_map[i] != '.' && 
						   _map[i] != '\0' && i<1033); i++) 
				prefixName[i] = _map[i];
		else
			strcpy(prefixName,"default");
		strcat(prefixName,".");
		strcat(prefixName,lrtsUnit->getName());
		
		// Produces the two file name
		// e.g., 16.LRTS(1,1.0,inf).003.txt
		// e.g., 16.LRTS(1,1.0,inf).003.dat
		char logFileName[1200];
		char datFileName[1200];
		sprintf(logFileName,"%s.%03d.%03d.txt",prefixName,run,_dist);
		sprintf(datFileName,"%s.%03d.%03d.dat",prefixName,run,_dist);
		
		// Open the main log file
		FILE *trialLog = fopen(logFileName,"w");
		unitSim->setLogFile(trialLog);
		
		// Now dump some statistics into a separate file
		FILE *datLog = fopen(datFileName,"w");
		fprintf(datLog,"%1.6f %1.6f\n",spDist,aStarDifficulty);
		fprintf(datLog,"%d %d\n",g->getNumNodes(), 
				map->getMapWidth()*map->getMapHeight());
		fprintf(datLog,"%d %d %d %d\n",xxs,yys,xxg,yyg);
		fclose(datLog);
		
		
		if (drawMap)
			printf("%c[2J", 27);
		
		// Now simulate until convergence (or time out)
		do 
		{
			currTime+=stepSize;
			unitSim->advanceTime(stepSize);
			
			if (drawMap)
				unitSim->print(false);
		}
		while (currTime < maxTime && !unitSim->done());
		
		// Clean things up
		if (unitSim != NULL) delete unitSim;
	}
	if (map != NULL) delete map;
}

