#include "map.h"
#include "abstractMap.h"
#include "unitSimulation.h"
#include "humanUnit.h"

#ifdef OS_MAC
#include <AGL/agl.h>
#endif

#ifndef COMMON_H
#define COMMON_H

enum
{
	kMainMenu = 500,
	kCloseMenuItem = 2,
	kInfoMenuItem = 4,
	kAnimateMenuItem = 5,
	kInfoState = 0,
	kAnimateState = 0
};

enum {
	kFSAAOff = 0,
	kFSAAFast = 1,
	kFSAANice = 2,
	kSamples = 4
};

// camera/context info
//typedef struct {
//	GLdouble x,y,z;
//} recVec;

typedef struct {
	recVec viewPos; // View position
	recVec viewDir; // View direction vector
	recVec viewUp; // View up direction
	recVec rotPoint; // Point to rotate about
	GLdouble aperture; // camera aperture
	GLint viewWidth,viewHeight; // current window/screen height and width
	GLfloat viewOriginX, viewOriginY; // always 0 
} recCamera;

// per view data
struct recContext
{
#ifdef OS_MAC
	AGLPixelFormat aglPixFmt;
	AGLContext aglContext;
	GLuint boldFontList;
	GLuint regFontList;
	EventLoopTimerRef timer;
	DMExtendedNotificationUPP  windowEDMUPP;
	AbsoluteTime time;
#endif

	bool animate;
	bool info;
	bool drawHelp;
	bool drawCaps;
	bool polygons;
	bool lines;
	bool points;
	bool showCredits;
	bool showLand;
	long lighting;
	bool paused;
	bool drawing;
	
	// spin
	GLfloat fRot [3];
	GLfloat fVel [3];
	GLfloat fAccel [3];
	
	// camera handling
	recCamera camera;
	GLfloat worldRotation [4];
	GLfloat objectRotation [4];
	GLfloat shapeSize;
	
	char message[256]; // buffer for message output
	float msgTime; // message posting time for expiration
	
	char * names;
	
	long surface; 
	long colorScheme;
	unsigned long subdivisions;
	unsigned long xyRatio;
	
	long modeFSAA;
	
	Map *map;
	abstractMap *abstrMap;
	unitSimulation *unitLayer;
	humanUnit *human;
};
typedef struct recContext recContext;
typedef struct recContext * pRecContext;

// single set of interaction flags and states
extern GLfloat gTrackBallRotation [4];
extern pRecContext gContextTrackingInfo;

void updateProjection(pRecContext pContextInfo);
void drawGL(pRecContext pContextInfo);
#ifdef OS_MAC
pRecContext GetCurrentContextInfo (WindowRef window);
#endif

bool doKeyboardCommand(pRecContext pContextInfo, char keyHit, bool shift);
void SetLighting(unsigned int mode);
void initialConditions(pRecContext pContextInfo);
void resetCamera(recCamera * pCamera);

void processCommandLineArgs(int argc, char *argv[]);
void handleBatchRTS(int argc, char *argv[], int mode, char *_map);
void handleBatchPRA(int argc, char *argv[], int mode, char *_map);
//void runBatchMode(char *mode, char *_map, int xxs, int yys, int xxg, 
//				  int yyg, int, double, double, unsigned int, bool _drawMap, double);
void runSimpleUnitTest(char *_map);
void runLRTS(char *_map, int xxs, int yys, int xxg, int yyg, 
			 int _d, double _gamma, double _LQ, unsigned _numRuns, 
			 bool _drawMap, int _dist);
void runPRLRTS(char *_map, int xxs, int yys, int xxg, int yyg, 
			 int _d, double _gamma, double _LQ, unsigned _numRuns, bool _drawMap,
			   int _dist);
void randomPathTesting(char *_map);
void pickStartEnd(int& xs, int& ys, int& xg, int& yg, 
				  Map *map, graph *g, abstractMap* aMap,
				  double& spDist, double& aStarDifficulty, int);
void scenarioGenerator(char *_map);
void discrepancy(char* scenName, char* algName, 
				 int _d=1, double _gamma=1.0, double _LQ=99999999999.0);

void runScenario(char* scenName, char* algName, 
				 int _d=1, double _gamma=1.0, double _LQ=99999999999.0);

void calculateAverageTimePerStep(char *_map, char *coordFile, int MAP_SIZE);

void calculateSubOptimality(char *_map, char *coordFile, int MAP_SIZE);
extern char gDefaultMap[1024];

#endif
