#include "Handler.H"
#include <cstdlib>
#include <iostream>
using namespace std;

Handler::Handler(const string &server_, sint4 port_,  QObject *receiver_)
{
  receiver = receiver_;
  server = server_;
  port = port_;
  
  client = NULL;
  
  loginOK      = false;
  buddyListOK  = false;
  inInvitation = false;
  usersListOk  = false;
}

Handler::~Handler(){
    delete client;
}

void Handler::run()
{
  string message;
  Net_Message m;
  //MessageEvent *m_event = NULL;
  
  // parsers
  Protocol::Login_Notify loginNotify;
  Protocol::Game_Start game_start;
  Protocol::Game_Info game_info;
  Protocol::Invitations invit;
  Protocol::Buddies buddies;
  Protocol::Player_List players;
  Protocol::Joined joined;
  Protocol::Left left_;
  Protocol::Who_List who;
  Protocol::Readied readied;
  Protocol::Unreadied unreadied;
  Protocol::Added_Buddy added_buddy;
  Protocol::Remed_Buddy remed_buddy;
  Protocol::Tell tell;
  Protocol::Aborted aborted;
  Protocol::Kicked kicked;
  Protocol::Error error;
  
  FOREVER {
    client->recv(message);
    m.setMessage(message);
    
    if(!client->read_is_good()){
      break;
    }
    
    if (m.getFrom() == "G")     { continue;}
    if (m.getFrom() != servicename)   { handle_chat(m);                  continue; }
    if (tell.parse(m))          { handle_tell(tell);               continue; }
    if (readied.parse(m))       { handle_readied(readied);         continue; }
    if (unreadied.parse(m))     { handle_unreadied(unreadied);     continue; }
    if (loginNotify.parse(m))   { handle_loginNotify(loginNotify); continue; }
    if (game_start.parse(m))    { handle_game_start(game_start);   continue; }
    if (game_info.parse(m))     { handle_game_info(game_info);     continue; }
    if (invit.parse(m))         { handle_invitations(invit);       continue; }
    if (aborted.parse(m))       { handle_aborted(aborted);         continue; }
    if (buddies.parse(m))       { handle_buddies(buddies);         continue; }
    if (players.parse(m))       { handle_playerList(players);      continue; }
    if (who.parse(m))           { handle_whoList(who);             continue; }
    if (joined.parse(m))        { handle_joined(joined);           continue; }
    if (left_.parse(m))         { handle_left (left_);             continue; }
    if (added_buddy.parse(m))   { handle_added(added_buddy);       continue; }
    if (remed_buddy.parse(m))   { handle_removed(remed_buddy);     continue; }
    if (kicked.parse(m))        { handle_kicked(kicked);           continue; }
    if (error.parse(m))         { handle_error(error);             continue; }
    
    cout << m.toString();
    cout << "\tmessage not parsed : \n\t\t" << message << endl;
  }
  
  cout << "\t************************* Disconnected ***********************************"<< endl;
  QApplication::postEvent(receiver, new MessageEvent("\t*** Disconnected ***", EventType::DISCONNECTED));

}

// teller functions
void Handler::tellService(const string& message){
    cout << "tellService:" << message << endl;
  client->send("tell " + servicename + " " + message);
}

void Handler::tellPeer(const string& peer, const string& message){
  client->send("tell " + peer + " " + message);
}

void Handler::tellBuddies(const string& message){
  FORALL(buddyList, i) client->send("tell " + (*i).first + " " + message);
}

void Handler::tellGGS(const string& message){
  client->send(message);
}



void Handler::handle_chat(const Net_Message& m){
  string context;
  
  if(m.getTo() == "U"){
    context = m.getFrom();
  }else{
    context = m.getTo();
  }
  createContext(context);
  string message;
  
  for(uint4 i = 0; i < m.getSize(); i++){
    message += m.getArgument(i);
    message += " ";
  }
  
  QApplication::postEvent(receiver, new TellEvent(m.getFrom(), context, message));
}


void Handler::handle_tell(const Tell& tell ){
  if(tell.get_iid() != current_iid){
    cout << "invalid tell command... user not currently in game with iid : " << tell.get_iid() << endl;
    return;
  }
  
  if(tell.get_pid() == current_pid ){
    return;
  }
  string sender = (playerList[tell.get_pid()]).first;
  QApplication::postEvent(receiver, new TellEvent(sender, "game", tell.get_out_msg()));
}



void Handler::GGSlogin(const string &nick_, const string &pass_, const string &server_ip, const string &port_, const string &servicename_)
{
  nick = nick_;
  pass = pass_;
  server = server_ip;
  port = atoi(port_.c_str());
  servicename = servicename_;
   
  // log on to ggs
  if(client == NULL){
    client = new Net_GGSClient(nick, pass, server, port, true);
  }else{
    client->close();
    terminate(); // kill the listen thread
    wait();
    delete client;
    client = new Net_GGSClient(nick, pass, server, port, true);
  }
  
  if(!(client->good())){
    // fixme QApplication::postEvent();
    QApplication::postEvent(receiver, new MessageEvent("Unable to connect to GGS:\n"+client->get_error(), EventType::ERROR));
    REM("unable to connect to GGS");
    return;
  }
  
   start();
}
/**
    logs onto the service using the given password and level
 */
void Handler::login(const string &nick_, const string &pass_, Level level_, const string &server_ip, const string &port_, const string &servicename_)
{
  nick = nick_;
  pass = pass_;
  level = level_;
  server = server_ip;
  port = atoi(port_.c_str());
  servicename = servicename_;
  
  // log on to ggs
  if(client == NULL){
    client = new Net_GGSClient(nick, pass, server, port, true);
  }else{
    client->close();
    terminate(); // kill the listen thread
    wait();
    delete client;
    client = new Net_GGSClient(nick, pass, server, port, true);
  }
  
  if(!(client->good())){
   QApplication::postEvent(receiver, new MessageEvent("Unable to connect to GGS:\n"+client->get_error(), EventType::ERROR));
       REM("unable to connect to GGS");
    return;
  }
  
  Finger finger;
  tellGGS(Finger::compose(servicename));
  awaitResponce(finger);
  
  if(finger.get("ERR") != "NULL"){
    QApplication::postEvent(receiver, new MessageEvent("The service is temporarily offline\nplease try again later", EventType::ERROR));
    return;
  }
  tellGGS(Finger::compose(nick));
  awaitResponce(finger);
  
  string ip = finger.get("host");
  if(ip == "NULL"){
    // fixme QApplication::postEvent();
    REM("unable to retrieve local host address");
    return;
  }
  
  // log on to the service
  tellService(Login::compose(level, pass, ip));
  
  start();
}

/**
  retrieves all the lists from the service
  */
void Handler::initHandler(){
  tellService(List::compose(INVITE_LIST));
  tellService(List::compose(BUDDY_LIST));
  tellService(List::compose(WHO_LIST));
}

/**
  creates a new chatting context and fires a NEW_CONTEXT event
*/
void Handler::createContext(const string& to){
  if(contextList.find(to) == contextList.end()){
    contextList[to] = to;
    QApplication::postEvent(receiver, new MessageEvent(to, EventType::NEW_CONTEXT));
  }
}

void Handler::deleteContext(const string& to){
  ChatContextList::iterator i = contextList.find(to);
  if( i != contextList.end()){
    contextList.erase(i);
  }
}

void Handler::handle_loginNotify(const Login_Notify& login)
{
  if(login.get_sender() != servicename) { return; }
  
  if(!loginOK){
    if(login.get_player_name() == nick){
      if(login.get_online()){
        QApplication::postEvent(receiver, new MessageEvent("log in ok", EventType::LOGIN_OK)); // Qt will delete
        initHandler();
        loginOK = true;
        return ;
      }else{
        QApplication::postEvent(receiver, new MessageEvent("login failed: incorrect information", EventType::LOGIN_FAIL)); // Qt will delete
        return;
      }
    }
  }
  
  // buddy notify
  if(buddyListOK){
    BuddyList::iterator i = buddyList.find(login.get_player_name());
    if(i != buddyList.end()){
      (*i).second = login.get_online();
      if(login.get_online()){
        QApplication::postEvent(receiver, new MessageEvent("buddy " + login.get_player_name() + " is online", EventType::REFRESH_BUDDIES));
      }else{
        if(i != whoList.end()){
          QApplication::postEvent(receiver, new MessageEvent( "buddy " + login.get_player_name() + " went offline ", EventType::REFRESH_BUDDIES));
        }
      }
    }
  }
  
  //user notify
  if(login.get_online()){
    whoList[login.get_player_name()] = true;
    QApplication::postEvent(receiver, new MessageEvent(login.get_player_name() + " is online", EventType::REFRESH_USERS));
  }else{
    WhoList::iterator i = whoList.find(login.get_player_name());
    whoList.erase(i);
    QApplication::postEvent(receiver, new MessageEvent(login.get_player_name() + " went offline ", EventType::REFRESH_USERS));
  }
  
}

void Handler::handle_readied (const Readied& readied){
  (playerList[readied.get_pid()]).second = true;
  QApplication::postEvent(receiver, new MessageEvent((playerList[readied.get_pid()]).first + " is READY", EventType:: REFRESH_PLAYERS));
}

void Handler::handle_unreadied (const Unreadied& unreadied){
  (playerList[unreadied.get_pid()]).second = false;
  QApplication::postEvent(receiver, new MessageEvent((playerList[unreadied.get_pid()]).first + " is UNREADY", EventType:: REFRESH_PLAYERS));
}


void Handler::handle_game_start(const Game_Start& start){
  REM("[------------GAME_START------------------]");
  if(start.get_sender() != servicename){    return;      }
  
  string command = "bin/ortsg -nosound " + current_info;
  cout << "COMMAND: " << command << endl;
  system(command.c_str());
}

void Handler::handle_game_info(const Game_Info& info){
  REM("[------------GAME_INFO------------------]");
  if(info.get_sender() != servicename){    return;      }
  current_info =  info.get_info();
}


void Handler::handle_invitations(const Invitations& inv){
  REM("{--------------------------INVITATIONS-----------------------------}");
  Invit_Disc current;
  current = inv.get();
  invitations [current.get_iid()] = make_pair( current , true);
    
  QApplication::postEvent(receiver, new MessageEvent("New invitation(s)", EventType::REFRESH_INVIT));
}



void Handler::handle_aborted(const Aborted& aborted){
  InvitationsList::iterator i = invitations.find(aborted.get_iid());
  if(i != invitations.end() ){
    invitations.erase(i);
  }
  stringstream s;
  s << aborted.get_iid();
  QApplication::postEvent(receiver, new MessageEvent(s.str(), EventType::REFRESH_INVIT));
}



void Handler::handle_buddies(const Buddies& buddies){
  for(uint4 i = 0; i < buddies.size(); i++) {
    buddyList[buddies.get_name(i)] = buddies.get_online(i) ;
  }
  buddyListOK = true;
  QApplication::postEvent(receiver, new MessageEvent("buddy list successfully retrieved", EventType::REFRESH_BUDDIES));
}


void Handler::handle_playerList    (const Player_List& players){
  REM("Players parsed");
  for(uint4 i = 0; i < players.size(); i++) {
    playerList[players.get_alias(i)] = make_pair ( players.get_name(i),players.get_ready(i) );
    cout << "adding : " <<  players.get_alias(i) << " "<< players.get_name(i) << " " << (bool)players.get_ready(i) << endl;
  }
  QApplication::postEvent(receiver, new MessageEvent("Buddies list", EventType:: REFRESH_PLAYERS));
}

void Handler::handle_whoList    (const Who_List& who){
  REM("Who parsed");
  if(usersListOk) { return; }
  usersListOk = true;
  for(uint4 i = 0; i < who.size(); i++) {
    whoList[who.get_name(i)] = true;
    cout << "(who)"<<  who.get_name(i) << endl;
  }
  QApplication::postEvent(receiver, new MessageEvent("Users list", EventType:: REFRESH_USERS));
}

void Handler::handle_joined    (const Joined& joined){
  REM("Joined parsed");
  
  // verify sender
  if(joined.get_sender() != servicename){ return ;}
  
  if(!inInvitation) { // this must be you
    inInvitation = true;
    current_pid = joined.get_pid();
    current_iid  = joined.get_iid();
    
    if(joined.get_pid() == 0) {
      //  u created this game
      QApplication::postEvent(receiver, new MessageEvent("created game successfully", EventType::CREATE_OK));
    }else{
      // not your game, just joined
      QApplication::postEvent(receiver, new MessageEvent("Joined game successfully", EventType::JOIN_OK));
    }
    
  } else { // someone joined a game that you are currently in
    playerList[joined.get_pid()] = make_pair(joined.get_player_alias(),false);
    QApplication::postEvent(receiver, new MessageEvent( "-----[ " + joined.get_player_alias() + " joined game ]--->", EventType::PLAYER_JOINED));
  }
}


void Handler::handle_left (const Protocol::Left& left_)
{
  REM("Parsed left");
  
  if(left_.get_pid() == current_pid){
    QApplication::postEvent(receiver, new MessageEvent("left game successfully", EventType::YOU_LEFT));
    clearInGameInfo();
  }else{
    PlayerList::iterator i = playerList.find(left_.get_pid());
    string name = ((*i).second).first;
    playerList.erase(i);
    QApplication::postEvent(receiver, new MessageEvent("<---[ " + name + " left game ]-----", EventType::PLAYER_LEFT));
  }
}

void Handler::handle_added(const Added_Buddy& added){
  REM("added parsed");
  buddyList[added.get_player_name()] = added.get_online();
  QApplication::postEvent(receiver, new MessageEvent("buddy " + added.get_player_name() + " has been added to your list", EventType::REFRESH_BUDDIES));
}

void Handler::handle_removed(const Remed_Buddy& remed){
  REM("remed parsed");
  buddyList.erase(buddyList.find(remed.get_player_name()));
  QApplication::postEvent(receiver, new MessageEvent("buddy " + remed.get_player_name() + " has been removed from your list", EventType::REFRESH_BUDDIES));
}

void Handler::handle_kicked(const Kicked& kicked){
  PlayerList::iterator i = playerList.find(kicked.get_pid());
  string name = ((*i).second).first;
  playerList.erase(i);
  QApplication::postEvent(receiver, new MessageEvent(name , EventType::PLAYER_KICKED));
  
  if(kicked.get_pid() == current_pid){
    QApplication::postEvent(receiver, new MessageEvent(name , EventType::U_KICKED));
    clearInGameInfo();
  }
}

void Handler::handle_error(const Error& error){
  QApplication::postEvent(receiver, new MessageEvent(error.get_msg() , EventType::ERROR));
}
    
bool Handler::awaitResponce(Message_Base& expected){
  string message;
  Net_Message m;
  
  FOREVER {
    client->recv(message);
    m.setMessage(message);
    if(expected.parse(m)){
      return true;
    }
  }
  return false;
}

void Handler::clearInGameInfo(){
  inInvitation = false;
  current_pid = -1;
  current_iid = -1;
  playerList.clear();
}
