/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you wish to add, delete or rename functions or slots use
** Qt Designer which will update this file, preserving your code. Create an
** init() function in place of a constructor, and a destroy() function in
** place of a destructor.
*****************************************************************************/

#include <qtextedit.h>
#include <string>
#include <qobject.h>
#include <iostream>

#include "Global.H"
#include "Net_GGSClient.H"
#include "Net_Message.H"
#include "Net_Protocol.H"

#include "Handler.H"
#include "close1.xpm"
#include "close2.xpm"
#include "close3.xpm"

using namespace std;

/*
  Called to set up GGSClient
*/

Handler *handler;
const QString WINDOWS_REGISTRY = "/QtExamples";
const QString APP_KEY = "/ortsNetGUI/";

int currentIID; /** the iid of the currently joined game */

typedef QMap<string, QTextEdit*> ContextList;
ContextList contextList;

char * ping_server(char *server_address);
char buffer[100]; // string buffer used for storing 1 line of text from ping output

/* 
  This function grabs the text in chatLineEdit and
  feeds it to historyTextEdit
*/
void ortsNetGUI::writeTo_History()
{
    // Retrieve context
    string context = contextTabWidget->tabLabel((contextTabWidget->currentPage()));
    fprintf(stderr, "PEER CONTEXT: %s\n", context.c_str());
    // Retrieve input from chatlineEdit
    string message = mainchatLineEdit->text();
    mainchatLineEdit->clear();
    if(context == "[ All Buddies ]"){
	handler->tellBuddies(message);
    }else{
	handler->tellPeer(context, message);
    }
    TellEvent* event = new TellEvent(":", context, "<font color=\"#FF0000\">" + message);
    QApplication::postEvent( this, event );
}


GQListBoxToolTip* tip;

void ortsNetGUI::init()
{    
    cout << "------- In Init ------- "<< endl;
    loadSettings();    
    login = 0;
    creategame = 0;
    addBuddyForm = 0;
    anonydialog = 0;
    error = new GenericErrorForm();
    currentIID = -1;
    string server = "opal.cs.ualberta.ca";
    int port = 5000;
    handler = new Handler(server, port, this);
    //tip= new GQListBoxToolTip(invitationListBox, handler);
    this->setDisabled(true);// disable main gui until user has successfully logged in
    QToolButton* tab_close_button = new QToolButton(contextTabWidget, "close_chat_tab_button");
    QIconSet* iconset = new QIconSet();
    iconset->setIconSize (QIconSet::Small, QSize(16,16));
    iconset->setPixmap (QPixmap (close1_xpm), QIconSet::Small, QIconSet::Normal);
    iconset->setPixmap (QPixmap (close2_xpm), QIconSet::Small, QIconSet::Active);
    iconset->setPixmap (QPixmap (close3_xpm), QIconSet::Small, QIconSet::Disabled);
    tab_close_button->setMinimumSize (16,16);
    tab_close_button->setMaximumSize (16,16);
    tab_close_button->setIconSet(*iconset);
    contextTabWidget->setCornerWidget(tab_close_button);
    connect(tab_close_button, SIGNAL(clicked()), this, SLOT(closeTab()));
    invitationTable->setColumnWidth(0, 60);
    invitationTable->setColumnWidth(1, 55);
    invitationTable->setColumnWidth(2, 55);
    invitationTable->setColumnWidth(3, 55);
    invitationTable->setColumnWidth(4, 60);
    invitationTable->setColumnWidth(5, 55);
    invitationTable->setColumnWidth(6, 55);
    invitationTable->setColumnWidth(7, 55);
    invitationTable->setColumnWidth(8, 55);
    startLoginDialog();
}  

void ortsNetGUI::customEvent( QCustomEvent* e )
{
    REM("-----------> Event received by ortsNetGUI");
    MessageEvent*  event = (MessageEvent*) e ;
    int type = event ->type();
    string message = event->getMessage();
    stringstream s;
    switch(type){
    case EventType::REFRESH_INVIT :
	REM("type == REFRESH_INVIT");
	refreshInviteList();
	int abort_iid;
	s.str(message);
	s>>abort_iid;
	if (currentIID==abort_iid) //we're in the game that was aborted
	{
	    currentIID=-1;
	    error->showError("Game aborted by owner");
	    switchToChatScreen();
	}
	break;
	
    case EventType::LOGIN_OK : // received login ok from handler;
	REM("type == LOGIN_OK");
	if(login->isVisible() ) {
	    login->hide();
	}
	this->setEnabled(true);
	break;
	
    case EventType::REFRESH_BUDDIES:
	REM("type == REFRESH_BUDDIES");
	refreshBuddyList();
	PagerTextEdit->append(event->getMessage());
	break;
	
    case EventType::REFRESH_USERS:
	REM("type == REFRESH_USERS");
	refreshUserList();
	break;
	
    case EventType::JOIN_OK :
	REM("type == JOIN_OK");
	switchToReadyScreen();
	break;
	
    case EventType::REFRESH_PLAYERS :
	REM("type == REFRESH_PLAYERS ");
	refreshPlayerList();
	break;
	
    case EventType::CREATE_OK :
	REM("type == CREATE_OK");
	// todo: enable creater buttons
	if(creategame->isVisible() ) {
	    creategame->hide(); 
	}
	switchToReadyScreen();
	break;
		
    case EventType::PLAYER_JOINED  :
	REM("type == PLAYER_JOINED ");
	PagerTextEdit->append("<font color=\"#0000FF\"><b> " + event->getMessage() + "<\b>");
	refreshPlayerList();
	break;
	
   case EventType::PLAYER_LEFT :
	REM("type == PLAYER_LEFT ");
	PagerTextEdit->append(event->getMessage());
	refreshPlayerList();
	break;
	
    case EventType::YOU_LEFT :
	REM("type == YOU_LEFT");
	switchToChatScreen();
	break;
	
    case EventType::NEW_CONTEXT:
	REM("type == NEW_CONTEXT");
	createContextTab(event->getMessage());
	break;
	
    case EventType::PLAYER_KICKED :
	REM("type == PLAYER_KICKED");
	PagerTextEdit->append("<font color=\"#0000FF\"><b> " + event->getMessage() + "<\b>");
	refreshPlayerList();
	break;	
    
    case EventType::U_KICKED :
	REM("type == U_KICKED");
	PagerTextEdit->append(event->getMessage() + "<\b>");
	switchToChatScreen();
	break;	

    case EventType::DISCONNECTED :
	REM("type == DISCONNECTED");
	PagerTextEdit->append(event->getMessage() + "<\b>");
	break;	
    
    case EventType::ERROR:
	REM("type == ERROR");
	error->showError(event->getMessage());
	break;	
	
    case EventType::TELL :
	REM("type == TELL");
	TellEvent* e = (TellEvent*) event;
	appendToContext(e->getContext(), e->getSender() + ": " + e->getMessage());
	break;
    } 
    //PagerTextEdit->append("- " + event->getMessage());
}

/**
  appends the give message to the appropriet context
*/
void ortsNetGUI::appendToContext(const string& context, const string& message)
{
    if(context == "game"){
	gameHistoryTextEdit->append(message);
	cout << message;
	return;
    }
    ContextList::iterator i = contextList.find(context);
    if(i != contextList.end()){
	i.data()->append(message);
    }
}

/**
  Show login dialog
*/
void ortsNetGUI::startLoginDialog()
{
    if(login == 0){
	REM("-------- CREATE NEW LOGINDIALOG ------- ");
	login = new LoginDialog(this);
    }
    login->setup(handler,this);
    login->exec();
}

void ortsNetGUI::startCreateDialog()
{
    if(creategame == 0){
	REM("-------- CREATE NEW CREATEGAME DIALOG ------- ");
	creategame = new CreateGame(this);
    }
    creategame->setup(handler);
    creategame->show();
}

void ortsNetGUI::refreshInviteList() {
    
    invitationTable->setNumRows(0); // clears Table
    InvitationsList list = handler->getInvitations(); // grab current list
    int index;
    stringstream ss;
    //string worker_info; // stl basic string version
    //char worker_info_char[256]; // c string version
    
    // cycle thru list of invitations and repopulate the game invites table with current list
    FORALL(list, i){
	if( ((*i).second).second ) {
	     Invit_Disc disc = ((*i).second).first;
	     
	     index = invitationTable->numRows();
	     fprintf(stderr, "INSERTING ROW AT INDEX: %d\n", index);
	     invitationTable->insertRows(index);
	     
	     ss.str("");
	     if(disc.get_passprotected() > 0) {
		 ss << "Yes";
	     } else {
		 ss << "No";
	     }
	     invitationTable->setText(index, 0, ss.str()); // set password
	     ss.str("");
	     ss << disc.get_iid();
	     invitationTable->setText(index, 1, ss.str()); // set iid
	     invitationTable->setText(index, 2, disc.get_game_name());
	     invitationTable->setText(index, 3, disc.get_owner());
	     invitationTable->setText(index, 4, "game type");
	     ss.str("");
	     ss << disc.get_num_players() << "/" << disc.get_max_players();
	     invitationTable->setText(index, 5, ss.str()); // set players
	     //disc.get_invisible()
	     //disc.get_anony_ids()
	     /*worker_info = handler->get_workerinfo();
	     strcpy(worker_info_char, worker_info.c_str());
	     if(worker_info == NULL || worker_info == "") {
		 invitationTable->setText(index, 6, "ping");
	     } else {
		 //tokenize the -shost value from the current_info (game info) string
		 char * worker_host = strtok (worker_info_char, " ");
		 worker_host = strtok (NULL, " "); // grab the 2nd token, that's the value we want
		 // call my ping function with host and get ping value
		 invitationTable->setText(index, 6, worker_host);
	     }*/
	     string myServer = handler->get_server();
	     char myServerChar[256];
	     strcpy(myServerChar, myServer.c_str());
	     invitationTable->setText(index, 6, ping_server(myServerChar));
	     invitationTable->setText(index, 7, "map");
	     invitationTable->setText(index, 8, disc.get_options());
	 }
     }
}

void ortsNetGUI::refreshUserList(){
    userListBox->clear();
    WhoList list = handler->getUsers();
    string current;
   
    FORALL(list, i){
	cout << current << endl;
	 current = (*i).first;
	 userListBox->insertItem (current);
     }
}

void ortsNetGUI::refreshBuddyList()
{
    buddyListBox->clear();
    BuddyList list = handler->getBuddies();
    string current;
   
    FORALL(list, i){
	 current = (*i).first;
	 if( (*i).second ) {
	     buddyListBox->insertItem (current);
	 }
     }
   
    FORALL(list, j){
	 current = (*j).first;
	 if(!(*j).second ) {
	     buddyListBox->insertItem ("- " + current);
	 }
     }
     buddyListBox->insertItem ("[ All Buddies ]");
 }



void ortsNetGUI::refreshPlayerList()
{
    playerListBox->clear();
    PlayerList list = handler->getPlayers();
    pair<string, uint1> current;
    
    FORALL(list, i){
	current = (*i).second;
	if( current.second ) {
	    playerListBox->insertItem( current.first + " [R]");
	}
     }
     FORALL(list, j){
	 current = (*j).second;
	if(! current.second ) {
	   playerListBox->insertItem(current.first);
	}
     }
}

/*****************************************************************************************
  *  This function makes the ready screen visible
  ****************************************************************************************/
void ortsNetGUI::switchToReadyScreen()
{
    widgetStack->raiseWidget(1);
}

/*****************************************************************************************
  *  This function makes the chat screen visible
  ****************************************************************************************/
void ortsNetGUI::switchToChatScreen()
{
    widgetStack->raiseWidget(0);
}

void ortsNetGUI::requestJoinGame()
{
    // get selected game index
    int index = invitationTable->currentRow();
    fprintf(stderr, "CURRENT ROW AT INDEX: %d\n", index);
    if(index < 0) return;
    // get iid
    InvitationsList list = handler->getInvitations();
    Invit_Disc disc = list[index].first;
    int iid = disc.get_iid();
    currentIID = iid;
    string user  = disc.get_owner();
    bool anony_ids = disc.get_anony_ids();
    bool passprotected = disc.get_passprotected();
    if( anony_ids ) cout << "GAME ANONYMIZES IDS" << endl;
    if( passprotected ) cout << "GAME IS PASSPROTECTED " << endl;
    // check if anonyid
    if( anony_ids || passprotected ) {
	if( anonydialog == 0){
	    REM("-------- CREATE NEW ANONYMIZING DIALOG ------- ");
	    anonydialog = new AnonymizingDialog(this);
	}
	anonydialog->setup( handler, iid, user, anony_ids, passprotected );
	anonydialog->show();
    }
    else{
	cout << "TEST JOIN: "<< Join::compose(iid, false, "null", false, "null") << endl;;
	handler->tellService(Join::compose(iid, false, "null ", false, "null"));	
    }
}


void ortsNetGUI::leaveInvitation()
{
    handler->tellService(Leave::compose(handler->get_iid()));
    currentIID=-1;
}


void ortsNetGUI::kickPlayer()
{
    REM("[kick]");
    PlayerList list = handler->getPlayers();
    pair<string, uint1> current;
    string selected = playerListBox->currentText();
    // get pid and iid 
   int iid  = handler->get_iid();
   int pid = -1; 
    // get pid
    FORALL(list, i) {
	current = (*i).second;
	if( current.first == selected ) {
	    pid = (*i).first;
	    break;
	}
    }
    if(pid < 0){ 
	cout << "name " << selected << " not found " << endl; 
	return; 
    }
   cout << "********************************************|" << Kick::compose(iid, pid) << endl;
   handler->tellService(Kick::compose(iid, pid));
}


void ortsNetGUI::createChatContext(QListBoxItem* m){
 handler->createContext(m->text());   
}

void ortsNetGUI::createContextTab(const string& name)
{
    QWidget* tab = new QWidget( contextTabWidget, "tab" );
    QVBoxLayout* tabLayout_ = new QVBoxLayout( tab, 10, 6, "tabLayout_"); 

    /*layout33 = new QHBoxLayout( 0, 0, 6, "layout33"); 
    QSpacerItem* spacer = new QSpacerItem( 32767, 10, QSizePolicy::Maximum, QSizePolicy::Minimum );
    layout33->addItem( spacer );*/

    /*closeToolButton = new QToolButton( tab, "closeToolButton" );
    closeToolButton->setEnabled( TRUE );
    closeToolButton->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, 0, 0, closeToolButton->sizePolicy().hasHeightForWidth() ) );
    closeToolButton->setPaletteForegroundColor( QColor( 255, 0, 0 ) );
    QFont closeToolButton_font(  closeToolButton->font() );
    closeToolButton_font.setPointSize( 6 );
    closeToolButton->setFont( closeToolButton_font ); 
    closeToolButton->setAutoRaise( TRUE );
    layout33->addWidget( closeToolButton );
    tabLayout_->addLayout( layout33 );*/
    
    mainhistoryTextEdit = new QTextEdit( tab, "mainhistoryTextEdit" );
    // add this text edit to the map
    contextList[name] = mainhistoryTextEdit;
    
    //closeToolButton->setText("X");
    QFont mainhistoryTextEdit_font(  mainhistoryTextEdit->font() );
    //QToolTip::add( closeToolButton, tr( "Close this tab" ) );
    mainhistoryTextEdit_font.setFamily( "Miscfixed" );
    mainhistoryTextEdit_font.setPointSize( 9 );
    mainhistoryTextEdit->setFont( mainhistoryTextEdit_font ); 
    mainhistoryTextEdit->setMouseTracking( TRUE );
    mainhistoryTextEdit->setWordWrap( QTextEdit::WidgetWidth );
    mainhistoryTextEdit->setReadOnly( TRUE );
    tabLayout_->addWidget( mainhistoryTextEdit );
    
    //connect( closeToolButton, SIGNAL( clicked() ), this, SLOT( closeTab() ) );
   
    contextTabWidget->insertTab(tab, name);
    int index = contextTabWidget->count() - 1;
    contextTabWidget->setCurrentPage(index);
}


void ortsNetGUI::closeTab(){
    handler->deleteContext(contextTabWidget->tabLabel((contextTabWidget->currentPage())));
    contextTabWidget->removePage(contextTabWidget->currentPage());
}

void ortsNetGUI::ready(bool ready)
{
    if(ready){
	handler->tellService(Ready::compose(handler->get_iid()));
    }else{
	handler->tellService(Unready::compose(handler->get_iid()));
    }
}

void ortsNetGUI::showAddBuddyForm()
{ 
    if(addBuddyForm == 0){
	REM("-------- CREATE NEW LOGINDIALOG ------- ");
	addBuddyForm = new AddBuddyForm(this);
    }
    addBuddyForm ->setup(handler);
    addBuddyForm ->show();
}


void ortsNetGUI::removeBuddy()
{    
    if( buddyListBox->currentItem()<0){ 
	REM("Select a buddy from the buddy list");
	error->showError("Select a buddy from the buddy list");
	return;
    }   
    string temp = buddyListBox->currentText();	
    if(temp == "[ All Buddies ]"){ 
	return; 
    }   
    cout << "BUDDY::::::: " << temp << endl;
    stringstream ss(temp);
    string name;
    ss >> name;    
    if(name == "-"){
	ss >> name;
    }
     handler->tellService(Buddy::compose(false, name));
}

void ortsNetGUI::tellGame()
{
    string message = gameChatLineEdit->text();
    gameChatLineEdit->clear();
    handler->tellService(Tell::compose(handler->get_iid(),handler->get_pid(), message));
    TellEvent* event = new TellEvent(":", "game", "<font color=\"#FF0000\">" + message);
    QApplication::postEvent( this, event );
}

/*****************************************************************************************
  * This function saves the current user settings 
  ****************************************************************************************/
void ortsNetGUI::saveSettings()
{
    REM("[ -- saving settings -- ]");
    QSettings settings;
    settings.insertSearchPath( QSettings::Windows, WINDOWS_REGISTRY );
    settings.writeEntry( APP_KEY + "WindowWidth", width() );
    settings.writeEntry( APP_KEY + "WindowHeight", height() );
    settings.writeEntry( APP_KEY + "WindowX", x() );
    settings.writeEntry( APP_KEY + "WindowY", y() );
}

/*****************************************************************************************
  * This function sets up ortsNetGUI to how it looked 
  * the previous time it was running
  ****************************************************************************************/
void ortsNetGUI::loadSettings()
{
    REM("[ -- loading settings -- ]");
    QSettings settings;
    settings.insertSearchPath( QSettings::Windows, WINDOWS_REGISTRY );
    int windowWidth = settings.readNumEntry( APP_KEY + "WindowWidth", 0 );
    int windowHeight = settings.readNumEntry( APP_KEY + "WindowHeight", 0 );
    int windowX = settings.readNumEntry( APP_KEY + "WindowX", 350 );
    int windowY = settings.readNumEntry( APP_KEY + "WindowY", 300 );
    
    resize( windowWidth, windowHeight );
    move( windowX, windowY );
}

/*****************************************************************************************
  * Overloaded function which is called when ortsNetGUI is closed
  ****************************************************************************************/
void ortsNetGUI::closeEvent( QCloseEvent *e )
{
    REM("[--- closeEVENT--- ]");
    saveSettings();
    e->accept();
}

void ortsNetGUI::userListContextMenu(QListBoxItem *item, const QPoint &p)
{
    if (item!=0)
    {
	QPopupMenu contextMenu(this);
	contextMenu.insertItem ("Add to Buddy List",this, SLOT(ContextAddBuddy()));
	contextMenu.exec(p);
    }
}

void ortsNetGUI::ContextAddBuddy()
{
    QListBoxItem *current = userListBox->selectedItem();
    cout<<"current item is "<<current->text()<<endl;
    if (current!=0)
    {
	string name = current->text();
	handler->tellService(Buddy::compose(true, name));
    }
}


void ortsNetGUI::chat_tab_changed()
{
    if (contextTabWidget->currentPageIndex()==0)
    {
	contextTabWidget->cornerWidget()->setDisabled(true);
    }
    else
    {
	contextTabWidget->cornerWidget()->setDisabled(false);
    }
}


void ortsNetGUI::buddyListContextMenu( QListBoxItem *item, const QPoint &p )
{
    if (item!=0)
    {
	QPopupMenu contextMenu(this);
	contextMenu.insertItem ("Remove from Buddy List",this, SLOT(removeBuddy()));
	contextMenu.exec(p);
    }
}

// this function calls system(ping host) and parses the output for an avg value
char * ping_server(char *server_address) {
    
    char cmd[100];
    sprintf(cmd, "ping -c 1 %s | grep avg > ping.output", server_address);
    
    system(cmd);
    
    FILE *pFile;
    //char myString[100];
    
    pFile = fopen("ping.output", "r");
    
    fgets(buffer, 100, pFile);
    
    char *pch = strtok(buffer, "/");
    pch = strtok (NULL, "/");
    pch = strtok (NULL, "/");
    pch = strtok (NULL, "/");
    pch = strtok (NULL, "/"); // grab 5th token, that's the avg we want
    
    fclose(pFile);
    
    return pch;
    
}
