// $Id: Decomp.C,v 1.11 2005/09/06 05:50:48 orts_furtak Exp $
 
// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Decomp.H"
#include "ShutDown.H"
#include "Options.H"
#include "SDLinit.H"
#include "GUI.H"

#define MY_DEBUG 0

// test routine constants
const sint4 GFX_MAG = 4;        // gfx magnification
const sint4 WX = 800, WY = 600; // window dimensions
const sint4 W=200, H=150;       // rectangle dimensions
const sint4 SS=W;               // sector size
const sint4 S=6;                // object size

REGISTER_TYPEOF(40, Vector<Decomp::Edge>);
REGISTER_TYPEOF(41, Vector<Decomp::Rect>);
REGISTER_TYPEOF(42, Vector<Decomp>);

#if !MY_DEBUG
#define TEST_OUT(x)
#else
#define TEST_OUT(x) cout << x << " " << p->id << ":" << p->sub_id << " ";  p->r.write(); cout << " r="; r.write(); cout << endl;
#endif

#if NODE_IDS
uint4 Decomp::Node::id0 = 1;
uint4 Decomp::Node::sub_id0 = 1;
#endif

#if NODE_IDS
#define NODE_INC_ID { \
  Decomp::Node::id0++; Decomp::Node::sub_id0 = 1; \
}
#else
#define NODE_INC_ID
#endif

// abstraction graph

class SearchNode {

public:

  struct Edge {
    SearchNode *succ;
    sint4 cost;
  };
  
  Vector<Edge> edges;

  uint4 visited;
  uint4 done;
  sint4 dist;
  SearchNode *parent;

  SearchNode() { visited = done = 0; }
};


// return true iff path found
// fixme: NOT TESTED!

bool dijkstra(SearchNode *start, SearchNode *goal)
{
  priority_queue<SearchNode*> heap;
  static uint4 magic = 1;

  ++magic; if (magic >= 1000000000) ERR("magic wrap around");

  start->visited = magic;
  start->dist = 0;
  start->parent = 0; // root
  heap.push(start);

  while (!heap.empty()) {

    SearchNode *u = heap.top();
    heap.pop();

    if (u == goal) return true;
    
    if (u->done == magic) continue;
    u->done = magic;

    FORALL (u->edges, i) {

      SearchNode::Edge &e = *i;
      assert(u->dist >= 0);
      
      if (e.succ->visited != magic) {

	e.succ->visited = magic;
	e.succ->parent = u;
	e.succ->dist = u->dist + e.cost;
	heap.push(e.succ);

      } else {

	sint4 new_dist = u->dist + e.cost;

	if (new_dist < e.succ->dist) {
	  e.succ->dist = new_dist;
	  e.succ->parent = u;
	  heap.push(e.succ);
	}
      }
    }
  }

  return false;
}



#if 0
Decomp::Node::Node()
{
  type = NONE; data = 0; magic = 0; v = 0; lt = ge = 0;
}

Decomp::Node::Node(GridCoor x0, GridCoor x1, GridCoor y0, GridCoor y1, Data d)
{
  assert(x0 < x1 && y0 < y1);

  type = R_NODE;
  r.x0 = x0; r.x1 = x1; r.y0 = y0; r.y1 = y1;
  data = d;

#if NODE_IDS
  id = id0;
  sub_id = sub_id0;
  ++sub_id0;
#endif

#if MY_DEBUG
  cout << "created "; write();
#endif
  
}

Decomp::Node::Node(Type type_, Data v_, Node *lt_, Node *ge_)
{
  type = type_;
  v = v_;
  lt = lt_;
  ge = ge_;
}    

#endif


Decomp::Node *Decomp::new_Node()
{
  Node *n = new Node;
  n->number = counter;
  counter++;
  nodes.push_back(n);
  return n;
}
  
Decomp::Node *Decomp::new_Node(GridCoor x0, GridCoor x1, GridCoor y0, GridCoor y1, Data d)
{
  assert(x0 < x1 && y0 < y1);

  Node *n = new_Node();
  
  n->type = Node::R_NODE;
  n->r.x0 = x0; n->r.x1 = x1; n->r.y0 = y0; n->r.y1 = y1;
  n->data = d;

#if NODE_IDS
  n->id = Node::id0;
  n->sub_id = Node::sub_id0;
  ++Node::sub_id0;
#endif

#if MY_DEBUG
  cout << "created "; n->write();
#endif
  
  return n;
}


Decomp::Node *Decomp::new_Node(Node::Type type_, Data v_, Node *lt_, Node *ge_)
{
  Node *n = new_Node();
  n->type = type_;
  n->v = v_;
  n->lt = lt_;
  n->ge = ge_;

  return n;
}    


Decomp::Node *Decomp::rect0()
{
  if (!root || root->type != Node::R_NODE) ERR("not a single rectangle");
  return root;
}


void Decomp::Node::write(std::ostream &os) const
{
  switch (type) {
  case X_NODE: os << "X_NODE"; break;
  case Y_NODE: os << "Y_NODE"; break;
  case R_NODE: os << "R_NODE"; break;
  default: os << "UKNOWN NODE TYPE"; break;
  }
  
#if NODE_IDS
  os << " " << id << ":" << sub_id;
#endif

  if (type == Node::R_NODE) {
    os << " [" <<  r.x0 << "," << r.x1 << "]x[" << r.y0 << "," << r.y1 << "] ";
#if NODE_IDS
    os << "N=";
    { FORALL (nl[N], i) os << (*i)->id << ":" << (*i)->sub_id << " "; }
    os << "E=";
    { FORALL (nl[E], i) os << (*i)->id << ":" << (*i)->sub_id << " "; }
    os << "S=";
    { FORALL (nl[S], i) os << (*i)->id << ":" << (*i)->sub_id << " "; }
    os << "W=";
    { FORALL (nl[W], i) os << (*i)->id << ":" << (*i)->sub_id << " "; }
#endif    
    os << endl;
    
  } else {
    os << "v: " << v << " " << lt << " " << ge << endl;
  }
}


bool Decomp::Node::check(const string &msg) const
{
  switch (type) {

  case X_NODE:
  case Y_NODE:
    if (lt == 0 || ge == 0) {
      write();
      ERR("pointer 0");
    }
    break;

  case R_NODE:
    {
      if (r.x0 >= r.x1 || r.y0 >= r.y1) {
	write();
	ERR(msg + ": coordinates not ordered");
      }

      // check neighbor lists

      // pointing to R_NODEs
    
      FORS (i, 4) {
	FORALL (nl[i], j) {
	  if ((*j)->type != R_NODE) {
	    write();
	    ERR2(msg + ": neighbor not R_NODE, list:", i);
	  }
	}
      }

      // edges aligned and ordered?

      if (!nl[N].empty()) {
	RList::const_iterator first = nl[N].begin();
	GridCoor prev = (*first)->r.x1;
	FORALL (nl[N], j) {
	  Rect &rj = (*j)->r;
	  if (rj.y0 != r.y1) { write(); ERR("north edge not aligned 1"); }
	  if (rj.x1 <= r.x0 || rj.x0 >= r.x1) { write(); ERR("north edge not aligned 2"); }
	  if (j != first) {
	    if (rj.x0 != prev) { write(); ERR("north edge not aligned 3"); }
	    prev = rj.x1;
	  }
	}	
      }
      if (!nl[E].empty()) {
	RList::const_iterator first = nl[E].begin();
	GridCoor prev = (*first)->r.y0;
	FORALL (nl[E], j) {
	  Rect &rj = (*j)->r;
	  if (rj.x0 != r.x1) { write(); ERR("east edge not aligned 1"); }
	  if (rj.y1 <= r.y0 || rj.y0 >= r.y1) { write(); ERR("east edge not aligned 2"); }
	  if (j != first) {
	    if (rj.y1 != prev) { write(); ERR("east edge not aligned 3"); }
	    prev = rj.y0;
	  }
	}	
      }
      if (!nl[S].empty()) {
	RList::const_iterator first = nl[S].begin();
	GridCoor prev = (*first)->r.x0;
	FORALL (nl[S], j) {
	  Rect &rj = (*j)->r;
	  if (rj.y1 != r.y0) { write(); (*j)->write(); ERR("south edge not aligned 1"); }
	  if (rj.x1 <= r.x0 || rj.x0 >= r.x1) { write(); ERR("south edge not aligned 2"); }
	  if (j != first) {
	    if (rj.x1 != prev) { write(); ERR("south edge not aligned 3"); }
	    prev = rj.x0;
	  }
	}	
      }
      if (!nl[W].empty()) {
	RList::const_iterator first = nl[W].begin();
	GridCoor prev = (*first)->r.y1;
	FORALL (nl[W], j) {
	  Rect &rj = (*j)->r;
	  if (rj.x1 != r.x0) { write(); ERR("west edge not aligned 1"); }
	  if (rj.y1 <= r.y0 || rj.y0 >= r.y1) { write(); ERR("west edge not aligned 2"); }
	  if (j != first) {
	    if (rj.y0 != prev) { write(); ERR("west edge not aligned 3"); }
	    prev = rj.y1;
	  }
	}	
      }

      break;
    }
    
  default: write(); ERR(msg + ": unknown type");
  }

  return true;
}

void Decomp::init()
{
  root = 0; r0.x0 = r0.x1 = r0.y0 = r0.y1 = 0;
#if NODE_IDS
  Node::id0 = 1; Node::sub_id0 = 1;
#endif  
}

Decomp::Decomp():counter(0) { init(); }

void Decomp::free()
{
  FORALL (nodes, i) delete *i;
  root = 0;
}


void Decomp::copy(const Decomp& x)
{
  if (x.root) ERR("non-trivial copy not yet implemented");

  init();
}


Decomp::Node *Decomp::locate(GridCoor x, GridCoor y)
{
  if (!root) return 0;

  Node *n = root;

  while (n->type != Node::R_NODE) {

    assert(n->type == Node::X_NODE || n->type == Node::Y_NODE);
    
    if (n->type == Node::X_NODE) {
      if (x < n->v) n = n->lt; else n = n->ge;
    } else {
      if (y < n->v) n = n->lt; else n = n->ge;
    }
  }

  return n;
}


void Decomp::insert(Node *p, const Rect &r, Data mask, Data data, set<Node*> &visited)
{
  if (visited.find(p) != visited.end()) return;
  visited.insert(p);
  
#if 0
    cout << "insert: p:"; p->write(); cout << " r:"; r.write(); cout<<endl;
  p->check("C");
  //check("insert");
#endif
  
  assert(p->type == Node::R_NODE);

  if (r.x0 >= p->r.x1 || r.x1 <= p->r.x0 || r.y0 >= p->r.y1 || r.y1 <= p->r.y0)
    ERR("no intersection!");
  
  assert((~mask & data) == 0); // no bits set outside mask

  // determine all successors at each side by intersecting all edge intervals with r

  bool cl[4];  // r covers left endpoint
  bool cr[4];  // r covers right endpoint
  bool ca[4];  // r covers any point

  // determine relevant successors and coverage for each edge
  
  Node::RList succ;

  north_edge(p, r, succ, cl[0], cr[0], ca[0]);
  east_edge (p, r, succ, cl[1], cr[1], ca[1]);
  south_edge(p, r, succ, cl[2], cr[2], ca[2]);
  west_edge (p, r, succ, cl[3], cr[3], ca[3]);  

#if 0  
  FORS (i, 4) cout << ca[i]; cout << endl;
  FORS (i, 4) cout << cl[i]; cout << endl; 
  FORS (i, 4) cout << cr[i]; cout << endl;  
#endif
  
  assert(p->type == Node::R_NODE);

  if ((p->data & mask) != data) {

    // different data -> split rectangle

    // cases

    if (!ca[0] && !ca[1] && !ca[2] && !ca[3]) fully_embedded(p, r, mask, data);
      
    else if (cl[0] && cl[1] && cl[2] && cl[3]) { 
    
      p->data = (p->data & ~mask) | data; // p->r covered by r: no split
      
    }

    else if (cl[0] && !ca[1] && !ca[2]) NW_corner(p, r, mask, data);
    else if (cl[1] && !ca[2] && !ca[3]) NE_corner(p, r, mask, data);
    else if (cl[2] && !ca[3] && !ca[0]) SE_corner(p, r, mask, data);
    else if (cl[3] && !ca[0] && !ca[1]) SW_corner(p, r, mask, data);
    
    else if (cl[0] && cr[0] && !ca[2]) full_N_edge(p, r, mask, data);
    else if (cl[1] && cr[1] && !ca[3]) full_E_edge(p, r, mask, data);
    else if (cl[2] && cr[2] && !ca[0]) full_S_edge(p, r, mask, data);
    else if (cl[3] && cr[3] && !ca[1]) full_W_edge(p, r, mask, data);

    else if (!ca[0] && !ca[2] && ca[1] && ca[3]) WE_embedded_strip(p, r, mask, data);
    else if (!ca[1] && !ca[3] && ca[0] && ca[2]) NS_embedded_strip(p, r, mask, data);

    else if (ca[0] && !ca[1] && !ca[2] && !ca[3]) N_middle_section(p, r, mask, data);
    else if (ca[1] && !ca[2] && !ca[3] && !ca[0]) E_middle_section(p, r, mask, data);
    else if (ca[2] && !ca[3] && !ca[0] && !ca[1]) S_middle_section(p, r, mask, data);
    else if (ca[3] && !ca[0] && !ca[1] && !ca[2]) W_middle_section(p, r, mask, data);

    else ERR("unhandled case");
  }

  // visit all relevant neighbors

  {
    FORALL (succ, i) insert(*i, r, mask, data, visited);
  }
}

void Decomp::insert(const Rect &r, uint4 mask, uint4 data)
{
  if (!root) {
    root = new_Node(r.x0, r.x1, r.y0, r.y1, data);
    r0 = r;
    return;
  }

#if 0
  // broken in SectorDecomp mode
  if (r.x0 < r0.x0 || r.x1 > r0.x1 || r.y0 < r0.y0 || r.y1 > r0.y1) {
    r.write(); r0.write();
    ERR("outside");
  }
#endif
  
  Node *l = locate(r.x0, r.y0);

#if 0  
  cout << "location of " << r.x0 << " " << r.y0 << ": ";
  l->write();
#endif

  set<Node*> visited;
  insert(l, r, mask, data, visited);
}

void Decomp::north_edge(Node *p, const Rect &r,
			Node::RList &succ, bool &cl, bool &cr, bool &ca)
{
  assert(p->type == Node::R_NODE);

#if MY_DEBUG  
  p->check("A");
#endif
  
  //  r.y1--------------
  //    
  //              ^
  // y=p->y1 --o--L--o------  part of L in interior of r?
  //           |     |
  //       i->r.x0  i->r.x1   
  //
  //  r.y0--------------

  uint4 y = p->r.y1;
  
  if (y < r.y1 && y > r.y0) { // r.y properly covers y
  
    FORALL (p->nl[Node::N], i) {
      
      assert((*i)->type == Node::R_NODE);
      
      // r crosses L
      if (!((*i)->r.x0 >= r.x1 || (*i)->r.x1 <= r.x0))  // x non-empty intersection
	succ.push_back(*i);
    }
  }

  //           _r.y1 
  //  p.x0     |     p.x1
  //   o-------------o y
  //   |   r   | 
  //   |-------|--r.y0
  //   r.x0   r.x1

  // r reaches boundary
  //       x non-empty intersection          &&  r.y at least reaches boundary
  ca = !(r.x0 >= p->r.x1 || r.x1 <= p->r.x0) && y <= r.y1 && y > r.y0;
  if (ca) {

    //  cl            cr
    //  p.x0         p.x1
    //   o-------------o y
    // | r   |       | r  | 
    // |-----|       |----|
    // r.x0  r.x1    r.x0 r.x1

    cl = p->r.x0 >= r.x0 && p->r.x0 <  r.x1; // left corner properly within r
    cr = p->r.x1 > r.x0  && p->r.x1 <= r.x1; // right corner properly within r   
  } else {
    cl = cr = false;
  }
}    

void Decomp::south_edge(Node *p, const Rect &r,
			Node::RList &succ, bool &cl, bool &cr, bool &ca)
{
  //
  //    r.y1  --------------
  //      
  //               
  // y= p->y0 --o--L--o------   part of interval in interior of r
  //            |  v  |
  //       i->r.x0  i->r.x1   
  //
  //    r.y0--------------
	 
  uint4 y = p->r.y0;

  if (y < r.y1 && y > r.y0) { // r.y properly covers y

    FORALL (p->nl[Node::S], i) {
      
      assert((*i)->type == Node::R_NODE);
      
      // r crosses L
      if (!((*i)->r.x0 >= r.x1 || (*i)->r.x1 <= r.x0))  // x non-empty intersection
      succ.push_back(*i);
    }
  }

  //   r.x0   r.x1
  //   |-------|--r.y1
  //   |   r   |   
  //   o-------------o y
  //  p.x0     |     p.x1
  //          --r.y0   

  // r reaches boundary
  //       x non-empty intersection          &&  r.y at least reaches boundary
  ca = !(r.x0 >= p->r.x1 || r.x1 <= p->r.x0) && y < r.y1 && y >= r.y0;
  if (ca) {

    // r.x0  r.x1    r.x0 r.x1
    // |-----|       |----|
    // | r   |       | r  | 
    //   o-------------o y
    //  p.x0         p.x1
    //  cr            cl

    // cr <-> cl (clockwise!)
    cr = p->r.x0 >= r.x0 && p->r.x0 <  r.x1; // left corner properly within r
    cl = p->r.x1 > r.x0  && p->r.x1 <= r.x1; // right corner properly within r   
  } else {
    cl = cr = false;
  }
}


void Decomp::east_edge(Node *p, const Rect &r,
		       Node::RList &succ, bool &cl, bool &cr, bool &ca)
{
  //  r.x0        p->x1  r.x1
  //             = x 
  //   |           |      |
  //               |      |- r.y1
  //    i->r.y1 -  o      |
  //               |      |
  //               ->     |
  //               |      |
  //    i->r.y0 -  o      |
  //               |      |- r.y0
  //               |      |

  uint4 x = p->r.x1;

  if (x < r.x1 && x > r.x0) { // r.x properly covers x

    FORALL (p->nl[Node::E], i) {
      
      assert((*i)->type == Node::R_NODE);

      // r crosses L
      if (!((*i)->r.y0 >= r.y1 || (*i)->r.y1 <= r.y0))  // y non-empty intersection
	succ.push_back(*i);
    }
  }
  
  //                r.x1
  //   r.y1 -----|--|
  //        |    o-p.y1  cl
  //        | -> |  |
  //        |    |  |
  //   r.y0 -----|--|
  //        |    o-p.y0  cr
  //       r.x0  |
  //             x
  
  // r reaches boundary
  //       y non-empty intersection          &&  r.x at least reaches boundary 
  ca = !(r.y0 >= p->r.y1 || r.y1 <= p->r.y0) && x <= r.x1 && x > r.x0;
  if (ca) {

    cl = p->r.y1 > r.y0  && p->r.y1 <= r.y1; // left corner properly within r
    cr = p->r.y0 >= r.y0 && p->r.y0 <  r.y1; // right corner properly within r   

  } else {
    cl = cr = false;
  }
}    

void Decomp::west_edge(Node *p, const Rect &r,
		       Node::RList &succ, bool &cl, bool &cr, bool &ca)
{
  // r.x0        p->x0  r.x1
  //            =x
  // |           |      |
  //             |      |
  //  i->r.y1 -  o      |
  //             |      |
  //            <-      |
  //             |      |
  //  i->r.y0 -  o      |
  //             |      |
  //             |      |

  uint4 x = p->r.x0;

  if (x < r.x1 && x > r.x0) {                     // r.x properly covers x

    FORALL (p->nl[Node::W], i) {

      assert((*i)->type == Node::R_NODE);
    
      if (!((*i)->r.y0 >= r.y1 || (*i)->r.y1 <= r.y0))  // y non-empty intersection
	succ.push_back(*i);
    }
  }

  //  r.x0     r.x1
  //    ---------- r.y1
  // cr | |o-p.y1 |  
  //    | |       |
  //    | | <-    |
  //    | |       |
  //    --|-------- r.y0
  //      |     
  //   cl o-p.y0
  //      x

  // r reaches boundary
  //       y non-empty intersection          &&  r.x at least reaches boundary 
  ca = !(r.y0 >= p->r.y1 || r.y1 <= p->r.y0) && x < r.x1 && x >= r.x0;
  if (ca) {

    cl = p->r.y0 >= r.y0 && p->r.y0 <  r.y1; // left corner properly within r   
    cr = p->r.y1 > r.y0  && p->r.y1 <= r.y1; // right corner properly within r

  } else {
    cl = cr = false;
  }
}    

void Decomp::delete_edges_to(Node *p)
{
  {
    FORALL (p->nl[Node::N], i) {
      assert(*i != p);
      Node::RList &rl = (*i)->nl[Node::S];
#ifndef NDEBUG
      uint4 os = rl.size();
#endif      
      FORALL (rl, j) if (*j == p) { rl.erase(j); break; }
      assert(rl.size() == os-1);
    }
  }
  {
    FORALL (p->nl[Node::E], i) {
      assert(*i != p);
      Node::RList &rl = (*i)->nl[Node::W];
#ifndef NDEBUG
      uint4 os = rl.size();
#endif      
      FORALL (rl, j) if (*j == p) { rl.erase(j); break; }
      assert(rl.size() == os-1);      
    }
  }
  {
    FORALL (p->nl[Node::S], i) {
      assert(*i != p);
      Node::RList &rl = (*i)->nl[Node::N];
#ifndef NDEBUG
      uint4 os = rl.size();
#endif      
      FORALL (rl, j) if (*j == p) { rl.erase(j); break; }
      assert(rl.size() == os-1);      
    }
  }
  {
    FORALL (p->nl[Node::W], i) {
      assert(*i != p);
      Node::RList &rl = (*i)->nl[Node::E];
#ifndef NDEBUG
      uint4 os = rl.size();
#endif      
      FORALL (rl, j) if (*j == p) { rl.erase(j); break; }
      assert(rl.size() == os-1);      
    }
  }
}


// important: for a given direction X, connect_X has to be called in
// counter clockwise order to append pointers in the right order

//            _____________
// example:   | A | B | C |
//
// ->  connect_north(p, C); connect_north(p, B); connect_north(p, A); 


// insert edges to all north neighbors of p that touch new rectangle (and vice versa)

void Decomp::connect_north(Node *p, Node *rnew)
{
  assert(p != rnew);
  assert(rnew->nl[Node::N].empty());
  
  FORALL (p->nl[Node::N], i) {
    assert((*i)->type == Node::R_NODE);
    if ((*i)->r.x1 <= rnew->r.x0) continue;
    if ((*i)->r.x0 >= rnew->r.x1) break;
    rnew->nl[Node::N].push_back(*i);
    Node::RList &rl = (*i)->nl[Node::S];
    Node::RList::iterator ppos = find(rl.begin(), rl.end(), p);
    assert(ppos != rl.end());
    rl.insert(ppos, rnew); // put edge to rnew right before p in list (p gets deleted later)
  }
}

// insert edges to all south neighbors of p that touch new rectangle (and vice versa)

/*
  ---------
  |       |p
  |       |
  |  new  |
  | |  |  |
  | |  |  | <- neighbor direction
--------------
*i |  |  |  | -> neighbor direction
--------------  
*/

void Decomp::connect_south(Node *p, Node *rnew)
{
  assert(p != rnew);
  assert(rnew->nl[Node::S].empty());
  
  FORALL (p->nl[Node::S], i) {
    assert((*i)->type == Node::R_NODE);    
    if ((*i)->r.x0 >= rnew->r.x1) continue;
    if ((*i)->r.x1 <= rnew->r.x0) break;
    rnew->nl[Node::S].push_back(*i);
    Node::RList &rl = (*i)->nl[Node::N];
    Node::RList::iterator ppos = find(rl.begin(), rl.end(), p);
    assert(ppos != rl.end());
    rl.insert(ppos, rnew);
  }
}

// insert edges to all east neighbors of p that touch new rectangle (and vice versa)

void Decomp::connect_east(Node *p, Node *rnew)
{
  assert(p != rnew);
  assert(rnew->nl[Node::E].empty());
  
  FORALL (p->nl[Node::E], i) {
    assert((*i)->type == Node::R_NODE);    
    if ((*i)->r.y0 >= rnew->r.y1) continue;
    if ((*i)->r.y1 <= rnew->r.y0) break;
    rnew->nl[Node::E].push_back(*i);
    Node::RList &rl = (*i)->nl[Node::W];
    Node::RList::iterator ppos = find(rl.begin(), rl.end(), p);
    assert(ppos != rl.end());
    rl.insert(ppos, rnew);
  }
}

// insert edges to all east neighbors of p that touch new rectangle (and vice versa)

void Decomp::connect_west(Node *p, Node *rnew)
{
  assert(p != rnew);
  assert(rnew->nl[Node::W].empty());
  
  FORALL (p->nl[Node::W], i) {
    assert((*i)->type == Node::R_NODE);    
    if ((*i)->r.y1 <= rnew->r.y0) continue;
    if ((*i)->r.y0 >= rnew->r.y1) break;
    rnew->nl[Node::W].push_back(*i);
    Node::RList &rl = (*i)->nl[Node::E];
    Node::RList::iterator ppos = find(rl.begin(), rl.end(), p);
    assert(ppos != rl.end());
    rl.insert(ppos, rnew);
  }
}


/*
           _ p->y1
|   |B|   |
|   |_|   |
| A |E| D |
|   |_|   |
|   | |   |
|   |C|   |_ p->y0


         x1 (r.x0)
        / \
       A   x2 (r.x1)
          / \
  (r.y1) y1  D
        / \
(r.y0) y2  B      
      / \
     C   E
    
*/


void Decomp::fully_embedded(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("fully_embedded");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);
  
  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0,    p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(r.x0,    r.x1,    r.y1,    p->r.y1, p->data);
  Node *C = new_Node(r.x0,    r.x1,    p->r.y0, r.y0,    p->data);
  Node *D = new_Node(r.x1,    p->r.x1, p->r.y0, p->r.y1, p->data);    
  Node *E = new_Node(r.x0,    r.x1,    r.y0,    r.y1,    (p->data & ~mask) | data);  

  // new decision nodes
  
  Node *yn2 = new_Node(Node::Y_NODE, r.y0, C, E);
  Node *yn1 = new_Node(Node::Y_NODE, r.y1, yn2, B);
  Node *xn2 = new_Node(Node::X_NODE, r.x1, yn1, D);
  
  // internal rectangle connections
  
  A->nl[Node::E].push_back(B);
  A->nl[Node::E].push_back(E);
  A->nl[Node::E].push_back(C);

  B->nl[Node::W].push_back(A);
  B->nl[Node::E].push_back(D);
  B->nl[Node::S].push_back(E);    
  
  C->nl[Node::W].push_back(A);
  C->nl[Node::E].push_back(D);
  C->nl[Node::N].push_back(E);
  
  D->nl[Node::W].push_back(C);
  D->nl[Node::W].push_back(E);
  D->nl[Node::W].push_back(B);

  E->nl[Node::N].push_back(B);  
  E->nl[Node::E].push_back(D);
  E->nl[Node::S].push_back(C);
  E->nl[Node::W].push_back(A);

  // connect new rectangles

  connect_north(p, D);  // counter clockwise!
  connect_north(p, B);
  connect_north(p, A);    

  connect_east(p, D);  

  connect_south(p, A); // counter clockwise!
  connect_south(p, C);  
  connect_south(p, D);

  connect_west(p, A);

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x0;
  p->lt = A;
  p->ge = xn2;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|C |      |
|__|      |
|  |      |
|B |   A  |
|  |      |
|  |      |_ p->y0
   r.x1

         x1 (r.x1)
        / \
(r.y0) y1  A
      / \
     B   C

*/
	   
void Decomp::NW_corner(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("NW_corner");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(r.x1,    p->r.x1, p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(p->r.x0, r.x1,    p->r.y0, r.y0,    p->data);
  Node *C = new_Node(p->r.x0, r.x1,    r.y0,    p->r.y1, (p->data & ~mask) | data);

  // new decision nodes
  
  Node *yn1 = new_Node(Node::Y_NODE, r.y0, B, C);
  
  // internal rectangle connections
  
  A->nl[Node::W].push_back(B);
  A->nl[Node::W].push_back(C);

  B->nl[Node::N].push_back(C);
  B->nl[Node::E].push_back(A);
  
  C->nl[Node::E].push_back(A);
  C->nl[Node::S].push_back(B);

  // connect new rectangles

  connect_north(p, A); // counter clockwise!
  connect_north(p, C);    

  connect_east(p, A);  

  connect_south(p, B);
  connect_south(p, A);  

  connect_west(p, C);
  connect_west(p, B);  

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x1;
  p->lt = yn1;
  p->ge = A;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|     |   |
|     | C |
|  A  |   |
|     |---| r.y0
|     | B |
|     |   |_ p->y0
   r.x0

         x1 (r.x0)
        / \
       A   y1 (r.y0)
          / \
         B   C

*/

void Decomp::NE_corner(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("NE_corner");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0,    p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(r.x0,    p->r.x1, p->r.y0, r.y0,    p->data);
  Node *C = new_Node(r.x0,    p->r.x1, r.y0,    p->r.y1, (p->data & ~mask) | data);

  // new decision nodes
  
  Node *yn1 = new_Node(Node::Y_NODE, r.y0, B, C);
  
  // internal rectangle connections
  
  A->nl[Node::E].push_back(C);
  A->nl[Node::E].push_back(B);

  B->nl[Node::N].push_back(C);
  B->nl[Node::W].push_back(A);
  
  C->nl[Node::W].push_back(A);
  C->nl[Node::S].push_back(B);

  // connect new rectangles

  connect_north(p, C);
  connect_north(p, A);    

  connect_east(p, B);
  connect_east(p, C);  

  connect_south(p, A);
  connect_south(p, B);  

  connect_west(p, A);  

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v  = r.x0;
  p->lt = A;
  p->ge = yn1;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|     |   |
|     | B |
|  A  |   |
|     |---| r.y1
|     | C |
|     |   |_ p->y0
   r.x0

         x1 (r.x0)
        / \
       A   y1 (r.y1)
          / \
         C   B

*/

void Decomp::SE_corner(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("SE_corner");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0,    p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(r.x0,    p->r.x1, r.y1,    p->r.y1, p->data);
  Node *C = new_Node(r.x0,    p->r.x1, p->r.y0, r.y1,    (p->data & ~mask) | data);

  // new decision nodes
  
  Node *yn1 = new_Node(Node::Y_NODE, r.y1, C, B);
  
  // internal rectangle connections
  
  A->nl[Node::E].push_back(B);
  A->nl[Node::E].push_back(C);

  B->nl[Node::W].push_back(A);
  B->nl[Node::S].push_back(C);

  C->nl[Node::N].push_back(B);
  C->nl[Node::W].push_back(A);
  
  // connect new rectangles

  connect_north(p, B);
  connect_north(p, A);    

  connect_east(p, C);
  connect_east(p, B);  

  connect_south(p, A);
  connect_south(p, C);  

  connect_west(p, A);  

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v  = r.x0;
  p->lt = A;
  p->ge = yn1;

  // remove all previous connections from and to outside

  delete_edges_to(p);  
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|B |      |
|__|r.y1  |
|  |      |
|C |   A  |
|  |      |
|  |      |_ p->y0
   r.x1

         x1 (r.x1)
        / \
(r.y1) y1  A
      / \
     C   B

*/

void Decomp::SW_corner(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("SW_corner");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(r.x1,    p->r.x1, p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(p->r.x0, r.x1,    r.y1,    p->r.y1, p->data);
  Node *C = new_Node(p->r.x0, r.x1,    p->r.y0, r.y1,    (p->data & ~mask) | data);

  // new decision nodes
  
  Node *yn1 = new_Node(Node::Y_NODE, r.y1, C, B);
  
  // internal rectangle connections
  
  A->nl[Node::W].push_back(C);
  A->nl[Node::W].push_back(B);

  B->nl[Node::E].push_back(A);
  B->nl[Node::S].push_back(C);

  C->nl[Node::N].push_back(B);
  C->nl[Node::E].push_back(A);
  
  // connect new rectangles

  connect_north(p, A);
  connect_north(p, B);    

  connect_east(p, A);  

  connect_south(p, C);
  connect_south(p, A);  

  connect_west(p, B);
  connect_west(p, C);  

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x1;
  p->lt = yn1;
  p->ge = A;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|    B    |
|_________| r.y0
|         |
|         |
|    A    |
|         |_ p->y0
       

         y1 (r.y0)
        / \
       A   B

*/

void Decomp::full_N_edge(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("full_N_edge");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, p->r.x1, p->r.y0, r.y0,    p->data);
  Node *B = new_Node(p->r.x0, p->r.x1, r.y0,    p->r.y1, (p->data & ~mask) | data);

  // internal rectangle connections
  
  A->nl[Node::N].push_back(B);

  B->nl[Node::S].push_back(A);

  // connect new rectangles

  connect_north(p, B);

  connect_east(p, A);
  connect_east(p, B);    

  connect_south(p, A);

  connect_west(p, B);
  connect_west(p, A);  

  // transform root node into y1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::Y_NODE;
  p->v = r.y0;
  p->lt = A;
  p->ge = B;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|    |    |
|    |    |
| A  | B  |
|    |    |
|    |    |
|    |    |_ p->y0
    r.x0

         x1 (r.x0)
        / \
       A   B

*/

void Decomp::full_E_edge(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("full_E_edge");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0,    p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(r.x0,    p->r.x1, p->r.y0, p->r.y1, (p->data & ~mask) | data);

  // internal rectangle connections
  
  A->nl[Node::E].push_back(B);

  B->nl[Node::W].push_back(A);

  // connect new rectangles

  connect_north(p, B);
  connect_north(p, A);  

  connect_east(p, B);

  connect_south(p, A);
  connect_south(p, B);  

  connect_west(p, A);

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x0;
  p->lt = A;
  p->ge = B;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}


/*
           _ p->y1
|    A    |
|_________| r.y1
|         |
|         |
|    B    |
|         |_ p->y0
       

         y1 (r.y1)
        / \
       B   A

*/

void Decomp::full_S_edge(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("full_S_edge");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, p->r.x1, r.y1,    p->r.y1, p->data);
  Node *B = new_Node(p->r.x0, p->r.x1, p->r.y0, r.y1,    (p->data & ~mask) | data);

  // internal rectangle connections
  
  B->nl[Node::N].push_back(A);

  A->nl[Node::S].push_back(B);

  // connect new rectangles

  connect_north(p, A);

  connect_east(p, B);
  connect_east(p, A);    

  connect_south(p, B);

  connect_west(p, A);
  connect_west(p, B);  

  // transform root node into y1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::Y_NODE;
  p->v = r.y1;
  p->lt = B;
  p->ge = A;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|    |    |
|    |    |
| B  | A  |
|    |    |
|    |    |
|    |    |_ p->y0
    r.x1

         x1 (r.x1)
        / \
       B   A

*/

void Decomp::full_W_edge(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("full_W_edge");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(r.x1,    p->r.x1, p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(p->r.x0, r.x1,    p->r.y0, p->r.y1, (p->data & ~mask) | data);
  
  // internal rectangle connections
  
  A->nl[Node::W].push_back(B);

  B->nl[Node::E].push_back(A);

  // connect new rectangles

  connect_north(p, A);
  connect_north(p, B);  

  connect_east(p, A);

  connect_south(p, B);
  connect_south(p, A);  

  connect_west(p, B);

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v  = r.x1;
  p->lt = B;
  p->ge = A;

  // remove all previous connections from and to outside

  delete_edges_to(p);  
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|    A    |
|_________| r.y1
|    C    |
|---------| r.y0
|    B    |
|         |_ p->y0
       

         y1 (r.y1)
        / \
(r.y0) y2  A
      / \
     B  C
*/

void Decomp::WE_embedded_strip(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("WE_embedded_strip");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, p->r.x1, r.y1,    p->r.y1, p->data);
  Node *B = new_Node(p->r.x0, p->r.x1, p->r.y0, r.y0,    p->data);
  Node *C = new_Node(p->r.x0, p->r.x1, r.y0,    r.y1,    (p->data & ~mask) | data);  

  // new decision nodes
  
  Node *yn2 = new_Node(Node::Y_NODE, r.y0, B, C);
  
  // internal rectangle connections
  
  A->nl[Node::S].push_back(C);

  C->nl[Node::N].push_back(A);
  C->nl[Node::S].push_back(B);    

  B->nl[Node::N].push_back(C);

  // connect new rectangles

  connect_north(p, A);

  connect_east(p, B);
  connect_east(p, C);      
  connect_east(p, A);    

  connect_south(p, B);

  connect_west(p, A);
  connect_west(p, C);
  connect_west(p, B);    

  // transform root node into y1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::Y_NODE;
  p->v = r.y1;
  p->lt = yn2;
  p->ge = A;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

/*
           _ p->y1
|   | |  |
|   | |  |
| A |C| B|
|   | |  |
|   | |  |
|   | |  |_ p->y0
 r.x0 r.x1

         x1 (r.x0)
        / \
       A   x2 (r.x1)
          / \
	 C   B
*/

void Decomp::NS_embedded_strip(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("NS_embedded_strip");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0, p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(r.x1, p->r.x1, p->r.y0, p->r.y1, p->data);
  Node *C = new_Node(r.x0,    r.x1, p->r.y0, p->r.y1, (p->data & ~mask) | data);  

  // new decision nodes
  
  Node *xn2 = new_Node(Node::X_NODE, r.x1, C, B);
  
  // internal rectangle connections
  
  A->nl[Node::E].push_back(C);

  C->nl[Node::W].push_back(A);
  C->nl[Node::E].push_back(B);    

  B->nl[Node::W].push_back(C);

  // connect new rectangles

  connect_north(p, B);
  connect_north(p, C);  
  connect_north(p, A);  

  connect_east(p, B);

  connect_south(p, A);
  connect_south(p, C);  
  connect_south(p, B);  

  connect_west(p, A);

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x0;
  p->lt = A;
  p->ge = xn2;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

#if 1

// vertical version

/*
           _ p->y1
|   | |   |
|   |D|   |
|   |_|   | r.y0
| A | | B |
|   |C|   |
|   | |   |_ p->y0
 r.x0 r.x1
 
         x1 (r.x0)
        / \
       A   x2 (r.x1)
          / \
  (r.y0) y1  B
        / \
       C   D      
*/


void Decomp::N_middle_section(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("N_middle_section");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0,    p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(r.x1,    p->r.x1, p->r.y0, p->r.y1, p->data);
  Node *C = new_Node(r.x0,    r.x1,    p->r.y0, r.y0,    p->data);
  Node *D = new_Node(r.x0,    r.x1,    r.y0,    p->r.y1, (p->data & ~mask) | data);    

  // new decision nodes
  
  Node *yn1 = new_Node(Node::Y_NODE, r.y0, C, D);
  Node *xn2 = new_Node(Node::X_NODE, r.x1, yn1, B);
  
  // internal rectangle connections (clockwise)
  
  A->nl[Node::E].push_back(D);
  A->nl[Node::E].push_back(C);

  B->nl[Node::W].push_back(C);
  B->nl[Node::W].push_back(D);
  
  C->nl[Node::N].push_back(D);
  C->nl[Node::W].push_back(A);  
  C->nl[Node::E].push_back(B);
  
  D->nl[Node::W].push_back(A);
  D->nl[Node::E].push_back(B);
  D->nl[Node::S].push_back(C);

  // connect new rectangles (anti-clockwise)

  connect_north(p, B); 
  connect_north(p, D);
  connect_north(p, A);    

  connect_east(p, B);    

  connect_south(p, A);
  connect_south(p, C);
  connect_south(p, B);  

  connect_west(p, A);  

  // transform root node into y1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x0;
  p->lt = A;
  p->ge = xn2;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

#else

// horizontal version

/*
           _ p->y1
|   | |   |
| A |D| B |
|___|_|___| r.y0
|         |
|    C    |
|         |_ p->y0
 r.x0 r.x1
 
         y1 (r.y0)
        / \
       C   x1 (r.x1)
          / \
  (r.x0) x2  B
        / \
       A   D      
*/


void Decomp::N_middle_section(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("N_middle_section");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0,    r.y0,    p->r.y1, p->data);
  Node *B = new_Node(r.x1,    p->r.x1, r.y0,    p->r.y1, p->data);
  Node *C = new_Node(p->r.x0, p->r.x1, p->r.y0, r.y0,    p->data);
  Node *D = new_Node(r.x0,    r.x1,    r.y0,    p->r.y1, (p->data & ~mask) | data);    

  // new decision nodes
  
  Node *xn2 = new_Node(Node::X_NODE, r.x0, A, D);
  Node *xn1 = new_Node(Node::X_NODE, r.x1, xn2, B);
  
  // internal rectangle connections
  
  A->nl[Node::E].push_back(D);
  A->nl[Node::S].push_back(C);

  B->nl[Node::W].push_back(D);
  B->nl[Node::S].push_back(C);
  
  C->nl[Node::N].push_back(A);
  C->nl[Node::N].push_back(D);  
  C->nl[Node::N].push_back(B);
  
  D->nl[Node::W].push_back(A);
  D->nl[Node::E].push_back(B);
  D->nl[Node::S].push_back(C);

  // connect new rectangles

  connect_north(p, B);
  connect_north(p, D);
  connect_north(p, A);    

  connect_east(p, C);
  connect_east(p, B);    

  connect_south(p, C);

  connect_west(p, A);  
  connect_west(p, C);

  // transform root node into y1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::Y_NODE;
  p->v = r.y0;
  p->lt = C;
  p->ge = xn1;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

#endif

/*
           _ p->y1
|     | A |
|     |___| r.y1
|     |   | 
|  C  | D |
|     |___| r.y0
|     | B |_ p->y0
    r.x0 p->x1
 
         x1 (r.x0)
        / \
       C   y1 (r.y1)
          / \
  (r.y0) y2  A
        / \
       B   D      
*/


void Decomp::E_middle_section(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("E_middle_section");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(r.x0,    p->r.x1, r.y1,    p->r.y1, p->data);
  Node *B = new_Node(r.x0,    p->r.x1, p->r.y0, r.y0,    p->data);
  Node *C = new_Node(p->r.x0, r.x0,    p->r.y0, p->r.y1, p->data);
  Node *D = new_Node(r.x0,    p->r.x1, r.y0,    r.y1,    (p->data & ~mask) | data);    

  // new decision nodes
  
  Node *yn2 = new_Node(Node::Y_NODE, r.y0, B, D);
  Node *yn1 = new_Node(Node::Y_NODE, r.y1, yn2, A);
  
  // internal rectangle connections
  
  A->nl[Node::S].push_back(D);
  A->nl[Node::W].push_back(C);

  B->nl[Node::N].push_back(D);
  B->nl[Node::W].push_back(C);
  
  C->nl[Node::E].push_back(A);
  C->nl[Node::E].push_back(D);  
  C->nl[Node::E].push_back(B);
  
  D->nl[Node::N].push_back(A);
  D->nl[Node::S].push_back(B);
  D->nl[Node::W].push_back(C);

  // connect new rectangles

  connect_north(p, A);
  connect_north(p, C);

  connect_east(p, B);
  connect_east(p, D);
  connect_east(p, A);      

  connect_south(p, C);  
  connect_south(p, B);

  connect_west(p, C);  

  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x0;
  p->lt = C;
  p->ge = yn1;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

#if 1

// vertical version

/*
           _ p->y1
|   | |   |
|   |C|   |
|   |_|   | r.y1
| A | | B |
|   |D|   |
|   | |   |_ p->y0
 r.x0 r.x1
 
         x1 (r.x0)
        / \
       A   x2 (r.x1)
          / \
  (r.y1) y1  B
        / \
       D   C      
*/


void Decomp::S_middle_section(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("S_middle_section");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0,    p->r.y0, p->r.y1, p->data);
  Node *B = new_Node(r.x1,    p->r.x1, p->r.y0, p->r.y1, p->data);
  Node *C = new_Node(r.x0,    r.x1,    r.y1,    p->r.y1, p->data);
  Node *D = new_Node(r.x0,    r.x1,    p->r.y0, r.y1,    (p->data & ~mask) | data);

  // new decision nodes
  
  Node *yn1 = new_Node(Node::Y_NODE, r.y1, D, C);
  Node *xn2 = new_Node(Node::X_NODE, r.x1, yn1, B);
  
  // internal rectangle connections (clockwise)
  
  A->nl[Node::E].push_back(C);
  A->nl[Node::E].push_back(D);

  B->nl[Node::W].push_back(D);
  B->nl[Node::W].push_back(C);
  
  C->nl[Node::E].push_back(B);
  C->nl[Node::S].push_back(D);
  C->nl[Node::W].push_back(A);  
  
  D->nl[Node::N].push_back(C);
  D->nl[Node::E].push_back(B);
  D->nl[Node::W].push_back(A);

  // connect new rectangles (anti-clockwise)

  connect_north(p, B); 
  connect_north(p, C);
  connect_north(p, A);    

  connect_east(p, B);    

  connect_south(p, A);
  connect_south(p, D);
  connect_south(p, B);  

  connect_west(p, A);  

  // transform root node into y1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x0;
  p->lt = A;
  p->ge = xn2;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

#else

// horizontal version

/*
           _ p->y1
|         |
|    C    |
|___ _ ___| r.y1
|   | |   |
| A |D| B |
|   | |   |_ p->y0
 r.x0 r.x1
 
         y1 (r.y1)
        / \
(r.x1) x1  C
      / \
(r.x0)x2 B
     / \
    A   D      
*/


void Decomp::S_middle_section(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("S_middle_section");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
  
  Node *A = new_Node(p->r.x0, r.x0,    p->r.y0, r.y1, p->data);
  Node *B = new_Node(r.x1,    p->r.x1, p->r.y0, r.y1, p->data);
  Node *C = new_Node(p->r.x0, p->r.x1, r.y1, p->r.y1, p->data);
  Node *D = new_Node(r.x0,    r.x1,    p->r.y0, r.y1, (p->data & ~mask) | data);    

  // new decision nodes
  
  Node *xn2 = new_Node(Node::X_NODE, r.x0, A, D);
  Node *xn1 = new_Node(Node::X_NODE, r.x1, xn2, B);
  
  // internal rectangle connections
  
  A->nl[Node::N].push_back(C);
  A->nl[Node::E].push_back(D);

  B->nl[Node::N].push_back(C);
  B->nl[Node::W].push_back(D);
  
  C->nl[Node::S].push_back(B);
  C->nl[Node::S].push_back(D);  
  C->nl[Node::S].push_back(A);
  
  D->nl[Node::N].push_back(C);
  D->nl[Node::W].push_back(A);
  D->nl[Node::E].push_back(B);

  // connect new rectangles

  connect_north(p, C);

  connect_east(p, B);
  connect_east(p, C);
  
  connect_south(p, A);
  connect_south(p, D);
  connect_south(p, B);    

  connect_west(p, C);  
  connect_west(p, A);

  // transform root node into y1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::Y_NODE;
  p->v  = r.y1;
  p->lt = xn1;
  p->ge = C;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}

#endif



/*
           _ p->y1
| A   |   |
|_____|   | r.y1
|     |   | 
| D   | C |
|_____|   | r.y0
| B   |   |_ p->y0
    r.x1 p->x1
 
         x1 (r.x1)
        / \
(r.y1) y1  C
       / \
(r.y0)y2  A
     / \
    B   D      
*/


void Decomp::W_middle_section(Node *p, const Rect &r, Data mask, Data data)
{
  TEST_OUT("W_middle_section");
  NODE_INC_ID;
  assert(p->type == Node::R_NODE);

  // new rectangles
   
  Node *A = new_Node(p->r.x0, r.x1,    r.y1,    p->r.y1, p->data);
  Node *B = new_Node(p->r.x0, r.x1,    p->r.y0, r.y0,    p->data);
  Node *C = new_Node(r.x1,    p->r.x1, p->r.y0, p->r.y1, p->data);
  Node *D = new_Node(p->r.x0, r.x1,    r.y0,    r.y1,    (p->data & ~mask) | data);    

  // new decision nodes
  
  Node *yn2 = new_Node(Node::Y_NODE, r.y0, B, D);
  Node *yn1 = new_Node(Node::Y_NODE, r.y1, yn2, A);
  
  // internal rectangle connections
  
  A->nl[Node::E].push_back(C);
  A->nl[Node::S].push_back(D);

  B->nl[Node::N].push_back(D);
  B->nl[Node::E].push_back(C);
  
  C->nl[Node::W].push_back(B);
  C->nl[Node::W].push_back(D);  
  C->nl[Node::W].push_back(A);
  
  D->nl[Node::N].push_back(A);
  D->nl[Node::E].push_back(C);
  D->nl[Node::S].push_back(B);

  // connect new rectangles

  connect_north(p, C);
  connect_north(p, A);

  connect_east(p, C); 

  connect_south(p, B);
  connect_south(p, C);

  connect_west(p, A); 
  connect_west(p, D); 
  connect_west(p, B); 
  
  // transform root node into x1
  // warning! must be done at the end because former R_NODE data gets destroyed

  p->type = Node::X_NODE;
  p->v = r.x1;
  p->lt = yn1;
  p->ge = C;

  // remove all previous connections from and to outside

  delete_edges_to(p);
  FORS (i, 4) p->nl[i].clear();
}



void Decomp::area(Node *p, set<Node*> &s, sint4 &a)
{
  if (s.find(p) != s.end()) return;
  s.insert(p);
  
  if (p->type != Node::R_NODE) ERR("not an R_NODE");

  a += (p->r.x1 - p->r.x0) * (p->r.y1 - p->r.y0);

  FORS (j, 4) {
    FORALL (p->nl[j], i) { area(*i, s, a); }
  }
}


void Decomp::collect_edges(Vector<Edge> &es)
{
  FORALL (nodes, p) {

    Node &n = **p;

    if (n.type != Node::R_NODE) continue;

    if (n.r.x0 > r0.x0 && n.nl[Node::W].empty()) { n.write(); ERR("no west edge"); }
    if (n.r.x1 < r0.x1 && n.nl[Node::E].empty()) { n.write(); ERR("no east edge"); }
    if (n.r.y0 > r0.y0 && n.nl[Node::S].empty()) { n.write(); ERR("no south edge"); }
    if (n.r.y1 < r0.y1 && n.nl[Node::N].empty()) { n.write(); ERR("no north edge"); } 
  
#if MY_DEBUG  
    n.check("F");
#endif
  
    FORS (j, 4) {
      FORALL (n.nl[j], i) {
	if (*i == 0) ERR("null edge!");
	if ((*i)->type != Node::R_NODE) ERR("no R_NODE");
	es.push_back(Edge(&n, *i));
	// collect_edges(*i, s, es);
      }
    }
  }
}

bool Decomp::check(const string &msg)
{
  cout << "check " << msg << endl;
  
  if (!root) ERR("root=0");

  // check connectivity: for each edge (a,b), (b,a) must also exist

  Vector<Edge> es;
  collect_edges(es);

  FORALL (es, i) {
    if (i->a == i->b) {
      i->a->write();
      ERR(msg+": loop");
    }
    if (i->a > i->b) swap(i->a, i->b);
  }

  sort(es.begin(), es.end());

  uint4 sz = es.size();

  // cout << sz << " edges" << endl;

  if (sz & 1) ERR(msg+": odd edge vector size");

  if (sz) {
    for (uint4 i = 0; i < sz-1; i += 2) {

      if (es[i] != es[i+1]) {
      
	cout << i << " "
	     << es[i].a << " "
	     << es[i].b << " "
	     << es[i+1].a << " "
	     << es[i+1].b << " "	
	     << endl;

	cout << "a: "; 
	es[i].a->write();
	cout << "b: ";
	es[i].b->write();

	cout << "a: "; 
	es[i+1].a->write();
	cout << "b: ";
	es[i+1].b->write();

	for (uint4 i = 0; i < sz; i++) {	
	  cout << es[i].a << " ";
#if NODE_IDS	  
	  cout << es[i].a->id << ":" << es[i].a->sub_id;
#endif	    
	  cout << " -> " << es[i].b << " ";
#if NODE_IDS	  	  
	  cout << es[i].b->id << ":" << es[i].b->sub_id;
#endif
	  cout << endl;
	}
	ERR(msg+": no edge twin found!");
      }
    }
  }

#if 0
  
  // check area (does not work if in SectorDecomp context (computes entire area then)
  
  sint4 a = 0;
  set<Node*> t;
  area(r, t, a);
  
  if (a != (r0.x1-r0.x0)*(r0.y1-r0.y0)) ERR(msg+": area not ok");

#endif
  
  return true;
}


void Decomp::dump(Node *p, set<Node*> &s)
{
  if (s.find(p) != s.end()) return;
  s.insert(p);
  
  cout << "rect node ";
  p->write();

  FORS (j, 4) {
    FORALL (p->nl[j], i) dump(*i, s);
  }
}


void Decomp::dump()
{
  Node *r = locate(0, 0);

  set<Node*> s;
  
  dump(r, s);
}


bool operator<(const Decomp::Rect &a, const Decomp::Rect &b)
{
  if (a.y0 < b.y0) return true;
  if (a.y0 > b.y0) return false;
  if (a.y1 < b.y1) return true;
  if (a.y1 > b.y1) return false;
  if (a.x0 < b.x0) return true;
  if (a.x0 > b.x0) return false;
  return a.x1 < b.x1;
}


static void draw(Decomp &d, SDL_Surface *screen)
{
  FORALL (d.nodes, i) {

    Decomp::Node *n = *i;

    if (n->type != Decomp::Node::R_NODE) continue;
    
    //  n->write();
  
    RGB rgb;

    if (n->data)
      rgb.from(255,255,255);
    else {

#if 0    
      rgb.from(rng.rand_uint4() % 100 + 50,
	       rng.rand_uint4() % 100 + 50,
	       rng.rand_uint4() % 100 + 50);
#else

      rgb.from(((uint4)n ^ 0x12f45d79) % 100 + 50,
	       ((uint4)n ^ 0x23456736) % 100 + 50,
	       ((uint4)n ^ 0x3281759d) % 100 + 50);
#endif
    }

#if 1  
    sint4 x1 = n->r.x0*GFX_MAG;
    sint4 y1 = (H-n->r.y1)*GFX_MAG;
    SDL_Rect rect;
    rect.x = x1;
    rect.y = y1;
    rect.w = GFX_MAG*(n->r.x1-n->r.x0);
    rect.h = GFX_MAG*(n->r.y1-n->r.y0);
    SDL_FillRect(screen, &rect, rgb.color());
#endif

#if 0
  
    // draw neighbor edges

    if (qt->w >= 8) {

      {
	// North connections

	QuadtreeNode::NList &nl = qt->nls[QuadtreeNode::N_N];
	rgb.from(0,0,0);

	FORALL (nl, i) {
	  QuadtreeNode &n = *(*i).node;
	  lineColor(screen, qt->xl+qt->w/2, qt->yu+qt->w/2, n.xl+n.w/2, n.yu+n.w/2,
		    rgb.color_alpha());
	}
      }

      {
	// East connections
	QuadtreeNode::NList &nl = qt->nls[QuadtreeNode::N_W];
	rgb.from(0,0,0);

	FORALL (nl, i) {
	  QuadtreeNode &n = *(*i).node;
	  lineColor(screen, qt->xl+qt->w/2, qt->yu+qt->w/2, n.xl+n.w/2, n.yu+n.w/2,
		    rgb.color_alpha());
	}
      }
    }
#endif

  }

  // mark degree >= 2, 2, 1, and 0 nodes

  {
    const sint4 SUCC_END = 10;
    Vector<sint4> num(SUCC_END);
      
    FORALL (d.nodes, i) {

      if ((*i)->type != Decomp::Node::R_NODE || (*i)->data != 0) continue;
	  
      vector<sint4> occupied;
	  
      FORS (j, 4) {
	if ((*i)->nl[j].empty()) occupied.push_back(1);
	else {
	  FORALL (((*i)->nl[j]), k) {
	    occupied.push_back((*k)->data != 0);
	  }
	}
      }

      assert(!occupied.empty());
      
      bool prev = occupied.back();
      sint4 toggle = 0;

      {
	FORALL (occupied, l) {
	  if (prev != *l) { prev = *l; ++toggle; }
	}
      }

#if 0
      FORALL (occupied, l) cout << *l;
      cout << " " << toggle/2 << endl;
#endif

      RGB rgb, border;

      toggle /= 2;

      if (toggle < SUCC_END-1) ++num[toggle]; else ++num[SUCC_END-1];
      
      switch (toggle) {
      case 0:  rgb.from(0,0,0);   border.from(0,0,0);   break;
      case 1:  rgb.from(255,0,0); border.from(128,0,0); break;
      case 2:  rgb.from(0,255,0); border.from(0,128,0); break;
      default: rgb.from(0,0,255); border.from(0,0,128); 
      }

      sint4 x1 = (*i)->r.x0*GFX_MAG;
      sint4 y1 = (H-(*i)->r.y1)*GFX_MAG;
      SDL_Rect rect;
      rect.x = x1;
      rect.y = y1;
      rect.w = GFX_MAG*((*i)->r.x1-(*i)->r.x0);
      rect.h = GFX_MAG*((*i)->r.y1-(*i)->r.y0);
      SDL_FillRect(screen, &rect, border.color());
      rect.x++;
      rect.w--;
      SDL_FillRect(screen, &rect, rgb.color());
    }

    { FORALL (num, i) { cout << *i << " "; } }
    cout << endl;
  }
}


void Decomp::count(sint4 &r, sint4 &x, sint4 &y, sint4 &l) const
{
  r = x = y = l = 0;
  
  FORALL (nodes, i) {
    switch ((*i)->type) {
    case Node::X_NODE: ++x; break;
    case Node::Y_NODE: ++y; break;
    case Node::R_NODE:
      ++r;
      l += ((*i)->r.x1-(*i)->r.x0) + ((*i)->r.y1-(*i)->r.y0);
      break;

    default: ERR("unknown node type");
    }
  }
}


void Decomp::test(Random &rng)
{
  signals_shut_down(false); // ctrl-c ends program

  cout << sizeof(Node) << endl;
  cout << sizeof(Node::RList) << endl;  

  sint4 path;
  Options::get("-path", path);

  if (path != 1) {

    SDL_Surface *screen = 0;
  
    SDLinit::video_init();
    screen = SDL_SetVideoMode(WX, WY, 0, SDL_HWSURFACE | SDL_ANYFORMAT);
    
    if (screen == NULL) {
      fprintf(stderr, "Couldn't set video mode");
      exit(1);
    }
    
    SDL_WM_SetCaption("test", 0);

    //    cout << "size: " << sizeof(QuadtreeNode) << endl;
    // cout << "dim: " << N << endl;
    //WX/GFX_MAG << " K=" << log(double(WX/GFX_MAG))/log(2.0) << endl;

    sint4 seed = 1;

    FOREVER {

      ++seed;
    
      cout << "SEED=" << seed << endl;
      //rng.srand(seed);

      SectorDecomp sd(W, H, SS);

      Vector<Rect> rs;

      FORS (i, path-1) {
      
	sint4 x0, y0;
      
	do {
	  x0 = rng.rand_uint4() % (W+1);
	  y0 = rng.rand_uint4() % (H+1);
	} while (x0 > W-S || y0 > H-S);
      
	Rect r;

	if (rng.rand_uint4() & 8)
	  r.set(x0, x0+S, y0, y0+1);
	else
	  r.set(x0, x0+1, y0, y0+S);
      
	rs.push_back(r);
      }
    
      FORALL (rs, j) {  

#if 0	
	cout << "insert [" << j->x0 << "," << j->x1 << "]x["
	     << j->y0 << "," << j->y1 << "]" << endl;
#endif
	
#if 0 // display decomposition after each insertion
	
	if (SDL_MUSTLOCK(screen) && SDL_LockSurface(screen) < 0 ) {
	  fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
	  return;
	}

	RGB rgb;
	rgb.from(0, 0, 0);
	SDL_Rect rect;
	rect.x = 0;
	rect.y = 0;
	rect.w = WX;
	rect.h = WY;
	SDL_FillRect(screen, &rect, rgb.color());

	FORALL (sd.sectors, i) draw(*i, screen);

	if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);
	SDL_UpdateRect(screen, 0, 0, 0, 0);
	cout << "wait" << endl;

	char input[300];
	cin.getline(input, 256); 
#endif
	
	sd.insert(*j, 1, 1);

	//cout << "done" << endl;
	//d.dump();
    
	// d.check("last");
      }

      if (SDL_MUSTLOCK(screen) && SDL_LockSurface(screen) < 0 ) {
	fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
	return;
      }

      RGB rgb;
      rgb.from(0, 0, 0);
      SDL_Rect rect;
      rect.x = 0;
      rect.y = 0;
      rect.w = WX;
      rect.h = WY;
      SDL_FillRect(screen, &rect, rgb.color());

      FORALL (sd.sectors, i) draw(*i, screen);

      if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);
      SDL_UpdateRect(screen, 0, 0, 0, 0);

      { FORALL (sd.sectors, i) (*i).check("foo"); }
      
#if 1      
      {
	sint4 r=0, x=0, y=0, l=0;
	
	FORALL (sd.sectors, i) {
	
	  sint4 dr, dx, dy, dl;
	  (*i).count(dr, dx, dy, dl);

	  r += dr; x += dx; y += dy; l += dl;
	}
	
	cout << "t=" << r+x+y << " x=" << x << " y=" << y << " r=" << r
	     << " l=" << l << endl;
      }
#endif

#if 0
      // TODO:
      // construct search graph

      vector<SearchNode*> search_graph;
      
      {
	FORALL (sd.sectors, i) {

	  FORALL (i->nodes, j) {
	    if ((*j)->type != Node::R_NODE || (*j)->data) continue;
	    SearchNode *n = new SearchNode;
	    search_graph.push_back(n);
	    // connect nodes ...
	  }
	}
      }

      // simplify search graph -> abstract graph
      
      // pick two random locations

      // find abstract nodes
      
      // compute shortest path in abstract graph

      // compute path in rectangle graph
      
      // draw path
      
#endif

      cout << "wait" << endl;

      char input[300];
      cin.getline(input, 256); 
      // cout << d.nodes.size() << endl;
    }
  }


#if 0

  // correctness
  
  sint4 seed = 1;

  FOREVER {

    ++seed;

    cout << "SEED=" << seed << endl;
    // rng.srand(seed);
    Decomp d;
    const sint4 W=128, H=128;
  
    Rect r = { 0, W, 0, H };

    d.insert(r, 0, 0);

    FORS (i, 50) {

      sint4 x0, x1, y0, y1;

      do {
	x0 = rng.rand_uint4() % (W+1);
	x1 = rng.rand_uint4() % (W+1);
	y0 = rng.rand_uint4() % (H+1);
	y1 = rng.rand_uint4() % (H+1);
      } while (x0 >= x1 || y0 >= y1);

      Rect r = { x0, x1, y0, y1 };

      cout << endl << "rect " << i << endl;

      cout << "insert [" << x0 << "," << x1 << "]x["<< y0 << "," << y1 << "]" << endl;
      d.insert(r, 1, 1);

      cout << "done" << endl;
      //d.dump();
      
      d.check("last");
    }
  }

#else

  // speed

  sint4 seed = 1;

  FORS (j, 1000) {

    ++seed;
    
    // cout << "SEED=" << seed << endl;
    // rng.srand(seed);
    Decomp d;
    const sint4 W=128, H=128;
    const sint4 S=4; // object size
    
    Rect r = { 0, W, 0, H };

    d.insert(r, 0, 0);

    Vector<Rect> rs;

    FORS (i, 1000) {
      
      sint4 x0, y0;
      
      do {
	x0 = rng.rand_uint4() % (W+1);
	y0 = rng.rand_uint4() % (H+1);
      } while (x0 > W-S || y0 > H-S);
      
      Rect r = { x0, x0+S, y0, y0+S };
      
      rs.push_back(r);
    }
    
    //sort(rs.begin(), rs.end());
    
    FORALL (rs, j) {  
      
      // cout << endl << "rect " << i << endl;
      // cout << "insert [" << x0 << "," << x1 << "]x["<< y0 << "," << y1 << "]" << endl;
      d.insert(*j, 1, 1);

      //cout << "done" << endl;
      //d.dump();
    
      //d.check("last");
    }

    // cout << d.nodes.size() << endl;
  }

#endif
  
  cout << "OK" << endl << endl;
}



SectorDecomp::SectorDecomp(Decomp::GridCoor w_, Decomp::GridCoor h_,
			   Decomp::GridCoor sector_size_)
{
  assert(sector_size_ > 0);
  w = w_; h = h_; sector_size = sector_size_;

  sx = w/sector_size; if (w % sector_size) ++sx;
  sy = h/sector_size; if (h % sector_size) ++sy;

  sectors = Vector<Decomp>(sx*sy);

  // initialize sectors
  
  FORS (y, sy) {
    FORS (x, sx) {
      Decomp::Rect r = { x*sector_size, (x+1)*sector_size, y*sector_size, (y+1)*sector_size };
      sectors[y*sx + x].insert(r, 0, 0);
    }
  }

  // connect sectors

  FORS (y, sy) {
    FORS (x, sx) {
      sint4 si = y*sx + x;

      // west
      if (x > 0) sectors[si].rect0()->nl[Decomp::Node::W].push_back(sectors[si-1].rect0());
      
      // east
      if (x < sx-1) sectors[si].rect0()->nl[Decomp::Node::E].push_back(sectors[si+1].rect0());
      
      // south
      if (y > 0) sectors[si].rect0()->nl[Decomp::Node::S].push_back(sectors[si-sx].rect0());
      
      // north
      if (y < sy-1) sectors[si].rect0()->nl[Decomp::Node::N].push_back(sectors[si+sx].rect0());
    }
  }

#if 0 && MY_DEBUG
  // OK
  FORU (i, sectors.size()) { stringstream ss; ss << i; sectors[i].check(ss.str().c_str()); }
#endif
}

Decomp::Node *SectorDecomp::locate(Decomp::GridCoor x, Decomp::GridCoor y)
{
  if (x >= w || y >= h) ERR("coordinate out of range");  // fixme: < 0 check if GridCoor signed!

  return sectors[(y/sector_size)*sx + x/sector_size].locate(x, y);
}

void SectorDecomp::insert(const Decomp::Rect &r, Decomp::Data mask, Decomp::Data data)
{
  if (r.x0 >= r.x1 || r.y0 >= r.y1 || r.x1 > w || r.y1 > h)
    ERR("rectangle out of range");  // fixme: < 0 check if GridCoor signed!
  
  sectors[(r.y0/sector_size)*sx + r.x0/sector_size].insert(r, mask, data);
}


