// $Id: GatherAI.C,v 1.8 2006/05/10 12:34:51 ddeutscher Exp $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "GatherAI.H"
#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameObj.H"
#include "GfxModule.H"
#include "TerrainModule.H"

using namespace std;

const sint4 GatherAI::FROM = EventFactory::new_who();
const sint4 GatherAI::GATHER_START = EventFactory::new_what();
const sint4 GatherAI::GATHER_STOP = EventFactory::new_what();

const std::string GatherAI::bp_name = "gather";

//===================================================================

GatherAI::GatherAI(GameStateModule &gsm_) : gsm(gsm_)
{
}

//===================================================================

GatherAI::~GatherAI()
{
}

//===================================================================

void GatherAI::stop(GameObj *obj)
{
  FIND (obj2dat, it, obj);
  if (it != obj2dat.end()) obj2dat.erase(it);
}

//===================================================================

void GatherAI::act(GameStateModule &/*gsm*/)
{
  //------------------------------

  for (set<GameObj*>::iterator i = mining.begin(); i != mining.end();) {
    GameObj *g = *i;
    if (g->component("pickaxe")->get_int("active") == 0) {
      need_act.push_back(g);
      obj2dat[g].state = GatherData::MOVE_BACK;
      set<GameObj*>::iterator j = i;
      ++i;
      mining.erase(j);
      continue;
    }
    ++i;
  }

  //------------------------------

  for (set<GameObj*>::iterator i = returning.begin(); i != returning.end();) {
    GameObj *g = *i;
    if (g->get_int("minerals") == 0) {
      need_act.push_back(g);
      obj2dat[g].state = GatherData::MOVE_TO;
      set<GameObj*>::iterator j = i;
      ++i;
      returning.erase(j);
      continue;
    }
    ++i;
  }

  //------------------------------

  FORALL (need_act, i) {
    FIND (obj2dat, it, *i);
    if (it == obj2dat.end()) continue;

    GatherData &gd = it->second;

    switch(gd.state) {

    case GatherData::MOVE_TO:
      {
        call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, 
                                               TerrainBase::Obj2ObjTask(*i,gd.res)));
      }
      continue;

    case GatherData::GATHER:
      {
        cout << "gather ai -> pickaxe" << endl;
        string bp = gd.res->bp_name();
        //        sint4 code = gd.res->code();
        ResBase &rb = res_map[bp];
        if (rb.resources.find(gd.res) == rb.resources.end()) {
          stop(*i);
          cout << "no resource?" << endl;
          continue;
        }
        Vector<sint4> args;
        args.push_back(gsm.get_game().get_cplayer_info().get_id(gd.res));
        it->first->component("pickaxe")->set_action("mine", args);
        mining.insert(*i);
      }
      continue;

    case GatherData::MOVE_BACK:
      {
        cout << "gather ai -> move back" << endl;
        string bp = gd.res->bp_name();
        //        sint4 code = gd.res->code();
        ResBase &rb = res_map[bp];

        if (rb.bases.size() == 0) continue;
        gd.base = *rb.bases.begin();
        
        call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, 
                                               TerrainBase::Obj2ObjTask(*i, gd.base)));
      }
      continue;

    case GatherData::RETURN:
      {
        cout << "gather ai -> return" << endl;
        string bp = gd.res->bp_name();
        //        sint4 code = gd.res->code();
        ResBase &rb = res_map[bp];
        if (rb.bases.find(gd.base) == rb.resources.end()) {
          stop(*i);
          continue;
        }
        Vector<sint4> args;
        args.push_back(gsm.get_game().get_cplayer_info().get_id(gd.base));
        it->first->set_action("return_resources", args);
        returning.insert(*i);
      }
      continue;

    default:
      ERR("??");
    }
  }

  need_act.clear();
}

//===================================================================


bool GatherAI::handle_event(const Event &e)
{
  if (e.get_what() == TerrainModule::PATH_SUCCESS) {

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "SUCCESS: not a PathStatusEvent");

    FIND (obj2dat, it, p->get_obj());
    if (it == obj2dat.end()) return false;

    GatherData::State &s = it->second.state;
    if (s == GatherData::MOVE_TO) s = GatherData::GATHER;
    if (s == GatherData::MOVE_BACK) s = GatherData::RETURN;
    need_act.push_back(p->get_obj());

    return true;
  }

  if (e.get_what() == TerrainModule::PATH_FAILURE) {
    cout << "gather ai got path failure" << endl;

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "FAILURE: not a PathStatusEvent");

    stop(p->get_obj());
    return true;
  }

  if (e.get_who() == GfxModule::FROM) {
    /*
    if (e.get_what() == GfxModule::ADD_CWIDGETS_MSG) {
      init_widget();
      return true;
    }
    */

    if (e.get_what() == GatherAI::GATHER_START) {
      REM("GATHER EVENT");
      const GatherEvent &ge = static_cast<const GatherEvent&>(e);
      if (ge.obj && ge.targ) {
        GatherData gd;
        gd.res = ge.targ;
        gd.base = 0; // compute later
        gd.state = GatherData::MOVE_TO;
        obj2dat[ge.obj] = gd;
        need_act.push_back(ge.obj);
      }
      return true;
    }

    if (e.get_what() == GatherAI::GATHER_STOP) {
      REM("STOP GATHER EVENT");
      const GatherEvent &ge = static_cast<const GatherEvent&>(e);
      stop(ge.obj);
      return true;
    }
  }

  if (e.get_who() == GameStateModule::FROM) {

    if (e.get_what() == GameStateModule::VIEW_MSG) {

      const sint4 me = gsm.get_game().get_client_player();
      
      // add new objects
      FORALL (gsm.get_changes().new_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (gob->bp_name() == "mineral")       add_res(gob);
        if (*gob->sod.owner != me) continue;
        if (gob->bp_name() == "controlCenter") add_base("mineral", gob);
      }
      
      // remove dead objects
      FORALL (gsm.get_changes().dead_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (gob->bp_name() == "mineral")       del_res(gob);
        if (*gob->sod.owner != me) continue;
        stop(gob);
        if (gob->bp_name() == "controlCenter") del_base("mineral", gob);
      }
      
      // compute actions
      act(gsm);

      return true;
    }
  }

  return false;
}

//===================================================================

void GatherAI::add_res(GameObj *gob) { res_map[gob->bp_name()].resources.insert(gob); }
void GatherAI::del_res(GameObj *gob) { res_map[gob->bp_name()].resources.erase(gob); }
void GatherAI::add_base(const string &res, GameObj *gob) { res_map[res].bases.insert(gob); }
void GatherAI::del_base(const string &res, GameObj *gob) { res_map[res].bases.erase(gob); }

//===================================================================

