#ifndef GatherAI_H
#define GatherAI_H

// $Id: GatherAI.H,v 1.5 2006/05/29 19:34:00 ddeutscher Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "LowAI.H"
#include "EventHandler.H"
#include "ClientModule.H"

class GameStateModule;

//===================================================================

class GatherData {
public:
  GameObj *res;
  GameObj *base;
  enum State { MOVE_TO, GATHER, MOVE_BACK, RETURN };
  State state;
};

//-------------------------------------------------------------------

class GatherAI : public LowAI<GatherData>, public EventHandler, public ClientModule
{
protected:

  struct ResBase {
    std::set<GameObj*> resources;
    std::set<GameObj*> bases;
  };

  std::map<std::string, ResBase> res_map; ///< Resources and places to take them.
  GameStateModule &gsm;
  Vector<GameObj*> need_act;
  std::set<GameObj*> mining;
  std::set<GameObj*> returning;

  void stop(GameObj *obj);

public:

  static const sint4 FROM;
  static const sint4 GATHER_START;
  static const sint4 GATHER_STOP;

  GatherAI(GameStateModule &gsm);
  ~GatherAI();

  void act(GameStateModule &gsm);
  static const std::string bp_name;

  void add(GameObj*) { ERR("?"); }
  void add_res(GameObj *gob);
  void add_base(const std::string &res, GameObj *gob);

  void del(GameObj*) { ERR("?"); }
  void del_res(GameObj *gob);
  void del_base(const std::string &res, GameObj *gob);

  bool handle_event(const Event &e);
};

//===================================================================

class GatherEvent : public Event
{
public:
  
  GatherEvent(sint4 who, sint4 what, GameObj *o, GameObj *t) :
    Event(who, what), obj(o), targ(t) {}

  ~GatherEvent() {}
  
  GameObj *obj, *targ;
};

//===================================================================

typedef std::map<GameObj*, GatherData> GatherDataMap;
REGISTER_TYPEOF(410, GatherDataMap::iterator);

#endif
