// $Id: DrawOnTerrain.C,v 1.5 2006/06/14 17:44:31 orts_frano Exp $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "DrawOnTerrain.H"
#include "GfxModule.H"
#include "Map.H"
#include "Game.H"
#include "GameTile.H"
#include "GfxTerrain.H"
#include "GfxTerrainTile.H"
#include "GameStateModule.H"

#include "GfxGlobal.H"
using namespace std;

//===================================================================

DrawOnTerrain::DrawOnTerrain(GfxModule &gm)
{
  gfxm = &gm;
}

//-------------------------------------------------------------------

DrawOnTerrain::~DrawOnTerrain()
{
}

//===================================================================

void DrawOnTerrain::begin_lines()
{
  /*
  glActiveTextureARB(GL_TEXTURE1);
  glDisable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE0);
  glDisable(GL_TEXTURE_2D);
  */

  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glEnable(GL_BLEND);
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_COLOR_MATERIAL);
  glDisable(GL_CULL_FACE);
  glEnable(GL_LIGHTING);
}

//-------------------------------------------------------------------

void DrawOnTerrain::end_lines()
{
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);
  //  glEnable(GL_CULL_FACE);

  glLineWidth(1.0);
}

//-------------------------------------------------------------------

void DrawOnTerrain::draw_line(real4 x1, real4 y1, real4 x2, real4 y2,
                              const Vec3<real4> &col, real4 width)
{
  glLineWidth(width);
  glColor4f(col.r, col.g, col.b, 1.0);

  glBegin(GL_LINES);
  glVertex3f(x1, gfxm->ground_height(x1, y1), y1);
  glVertex3f(x2, gfxm->ground_height(x2, y2), y2);
  glEnd();
}

//-------------------------------------------------------------------

void DrawOnTerrain::draw_circle(real4 cx, real4 cy, real4 r,
                                const Vec3<real4> &col, real4 width)
{
  const sint4 seg = 32;

  glLineWidth(width);
  glColor4f(col.r, col.g, col.b, 1.0);

  glBegin(GL_LINE_LOOP);

  real4 arc = 2.0 * M_PI / (real4)seg;
  FORU (i, seg+1) {
    real4 a = arc*i;
    real4 x = cos(a) * r + cx;
    real4 y = sin(a) * r + cy;
    glVertex3f(x, gfxm->ground_height(x , y), y);
  }

  glEnd();
}

//===================================================================

void DrawOnTerrain::begin_tiles()
{
  glActiveTextureARB(GL_TEXTURE1);
  glDisable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE0);
  glDisable(GL_TEXTURE_2D);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glEnable(GL_BLEND);
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_COLOR_MATERIAL);
  glDisable(GL_CULL_FACE);
  glEnable(GL_LIGHTING);
}

//-------------------------------------------------------------------

void DrawOnTerrain::end_tiles()
{
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
}

//-------------------------------------------------------------------

void DrawOnTerrain::draw_tile(sint4 tx, sint4 ty, const Vec3<real4> &col)
{
  const Map<GameTile> &map = gfxm->get_gsm().get_game().get_map();

  if (tx < 0 || tx >= map.get_width()) return;
  if (ty < 0 || ty >= map.get_height()) return;

  sint4 tind = map.xy2ind(tx, ty);

  GfxTerrainTile *t = &gfxm->terrain->tiles[tind];
  if (!t) return;
  if (!t->is_initialized()) return;

  glColor4f(col.r, col.g, col.b, 0.5);
  t->draw_flat();
}

//===================================================================

void DrawOnTerrain::begin_squares()
{
  glActiveTextureARB(GL_TEXTURE1);
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);
  glDisable(GL_TEXTURE_2D);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  glActiveTextureARB(GL_TEXTURE0);
  glDisable(GL_TEXTURE_2D);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glEnable(GL_BLEND);
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_COLOR_MATERIAL);
  glDisable(GL_CULL_FACE);
  glDisable(GL_LIGHTING);
}

//-------------------------------------------------------------------

void DrawOnTerrain::end_squares()
{
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
}

//-------------------------------------------------------------------

void DrawOnTerrain::draw_square(real4 x1, real4 y1, real4 x2, real4 y2,
                                const Vec3<real4> &col)
{
  if (x1 > x2) swap(x1, x2);
  if (y1 > y2) swap(y1, y2);

  glColor4f(col.r, col.g, col.b, 1.0);

  glBegin(GL_QUAD_STRIP);
  glVertex3f(x2, gfxm->ground_height(x2, y1), y1);
  glVertex3f(x1, gfxm->ground_height(x1, y1), y1);
  glVertex3f(x2, gfxm->ground_height(x2, y2), y2);
  glVertex3f(x1, gfxm->ground_height(x1, y2), y2);
  glEnd();
}

//===================================================================

