// $Id: GfxDrawPick.C,v 1.79 2006/05/25 22:21:16 orts_furtak Exp $

// This is an ORTS file
// (c) Michael Buro  
// (c) Tim Furtak    
// (c) Keith Yerex
// (c) Erik Benner
// licensed under the GPL

#include "GfxGlobal.H"
#include "Global.H"
#include "GameStateModule.H"
#include "GfxModule.H"
#include "GfxTerrainTile.H"
#include "GfxTerrain.H"
#include "WaterTile.H"
#include "GfxAnimesh.H"
#include "GfxEmitter.H"
#include "LayoutHV.H"
#include "Octree.H"
#include "Game.H"
#include "CycleCounter.H"
#include "Minimap.H"
#include "GameTile.H"
#include "GameObj.H"
#include "Options.H"
#include "Profiler.H"

using namespace std;

#define NARROW 5  // sel_view |w,h| <=  : pick one object 
#define TIMING 0

const sint4 NON_ACTIVE_REFRESH = 3; // refresh non-active view once in this period

REGISTER_TYPEOF(230, list<GfxModule::ColRect*>::iterator);
REGISTER_TYPEOF(232, list<GfxModule::ColorTileSet*>::iterator);
REGISTER_TYPEOF(234, Vector<GfxTerrainTile*>::iterator);

//===================================================================

void GfxModule::update_gfxobjs() {
  fps.update();
  real4 dt = fps.getDt();
  FORALL (game_objs, it) (*it)->update(dt); // update interpolated position
  WaterTile::anim(dt);

  // go through GfxObjects; remove dead ones, update moved ones
  Vector<GfxObject*> to_be_removed;
  FORALL (gfx_objs, it) {
    GfxObject *go = *it;
    
    if (go->dead) {
      
      GfxGameObject *ggo = GfxGameObject::get_gfx(go);
      if (ggo) ggo->gfx_obj = 0;
      to_be_removed.push_back(go);
      octree->remove(go);
      
    } else {
      
      if (go->hidden == 0) {
        go->update(dt);
        if (go->moved) {
          octree->update(go);
          go->moved = false;
        }
      } else if (go->hidden == 3) {
        octree->remove(go);
        go->hidden = 2;
      } else if (go->hidden == 1) {
        octree->add(go);
        go->moved = false;
        go->hidden = 0;
      }

    }
  }
  
  FORALL (to_be_removed, it) {
    gfx_objs.erase(*it);
    delete *it;
  }
}

//-------------------------------------------------------------------

void GfxModule::draw()
{
  Profiler::tick();
  ++frames;
  glerror("draw");

  static GLint T0     = 0;
  static GLint Frames = 0;
  static real8 rfps = 0.0;

  if (!pause_draw) update_gfxobjs();

  sint4 slow_view = -1; // <0: no slow view

#if 0
  // pick a view different from the active view
  
  if (views.size() > 1) {

    sint4 f = slow_frames % ::max(sint4(views.size()-1), NON_ACTIVE_REFRESH);
    if (f < sint4(views.size()-1)) {
      slow_view = f;
      if (sint4(active_view) <= slow_view) ++slow_view;
    }

    ++slow_frames;
  }

  assert(slow_view < sint4(views.size()));

  draw_view(views[active_view], true);
#endif

  // any scripted changes to scene geometry
  {
    SActParam p;
    p.obj.push_back(gui_obj);
    ObjAction *d = gui_obj->action("pre_draw");
    if (d) d->Act(p);
  }

  // draw each of the views
  FORU (i, views.size()) {
    if (i != active_view) {
      draw_view(views[i], false);
    } else {
      draw_view(views[active_view], true);
    }
  }

  //  draw_view(views[active_view], true);
  
  if (net_active) {

    SDL_Rect r;
    compute_rect(button1_down_x, button1_down_y, last_mouse_x, last_mouse_y, r);

    // intersect net rectangle with active viewport

    rect_intersection(views[active_view].viewport, r, r);
    if (r.w > 0 && r.h > 0) draw_rectangle(r, 1, Col3(0,0,0));
    r.x++;
    r.y++;
    rect_intersection(views[active_view].viewport, r, r);
    if (r.w > 0 && r.h > 0) draw_rectangle(r, 1, Col3(1,1,1));
  }

  if (slow_view >= 0) {
    draw_view(views[slow_view], false);
  }

  
  octree->add_count = 0;
  octree->remove_count = 0;    


  gui_obj->set_int("tick", game->tick());

  static sint4 start = time(0);
  gui_obj->set_int("time", time(0)-start);
  
  // draw the scripted 2d overlay
  //==============================
  SActParam p;
  p.obj.push_back(gui_obj);
  ObjAction *d = gui_obj->action("draw");

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  glOrtho(0, opt.win_w, 0, opt.win_h, -1000, 1000);    
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  glViewport(0, 0, opt.win_w, opt.win_h);
  
  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_DEPTH_TEST);

  glEnable(GL_TEXTURE_2D);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glColor4f(1,1,1,1);

  if (d) d->Act(p);

  if (opt.fps) {
    float foo = rfps;
    ostringstream oss;
    form(oss, "%8.2f FPS", foo);
    
    glColor3f(1,1,1);
    glDisable(GL_TEXTURE_2D);
    render_bitmap_string(20, opt.win_h-100, GLUT_BITMAP_8_BY_13, oss.str().c_str());
    glEnable(GL_TEXTURE_2D);
  }


  
  glDisable(GL_BLEND);

  glDisable(GL_TEXTURE_2D);

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  //==============================
  //==============================
  
  if (game->tick() % 1 == 0) minimap->update_texture();

  if (opt.minimap_on) draw_minimap();

  glutSwapBuffers();
  ++gfx_frames;
  ++Frames;

  {
    GLint t = SDL_GetTicks();
    if (t - T0 >= 200) {
      GLfloat seconds = (t - T0) / 1000.0;
      rfps = Frames / seconds;

      //      printf("%d frames in %g seconds = %g FPS\n", Frames, seconds, fps);
      T0 = t;
      Frames = 0;

      gui_obj->set_int("fps", static_cast<sint4>(rfps+0.5));
    }
  }
}


// compares the z values of picked objects 

struct IdZ_Cmp {

  bool operator()(const IdZ &p1, const IdZ &p2) { return p1.z < p2.z; }

};

//===================================================================

// applies to last view drawn!

static void get_click_spaghetti(const SDL_Rect &vp,
                                sint4 mx, sint4 my,
                                Vec3<real4> &l1, Vec3<real4> &l2)
{
  double ModelMatrix[16]; 
  glGetDoublev(GL_MODELVIEW_MATRIX, ModelMatrix);
  double ProjMatrix[16];
  glGetDoublev(GL_PROJECTION_MATRIX, ProjMatrix);

  GLint viewport[4] = { vp.x, vp.y, vp.w, vp.h };
  GLdouble x, y, z;

  // click-spaghetti (l1,l2)

  gluUnProject(mx, my, 0.5, ModelMatrix, ProjMatrix, viewport, &x, &y, &z);
  l1 = Vec3<real4>(x,y,z);

  gluUnProject(mx, my, 1.5, ModelMatrix, ProjMatrix, viewport, &x, &y, &z);
  l2 = Vec3<real4>(x,y,z);    
}

//===================================================================

uint4 GfxModule::pick(GfxView &v, const SDL_Rect &sel_view_,
                      Vec3<real4> &intersection, Vector<IdZ> &ids, bool pick_objs)
{
  SDL_Rect sel_view = sel_view_;

  ids.clear();
  Vector<IdZ> g_ids;
  
  intersection.x = intersection.y = intersection.z = -1;
  
  const sint4 BUFSIZE = 4096;
  GLuint select_buffer[BUFSIZE];
  glSelectBuffer(BUFSIZE, select_buffer);
  glRenderMode(GL_SELECT);
  glInitNames();
  glPushName(0);

  select_view(v, &sel_view, pick_objs);

  glPopName();
  sint4 hits = glRenderMode(GL_RENDER);
  //  cout << "hits = " << hits << endl;

  if (hits <= 0) {

    // 0 hits => determine intersection with ground plane

    if (ground_intersection(v, sel_view.x, sel_view.y, intersection.x, intersection.z) == 0) {
      REM("OOPS: no intersection?!");
      return 0;
    }

    intersection.y = 0;
    return 1;
  }

  {
    sint4 i, j, names;
    GLuint *ptr;
    
    // printf ("hits = %d\n", hits);
    ptr = static_cast<GLuint *>(select_buffer);
    FOR (i, hits) { /*  for each hit  */
      names = static_cast<int>(*ptr);
      //printf (" number of names for hit = %d\n", names);
      ++ptr;
      uint4 z = *ptr;
      //printf("  z1 is %g;", (real4) *ptr/0x7fffffff);
      ++ptr;
      //printf(" z2 is %g\n", (real4) *ptr/0x7fffffff);
      ++ptr;
      for (j = 0; j < names; ++j, ++ptr) {      /*  for each name */
        //printf ("  hit %d's name is: %d\n", i, *ptr);
        IdZ h = { *ptr, z };
        ids.push_back(h);
      }
    }
  }

  if (ids.empty()) return 0;
  
  sort(ids.begin(), ids.end(), IdZ_Cmp());

  //  FORALL (ids, i) cout << (*i).id << "  z=" << real4((*i).z)/max_sint4 << endl;

  bool narrow = sel_view.w <= NARROW && sel_view.h <= NARROW;
  
  sel_view.x--;
  sel_view.y--;
  sel_view.w += 2;
  sel_view.h += 2;

  if (!ids.empty()) {
    uint4 obj_id = ids[0].id;
    GfxObject *po = GfxObject::id2obj(obj_id >> GfxObject::SUB_ID_LOG);
    
    if (narrow) {
      ids.erase(ids.begin()+1, ids.end()); // select one object only
      intersection = intersect_ray(v, po, &sel_view);
    }
    
    return obj_id;
  }

  uint4 obj_id = ids[0].id;
  GfxObject *po = GfxObject::id2obj(obj_id >> GfxObject::SUB_ID_LOG);
  
  if (narrow) {
    ids.erase(ids.begin()+1, ids.end()); // select one object only
    intersection = intersect_ray(v, po, &sel_view);
  }
  
  return obj_id;
}

//===================================================================

// return true iff error
// works on most recent view drawn

sint4 GfxModule::ground_intersection(const GfxView &v, sint4 mx, sint4 my,
                                     real4 &wx, real4 &wz, real4 height)
{
  Vec3<real4> l1, l2;
  real4 t;

  get_click_spaghetti(v.viewport, mx, my, l1, l2);
  Vec3<real4> point(0,height,0), normal(0,1,0), inter;

  wx = wz = 0; // default when error
  
  // intersect (l1,l2) with ground plane (point,normal)

  if (intersect_line_plane(l1, l2, point, normal, inter, t) != 0) return 0;

  wx = inter.x; wz = inter.z;
  return t < 0 ? -1 : 1;
}

//===================================================================

/** Intersects the current view with the xz plane (x0,y0,x1,y1) at height fh.
    Sets points to the convex polygon that results.
*/

void GfxModule::view_to_plane(real4 x0, real4 x1, real4 y0, real4 y1, real4 fh, Vector<Coor3> &points)
{
  Vector<real4> x;
  Vector<real4> y;

  x.push_back(x0);
  x.push_back(x1);
  x.push_back(x1);
  x.push_back(x0);
  x.push_back(x0);

  y.push_back(y0);
  y.push_back(y0);
  y.push_back(y1);
  y.push_back(y1);
  y.push_back(y0);

  Frustum fr;
  fr.compute();

  FORU (j, 4) { // for each frustum plane (ignore near/far clipping planes)
    
    Vector<real4> x_;
    Vector<real4> y_;

    // for each line segment of the field
    sint4 s = x.size()-1;

    FORS (i, s) {
      const real4* fp = fr.m_Frustum[j];

      Vec3<real4> p1(x[i],fh,y[i]);
      Vec3<real4> p2(x[i+1],fh,y[i+1]);

      real4 k1 = -(fp[0]*p1.x + fp[1]*p1.y + fp[2]*p1.z + fp[3]);
      real4 k2 = -(fp[0]*p2.x + fp[1]*p2.y + fp[2]*p2.z + fp[3]);

      if (k1 > 0 && k2 > 0) continue; // both points are outside

      if (k1 > 0) { // first point is outside
        Vec3<real4> u = p2 - p1;
        Vec3<real4> n(fp[0],fp[1],fp[2]);
        real4 d = n.dot(u);
        
        if (fabs(d) < 1e-6) continue; // almost parallel -> added as end of last segment
        
        real4 t = k1/d;
        p1 += u * t; // intersection point

        x_.push_back(p1.x);
        y_.push_back(p1.z);
        continue;
      }

      // first point is inside
      x_.push_back(x[i]);
      y_.push_back(y[i]);

      if (k2 > 0) { // second point is outside

        Vec3<real4> u = p2 - p1;
        Vec3<real4> n(fp[0],fp[1],fp[2]);
        real4 d = n.dot(u);

        // if almost parallel, p2 is close enough

        if (fabs(d) >= 1e-6) { // not almost parallel
          real4 t = k1/d;
          p2 = p1 + u * t; // intersection point
        }
        x_.push_back(p2.x);
        y_.push_back(p2.z);
      }
    }
    if (x_.size() > 0) {
      x_.push_back(x_[0]);
      y_.push_back(y_[0]);
    }
    x = x_;
    y = y_;
  }

  points.clear();
  FORS (i, x.size()-1) {
    points.push_back(Coor3(x[i], fh, y[i]));
  }
}

//===================================================================

struct CompareGfxObj {

  bool operator()(GfxObject * const &p1, GfxObject * const &p2) {
    return p1->order()<p2->order();
  }

};

//-------------------------------------------------------------------

void GfxModule::select_view(GfxView &v, const SDL_Rect *sel_view, bool draw_objs)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();


  GLint viewport[4] = { v.viewport.x, v.viewport.y, v.viewport.w, v.viewport.h };

  gluPickMatrix(sel_view->x+sel_view->w/2,
                sel_view->y+sel_view->h/2,
                sel_view->w, sel_view->h, viewport);

  init_view_camera(v);

  Vector<GfxObject*> to_be_drawn;
  to_be_drawn.reserve(10000);
 
  // draw models
  {
    glPushMatrix();

    octree->draw(to_be_drawn);
    if (draw_objs) {
      FORALL (to_be_drawn, i) (*i)->pick();
    } else {
      FORALL (to_be_drawn, i) if ((*i)->order() <= GfxObject::O_WATER) (*i)->pick();
    }

    glPopMatrix();
  }

  glerror("select_view");
}

//===================================================================

// find the intersection of a ray with an object  

Vec3<real4> GfxModule::intersect_ray(GfxView &v, GfxObject *o, const SDL_Rect *sel_view)
{
  assert(o);
  {
    GameObj *g = GfxGameObject::get_gob(o);
    if (g && *g->sod.shape == Object::CIRCLE) return o->midpoint();
  }
  glDisable(GL_LIGHTING);
  glActiveTextureARB(GL_TEXTURE1);
  glDisable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE0);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_COLOR_MATERIAL);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glViewport(v.viewport.x, v.viewport.y, v.viewport.w, v.viewport.h);
  
  init_view_camera(v);

  glScissor(sel_view->x, sel_view->y, sel_view->w, sel_view->h);
  glEnable(GL_SCISSOR_TEST);

  glClearColor(0,0,0,1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // draw model
  o->pick();

  glDisable(GL_SCISSOR_TEST);

//  Vec3<real4> pixels[sel_view->w][sel_view->h];
  Vec3<real4> *pixels = new Vec3<real4>[sel_view->w * sel_view->h];
//  glReadPixels(sel_view->x, sel_view->y, sel_view->w, sel_view->h, GL_RGB, GL_FLOAT, pixels[0][0]); 
  glReadPixels(sel_view->x, sel_view->y, sel_view->w, sel_view->h, GL_RGB, GL_FLOAT, pixels[0].ptr()); 

  glerror("intersect_ray");

  sint4 w = sel_view->w;

  for(int i=0;i<sel_view->w;i++)
    for(int j=0;j<sel_view->h;j++)
    {
//      if(pixels[i][j].r || pixels[i][j].g || pixels[i][j].b)
      if (pixels[i*w+j].r || pixels[i*w+j].g || pixels[i*w+j].b) {
	return o->aabb_o + o->aabb_d * pixels[i*w+j]; //pixels[i][j];
      }
    }
  return o->midpoint();
}

//===================================================================

void GfxModule::draw_sky(GfxView &v) {
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  
  const real4 f = 1.0;
  glOrtho(-f, f, -f, f, 0, 10);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  glDepthMask(GL_FALSE);
  glDisable(GL_LIGHTING);

  glRotatef(v.x_angle, 1.0, 0.0, 0.0);
  //  glRotatef(v.y_angle, 0.0, 1.0, 0.0);

  const real4 s = 4.0;

  real4 c0[4] = { 0.0, 0.0, 0.7, 1.0 }; // zenith
  real4 c1[4] = { 0.5, 0.5, 1.0, 1.0 }; // horizon

  real4 c2[4] = { 0.2, 0.25, 0.2, 1.0 }; // horizon 2
  real4 c3[4] = { 0.0, 0.0, 0.0, 1.0 }; // nadir

  glTranslatef(0, -0.4, 0); // raise the terrain

  glBegin(GL_QUADS);
  glColor4fv(c0);
  glVertex3f(-s, s, 0);
  glVertex3f( s, s, 0);
  glColor4fv(c1);
  glVertex3f( s, 0, s);
  glVertex3f(-s, 0, s);

  glColor4fv(c2);
  glVertex3f(-s, 0, s);
  glVertex3f( s, 0, s);
  glColor4fv(c3);
  glVertex3f( s, -s, 0);
  glVertex3f(-s, -s, 0);


  glColor4fv(c0);
  glVertex3f( s, s, 0);
  glVertex3f(-s, s, 0);
  glColor4fv(c1);
  glVertex3f(-s, 0, -s);
  glVertex3f( s, 0, -s);

  glColor4fv(c2);
  glVertex3f( s, 0, -s);
  glVertex3f(-s, 0, -s);
  glColor4fv(c3);
  glVertex3f(-s, -s, 0);
  glVertex3f( s, -s, 0);
  glEnd();

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  glEnable(GL_LIGHTING);
  glDepthMask(GL_TRUE);
}

//===================================================================

/** Set position and parameters of light0 and ambient light. */

void GfxModule::init_light0()
{
  // ambient light parameters
  GLfloat ambientLight[] = { opt.ambient_light, opt.ambient_light, opt.ambient_light, 1.0 };
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight);

  // diffuse light parameters  
  GLfloat diffuseLight[] = { opt.diffuse_light, opt.diffuse_light, opt.diffuse_light, 1.0 };  
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);

  // specular light parameters
  GLfloat specularLight[] = { opt.specular_light, opt.specular_light, opt.specular_light, 1.0 };
  glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);

  GLfloat lightPos[] = { -1, 2, -1, 0 };         // light position
  glLightfv(GL_LIGHT0, GL_POSITION, lightPos);

  GLfloat specularReflection[] = { 0, 0, 0, 0};  // set specularity
  glMaterialfv(GL_FRONT, GL_SPECULAR, specularReflection);
  glMateriali(GL_FRONT, GL_SHININESS, 0);        // specular exponent

  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientLight);
}

//===================================================================

void GfxModule::init_view_camera(GfxView &v)
{
  real4 ztrans = 0;

  if (opt.perspective_mode) {
    real4 fov = 35;
    gluPerspective(fov, real4(v.viewport.w)/real4(v.viewport.h), 1, 1000);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    ztrans = -0.4*max_dim/tan(fov*(M_PI/180.0)*0.5)*v.scale;
    glTranslatef(0, 0, ztrans);
  } else {
    
    // [-1,1]^2 is mapped to (x,y)-(x+width,y+height)
    // scaling corrects aspect ratio
    sint4 d = v.viewport.w - v.viewport.h;
    if      (d > 0) glScalef(1.0, real4(v.viewport.w)/v.viewport.h, 1.0);
    else if (d < 0) glScalef(real4(v.viewport.h)/v.viewport.w, 1.0, 1.0);

    real4 f = 0.5*0.75*v.scale;

    glOrtho(-max_dim*f, max_dim*f,
            -max_dim*f, max_dim*f,
            -max_dim*10, max_dim*10);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
  }

  glRotatef(v.x_angle, 1.0, 0.0, 0.0);
  glRotatef(v.y_angle, 0.0, 1.0, 0.0);

  glTranslatef(v.x_off-dim.x*0.5,
               v.y_off,
               v.z_off-dim.z*0.5);
}

//===================================================================

void GfxModule::draw_view(GfxView &v, bool active)
{
  PROFILE("GfxModule::draw_view");

  cur_view = &v;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  glScissor(v.viewport.x, v.viewport.y, v.viewport.w, v.viewport.h);
  glEnable(GL_SCISSOR_TEST);

  init_view_camera(v);

  init_light0(); // sets light position and ambient light
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);


  Vector<GfxObject*>::iterator i;
  Vector<GfxObject*> to_be_drawn;
  to_be_drawn.reserve(10000);
  real4 minb[2];
  real4 maxb[2];
  sint4 shadow_width=0, shadow_height=0;

  {
    PROFILE("build shadows");
    static uint8 it = 0;
    ++it;
    
    CycleCounter c1;

    // compute corners (for the minimap)

    const SDL_Rect &r = v.viewport;
    ground_intersection(r, r.x,     r.y,     v.c1.x, v.c1.y);
    ground_intersection(r, r.x,     r.y+r.h, v.c2.x, v.c2.y);
    ground_intersection(r, r.x+r.w, r.y+r.h, v.c3.x, v.c3.y);
    ground_intersection(r, r.x+r.w, r.y,     v.c4.x, v.c4.y);

    // constrain the shadow map to the gamefield
    real4 max_w = game->get_map().get_width();
    real4 max_h = game->get_map().get_height();


    Vector<Coor3> points;
    view_to_plane(0, max_w, 0, max_h, 5.0, points);

    minb[0] = max_w;
    minb[1] = max_h;
    maxb[0] = 0;
    maxb[1] = 0;

    FORALL (points, i) {
      minb[0] = min(minb[0], (*i).x);
      minb[1] = min(minb[1], (*i).z);
      maxb[0] = max(maxb[0], (*i).x);
      maxb[1] = max(maxb[1], (*i).z);
    }

    view_to_plane(0, max_w, 0, max_h, 0.0, points);

    FORALL (points, i) {
      minb[0] = min(minb[0], (*i).x);
      minb[1] = min(minb[1], (*i).z);
      maxb[0] = max(maxb[0], (*i).x);
      maxb[1] = max(maxb[1], (*i).z);
    }

    v.corners = points;

    // compute the bounds of the visible region

    if (minb[0] > maxb[0]) minb[0] = maxb[0];
    if (minb[1] > maxb[1]) minb[1] = maxb[1];

    if (minb[0] == maxb[0]) maxb[0] = minb[0] + 1;
    if (minb[1] == maxb[1]) maxb[1] = minb[1] + 1;

    real4 center[2] = { (maxb[0]+minb[0])/2, (maxb[0]+minb[0])/2 };

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(minb[0]-center[0],
            maxb[0]-center[0],
            minb[1]-center[1],
            maxb[1]-center[1],
            -100, 100);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glRotatef(-90,1,0,0);
    glTranslatef(-center[0],0,-center[1]);

    // build new shadow textures ===================

    shadow_width = min(static_cast<sint4>(v.viewport.w), opt.shadow_res);
    shadow_height = min(static_cast<sint4>(v.viewport.h), opt.shadow_res);
    glViewport(v.viewport.x, v.viewport.y, shadow_width, shadow_height);

    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);

    // draw the minimap visibility texture ---------
    glDisable(GL_BLEND);
    glColor4f(0,0,0,1);
    glBindTexture(GL_TEXTURE_2D, minimap->get_visible_tex_id());

    glBegin(GL_QUADS);
    
    glTexCoord2f(minimap->scale_w, 0);
    glVertex3f(game->get_map().get_width(), 0, 0);
    glTexCoord2f(minimap->scale_w, minimap->scale_h);
    glVertex3f(game->get_map().get_width(), 0, game->get_map().get_height());
    glTexCoord2f(0, minimap->scale_h);
    glVertex3f(0, 0, game->get_map().get_height());
    glTexCoord2f(0, 0);
    glVertex3f(0, 0, 0);

    glEnd();
    glColor4f(0,0,0,1);

    // ---------------------------------------------
  
    glEnable(GL_BLEND);

    octree->draw(to_be_drawn);

    // ------- find object type transitions, update, draw shadows -------

    sort(to_be_drawn.begin(), to_be_drawn.end(), CompareGfxObj());
    Vector<GfxObject*>::iterator i = to_be_drawn.begin();

    // skip terrain
    for(; i != to_be_drawn.end() && (*i)->order() == GfxObject::O_TERRAIN; ++i);

    // animate water
    for (; i != to_be_drawn.end() && (*i)->order() == GfxObject::O_WATER; ++i);// (*i)->update(dt);


    bool one_pass_shadows = glBlendEquationSeparateEXT && glBlendFuncSeparateEXT;

    if (one_pass_shadows) {
      glBlendEquationSeparateEXT(GL_FUNC_ADD, GL_MAX);
      glBlendFuncSeparateEXT(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE);
    } else {
      glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_FALSE);
    }

    Vector<GfxObject*>::iterator objects_start=i;

    // draw shadows for ground objects
    for (; i != to_be_drawn.end() && (*i)->order() != GfxObject::O_IN_AIR; i++) {
      (*i)->draw_shadow();
    }

    // draw shadows for flying
    Vector<GfxObject*>::iterator flying_start=i;
    for (; i != to_be_drawn.end() && (*i)->order() == GfxObject::O_IN_AIR; ++i) {
      (*i)->draw_shadow();
    }

    Vector<GfxObject*>::iterator objects_end=i;

    // update particle systems
    for(; i != to_be_drawn.end(); ++i); // (*i)->update(dt);

    // ------------------------------------------------------------------

    // add the alpha channel to the shadow texture (if doing a 2nd pass)
    
    if (!one_pass_shadows) {

      glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_TRUE);
      glBlendFunc(GL_ONE,GL_ONE);
      
      if (glBlendEquationEXT) {
        glBlendEquationEXT(GL_MAX_EXT);
      } else {
        glBlendEquation(GL_MAX);
      }
      
      for (i = objects_start; i != objects_end; ++i) (*i)->draw_shadow();
      
      glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);

    }

    // draw special things on the ground -----------

    const sint4 TP  = gsm->get_game().get_tile_points();
    const real4 ITP = 1.0/TP;

    glDisable(GL_CULL_FACE);
    glActiveTextureARB(GL_TEXTURE1);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0);
    glDisable(GL_TEXTURE_2D);


    FORALL (marked_rectangles, it) {
      // coloured rectangles with coordinates in tile points
      ColRect *t = *it;

      glColor4f(t->r, t->g, t->b, 0.5);
      glBegin(GL_QUADS);
      glVertex3f(t->x0 * ITP, 0, t->y0 * ITP);
      glVertex3f(t->x1 * ITP, 0, t->y0 * ITP);
      glVertex3f(t->x1 * ITP, 0, t->y1 * ITP);
      glVertex3f(t->x0 * ITP, 0, t->y1 * ITP);
      glEnd();
    }

    glEnable(GL_CULL_FACE);

    // create the new ground shadow texture --------
    glBindTexture(GL_TEXTURE_2D, shadow_tex[0]);
    glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
                        v.viewport.x, v.viewport.y,
                        shadow_width,shadow_height);


    // draw the minimap visibility texture ---------
    //   (Visibility is an alpha texture.  With blending disabled
    //   drawing clears to black with opacity only for unseen tiles)

    glActiveTextureARB(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glColor4f(0,0,0,1);
    glBindTexture(GL_TEXTURE_2D, minimap->get_visible_tex_id());

    glBegin(GL_QUADS);
    
    glTexCoord2f(minimap->scale_w, 0);
    glVertex3f(game->get_map().get_width(), 0, 0);
    glTexCoord2f(minimap->scale_w, minimap->scale_h);
    glVertex3f(game->get_map().get_width(), 0, game->get_map().get_height());
    glTexCoord2f(0, minimap->scale_h);
    glVertex3f(0, 0, game->get_map().get_height());
    glTexCoord2f(0, 0);
    glVertex3f(0, 0, 0);

    glEnd();
    glColor4f(0,0,0,1);
    glEnable(GL_BLEND);

    // ---------------------------------------------
  
    // draw shadows for flying
    //   with 2-pass shadows glBlendEquation is still GL_MAX for all channels
    //   and glBlendFunc is GL_ONE, GL_ONE

    for(i=flying_start;i != to_be_drawn.end() && (*i)->order() == GfxObject::O_IN_AIR; ++i) {
      (*i)->draw_shadow();
    }

    // create the new flying shadow texture
    glBindTexture(GL_TEXTURE_2D, shadow_tex[1]);
    glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
                        v.viewport.x, v.viewport.y,
                        shadow_width, shadow_height);

    // =============================================

    // done with creating shadows; restore standard blending, etc.

    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

    if (glBlendEquationEXT) {
      glBlendEquationEXT(GL_FUNC_ADD_EXT);
    } else {
      glBlendEquation(GL_FUNC_ADD);
    }

    glDisable(GL_BLEND);
    glEnable(GL_LIGHTING);
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glViewport(v.viewport.x, v.viewport.y, v.viewport.w, v.viewport.h);
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  } // end build shadows

  // Render blur effect
  if (opt.blur > 0) {
    // Render special effects to the blur texture
    glViewport(0, 0, opt.blur_res, opt.blur_res);
    glClearColor(0,0,0,0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    // Render terrain to the depth buffer
    glDepthMask(GL_TRUE);
    glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
    terrain->beginDraw();
    for (i=to_be_drawn.begin();
         i != to_be_drawn.end() && (*i)->order()==GfxObject::O_TERRAIN;
         ++i)
    {
      (*i)->draw();
    }
    terrain->endDraw();
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

    // Render particle effects to texture
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glEnable(GL_BLEND);
    
    //FORALL (to_be_drawn,i2) (*i2)->draw_damage();
    
    GfxEmitter::beginDraw();
    FORALL (to_be_drawn, i2) (*i2)->draw_emitters();
    GfxEmitter::endDraw();
    
    glDisable(GL_BLEND); 
    
    // Fill-up the accumulation buffer if it's not already full
    if (blur_buffer < opt.blur)
      while (blur_buffer++ < opt.blur)
        glAccum(GL_ACCUM, 1.f/opt.blur);
    
    // Add the scene to the accumulation buffer
    glAccum(GL_ACCUM, 1.f/opt.blur);
    // Fetch and display with a slight gamma boost
    glAccum(GL_RETURN, 1.f + (1.f- (0.5f/opt.blur)));
    // Capture the scene to texture
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, blur_tex);
    glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
                        0, 0,                        // x, y
                        opt.blur_res, opt.blur_res); // width, height
    
    // Subtract the last scene from the accumulation buffer
    // to keep the real-time blur continuous.
    glAccum(GL_ACCUM, -1.f/opt.blur);
    
    glerror("draw_view blur effect");
  }
  
  // Reset frame buffer to its normal size and clear
  glViewport(v.viewport.x, v.viewport.y, v.viewport.w, v.viewport.h);
  glClearColor(0,0,0,1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  //  draw_sky(v);
  
  {
    static uint8 it = 0;
    ++it;

    glActiveTextureARB(GL_TEXTURE1);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, shadow_tex[0]); // shadows cast on the ground

    real4 scx = static_cast<real4>(shadow_width) / opt.shadow_res;
    real4 scy = static_cast<real4>(shadow_height) / opt.shadow_res;

    GLfloat texgen_s[] = { scx/(maxb[0]-minb[0]), 0, 0, -minb[0]*scx/(maxb[0]-minb[0]) };
    GLfloat texgen_t[] = { 0, 0, scy/(maxb[1]-minb[1]), -minb[1]*scy/(maxb[1]-minb[1]) };
    
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    
    glTexGeni (GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
    glTexGenfv(GL_S, GL_EYE_PLANE, texgen_s);
    glTexGeni (GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
    glTexGenfv(GL_T, GL_EYE_PLANE, texgen_t);

    if (wireframe)
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    else
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    // terrain --------------------

    {
      PROFILE("terrain");
      terrain->beginDraw();

      glActiveTextureARB(GL_TEXTURE1);
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
      glActiveTextureARB(GL_TEXTURE0);
      
      for(i = to_be_drawn.begin();
          i != to_be_drawn.end() && (*i)->order() == GfxObject::O_TERRAIN;
          ++i)
        {
          (*i)->draw();
        }
      terrain->endDraw();
    }

    // water tiles ----------------

    {
      PROFILE("water tiles");
      WaterTile::beginDraw();
      for(;i != to_be_drawn.end() && (*i)->order() == GfxObject::O_WATER; ++i) {
        (*i)->draw();
      }
      WaterTile::endDraw();
    }
    
    // marked tiles ---------------
    {
      PROFILE("marked tiles");
      glActiveTextureARB(GL_TEXTURE1);
      glDisable(GL_TEXTURE_2D);
      glActiveTextureARB(GL_TEXTURE0);
      glDisable(GL_TEXTURE_2D);
      
      glBlendFunc(GL_SRC_ALPHA, GL_ONE);
      glEnable(GL_BLEND);
      glDisable(GL_DEPTH_TEST);
      glDisable(GL_COLOR_MATERIAL);
      glDisable(GL_CULL_FACE);
      
      //    glDisable(GL_LIGHTING);
      
      FORALL (marked_tiles, it) {
        
        ColorTileSet *t = *it;
        
        FORALL (t->tiles, i) if ((*i)->is_initialized()) {
          glColor4f(t->r, t->g, t->b, 0.5); // color is reset by draw_outline()
          (*i)->draw_flat();
        }
      }
      
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glDisable(GL_BLEND);
      glEnable(GL_DEPTH_TEST);
      glEnable(GL_COLOR_MATERIAL);
      glEnable(GL_CULL_FACE);
    }
    
    //-----------------------------

    glerror("draw_view 1");
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GEQUAL, 0.3f);

    // draw models
    {
      PROFILE("models");
      glPushMatrix();

      GfxAnimesh::beginDraw();

      // land objects

      {
        PROFILE("land objects");

        glActiveTextureARB(GL_TEXTURE1);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D,shadow_tex[1]); // shadows cast by flying objects
        glActiveTextureARB(GL_TEXTURE0);
        
        if (wireframe) {
          glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
          FORALL (game_objs, it) {
            (*it)->draw_shape();
          }
          glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        } else {
          
          Vector<GfxObject*>::iterator objects_start = i;
          
          if (opt.draw_occluded) {
            
            glCullFace(GL_BACK);
            //glBlendFunc(GL_ONE, GL_SRC_ALPHA);
            glBlendFunc(GL_DST_COLOR, GL_SRC_ALPHA);
            
            glDisable(GL_DEPTH_TEST);
            
            glActiveTextureARB(GL_TEXTURE1);
            glDisable(GL_TEXTURE_2D);
            glActiveTextureARB(GL_TEXTURE0);
            glDisable(GL_TEXTURE_2D);
            
            const Vector<Col3> &col = get_player_colors();
            
            for (; i != to_be_drawn.end() && (*i)->zcat == GfxObject::O_ON_LAND; ++i) {
              
              GameObj *gob = GfxGameObject::get_gob(*i);
              const sintptr *o = 0;
              if (gob) o = gob->get_const_int_ptr("owner");
              
              if (o) {
                glColor4f(col[*o].r, col[*o].g, col[*o].b, 0.5);
              } else {
                glColor4f(1, 1, 1, 0.5);
              }
              (*i)->draw();
            }
            
            glCullFace(GL_FRONT);
            glColor4f(1, 1, 1, 1);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            glEnable(GL_DEPTH_TEST);
            
            glActiveTextureARB(GL_TEXTURE1);
            glEnable(GL_TEXTURE_2D);
            glActiveTextureARB(GL_TEXTURE0);
            glEnable(GL_TEXTURE_2D);
          }
          
          for (i = objects_start; i != to_be_drawn.end() && (*i)->zcat == GfxObject::O_ON_LAND; ++i) {
            (*i)->draw();
          }
        }
      }

      glActiveTextureARB(GL_TEXTURE1);
      glDisable(GL_TEXTURE_2D);
      glActiveTextureARB(GL_TEXTURE0);

      // air objects

      {
        PROFILE("air objects");

        if (wireframe) {
          
          glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
          for(; i != to_be_drawn.end() && (*i)->zcat == GfxObject::O_IN_AIR; ++i) {
            GfxGameObject *gfx = GfxGameObject::get_gfx(*i);
            gfx->draw_shape();
          }
          glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
          
        } else {
          
          for (; i != to_be_drawn.end() && (*i)->zcat == GfxObject::O_IN_AIR; ++i) {
            (*i)->draw();
          }
        }
      }
      
      GfxAnimesh::endDraw();
      glPopMatrix();
      glDisable(GL_ALPHA_TEST);
      
      glActiveTextureARB(GL_TEXTURE1);
      glDisable(GL_TEXTURE_2D);
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); // restore texture blending
      
      // Draw emitters and damage indicators
      glEnable(GL_BLEND);
      FORALL (to_be_drawn, i2) {
        GfxGameObject *gfx = GfxGameObject::get_gfx(*i2);
        if (gfx) gfx->draw_damage();
      }
      
      // If blur effect is disabled, render effects normally
      if (opt.blur <= 0) {
      	GfxEmitter::beginDraw();
      	FORALL (to_be_drawn,i2) (*i2)->draw_emitters();
      	GfxEmitter::endDraw();
        
      } else {
        // Draw special effects blur texture over the entire screen
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix(); glLoadIdentity();
        glMatrixMode(GL_PROJECTION);
        glPushMatrix(); glLoadIdentity();
        
        glOrtho(0, 1, 0, 1, 0, 1);
        
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_CULL_FACE);
        
        glActiveTextureARB(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D,blur_tex);
        glEnable(GL_TEXTURE_2D);
        
        glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
        
        glBegin(GL_QUADS);	
        glTexCoord2f(0.f, 0.f);
        glVertex2f  (0.f, 0.f); // top left
        
        glTexCoord2f(0.f, 1.f);
        glVertex2f  (0.f, 1.f); // top right
        
        glTexCoord2f(1.f, 1.f);
        glVertex2f  (1.f, 1.f); // bottom right
        
        glTexCoord2f(1.f, 0.f);
        glVertex2f  (1.f, 0.f); // bottom left
        glEnd();
        
        glEnable(GL_DEPTH_TEST);
        glDisable(GL_TEXTURE_2D);
        
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
        glPopMatrix();
      }
      glDisable(GL_BLEND);
      
    }
    glerror("draw_view 2");

    glActiveTextureARB(GL_TEXTURE1);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glActiveTextureARB(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);

    //-------------------------------------
    // signal listeners to draw
    {
      PROFILE("DRAW_MSG");
      call_handlers(Event(FROM, DRAW_MSG));
    }
    //-------------------------------------

    glActiveTextureARB(GL_TEXTURE1);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glActiveTextureARB(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);

    // scripted view overlay

    {
      PROFILE("script overlay");
      
      ObjAction *d = gui_obj->action("draw_view");
      
      if (d) {
        SActParam p;
        p.obj.push_back(gui_obj);
        
        glDisable(GL_DEPTH_TEST);
        glEnable(GL_COLOR_MATERIAL);
        glColor3f(1, 1, 1);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        
        d->Act(p); // view overlay
        
        glEnable(GL_DEPTH_TEST);
        glEnable(GL_COLOR_MATERIAL);
      }
      
      glDisable(GL_BLEND);
      glEnable(GL_LIGHTING);
      glDisable(GL_TEXTURE_2D);
    }

    //  glScissor(0, 0, win_w, win_h);

    glDisable(GL_SCISSOR_TEST);

    // draw frame
    SDL_Rect r = v.viewport;
    sint4 width = 2;
    r.x += 1; r.w -= 2;
    r.y += 1; r.h -= 2;
    
    // glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    draw_rectangle(r, width, active ? view_colors[0] : v.frame_color);
  }

  glerror("draw_view");
}

//===================================================================

// draw 2d rectangle on top of views

void GfxModule::draw_rectangle(const SDL_Rect &r, sint4 width, const Col3 &col)
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glViewport(0, 0, opt.win_w, opt.win_h);
  glOrtho(0, opt.win_w, 0, opt.win_h, 0, 1 );

  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);

  glLoadName(NO_OBJ_ID);

  glColor3f(col.r, col.g, col.b);

  glBegin(GL_QUAD_STRIP);

  glVertex2f(r.x,           r.y+r.h); // top left
  glVertex2f(r.x+width,     r.y+r.h-width);

  glVertex2f(r.x+r.w,       r.y+r.h); // top right
  glVertex2f(r.x+r.w-width, r.y+r.h-width);

  glVertex2f(r.x+r.w,       r.y);     // bottom right
  glVertex2f(r.x+r.w-width, r.y+width);

  glVertex2f(r.x,           r.y);     // bottom left
  glVertex2f(r.x+width,     r.y+width);

  glVertex2f(r.x,           r.y+r.h); // top left
  glVertex2f(r.x+width,     r.y+r.h-width);

  glEnd();
  glEnable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();  
}

//===================================================================
