#ifndef GfxModule_H
#define GfxModule_H

// $Id: GfxModule.H,v 1.36 2006/06/15 05:26:38 orts_mburo Exp $

// This is an ORTS file
// (c) Michael Buro
// (c) Tim Furtak
// (c) Keith Yerex
// (c) Sami Wagiaalla
// (c) Erik Benner
// licensed under the GPL

// the former GfxClient

#include "GfxGlobal.H"
#include "ClientModule.H"
#include "Action.H"
#include "FPS.H"
#include "SplashScreen.H"
#include "Minimap.H"
#include "SoundMixer.H"
#include "GfxView.H"
#include "LoopFunctor.H"

#define NO_OBJ_ID 1 // openGL object id for drawing things that aren't objects
#define USE_LIGHT 1

#define GFX_VERSION "0.95"

class CWidget;
class Octree;
class Tile;
class GfxTerrain;
class GfxTerrainTile;
class Minimap;
class Tokenizer;
class ScriptFunctionDef;
class GfxGlutAdaptor;
class GameStateModule;
class Options;
class GfxObject;
class GfxGameObject;

// object id + z value

struct IdZ {
  uint4 id;
  uint4 z;
};

//===================================================================
//===================================================================

extern GfxModule *GUI_gfxm;  // gives GUI functions access to options (fixme: hack)

stype GUI_strcmp           (const SActParam&);
stype GUI_printf           (const SActParam&);
stype GUI_text             (const SActParam&);
stype GUI_textf            (const SActParam&);
stype GUI_Color            (const SActParam&);
stype GUI_Bitmap           (const SActParam&);
stype GUI_Rect             (const SActParam&);
//stype GUI_SetTexture       (const SActParam&);

stype GUI_Outline          (const SActParam&);

//sami-------------------------------------------------------
stype GUI_psin             (const SActParam&);
stype GUI_pcos             (const SActParam&);

stype GUI_set_cursor       (const SActParam&);
stype GUI_error            (const SActParam&);

void* get_GLUT_Font(sint4 script_font);
void render_bitmap_string(GLfloat, GLfloat, void*, const char*);
//-----------------------------------------------------------


GLenum glerror(char *);

//===================================================================
//===================================================================

class GfxModule : public ClientModule {

public:

  // public to allow REGISTER_TYPEOF definitions
  typedef std::map<std::string, Sound*> Name2SoundMap;
  typedef std::map<GameObj*, GfxGameObject*> GameObj2GfxGameObjMap;
  typedef std::map<std::string, std::string> String2StringMap;

  struct Options {

    sint4 win_w;
    sint4 win_h;
    bool  perspective_mode;
    bool  minimap_on;
    bool  minimap_rot;
    sint4 shadow_res;
    sint4 texture_level;
    sint4 blur_res;
    sint4 blur;  // # of frames to include in the blur effect
    bool  no_sound;
    bool  splash;
    sint4 model_mode;
    sint4 texture_mode;
    bool  cheap_shadows;
    real8 terrain_gamma;
    real8 model_gamma;
    real8 terrain_noise1;
    real8 terrain_noise2;
    std::string terrain_vertex_shader;
    std::string terrain_fragment_shader;
    real8 ambient_light;
    real8 diffuse_light;
    real8 specular_light;
    std::string gui_file;
    std::string path;
    std::string gfx_path;
    std::string db_file;
    bool  full_screen;
    bool  fps;
    bool  draw_occluded;
    bool  verbose_load;
    sint4 seed;
    bool  refresh;
    std::string tileset;
    bool  clear_fast;
    
    Options();
    static void add();
  };

  // messages
  static const sint4 FROM; 
  // static const sint4 ADD_CWIDGETS_MSG; // now cwidgets can be addedstatic
  static const sint4 DRAW_MSG;
  static const sint4 SCRIPT_MSG;

  GfxModule();
  virtual ~GfxModule();

  std::string version() const { return GFX_VERSION; }
  
  Options opt;
  Random rand;
  
  void init(GameStateModule &gsm_, const Options &opt_);
  void load();

  GameStateModule &get_gsm() { return *gsm; }
  bool get_quit() const { return quit; }
  bool get_loaded() const { return loading; }
  
  void save_terrain();

  void draw();
  void splash();
  void loop();
  void start_loop(LoopFunctor* l);

  // GLUT events ----------------------------
  void event_key_down(unsigned char key, int x, int y);
  void event_key_up  (unsigned char key, int x, int y);

  void event_special_key_down(int key, int x, int y);
  void event_special_key_up  (int key, int x, int y);

  void event_mouse_down(int btn, int state, int x, int y);
  void event_mouse_up  (int btn, int state, int x, int y);
  void event_mouse_motion(int x, int y);
  void event_resize(int width, int height);
  //  void event_expose();
  //  void event_active();
  // ----------------------------------------

  // called after view has arrived, updates gfx world view
  void process_changes();
  void update_selected();
  void update_id_list();

  void add_GfxObject(GfxObject *go);
  void add_GfxGameObject(GfxGameObject *ggo);
  void del_GfxObject(GfxObject *go);
  void del_GfxGameObject(GfxGameObject *ggo);

  const Vector<Col3>& get_player_colors() const { return player_colors; }

  stype GUI_CreateView    (const SActParam&);
  stype GUI_ResizeView    (const SActParam&);
  stype GUI_DeleteView    (const SActParam&);

  stype GUI_Overlay       (const SActParam&);
  stype GUI_start_overlay (const SActParam&);
  stype GUI_end_overlay   (const SActParam&);

  stype GUI_Screen2World  (const SActParam&);
  stype GUI_Screen2Obj    (const SActParam&);
  stype GUI_MinimapPos    (const SActParam&);
  stype GUI_DrawModel2D   (const SActParam&);

  stype GUI_MarkTiles     (const SActParam&);
  stype GUI_UnmarkTiles   (const SActParam&);
  stype GUI_MarkRect      (const SActParam&);
  stype GUI_EditRect      (const SActParam&);
  stype GUI_UnmarkRect    (const SActParam&);
  stype GUI_FlattenTerrain(const SActParam&);
  stype GUI_UnFlattenTerrain(const SActParam&);

  stype GUI_LoadBP        (const SActParam&);

  stype GUI_Deselect_id   (const SActParam&);

  stype GUI_Get_DBVal     (const SActParam&);
  stype GUI_Set_DBVal     (const SActParam&);
  stype GUI_Get_DBVec     (const SActParam&);
  stype GUI_Set_DBVec     (const SActParam&);

  stype GUI_load_sound    (const SActParam&);
  stype GUI_play_sound    (const SActParam&);
  stype GUI_quitGame      (const SActParam&);
  stype GUI_toggle_pathfinding (const SActParam&);

  stype GUI_broadcast     (const SActParam&);

  void overlay_mode();
//  void overlay_mode_end();

  void add_cwidget(CWidget &widget);
  void rem_cwidget(CWidget &widget);
  void finish_cwidget();
  Vector<CWidget*> cwidget_queue; ///< Waiting to be added.

  FPS fps;

  void play_sound(Sound &sound, real4 lvol, real4 rvol);
  Sound* load_sound(const std::string &filename);

  //---------------------

  struct ColorTileSet {
    real4 r,g,b;
    Vector<GfxTerrainTile*> tiles;
  };
  
  struct ColRect {
    real4 r,g,b;
    sint4 x0, x1, y0, y1;
  };

  //---------------------

  static const sint4 STYPE_GFXOBJ_ID;
  static const sint4 STYPE_SOUND_ID;
  static const sint4 STYPE_COLORTILESET_ID;
  static const sint4 STYPE_COLRECT_ID;   
  
  static GfxObject *stype_to_GfxObject(const stype &s) {
    return static_cast<GfxObject*>(s.to_special(STYPE_GFXOBJ_ID));
  }
  
  static Sound *stype_to_Sound(const stype &s) {
    return static_cast<Sound*>(s.to_special(STYPE_SOUND_ID));
  }
  
  static ColorTileSet *stype_to_ColorTileSet(const stype &s) {
    return static_cast<ColorTileSet*>(s.to_special(STYPE_COLORTILESET_ID));
  }
  
  static ColRect *stype_to_ColRect(const stype &s) {
    return static_cast<ColRect*>(s.to_special(STYPE_COLRECT_ID));
  }

protected:

  void add_script_functions();

  static const char *APP_TITLE; //!< Title for the application window

  SoundMixer mixer;
  Name2SoundMap name2sound;
 
  GameStateModule *gsm;
  Game *game; // = gsm->game
  GfxTerrain *terrain;

  SplashScreen splashScreen;
  LoopFunctor *looper;
  
  bool quit; // true <=> exit
  bool loading; // true <=> exit
  
  Octree *octree;
  Vector<Col3> view_colors; // 0: active
  Vector<Col3> player_colors; // 0: active

  std::list<ColorTileSet*> marked_tiles;
  std::list<ColRect*> marked_rectangles;

  void init_GL();
  void init_GLUT_window();
  void init_GLUT_events();

  sint4 win_bpp, win_flags;
  bool wireframe;
  bool use_textures;
  bool use_light;
  bool pathfinding_on;

  sint4    border;
  SDL_Rect views_rect; // all playfield views
  SDL_Rect mm_rect;    // minimap

  sint4 frames;
  sint4 slow_frames; // determines which non-active view to draw
  
  GLuint ground_tex;
  GLuint shadow_tex[2];
  
  sint4  blur_buffer;  // # of frames currently in the buffer
  GLuint blur_tex;

  
  Coor3 dim;  // world size (dx,dy,dz)
  real4 max_dim; // largest of dx, dz
  Vector<GfxView> views;
  GfxView *cur_view; ///< The view currently being drawn.

  ScriptObj *gui_obj;

  String2StringMap script_db; ///< String pairs for storing settings in the script.

  uint4 active_view;
  uint4 last_mouse_x, last_mouse_y; // the position of the last event involving the mouse
  uint4 button1_down_x, button1_down_y; // the last position button1 was clicked
  bool net_active;
 
  bool ctrl_down;
  bool shift_down;
  bool alt_down;

  bool right_mouse_down;
  bool middle_mouse_down;
  bool left_mouse_down;

  bool pause_draw;

  int mouse_down_x;
  int mouse_down_y;

  enum CLICK_STATE { NORMAL, JUST_CLICKED };
  CLICK_STATE cur_click_state;

 
  uint4 gfx_frames;
  uint4 num_tiles;
  real4 lx_angle, ly_angle; // light

  std::map<std::string, GLuint> gl_textures;
  Vector<GLuint> textures; // loaded by the script
  Vector<GLuint> models;   // loaded by the script

  Minimap *minimap;

  std::string prev_layout_name;
  
  //---------- model stuff ------------------

  std::set<GfxObject*> gfx_objs; // except terrain
  std::set<GfxGameObject*> sel_objs; // selected objs
  std::set<GfxGameObject*> game_objs;

  Vector<stype> *sel_ids; // ids of selected objects
  Vector<stype> *all_ids; // ids of all objects
  
  GameObj2GfxGameObjMap game2gfx;
  typedef GameObj2GfxGameObjMap::iterator game2gfx_iterator;
  Vector<GfxGameObject*> static_objects; ///< Objects to be kept around when not visible.

  void check_static_objects();

  // update active_view by looking at last_mouse_x and last_mouse_y
  void update_active_view();

  void select_view(GfxView &v, const SDL_Rect *sel_view, bool draw_objs);
  Vec3<real4> intersect_ray(GfxView &v, GfxObject *o, const SDL_Rect *sel_view);

  // draws view (also computes corners in v)
  void draw_view(GfxView &v, bool active);
  void draw_sky(GfxView &v);
  void draw_rectangle(const SDL_Rect &rect, sint4 width, const Col3 &col);
  void update_gfxobjs();

  void init_light0();
  void init_view_camera(GfxView &v);
  void initialize_colors();

  // change playing field view layout
  void update_layout(const std::string &layout_name, bool force = false);
  
  // returns id of closest object intersecting the click spaghetti;
  // also computes intersection coordinate
  uint4 pick(GfxView &v, const SDL_Rect &sel_view, Vec3<real4> &intersec, Vector<IdZ> &ids,
             bool pick_objs = true);

  // return 0 on error, -1 if intersection is behind, 1 otherwise
  sint4 ground_intersection(const GfxView &v, sint4 mx, sint4 my, real4 &x, real4 &z, real4 height=0);
  void view_to_plane(real4 x0, real4 x1, real4 y0, real4 y1, real4 fh, Vector<Coor3> &points);

  real4 ground_height(real4 x, real4 y);

  // must be called after draw_view() (which computes view corners)
  void draw_minimap();

  void load_tileset(const std::string &filename);
  void load_bitmap_list(Tokenizer &tokenizer, const std::string &path="");
  void load_model_list(Tokenizer &tokenizer, const std::string &path="");
  void load_database(const std::string &db_name);
  void save_database(const std::string &db_name);

  Vector<std::string> loaded_bp_files;
  bool load_interface(const std::string &fname, Tokenizer *t = 0);
  static std::string collapse_file_path(const std::string &p);

  /// convert minimap coordinates into real, centered playfield coordinates; midpoint is (0,0)
  void mm_to_playfield(sint4 mm_x, sint4 mm_y, real4 &pf_x, real4 &pf_y);

  /// send objects to (locx, locy) [real valued, tile-based coordinates]
  void move_objects(real4 locx, real4 locy);

  void set_mods();

  friend class ScriptIdSet;
  friend class ScriptModel;
  friend class ScriptEmitter;
  friend class GfxGlutAdaptor;
  friend class GfxGameObject;
  friend class DrawOnTerrain;
};

//===================================================================

class GfxGlutAdaptor
{
private:
  
  static GfxModule *c;

public:

  static void set_client(GfxModule *c_) {
    c = c_;
  }

  static void loop() { c->loop(); }
 
  static void draw() { c->draw(); }
  
  static void splash() { c->splash(); }

  
  static void event_resize(int width, int height) {
    c->event_resize(width, height);
  }
  
  static void event_key_down(unsigned char key, int x, int y) {
    c->event_key_down(key, x, y);
  }
  
  static void event_special_key_down(int key, int x, int y) {
    c->event_special_key_down(key, x, y);
  }
  
  void event_special_key_up(int key, int x, int y) {
    c->event_special_key_up(key, x, y);
  }

  static void event_key_up(unsigned char key, int x, int y) {
    c->event_key_up(key, x, y);
  }
  
  static void event_mouse_down(int btn, int state, int x, int y) {
    c->event_mouse_down(btn, state, x, y);
  }
  
  static void event_mouse_up(int btn, int state, int x, int y) {
    c->event_mouse_up(btn, state, x, y);
  }
  
  static void event_mouse_motion(int x, int y) {
    c->event_mouse_motion(x, y);
  }

};

REGISTER_TYPEOF(210, Vector<IdZ>::iterator);
REGISTER_TYPEOF(212, Vector<GfxView>::iterator);
REGISTER_TYPEOF(214, Vector<SDL_Rect>::iterator);
REGISTER_TYPEOF(216, Vector<Coor3>::iterator);
REGISTER_TYPEOF(220, std::list< Vector<sint4> >::iterator);
// 222 is defined in GameModule.C
REGISTER_TYPEOF(223, GfxModule::Name2SoundMap::iterator);
REGISTER_TYPEOF(224, GfxModule::Name2SoundMap::const_iterator);
REGISTER_TYPEOF(225, GfxModule::GameObj2GfxGameObjMap::iterator);
REGISTER_TYPEOF(226, GfxModule::GameObj2GfxGameObjMap::const_iterator);
REGISTER_TYPEOF(227, GfxModule::String2StringMap::iterator);
REGISTER_TYPEOF(228, GfxModule::String2StringMap::const_iterator);
#endif
