#ifndef GfxTerrain_H
#define GfxTerrain_H

// $Id: GfxTerrain.H,v 1.17 2005/08/26 23:23:51 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "GfxGlobal.H"
#include "GfxTerrainTile.H"
#include "Game.H"
#include "GfxObject.H"

class Octree;
class GfxModule;

/** Consolidates a group of tiles into one object to speed up rendering. */

class TerrainBlock : public GfxObject
{
public:
  static const sint4 sector_size = 4;

  GfxTerrainTile* tiles[sector_size+2][sector_size+2];
  sint4 empty; // the number of tiles not used (including a 1-tile fringe)
  GLuint dlist, olist; // display lists for the tile and the tile's outline

  TerrainBlock();
  virtual ~TerrainBlock();

  void new_tile(sint4 x, sint4 y, GfxTerrainTile *t, Octree *oct);
  void build_list();
  virtual void draw();
  virtual void pick();
  virtual void update(float){};
  DrawOrder order() const { return O_TERRAIN; }
};


/** Collection of GfxTerrainTiles for rendering and height adjustments. */

class GfxTerrain
{
public:
  
  GfxTerrain(GfxModule &gfxm_, sint4 tw, sint4 th, GLuint tex, Octree *o);
  ~GfxTerrain();

  GfxModule &gfxm;

  void new_tile(sint4 index, Game *game);
  
  void flatten(real4 x1, real4 y1, real4 x2, real4 y2);
  void unflatten(real4 x1, real4 y1, real4 x2, real4 y2);

  GfxTerrainTile* tile(sint4 x, sint4 y) const {
    if (x < 0 || y < 0 || x >= w || y >= h) return 0;
    return &tiles[w*y+x];
  }

  sint4 getWidth()  const { return w; }
  sint4 getHeight() const { return h; }

  void beginDraw();
  void endDraw();
  
  GfxTerrainTile *tiles;
  GLuint ground_tex;

private:

  sint4 w, sw;
  sint4 h, sh;
  
  TerrainBlock *sectors;
  Octree *octree;

  void uflatten(real4 x1, real4 y1, real4 x2, real4 y2, sint4 f);
  
};

#endif
