// $Id: GfxTerrainTile.C,v 1.52 2006/05/21 09:01:20 orts_furtak Exp $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxTerrainTile.H"
#include "GfxTerrain.H"
#include "Game.H"
#include "GameTile.H"
#include "Image.H"
#include "GfxModule.H"
#include "WaterTile.H"

using namespace std;

//===================================================================

Vec3<real4> GfxTerrainTile::noise_scale1;
Vec3<real4> GfxTerrainTile::noise_scale2;
Image GfxTerrainTile::noise;
real4 GfxTerrainTile::sz = 1.0/8.0;
real4 GfxTerrainTile::eps = 1.0/512.0;
bool GfxTerrainTile::outline = false;
Vector<GfxTerrainTile::GeometryType> GfxTerrainTile::types;
ShaderProgram GfxTerrainTile::tile_shader;

//===================================================================

GfxTerrainTile::GfxTerrainTile()
{
  initialized = false;
  flattened[0] = 0;
  flattened[1] = 0;
  flattened[2] = 0;
  flattened[3] = 0;
  tx = 0;
  ty = 0;
}

//-------------------------------------------------------------------

void GfxTerrainTile::set_types(const Vector<GfxTerrainTile::GeometryType> types_)
{
  types = types_;
}

//-------------------------------------------------------------------

bool GfxTerrainTile::load_noise(const string &filename, const Vec3<real4> &scale1, const Vec3<real4> &scale2)
{
  //sint4 w, h;
  //sint4 bpp;  // bytes per pixel

	//Image image;
	if(!noise.loadTGA(filename.c_str()))
	{
		cout << "Unable to load terrain noise texture " << filename << endl;
    return 0;
	}
	
	//bpp = image.bpp();
  //noise = load_tga_data(filename.c_str(), bpp, w, h);
	//noise = image.pixels();
  //noise_w = image.width();
  //noise_h = image.height();

  if(noise.bpp() != 1) {
    cout << "Terrain noise texture is not 1bpp grayscale " << filename << endl;
    return 0;
  }
  
  noise_scale1 = scale1;
  noise_scale2 = scale2;
  return 1;
}

//-------------------------------------------------------------------

real4 GfxTerrainTile::get_noise_at(real4 x, real4 y)
{
  if (noise.width() == 0 || noise.height() == 0) return 0.0;

  real4 px = noise_scale1.x * x;
  real4 py = noise_scale1.z * y;
  
  sint4 nx = (sint4)(floor(px)) % noise.width(); if (nx < 0) nx += noise.width();
  sint4 ny = (sint4)(floor(py)) % noise.height(); if (ny < 0) ny += noise.height();
  real4 fx = px-floor(px);
  real4 fy = py-floor(py);
  
  real4 n1 = noise_scale1.y * (real4)(noise[noise.width()*ny+nx]-128);
  real4 n2 = noise_scale1.y * (real4)(noise[noise.width()*ny+(nx+1) % noise.width()]-128);
  real4 n3 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + nx]-128);
  real4 n4 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + (nx+1) % noise.width()]-128);
  real4 n  = n1*(1-fx)*(1-fy) + n2*fx*(1-fy) + n3*(1-fx)*fy + n4*fx*fy;

/*  ny = (sint4)(floor(noise_scale2.z*y)) % noise_h;
  nx = (sint4)(floor(noise_scale2.x*x)) % noise_w;
  n += noise_scale2.y*(real4)(noise[noise_w*ny+nx]-128); */
  return n;
}

//-------------------------------------------------------------------

void GfxTerrainTile::compute_tc()
{
  int u=0, v=2;
  if (trans) { u=2; v=0;}

  FORS (i, 4) {
    FORS (j, 4) {
      texcoord[i][j][0] = hf*(vert[i][j][u]-vert[0][0][u])*(sz-eps)+tt1[0] + eps*hf;
      texcoord[i][j][1] = vf*(vert[i][j][v]-vert[0][0][v])*(sz-eps)+tt1[1] + eps*vf;
    }
  }
}

//-------------------------------------------------------------------

// fixme: explain what's going on

void GfxTerrainTile::initverts()
{
  real4 miny = numeric_limits<real4>::max();
  real4 maxy = numeric_limits<real4>::min();
  real4 x = aabb_o.x;
  real4 y = aabb_o.z;
  const real4 a = 0.4;
  const real4 b = 0.3;
  
  real4 tile_points[11][4][4][2] ={
   // flat tile (regular grid) //
   { {{0.0, 0.0}, {0.5, 0.0}, {0.0, 0.5}, {0.5, 0.5}},
     {{0.5, 0.0}, {1.0, 0.0}, {0.5, 0.5}, {1.0, 0.5}},
     {{0.0, 0.5}, {0.5, 0.5}, {0.0, 1.0}, {0.5, 1.0}},
     {{0.5, 0.5}, {1.0, 0.5}, {0.5, 1.0}, {1.0, 1.0}} },
   // cliff in Y direction //
   { {{0.0, 0.0},   {0.5, 0.0},   {0.0, 0.5-a}, {0.5, 0.5-b}},
     {{0.5, 0.0},   {1.0, 0.0},   {0.5, 0.5-b}, {1.0, 0.5-a}},
     {{0.0, 0.5+a}, {0.5, 0.5+b}, {0.0, 1.0},   {0.5, 1.0}},
     {{0.5, 0.5+b}, {1.0, 0.5+a}, {0.5, 1.0},   {1.0, 1.0}} },
   // cliff in X direction //
   { {{0.0,   0.0},  {0.5-a, 0.0},  {0.0,   0.5},  {0.5-b, 0.5}},
     {{0.5+a, 0.0},  {1.0,   0.0},  {0.5+b, 0.5},  {1.0,   0.5}},
     {{0.0,   0.5},  {0.5-b, 0.5},  {0.0,   1.0},  {0.5-a, 1.0}},
     {{0.5+b, 0.5},  {1.0,   0.5},  {0.5+a, 1.0},  {1.0,   1.0}} },
   // 8 types of diagonal cliffs //
   { {{0.0,   0.0},    {0.5-b/2, 0.0},  {0.0, 0.5-b/2},  {0.3-b/2, 0.3-b/2}},
     {{0.5+b, 0.0},    {1.0,     0.0},  {0.5, 0.5},      {1.0,     0.5}},
     {{0.0,   0.5+b},  {0.5,     0.5},  {0.0, 1.0},      {0.5,     1.0}},
     {{0.5,   0.5},    {1.0,     0.5},  {0.5, 1.0},      {1.0,     1.0}} },
   
   { {{0.0,      0.0},       {0.5-b/2,  0.0},       {0.0,      0.5-b/2},   {0.25-b/2, 0.25-b/2}},
     {{0.5+b/2,  0.0},       {1.0,      0.0},       {0.25+b/2, 0.25+b/2},  {1.0,      0.5}},
     {{0.0,      0.5+b/2},   {0.25+b/2, 0.25+b/2},  {0.0,      1.0},       {0.5,      1.0}},
     {{0.25+b/2, 0.25+b/2},  {1.0,      0.5},       {0.5,      1.0},       {1.0,      1.0}} },

   { {{0.0,     0.0},  {0.5-b, 0.0},    {0.0,   0.5},    {0.5, 0.5}},
     {{0.5+b/2, 0.0},  {1.0,   0.0},    {0.5+b, 0.5-b},  {1.0, 0.5-b/2}},
     {{0.0,     0.5},  {0.5,   0.5},    {0.0,   1.0},    {0.5, 1.0}},
     {{0.5,     0.5},  {1.0,   0.5+b},  {0.5,   1.0},    {1.0, 1.0}} },

   { {{0.0,      0.0},       {0.5-b/2,  0.0},       {0.0,      0.5},       {0.75-b/2, 0.25+b/2}},
     {{0.5+b/2,  0.0},       {1.0,      0.0},       {0.75+b/2, 0.25-b/2},  {1.0,      0.5-b/2}},
     {{0.0,      0.5},       {0.75-b/2, 0.25+b/2},  {0.0,      1.0},       {0.5,      1.0}},
     {{0.75-b/2, 0.25+b/2},  {1.0,      0.5 +b/2},  {0.5,      1.0},       {1.0,      1.0}} },

   { {{0.0, 0.0},    {0.5,   0.0},    {0.0,   0.5-b},  {0.5,   0.5}},
     {{0.5, 0.0},    {1.0,   0.0},    {0.5,   0.5},    {1.0,   0.5}},
     {{0.0, 0.5+b},  {0.5-b, 0.5+b},  {0.0,   1.0},    {0.5-b, 1.0}},
     {{0.5, 0.5},    {1.0,   0.5},    {0.5+b, 1.0},    {1.0,   1.0}} },

   { {{0.0,      0.0},       {0.5,      0.0},       {0.0,      0.5-b/2},   {0.25+b/2, 0.75-b/2}},
     {{0.5,      0.0},       {1.0,      0.0},       {0.25+b/2, 0.75-b/2},  {1.0,      0.5}},
     {{0.0,      0.5+b/2},   {0.25-b/2, 0.75+b/2},  {0.0,      1.0},       {0.5-b/2,  1.0}},
     {{0.25+b/2, 0.75-b/2},  {1.0,      0.5},       {0.5+b/2,  1.0},       {1.0,      1.0}} },

   { {{0.0,   0.0},    {0.5, 0.0},  {0.0,     0.5},  {0.5,   0.5}},
     {{0.5,   0.0},    {1.0, 0.0},  {0.5,     0.5},  {1.0,   0.5-b}},
     {{0.0,   0.5},    {0.5, 0.5},  {0.0,     1.0},  {0.5-b, 1.0}},
     {{0.5+b, 0.5+b},  {1.0, b/2},  {0.5+b/2, 1.0},  {1.0,   1.0}} },

   { {{0.0,      0.0},       {0.5,      0.0},       {0.0,      0.5},       {0.75-b/2, 0.75-b/2}},
     {{0.5,      0.0},       {1.0,      0.0},       {0.75-b/2, 0.75-b/2},  {1.0,      0.5-b/2}},
     {{0.0,      0.5},       {0.75-b/2, 0.75-b/2},  {0.0,      1.0},       {0.5-b/2,  1.0}},
     {{0.75+b/2, 0.75+b/2},  {1.0,      0.5},       {0.5,      1.0},       {1.0,      1.0}} }
  };
  
  
  sint4 tile_type = 0;
  
  if (has_cliffs) {
    if (h[0] == h[1] && h[2] == h[3]) tile_type = 1;
    else if (h[0] == h[2] && h[1] == h[3]) tile_type = 2;
    else if (h[1] == h[2] && h[1] == h[3]) {
      if (h[0] > h[1]) tile_type = 3;
      else             tile_type = 4;
    } else if (h[0] == h[2] && h[0] == h[3]) {
      if (h[1] > h[0]) tile_type = 5;
      else             tile_type = 6;
    } else if (h[0] == h[1] && h[0] == h[3]) {
      if (h[2] > h[0]) tile_type = 7;
      else             tile_type = 8;
    } else if (h[0] == h[1] && h[0] == h[2]) {
      if (h[3] > h[0]) tile_type = 9;
      else tile_type = 10;
    } else
      tile_type = 0;
  } else
    tile_type = 0;

  bool all_flat = (flattened[0] && flattened[1] && flattened[2] && flattened[3]);

  FORS (i, 4) {
    FORS (j, 4) {
      Vec3<real4> &v = vert[i][j];

      if (has_cliffs) {
	v.x = tile_points[tile_type][i][j][0];
	v.z = tile_points[tile_type][i][j][1];
       	v.y = h[i];
    	
//	v.x += get_noise_at(v.x+10, v.z+10+h[i]);
//	v.z += get_noise_at(v.x+20, v.z+20+h[i]);

      } else {

	v.x = tile_points[0][i][j][0];
	v.z = tile_points[0][i][j][1];
   
	v.y = (1-v.x)*(1-v.z)*h[0] + v.x*(1-v.z)*h[1]
	    + (1-v.x)*v.z*h[2]     + v.x*v.z*h[3];
      }
      // offset for tile position
      v.x += x;
      v.z += y;

      if (!all_flat) {
        bool flat = false;
        if (flattened[i] && flattened[j]) {
          if (i == 0 && j != 3) flat = true;
          if (i == 1 && j != 2) flat = true;
          if (i == 2 && j != 1) flat = true;
          if (i == 3 && j != 0) flat = true;
        }
        if (!flat) v.y += get_noise_at(v.x, v.z);
      }

      // calculate normal (not quite correct yet)
      real4 eps = 0.1;
      Vec3<real4> t1(v.x-eps, get_noise_at(v.x-eps, v.z), v.z);
      Vec3<real4> t2(v.x+eps, get_noise_at(v.x+eps, v.z), v.z);
      Vec3<real4> t3(v.x,     get_noise_at(v.x, v.z-eps), v.z-eps);
      Vec3<real4> t4(v.x,     get_noise_at(v.x, v.z+eps), v.z+eps);
      norm[i][j] = (t3-t4).cross(t1-t2);
      norm[i][j].normalize();

      maxy = max(maxy, v.y);
      miny = min(miny, v.y);
    }
  }

  aabb_o.y = miny;
  aabb_d.y = maxy-miny;
}

//-------------------------------------------------------------------

void GfxTerrainTile::initblack(sint4 index, GfxTerrain *t)
{
  if (initialized) return;
  ter = t;
  sint4 x = index % t->getWidth();
  sint4 y = index / t->getWidth();

  tx = x;
  ty = y;

  if (x > 0 && ter->tile(x-1, y)->initialized) {
    h[0] = ter->tile(x-1, y)->h[1];
    h[2] = ter->tile(x-1, y)->h[3];
  }
  if (x+1 < ter->getWidth() && ter->tile(x+1, y)->initialized) {
    h[1] = ter->tile(x+1, y)->h[0];
    h[3] = ter->tile(x+1, y)->h[2];
  }
  if (y > 0 && ter->tile(x, y-1)->initialized) {
    h[0] = ter->tile(x, y-1)->h[2];
    h[1] = ter->tile(x, y-1)->h[3];
  }
  if (y+1 < ter->getHeight() && ter->tile(x, y+1)->initialized) {
    h[2] = ter->tile(x, y+1)->h[0];
    h[3] = ter->tile(x, y+1)->h[1];
  }
  if (x > 0 && y > 0 && ter->tile(x-1, y-1)->initialized) {
    h[0] = ter->tile(x-1, y-1)->h[3];
  }
  if (x+1 < ter->getWidth() && y+1 < ter->getHeight() && ter->tile(x+1, y+1)->initialized) {
    h[3] = ter->tile(x+1, y+1)->h[0];
  }
  if (y > 0 && x+1 < ter->getWidth() && ter->tile(x+1, y-1)->initialized) {
    h[1] = ter->tile(x+1, y-1)->h[2];
  }
  if (y+1 < ter->getHeight() && x > 0 && ter->tile(x-1, y+1)->initialized) {
    h[2] = ter->tile(x-1, y+1)->h[1];
  }

  has_cliffs = false;

  // pick texture tiles

  hf = 1;
  vf = 1;
  trans = 0;
  tt1[0] = 0;
  tt1[1] = sz;
  
  aabb_o.x = x;
  aabb_o.z = y;
  aabb_d.x = 1;
  aabb_d.z = 1;

  initverts();

//  initialized = true;
}

//-------------------------------------------------------------------

WaterTile* GfxTerrainTile::new_WaterTile()
{
#if 0
  const real4 depth = 0.25;
  Vec3<real4> watervert[4];

  watervert[0] = Vec3<real4>(tx+1, h[1]-depth, ty);
  watervert[1] = Vec3<real4>(tx,   h[0]-depth, ty);
  watervert[2] = Vec3<real4>(tx+1, h[3]-depth, ty+1);
  watervert[3] = Vec3<real4>(tx,   h[2]-depth, ty+1);
#else
  const real4 depth = -0.5;
  Vec3<real4> watervert[4];

  real4 m = h[3];
  FORU (i, 3) if (h[i] < m) m = h[i];

  watervert[0] = Vec3<real4>(tx+1, m-depth, ty);
  watervert[1] = Vec3<real4>(tx,   m-depth, ty);
  watervert[2] = Vec3<real4>(tx+1, m-depth, ty+1);
  watervert[3] = Vec3<real4>(tx,   m-depth, ty+1);
#endif

  //    FORU(i, 4) watervert[i].y += get_noise_at(watervert[i].x, watervert[i].z);

  return new WaterTile(watervert);
}

//-------------------------------------------------------------------

void GfxTerrainTile::init(sint4 index, Game *game, GfxTerrain *t)
{
  set_gfxm(t->gfxm);
  ter = t;
  this->index = index;

  const real4 inv_h = 1.0/GameConst::HEIGHT_MULT;
  sint4 x = game->get_map().ind2x(index);
  sint4 y = game->get_map().ind2y(index);

  tx = x;
  ty = y;

  const GameTile &tile = game->get_map()(index);

  Random &rng = gfxm->rand;
  
  sint4 htemp[4];
  tile.get_topo(htemp[0], htemp[1], htemp[3], htemp[2], split);
  FORS (i,4) h[i] = inv_h*htemp[i];
  
  tile.get_type(typeW, typeE);
  GeometryType tW = types[typeW];
  GeometryType tE = types[typeE];

  has_cliffs = (tW == CLIFF || tE == CLIFF);

  if (has_water()) {
    FORS (i, 4) h[i] -= 0.8;
  }

  // pick texture tiles
  hf = 1;
  vf = 1;
  trans = 0;
  if (tW == WATER && tE != WATER) {

    tt1[0] = sz*2;
    tt1[1] = sz*5;
    if (split == Tile::BT_SPLIT) vf = -1;
    
  } else if (tE == WATER && tW != WATER) {
    tt1[0] = sz*2;
    tt1[1] = sz*5;
    hf = -1;
    if (split==Tile::TB_SPLIT) vf = -1;
    
  } else if (has_water()) {

    tt1[0] = sz * (3+(rng.rand_uint4() % 5));
    tt1[1] = sz*5;
    
  } else {

    tt1[0] = sz * (rng.rand_uint4() % 8);
    tt1[1] = sz * min(int(rng.rand_uint4() % 5), 1);
    hf = (2*(float)(rng.rand_uint4() % 2))-1;
    vf = (2*(float)(rng.rand_uint4() % 2))-1;
    trans=(rng.rand_uint4() % 2);
  }
  
  tt1[0] += sz*(hf<0);
  tt1[1] += sz*(vf<0);

  tt2[0] = sz * (rng.rand_uint4() % 8);
  tt2[1] = 6*sz;
  
  aabb_o.x = x;
  aabb_o.z = y;
  aabb_d.x = 1;
  aabb_d.z = 1;

  initverts();

  initialized = true;
}

//===================================================================

void GfxTerrainTile::flatten(sint4 f)
{
  flattened[0] += f;
  flattened[1] += f;
  flattened[2] += f;
  flattened[3] += f;
  
  //  if (initialized)
  {
    initverts();

    //    return;
    // align surrounding tiles
    int i = tx;
    int j = ty;

    if (i > 0) {
      {
        GfxTerrainTile *t = ter->tile(i-1, j); // W tile
        if (t) {
          t->flattened[1] += f;
          t->flattened[3] += f;
          t->initverts();
        }
      }
      if (j > 0) {
        GfxTerrainTile *t = ter->tile(i-1, j-1); // NW tile
        if (t) {
          t->flattened[3] += f;
          t->initverts();
        }
      }

      if (j+1 < ter->getHeight()) {
        GfxTerrainTile *t = ter->tile(i-1, j+1); // SW tile
        if (t) {
          t->flattened[1] += f;
          t->initverts();
        }
      }
    }

    if (j > 0) {
      GfxTerrainTile *t = ter->tile(i, j-1); // N tile
      if (t) {
        t->flattened[2] += f;
        t->flattened[3] += f;
        t->initverts();
      }
    }
    if (j+1 < ter->getHeight()) {
      GfxTerrainTile *t = ter->tile(i, j+1); // S tile
      if (t) {
        t->flattened[0] += f;
        t->flattened[1] += f;
        t->initverts();
      }
    }

    if (i+1 < ter->getWidth()) {
      {
        GfxTerrainTile *t = ter->tile(i+1, j); // E tile
        if (t) {
          t->flattened[0] += f;
          t->flattened[2] += f;
          t->initverts();
        }
      }
      if (j > 0) {
        GfxTerrainTile *t = ter->tile(i+1, j-1); // NE tile
        if (t) {
          t->flattened[2] += f;
          t->initverts();
        }
      }
      if (j+1 < ter->getHeight()) {
        GfxTerrainTile *t = ter->tile(i+1, j+1); // SE tile
        if (t) {
          t->flattened[0] += f;
          t->initverts();
        }
      }
    }
  }
}

//===================================================================

void GfxTerrainTile::setTexture(real4 row, real4 col, sint4 ncol,
				float hflip, float vflip, bool transpose)
{
  tt1[0] = sz * (col+(gfxm->rand.rand_uint4() % ncol));
  tt1[1] = row*sz;
  trans = transpose;
  hf = hflip;
  vf = vflip;
  tt1[0] += sz*(hf<0);
  tt1[1] += sz*(vf<0);
}

//-------------------------------------------------------------------

void GfxTerrainTile::br_neighbor(GfxTerrainTile *t)
{
  if (t->initialized && initialized) {
    if (has_water() && !t->has_water()) {
      h[3] = t->h[0];
      initverts();
    }
    if (!has_water() && t->has_water()) {
      t->h[0] = h[3];
      t->initverts();
    }
  }
}

//-------------------------------------------------------------------

void GfxTerrainTile::bl_neighbor(GfxTerrainTile *t)
{
  if (t->initialized && initialized) {
    if (has_water() && !t->has_water()) {
      h[2] = t->h[1];
      initverts();
    }
    if (!has_water() && t->has_water()) {
      t->h[1] = h[2];
      t->initverts();
    }
  }
}

//-------------------------------------------------------------------

void GfxTerrainTile::b_neighbor(GfxTerrainTile *t)
{
  if (t->initialized && initialized) {
      if (has_water() && !t->has_water()) {
	h[2] = t->h[0];
	h[3] = t->h[1];
	initverts();
	if (has_water() == 2) {
	  setTexture(5, 0, 2, 1, -1, false);
	  compute_tc();
	}
      }
      if (!has_water() && t->has_water()) {
	t->h[0] = h[2];
	t->h[1] = h[3];
	t->initverts();
	if (t->has_water() == 2) {
	  t->setTexture(5, 0, 2, 1, 1, false);
	  t->compute_tc();
	}
      }
      if (has_cliffs || t->has_cliffs) {
	t->vert[0][0] = vert[2][2] = (vert[2][2] + t->vert[0][0])*0.5;
	t->vert[0][1] = vert[2][3] = (vert[2][3] + t->vert[0][1])*0.5;
	t->vert[1][0] = vert[3][2] = (vert[3][2] + t->vert[1][0])*0.5;
      }

//      vert[2][2]=t->vert[0][0];
//      vert[2][3]=t->vert[0][1];
//      vert[3][2]=t->vert[1][0];

  }
}

//-------------------------------------------------------------------

void GfxTerrainTile::r_neighbor(GfxTerrainTile *t)
{
  if (t->initialized && initialized) {
      if (has_water() && !t->has_water()) {
	h[1] = t->h[0];
	h[3] = t->h[2];
	initverts();
	if (has_water() == 2) {
	  setTexture(5, 0, 2, 1, -1, true);
	  compute_tc();
	}
      }
      if (!has_water() && t->has_water()) {
	t->h[0] = h[1];
	t->h[2] = h[3];
	t->initverts();
	if (t->has_water() == 2) {
	  t->setTexture(5, 0, 2, 1, 1, true);
	  t->compute_tc();
	}

      }

      if (has_cliffs || t->has_cliffs) {
	t->vert[0][0] = vert[1][1] = (vert[1][1] + t->vert[0][0])*0.5;
	t->vert[0][2] = vert[1][3] = (vert[1][3] + t->vert[0][2])*0.5;
	t->vert[2][0] = vert[3][1] = (vert[3][1] + t->vert[2][0])*0.5;
      }

//      vert[1][1]=t->vert[0][0];
//      vert[1][3]=t->vert[0][2];
//      vert[3][1]=t->vert[2][0];
  }
}

//===================================================================

#define CNV(i,j)\
glColor3f(vert[i][j].y/8+0.35, vert[i][j].y/8+0.35, vert[i][j].y/8+0.35);\
glNormal3fv(norm[i][j].ptr());\
glVertex3fv(vert[i][j].ptr());
void GfxTerrainTile::draw()
{
  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);
    glTexCoord2fv(texcoord[0][2]); CNV(0,2);
    glTexCoord2fv(texcoord[0][1]); CNV(0,1);
    glTexCoord2fv(texcoord[0][3]); CNV(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[1][0]); CNV(1,0);
    glTexCoord2fv(texcoord[1][2]); CNV(1,2);
    glTexCoord2fv(texcoord[1][1]); CNV(1,1);
    glTexCoord2fv(texcoord[1][3]); CNV(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[2][0]); CNV(2,0);
    glTexCoord2fv(texcoord[2][2]); CNV(2,2);
    glTexCoord2fv(texcoord[2][1]); CNV(2,1);
    glTexCoord2fv(texcoord[2][3]); CNV(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[3][0]); CNV(3,0);
    glTexCoord2fv(texcoord[3][2]); CNV(3,2);
    glTexCoord2fv(texcoord[3][1]); CNV(3,1);
    glTexCoord2fv(texcoord[3][3]); CNV(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[0][1].y, tt2[1]+sz-eps); CNV(0,1);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+sz/2);   CNV(0,3);
    glTexCoord2f(tt2[0]+sz*vert[1][0].y, tt2[1]+sz-eps); CNV(1,0);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+sz/2);   CNV(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/2); CNV(2,1);
    glTexCoord2f(tt2[0]+sz*vert[2][3].y, tt2[1]+eps);  CNV(2,3);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/2); CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[3][2].y, tt2[1]+eps);  CNV(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[2][0].y, tt2[1]+eps);  CNV(2,0);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/2); CNV(2,1);
    glTexCoord2f(tt2[0]+sz*vert[0][2].y, tt2[1]+eps);  CNV(0,2);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+sz/2); CNV(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/2);   CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[3][1].y, tt2[1]+sz-eps); CNV(3,1);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+sz/2);   CNV(1,2);
    glTexCoord2f(tt2[0]+sz*vert[1][3].y, tt2[1]+sz-eps); CNV(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+eps);  CNV(1,2);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+eps);  CNV(0,3);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/4); CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/4); CNV(2,1);
    glEnd();

  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    glTexCoord2fv(texcoord[0][3]); CNV(0,3);
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);
    glTexCoord2fv(texcoord[0][2]); CNV(0,2);
    glTexCoord2fv(texcoord[2][0]); CNV(2,0);
    glTexCoord2fv(texcoord[2][2]); CNV(2,2);
    glTexCoord2fv(texcoord[2][3]); CNV(2,3);
    glTexCoord2fv(texcoord[3][2]); CNV(3,2);
    glTexCoord2fv(texcoord[3][3]); CNV(3,3);
    glTexCoord2fv(texcoord[3][1]); CNV(3,1);
    glTexCoord2fv(texcoord[1][3]); CNV(1,3);
    glTexCoord2fv(texcoord[1][1]); CNV(1,1);
    glTexCoord2fv(texcoord[1][0]); CNV(1,0);
    glTexCoord2fv(texcoord[0][1]); CNV(0,1);
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);
    glEnd();
  }

#define FOO(i,j)\
          glTexCoord2fv(texcoord[0][0].ptr());\
                                        \
          glColor3f(0.5, 0.5, 0.5);\
          glNormal3fv(norm[i][j].ptr());\
          glVertex3fv(vert[i][j].ptr());\
                                  \
          glColor3f(0,0,0);       \
          glVertex3f(vert[i][j].x, vert[i][j].y - 50, vert[i][j].z);

  /*
  glActiveTextureARB(GL_TEXTURE1);
  glDisable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE0);

  glBegin(GL_TRIANGLE_STRIP);
  FOO(0,0);
  FOO(0,2);
  FOO(2,0);
  FOO(2,2);
  FOO(2,3);
  FOO(3,2);
  FOO(3,3);
  FOO(3,1);
  FOO(1,3);
  FOO(1,1);
  FOO(1,0);
  FOO(0,1);
  FOO(0,0);
  glEnd();

  glActiveTextureARB(GL_TEXTURE1);
  glEnable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE0);
  */

  if (outline)draw_outline();
}

//-------------------------------------------------------------------

#define CNVf(i,j)\
glNormal3fv(norm[i][j].ptr());\
glVertex3fv(vert[i][j].ptr());

/** Draw the tile while ignoring texture.
    Combine with blending to highlight certain tiles.
*/

void GfxTerrainTile::draw_flat()
{
  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVf(0,0);
    CNVf(0,2);
    CNVf(0,1);
    CNVf(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(1,0);
    CNVf(1,2);
    CNVf(1,1);
    CNVf(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,0);
    CNVf(2,2);
    CNVf(2,1);
    CNVf(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(3,0);
    CNVf(3,2);
    CNVf(3,1);
    CNVf(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVf(0,1);
    CNVf(0,3);
    CNVf(1,0);
    CNVf(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,1);
    CNVf(2,3);
    CNVf(3,0);
    CNVf(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,0);
    CNVf(2,1);
    CNVf(0,2);
    CNVf(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(3,0);
    CNVf(3,1);
    CNVf(1,2);
    CNVf(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(1,2);
    CNVf(0,3);
    CNVf(3,0);
    CNVf(2,1);
    glEnd();
  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    CNVf(0,3);
    CNVf(0,0);
    CNVf(0,2);
    CNVf(2,0);
    CNVf(2,2);
    CNVf(2,3);
    CNVf(3,2);
    CNVf(3,3);
    CNVf(3,1);
    CNVf(1,3);
    CNVf(1,1);
    CNVf(1,0);
    CNVf(0,1);
    CNVf(0,0);
    glEnd();
  }
  if (outline)draw_outline();
}

//-------------------------------------------------------------------

#define CNVp(i,j)\
glColor3fv(((vert[i][j]-aabb_o)/aabb_d).ptr());\
glVertex3fv(vert[i][j].ptr());

void GfxTerrainTile::pick()
{
  glLoadName(my_id << GfxObject::SUB_ID_LOG);

  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVp(0,0);
    CNVp(0,2);
    CNVp(0,1);
    CNVp(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(1,0);
    CNVp(1,2);
    CNVp(1,1);
    CNVp(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,0);
    CNVp(2,2);
    CNVp(2,1);
    CNVp(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(3,0);
    CNVp(3,2);
    CNVp(3,1);
    CNVp(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVp(0,1);
    CNVp(0,3);
    CNVp(1,0);
    CNVp(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,1);
    CNVp(2,3);
    CNVp(3,0);
    CNVp(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,0);
    CNVp(2,1);
    CNVp(0,2);
    CNVp(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(3,0);
    CNVp(3,1);
    CNVp(1,2);
    CNVp(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(1,2);
    CNVp(0,3);
    CNVp(3,0);
    CNVp(2,1);
    glEnd();
  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    CNVp(0,3);
    CNVp(0,0);
    CNVp(0,2);
    CNVp(2,0);
    CNVp(2,2);
    CNVp(2,3);
    CNVp(3,2);
    CNVp(3,3);
    CNVp(3,1);
    CNVp(1,3);
    CNVp(1,1);
    CNVp(1,0);
    CNVp(0,1);
    CNVp(0,0);
    glEnd();
  }
}

//-------------------------------------------------------------------

/** Draws an outline along the South and West sides. */

void GfxTerrainTile::draw_outline()
{
  glLoadName(my_id << GfxObject::SUB_ID_LOG);

  real4 oset = 0.01;
  
  glColor3f(0.1, 0.1, 0.1);
  glBegin(GL_LINE_STRIP);
  glVertex3f(vert[0][0].x, vert[0][0].y+oset, vert[0][0].z);
  glVertex3f(vert[0][2].x, vert[0][2].y+oset, vert[0][2].z);
  glVertex3f(vert[2][0].x, vert[2][0].y+oset, vert[2][0].z);
  glVertex3f(vert[2][2].x, vert[2][2].y+oset, vert[2][2].z);
  glVertex3f(vert[2][3].x, vert[2][3].y+oset, vert[2][3].z);
  glVertex3f(vert[3][2].x, vert[3][2].y+oset, vert[3][2].z);
  glVertex3f(vert[3][3].x, vert[3][3].y+oset, vert[3][3].z);
  glEnd();
}

/*
bool GfxTerrainTile::line_intersection(const Coor3 &l1, const Coor3 &l2, Coor3 &intersec)
{
  return true;
}
*/
