#ifndef GfxTerrainTile_H
#define GfxTerrainTile_H

// $Id: GfxTerrainTile.H,v 1.33 2005/08/30 05:49:37 orts_furtak Exp $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxGlobal.H"
#include "GfxObject.H"
#include "Image.H"
#include "Tile.H"
#include "Shader.H"

class GfxTerrain;
class WaterTile;

class GfxTerrainTile : public GfxObject
{
public:

  real4 h[4]; // corner heights (SW, SE, NW, NE)
  sint1 flattened[4]; ///< If the corner is ignoring noise.  Indices match h[4].


  /// Affects how the tile geometry is created and how texture is applied.
  /// Water tiles have trenches, ground and cliff tiles do not.
  enum GeometryType { UNKNOWN, WATER, GROUND, CLIFF };

  GfxTerrainTile();

  void init(sint4 index, Game *game, GfxTerrain *t);
  void initblack(sint4 index, GfxTerrain *t);
  void initverts();

  void r_neighbor(GfxTerrainTile *t);  // set right neighbour
  void b_neighbor(GfxTerrainTile *t);  // set bottom 
  void br_neighbor(GfxTerrainTile *t); // set bottom-right neighbor
  void bl_neighbor(GfxTerrainTile *t); // set bottom-left neighbor
  virtual void update(float) {};
  virtual void draw();
  void draw_flat();
  virtual void pick();
  virtual void draw_outline();
  bool line_intersection(const Coor3 &l1, const Coor3 &l2, Coor3 &intersec);

  static bool load_noise(const std::string &file,
			 const Vec3<real4> &scale1, const Vec3<real4> &scale2);
  
  static real4 get_noise_at(real4 x, real4 y);
  bool is_initialized() const { return initialized; }

  void setTexture(real4 row, real4 col=0,sint4 ncol=8,
		  float hflip=1, float vflip=1, bool transpose=0);

  void flatten(sint4 f);

  DrawOrder order() const { return O_TERRAIN; }
  virtual ~GfxTerrainTile() {}

  void compute_tc();

  sint4 has_water() { return (types[typeW] == WATER) + (types[typeE] == WATER); }
  WaterTile *new_WaterTile();

  static void set_types(const Vector<GeometryType> types_);

  static bool outline;
  static ShaderProgram tile_shader;

protected:
  GfxTerrain *ter;
  sint4 index; ///< This tile's index in GfxTerrain.
  sint4 tx;
  sint4 ty;

  Vec3<real4> vert[4][4];
  Vec3<real4> norm[4][4];
  real4 texcoord[4][4][2];

  real4 tt1[2]; // texture tile 1, offset to grass (or other) texture
  real4 hf;     // texture tile 1 horizontal flip
  real4 vf;     // texture tile 1 vertical flip
  bool trans;   // texture tile 1 transpose 
  real4 tt2[2]; // texture tile 2, offset to cliff texture

  bool initialized;
  bool has_cliffs;
  Tile::Split split;
  Tile::Type typeW, typeE;


  inline GeometryType get_typeW() const { return types[typeW]; }
  inline GeometryType get_typeE() const { return types[typeE]; }
  // tE should be the same as tW for unsplit tiles, but don't take chances if split == NO_SPLIT
  inline GeometryType get_typeN() const { return (split == Tile::TB_SPLIT) ? types[typeE] : types[typeW]; }
  inline GeometryType get_typeS() const { return (split == Tile::BT_SPLIT) ? types[typeE] : types[typeW]; }

  static real4 sz;  // texture tile size
  static real4 eps; // small value (size of a pixel in the texture)

  static Image noise;
  static Vec3<real4> noise_scale1;
  static Vec3<real4> noise_scale2;

  static Vector<GeometryType> types;

  friend class GfxTerrain;
};

#endif
