#ifndef Shader_H
#define Shader_H

// $Id: Shader.H,v 1.7 2005/10/03 19:47:29 orts_furtak Exp $

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "GfxGlobal.H"

//===================================================================

class Shader
{
public:
  GLuint shader;

  void clear();
  static char* read_source(const std::string &filename);
  void compile();
};

//-------------------------------------------------------------------

class VertexShader : public Shader
{
public:
  VertexShader() { shader = 0; }
  VertexShader(const std::string &filename) { load(filename); }

  bool load(const std::string &filename);

  ~VertexShader() {
    if (shader != 0) {
      glDeleteObjectARB(shader);
    }
  }
};

//-------------------------------------------------------------------

class FragmentShader : public Shader
{
public:
  FragmentShader() { shader = 0; }
  FragmentShader(const std::string &filename) { load(filename); }

  bool load(const std::string &filename);

  ~FragmentShader() {
    if (shader != 0) {
      glDeleteObjectARB(shader);
    }
  }
};

//-------------------------------------------------------------------

class ShaderProgram
{
protected:
  GLuint prog;
  FragmentShader *fs;
  VertexShader *vs;

public:
  ShaderProgram();

  GLuint id() const { return prog; }
  void fragment_shader(const std::string &filename);
  void vertex_shader(const std::string &filename);

  void link();

  bool begin() {
    if (!glUseProgram || !prog) return false;
    glUseProgram(prog);
    return true;
  }

  void end() {
    if (glUseProgram) glUseProgram(0);
  }
};

//===================================================================

#endif
