/** @file Action.C
    @see Action.H

    $Id: Action.C,v 1.45 2006/05/29 16:59:38 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Action.C,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Action.H"
#include "Game.H"
#include "PlayerInfo.H"
#include "ScriptObj.H"
#include "GameObj.H"
#include "Script.H"
#include "ByteData.H"
#include "Token.H"
#include "Random.H"
#include "ScriptParser.H"

using namespace std;

REGISTER_TYPEOF(70, Vector<BaseActions*>::iterator);
REGISTER_TYPEOF(71, Vector<BaseActions*>::const_iterator);
REGISTER_TYPEOF(72, Vector<MoveActions::maction>::iterator);
REGISTER_TYPEOF(73, Vector<MoveActions::maction>::const_iterator);

#define TEST_ERR ERR // change to REM for competition

#define PARSE_ERR(s) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR(s);\
}

#define PARSE_ERR2(s1, s2) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR2(s1,s2);\
}

Vector<ObjAction*> ObjAction::all_actions;
stack<string> ObjAction::dir_path;

//===================================================================

void Action::set(sint4 id_, const Vector<sint4> &params_, sint4 priority_)
{
  id = id_;
  params = params_;
  priority = priority_;
}

//-------------------------------------------------------------------

void Action::clear()
{
  id = NOOP;
  params.clear();
  priority = -1;
  client_action_id = 0;
}

//===================================================================

void SActParam::add_ref() {
  FORALL (obj, it) (*it)->add_ref();
}

void SActParam::del_ref() {
  FORALL (obj, it) (*it)->del_ref();
}

//===================================================================

void SavedActParam::add_ref() {
  FORALL (obj, it) (*it)->add_ref();
}

void SavedActParam::del_ref() {
  FORALL (obj, it) (*it)->del_ref();
}

void SavedActParam::save(ostream &os) const
{
  SaveLoad::save(obj, os);
  SaveLoad::save(val, os);

#if 0
  // old code
  sint4 s;
  s = obj.size();
  OS_WRITE(s);
  if (s > 0) os.write((char*)&obj[0], sizeof(obj[0])*s);

  s = val.size();
  OS_WRITE(s);
  if (s > 0) os.write((char*)&val[0], sizeof(val[0])*s);
#endif  
}

void SavedActParam::load(istream &is)
{
  SaveLoad::load(obj, is);
  SaveLoad::load(val, is);  

#if 0
  // old code
  sint4 s;
  IS_READ(s);
  obj.clear();
  FORU (i, s) {
    ScriptObj::Ptr p(0);
    IS_READ(p);
    obj.push_back(p);
  }

  is.read((char*)&s, sizeof(s));
  val.resize(s);
  if (s > 0) is.read((char*)&val[0], sizeof(val[0])*s);
#endif
  
}

//===================================================================

void ObjActionDefinition::parse_action_def(Tokenizer &tokenizer)
{
  string token;

  file_name = tokenizer.file_name();
  path = tokenizer.file_path();

  token = tokenizer.next_token();
  if (token != "(") PARSE_ERR2 ("'(' expected, found: ", token.c_str());

  FOREVER { // object id names
    token = tokenizer.next_token();
    if (token == ";") break;
    if (tokenizer.token_type() != Tokenizer::ID)
      PARSE_ERR2("object name expected", token.c_str());

    obj_id_names.push_back(token);
    
    token = tokenizer.next_token();
    if (token == ";") break;
    if (token != ",") PARSE_ERR("';' or ',' expected in object list");
  }

  FOREVER { // parameter list
    token = tokenizer.next_token();
    if (token == ";") break;
    if (tokenizer.token_type() != Tokenizer::ID)
      PARSE_ERR2("parameter name expected", token.c_str());
    
    param_names.push_back(token);
    
    token = tokenizer.next_token();
    if (token == ";") break;
    if (token != ",") PARSE_ERR("';' or ',' expected in parameter list");
  }

  FOREVER { // state parameters
    token = tokenizer.next_token();
    if (token == ")") break;
    if (tokenizer.token_type() != Tokenizer::ID)
      PARSE_ERR2("state name expected", token.c_str());

    state_param_names.push_back(token);
    
    token = tokenizer.next_token();
    if (token == ")") break;
    if (token != ",") PARSE_ERR("')' or ',' expected in state list");
  }
  
  num_obj_ids      = obj_id_names.size();
  num_params       = param_names.size();
  num_state_params = state_param_names.size();

  token = tokenizer.next_token();

  if (num_obj_ids == 0 && num_params == 0)
    format = ObjActionDefinition::NONE;

  else if (num_obj_ids == 1 && num_params == 0)
    format = ObjActionDefinition::OBJ_ID;

  else if (num_obj_ids == 0 && num_params == 2 && token == "coord") {
    format = ObjActionDefinition::COORD;
    token = tokenizer.next_token();

  } else
    format = ObjActionDefinition::OTHER;
  
  if (token != "{") PARSE_ERR("'{' expected at start of script");
  
  script = "{ ";
  int b_count = 1; // brace count
  
  while (b_count > 0 && token != "") {
    token = tokenizer.next_raw_token();
    script += token;
    
    if (token == "{") ++b_count;
    if (token == "}") --b_count;
  }
}

//===================================================================

ObjAction::ObjAction(ActionFunction action, ObjActionDefinition &def)
{
  t = COMPILED;
  hidden = def.hidden;
  format = def.format;
  CFunc = action;
  SFunc = NULL;
  Script = NULL;
  num_obj_ids = def.num_obj_ids;
  num_params = def.num_params;
  num_state_params = def.num_state_params;
  all_actions.push_back(this);
}

ObjAction::ObjAction(ActionScript *action, ObjActionDefinition &def)
{
  t = SCRIPTED;
  hidden = def.hidden;
  format = def.format;
  CFunc = NULL;
  SFunc = NULL;
  Script = action;
  num_obj_ids = def.num_obj_ids;
  num_params = def.num_params;
  num_state_params = def.num_state_params;
  all_actions.push_back(this);
}

ObjAction::ObjAction(ObjFunction action, ObjActionDefinition &def)
{
  t = CLASS_FUNC;
  hidden = def.hidden;
  format = def.format;
  CFunc = NULL;
  SFunc = action;
  Script = NULL;
  num_obj_ids = def.num_obj_ids;
  num_params = def.num_params;
  num_state_params = def.num_state_params;
  all_actions.push_back(this);
}

/** Creates an action from an action definition and points that definition to the
    new action.
*/

void ObjAction::create(ObjActionDefinition &def, PlayerInfo &pi, Game &game,
                       std::string &bp_name)
{
  ScriptParser parser;
  parser.set_context(pi, game, def.file_name, def.line_num);
  parser.current_action_name = bp_name + "::" + def.name;
  
  ActionScript *script = parser.Parse(def);
  SCRIPT_ASSERT(script && "script ptr 0");

  ObjAction *a = new ObjAction(script, def);
  SCRIPT_ASSERT(a && "action ptr 0");
  def.func = a;
  def.script.clear();
}

void ObjAction::clear()
{
  FORALL (all_actions, i) delete *i;
}

sint4 ObjAction::act_index(ObjAction *a)
{
  FORALL (all_actions, i) if (*i == a) return i-all_actions.begin();
  SCRIPT_ERR("can't find action");
  return -1;
}

ObjAction* ObjAction::get_action(sint4 i)
{
  SCRIPT_ASSERT(i >= 0 && i < (sint4)(all_actions.size()));
  return all_actions[i];
}

//-------------------------------------------------------------------

stype ObjAction::Act(const SActParam &p) {
  stype result;

  if (t == COMPILED) {
    result = CFunc(p);
  } else if (t == SCRIPTED) {
    dir_path.push(Script->get_pwd());
    result = Script->Act(p);
    dir_path.pop();
  } else if (t == CLASS_FUNC) {
    SActParam q = p;
    SCRIPT_ASSERT(p.obj.size() == q.obj.size() && "obj# mismatch");
    SCRIPT_ASSERT(p.val.size() == q.val.size() && "val# mismatch");    
    ScriptObj *b = q.obj[0];
    q.obj.erase(q.obj.begin());
    result = (b->*SFunc)(q);
  } else {
    SCRIPT_ERR("empty action");
  }

  // calling the action increments the reference count of the parameters
  // remove those references
  return result;
}

//-------------------------------------------------------------------

string ObjAction::get_pwd()
{
  if (dir_path.empty()) return "";
  return dir_path.top();
}

//===================================================================

string Action::to_string(ObjId cid) const
{
  if (id == NOOP) return "";

  string u;

  write_b_int4(u, cid);
  write_b_int4(u, id);
  FORALL (params, i) write_b_int4(u, (*i));

  return u;
}

//===================================================================

/** Read an encoded list of actions for each player, add those actions to
    the appropriate queue (motion or action), and then evaluate each queue.
    Updates the list of dead objects.
*/

void Game::exec_actions(Vector<string> &commands)
{
  if (commands.size() != static_cast<uint4>(player_n)) ERR("wrong action vector size");

  set<ScriptObj*> dead;
  ScriptObj::collect_dead(dead); // call destructors for dead objects

  //  remove dead objects from the object list

  FORS (q, player_n+1) {

    ObjCont &obj = objs[q];
    ObjCont::iterator it = obj.begin();

    while (it != obj.end()) {
      if (dead.find(*it) != dead.end()) {
        it = obj.erase(it);
        continue;
      }
      ++it;
    }
  }

  ScriptObj::collect_destroyed(); // delete dead objects


  // parse the action messages from the client
  // "object id" "action number" [parameters]

  FORS (i, player_n) {

    if (commands[i].empty()) continue;
    const char *b = commands[i].c_str();
    const char *ss = b;
    sint4 length = commands[i].size();

    string token;
    sint4 x_id;
    ScriptObj *x;

    //    cout << "Player:" << i << " : " << ss.str() << endl;

    while (ss - b < length) {

      x_id = read_b_int4(ss);
      x = players[i]->get_obj(x_id);
      if (x == 0) TEST_ERR (x_id << " NOT AN OBJECT\n");
      if (x->get_int("owner") != i) {
        cerr << "owner = " << x->get_int("owner") << endl;
        cerr << "player = " << i << endl;
        TEST_ERR("NOT THE OWNER\n");
      }
      sint4 action_id = read_b_int4(ss);

      if (action_id <= -1) { // move

        GameObj *gob = x->get_GameObj();
        if (gob != 0) {
          sint4 speed = -1 - action_id;
          sint4 tx = read_b_int4(ss);
          sint4 ty = read_b_int4(ss);
          if (players[i]->valid_id(x_id)) move_actions.add(gob, tx, ty, speed);
        } else {
          REM("BUH?  Not a GameObj???");
        }

      } else {
	
	Vector<stype> params;
        ObjAction &a = *x->action(action_id);
        FORU (j, a.num_obj_ids + a.num_params) {
          sint4 p = read_b_int4(ss);
          params.push_back(stype(p));
        }
        if (players[i]->valid_id(x_id))
          simple_actions.add(x, action_id, params, 0);
        //	ERR("unknown action");
      }
    }
  }

  FORALL (actions, i) (*i)->Act(); // movement and object actions

  if (opt.cpind == -1) {
    FORS (q, player_n+1) { // mark dead objects
      ObjCont &obj = objs[q];
      FORALL (obj, it) {
        const sintptr *hp = (*it)->get_const_int_ptr("hp");
        if (hp && *hp < 1) (*it)->kill();
      }
    }
  }
}

//===================================================================

/** Add action requests from the client.
    @param x       The parent object.
    @param act_id  The action id for the parent (could refer to a child's action).
    @param param   A vector of object ids and integer parameters for the action.
    @param delay   How many ticks from now the action should be executed.
*/

void SimpleActions::add(ScriptObj *x, sint4 act_id, Vector<stype> &param, uint4 delay)
{
  if (!x->action_enabled(act_id)) { cout << "action not enabled" << endl; return; }
  if (x->action_hidden(act_id)) { cout << "hidden action" << endl; return; }

  saction *q = new saction;
  q->active = true;
  q->zero_tick = game->tick() + delay;
  q->sec_key = game->rand.rand_uint4();
  q->a = x->action(act_id);
  // copy the non-object id parameters
  for (uint4 i = q->a->num_obj_ids; i < q->a->num_obj_ids + q->a->num_params; ++i) {
    q->p.val.push_back(param[i]);
  }
  // state parameters are 0
  FORU (i, q->a->num_state_params) q->p.val.push_back(stype());

  q->obj_ref = true;
  q->queue_ref = true;

  q->p.obj.push_back(x->action_component(act_id)); // "this"

  //  const PlayerInfo &pi = game->get_player_info(x->get_owner());
  const PlayerInfo &pi = game->get_player_info(x->get_int("owner"));

  // obj parameters, not including "this"
  FORU (i, q->a->num_obj_ids) {
    ScriptObj *gob = pi.get_obj(param[i].to_int());
    if (gob == 0) {
      cout << "obj parameter " << i << " : " << q->p.val[i].to_int() << " failed" << endl;
      delete q;
      return;
    }
    if (!pi.valid_id(param[i].to_int())) {
      cout << "id not valid" << endl;
      delete q;
      return;
    }
    q->p.obj.push_back(gob);
  }

  GameObj *g = x->get_GameObj();
  if (g) *g->last_command[act_id] = q;
  q->obj_act_id = -1;

  q->p.add_ref();
  actions.push(q);
}

//-------------------------------------------------------------------
// local action requests

void SimpleActions::spawn(saction *q)
{
  q->p.add_ref();
  actions.push(q);
}

//-------------------------------------------------------------------

// local action requests

stype SimpleActions::spawn_now(saction *sa)
{
  stype res;

  if (sa->active) {
    if (sa->obj_act_id == -1) {
      res = sa->a->Act(sa->p);
    } else {
#if 0
      cout << sa->p.obj[0]->bp_name() << " CAN'T ACT '" << sa->name << "'" << endl;
#else
      // can't act, set the action as outgoing
      Vector<sint4> params;
      for (uint4 i = 1; i < sa->p.obj.size(); ++i)
        params.push_back(game->get_cplayer_info().get_id(sa->p.obj[i]));
      
      FORALL (sa->p.val, it) params.push_back(it->to_int());
      
      //        if (*sa->p.obj[0]->parent->sod.owner == game->get_client_player())
      sa->p.obj[0]->parent->set_action(sa->obj_act_id, params);
#endif
    }
  }

  if (!sa->obj_ref) delete sa;
  else              sa->queue_ref = false;

  return res;
}

//-------------------------------------------------------------------

void SimpleActions::Act()
{
  while (!actions.empty()) {

    saction* const sa = actions.top(); // constant pointer to a simple action on the heap
    if (sa->zero_tick > game->tick()) break;

    // The action is going to be executed, and may place other actions on the queue;
    // make sure that we remove it now, because it might not be first anymore if that happens.
    actions.pop();
    
    sa->p.del_ref();

    if (sa->p.obj[0]->parent->is_dead()) { // the object died before it could execute the action
      if (!sa->obj_ref) delete sa;
      else              sa->queue_ref = false;
      continue;
    }
    
    if (sa->active) {
      if (sa->obj_act_id == -1) {
        sa->a->Act(sa->p);
      } else {
        // can't act, set the action as outgoing
        Vector<sint4> params;
        for (uint4 i = 1; i < sa->p.obj.size(); ++i)
          params.push_back(game->get_cplayer_info().get_id(sa->p.obj[i]));
        
        FORALL (sa->p.val, it) params.push_back(it->to_int());
        
        //        if (*sa->p.obj[0]->parent->sod.owner == game->get_client_player())
        sa->p.obj[0]->parent->set_action(sa->obj_act_id, params);
      }
    }
    
    if (!sa->obj_ref) delete sa;
    else              sa->queue_ref = false;
  }
}

//-------------------------------------------------------------------

SimpleActions::~SimpleActions()
{
  while (!actions.empty()) {
    saction* const sa = actions.top(); // assume that the queue owns everything
    if (!sa->obj_ref) delete sa;             // the object has abandoned the action
    else              sa->queue_ref = false; // remove our claim
    actions.pop();
  }
}

//-------------------------------------------------------------------

void SimpleActions::save(ostream &/*os*/) const
{
  ERR("FIXME: save has side effects, must be const");

#if 0  


  sint4 flag = 1;

  while (!actions.empty()) {

    OS_WRITE(flag);

    saction* const sa = actions.top(); // constant pointer to a simple action on the heap
    actions.pop();
    
    sa->p.del_ref();

    if (sa->p.obj[0]->parent->is_dead()) { // the object died before it could execute the action
      if (!sa->obj_ref) delete sa;
      else              sa->queue_ref = false;
      continue;
    }
    
    if (sa->active) {
      sint4 s = sa->name.size();
      OS_WRITE(s);
      os.write(sa->name.c_str(), s);
      OS_WRITE(sa->zero_tick);
      OS_WRITE(sa->sec_key);

      sint4 i = ObjAction::act_index(sa->a);
      OS_WRITE(i);
      OS_WRITE(sa->obj_act_id);

      sa->p.save(os); // save parameter list

      OS_WRITE(sa->obj_ref);
      OS_WRITE(sa->queue_ref);
    }
  }

  flag = 0;
  OS_WRITE(flag);

#endif
}

//-------------------------------------------------------------------

void SimpleActions::load(istream &is)
{
  sint4 flag;
  FOREVER {
    IS_READ(flag);
    if (flag == 0) break;

    saction *sa = new saction;
    sa->active = true;

    sint4 s;
    IS_READ(s);
    FORU (i, s) {
      char c;
      IS_READ(c);
      sa->name += c;
    }
    IS_READ(sa->zero_tick);
    IS_READ(sa->sec_key);

    sint4 i;
    IS_READ(i);
    sa->a = ObjAction::get_action(i);
    IS_READ(sa->obj_act_id);

    sa->p.load(is);

    IS_READ(sa->obj_ref);
    IS_READ(sa->queue_ref);

    sa->p.add_ref();
    actions.push(sa);
  }
}

//===================================================================

void MoveActions::Act()
{
  if (game->get_client_player() == -1) {
    FORALL (moves, i) game->obj_move_to(*i->a, i->t, i->speed);
  } else {
    FORALL (moves, i) {
      //      cout << "SENDING MOVE ACTION" << endl;
      Vector<sint4> params;
      params.push_back(i->t.x);
      params.push_back(i->t.y);
      params.push_back(i->speed);
      i->a->set_action("move", params);
    }
  }
  moves.clear();
}

//===================================================================

stype CompiledAction1(Vector<GameObj*> &objs, Vector<stype> &)
{
  GameObj &gob = *objs[0];
  sint4 i = gob.get_int("radius") + 2;
  gob.set_int("radius", i);
  //  sint4 i = gob.get_int("marker") +1;
  //  if (i > 5) i = 1;
  //  gob.set_int("marker", i);
  return stype(sint4(1));
}

//===================================================================
//===================================================================
//===================================================================

/** Creates a list of movement actions for each player's objects. */

void Game::gen_actions(Vector<string> &actions)
{
  FORS (p, player_n) {

    actions[p].clear();
    ObjCont &obj = objs[p];

    FORALL (obj, it) {

      GameObj *foo = (*it)->get_GameObj();
      if (!foo) continue;
      GameObj &gob = *foo;

      if (*gob.sod.speed == 0 && !gob.is_dead()) {

        ScalarPoint t;

        if ((rand.rand_uint4() % 100) < 33) {

          // random target on fringe (x or y unchanged)
	    
          switch (rand.rand_uint4() & 3) {
          case 0: t.x = *gob.sod.x; t.y = 0; break;
          case 1: t.x = *gob.sod.x; t.y = points_y-1; break;
          case 2: t.x = 0; t.y = *gob.sod.y; break;	      
          case 3: t.x = points_x-1; t.y = *gob.sod.y; break;
          }
	  
        } else {

          // random target on fringe

          switch (rand.rand_uint4() & 3) {

          case 0:
            t.x = 0;
            t.y = rand.rand_uint4() % (pf.get_width()*tile_points);
            break;

          case 1:
            t.x = pf.get_height()*tile_points - 1;
            t.y = rand.rand_uint4() % (pf.get_height()*tile_points);
            break;
	    
          case 2:
            t.x = rand.rand_uint4() % (pf.get_width()*tile_points);
            t.y = 0;
            break;

          case 3:
            t.x = rand.rand_uint4() % (pf.get_height()*tile_points);
            t.y = pf.get_height()*tile_points - 1;
            break;
          }
        }

        sint4 a = players[p]->get_id(&gob);

        Vector<sint4> params;
        params.push_back(t.x);
        params.push_back(t.y);
        params.push_back(*gob.sod.max_speed);
	
        gob.set_action("move", params);
	actions[p] += gob.cur_action.to_string(a);
      }
    }
  }
}

void Game::gen_client_actions()
{
  ObjCont &obj = objs[opt.cpind];

  FORALL (obj, it) {

    GameObj *foo = (*it)->get_GameObj();
    if (!foo) continue;
    GameObj &gob = *foo;

#if 0
    
    if (*gob.sod.speed)
    if (*gob.sod.shape == Object::CIRCLE) {
      if (gob.components.find("weapon") != gob.components.end()) {
        ScriptObj *w = gob.components["weapon"];
        sint4 r = w->get_int("max_ground_range");
        sint4 x0 = fine_grid(*gob.sod.x - r) / sec_wh;
        sint4 x1 = fine_grid(*gob.sod.x + r) / sec_wh;
        sint4 y0 = fine_grid(*gob.sod.y - r) / sec_wh;
        sint4 y1 = fine_grid(*gob.sod.y + r) / sec_wh;
        if (x0 < 0) x0 = 0;
        if (y0 < 0) y0 = 0;
        if (x1 >= sec_nx) x1 = sec_nx - 1;
        if (y1 >= sec_ny) y1 = sec_ny - 1;

        GameObj *targ = 0;
        sint4 min_hp = 1000000;

        for (sint4 i = x0; i <= x1; ++i) {
          for (sint4 j = y0; j <= y1; ++j) {
            sint4 ind = j * sec_nx + i;
            Array<GameObj*> &a = motion_secs.secs[ind];
            FORALL (a, i) {
              GameObj *it = *i;
              if (*it->sod.owner < player_n && *it->sod.owner != *gob.sod.owner) {
                if (distance(gob, *it) <= r) {
                  if (it->has_attr("hp") && (!targ || it->get_int("hp") < min_hp)) {
                    if (can_see_obj(gob, *it)) {
                      targ = it;
                      min_hp = it->get_int("hp");
                    }
                  }
                }
              }
            }
          }
        }
        if (targ) {
          Vector<sint4> params;
          params.push_back( players[*gob.sod.owner]->get_id(targ) );
          w->set_action("attack", params);
          continue;
        }
      }
    }
#endif
    
    if (*gob.sod.max_speed == 0) continue;
    sintptr *ptr = gob.get_int_ptr("speed");

    if (ptr && !*ptr) {

      ScalarPoint t;

      if ((rand.rand_uint4() % 1000) < 333) {

        // random target on fringe (x or y unchanged)
	
        switch (rand.rand_uint4() & 3) {
        case 0: t.x = *gob.sod.x; t.y = 0; break;
        case 1: t.x = *gob.sod.x; t.y = points_y-1; break;
        case 2: t.x = 0; t.y = *gob.sod.y; break;	      
        case 3: t.x = points_x-1; t.y = *gob.sod.y; break;
        }
        
      } else {

        // random target on fringe
	
        switch (rand.rand_uint4() & 3) {

        case 0:
          t.x = 0;
          t.y = rand.rand_uint4() % (pf.get_height()*tile_points);
          break;

        case 1:
          t.x = pf.get_width()*tile_points - 1;
          t.y = rand.rand_uint4() % (pf.get_height()*tile_points);
          break;

        case 2:
          t.x = rand.rand_uint4() % (pf.get_width()*tile_points);
          t.y = 0;
          break;

        case 3:
          t.x = rand.rand_uint4() % (pf.get_width()*tile_points);
          t.y = pf.get_height()*tile_points - 1;
          break;
        }
      }

      Vector<sint4> params;
      params.push_back(t.x);
      params.push_back(t.y);
      params.push_back(*gob.sod.max_speed);
      
      gob.set_action("move", params);
    }
  }
}

//===================================================================

