#ifndef ByteData_H
#define ByteData_H

/** @file ByteData.H
    Binary encoding of integers.

    $Id: ByteData.H,v 1.4 2005/07/09 00:42:52 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/ByteData.H,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"


inline void write_b_int4(char* &s, sint4 num)
{
  ERR("foo");
  char x;

  if (num < 0) { x = 0x40; num = -num; }
  else         { x = 0x00; }

  x += num & 0x3F; num >>= 6;

  while (num) {
    x += 0x80; *s++ = x;
    x = num & 0x7F; num >>= 7;
  }
  *s++ = x;
}

//-------------------------------------------------------------------

inline void write_b_int8(char* &s, sint8 num)
{
  ERR("foo");
  char x;

  if (num < 0) { x = 0x40; num = -num; }
  else         { x = 0x00; }

  x |= num & 0x3F; num >>= 6;

  while (num) {
    x |= 0x80; *s++ = x;
    x = num & 0x7F; num >>= 7;
  }
  *s++ = x;
}

//-------------------------------------------------------------------

inline void write_b_flat4(char* &s, uint4 num)
{
  ERR("foo");
  FORU(i, 4) {
    s[3-i] = num & 255;
    num >>= 8;
  }
  s+=4;
}

//-------------------------------------------------------------------

inline void write_b_int4(std::string &s, sint4 num)
{
  char x;

  if (num < 0) { x = 0x40; num = -num; }
  else         { x = 0x00; }

  x += num & 0x3F; num >>= 6;

  while (num) {
    x += 0x80; s += x;
    x = num & 0x7F; num >>= 7;
  }
  s += x;
}

//-------------------------------------------------------------------

inline void write_b_int8(std::string &s, sint8 num)
{
  char x;

  if (num < 0) { x = 0x40; num = -num; }
  else         { x = 0x00; }

  x |= num & 0x3F; num >>= 6;

  while (num) {
    x |= 0x80; s += x;
    x = num & 0x7F; num >>= 7;
  }
  s += x;
}

//-------------------------------------------------------------------

inline void write_b_flat4(std::string &s, uint4 num)
{
  char u[4];
  FORU (i, 4) {
    u[3-i] = num & 255;
    num >>= 8;
  }
  s.append(u, 4);
}

//===================================================================

/** Read an integer from a byte string and advance the pointer. */

inline sint4 read_b_int4(const char* &s)
{
  bool negate = *s & 0x40;
  sint4 shift = 0;

  sint4 v = *s & 0x3F; shift += 6;

  while (*s & 0x80) {
    s++;
    if (shift > 32) ERR("danger");
    
    v += sint4(*s & 0x7F) << shift;
    shift += 7;
  }
  s++;

  if (negate) v = -v;
  return v;
}

//-------------------------------------------------------------------

inline sint8 read_b_int8(const char* &s)
{
  bool negate = *s & 0x40;
  sint4 shift = 0;

  sint8 v = *s & 0x3F; shift += 6;

  while (*s & 0x80) {
    s++;
    if (shift > 57) ERR("more danger");

    v += sint8(*s & 0x7F) << shift;
    shift += 7;
  }
  s++;

  if (negate) v = -v;
  return v;
}

//-------------------------------------------------------------------

inline uint4 read_b_flat4(const char* &s)
{
  uint4 x = uint1(*s++);
  FORU(i,3) {
    x <<= 8;
    x += uint1(*s++);
  }
  return x;
}

//===================================================================

#endif
