#ifndef GUI_H
#define GUI_H

/** @file GUI.H
    Basic/debug server display.

    $Id: GUI.H,v 1.6 2005/09/06 05:50:48 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/GUI.H,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"

class Options;
class GameObj;
class Game;

struct RGB {
  uint1 R, G, B;
  uint4 col;

  void from(sint4 r, sint4 g, sint4 b=0) {
    R = r; G = g; B = b;
    col = ((((R << 8)+G) << 8) + B);
  }
  
  RGB(sint4 r=0, sint4 g=0, sint4 b=0) : R(r), G(g), B(b) {
    //col = (((((R << 8)+G) << 8) + B) << 8) + 255;
    col = ((((R << 8)+G) << 8) + B);    
  }
  
  uint4 color() const { return col; }
  uint4 color_alpha(sint4 alpha=255) const {
    assert(alpha >=0 && alpha < 256);
    return (col << 8) + alpha;
  }  
};


struct GfxView2 {
  sint4 x, y, w, h; // rectangle on screen
  sint4 mx, my;     // upper left pixel on map
  sint4 mag;
};


/** Basic/debug server display */

class GUI {

protected:

  // options
  sint4 DISP_VISI, DISP_SEEN, DISP_KNOW, DISP_GRID;

public:

  enum { MARKER_NO, MARKER_H, MARKER_V, MARKER_HV, MARKER_DOT, MARKER_CIRCLE };

  Game *game;
  SDL_Surface *screen;
  sint4 s_x, s_y;
  sint4 gfx_mag_frac;
  sint4 adj_x, adj_y;
  Vector<GfxView2> views;
  sint4 active_view;
  GfxView2 minimap;
  bool quit;  // true <=> quit application
  sint4 scale(sint4 x, sint4 mag) const;
  
  GUI();
  ~GUI();
  void init(sint4 width, sint4 height, Game &g);

  void event();  
  bool adjust_view_xy(sint4 dx, sint4 dy);
  bool adjust_view_mag(sint4 dm);
  bool correct_view();
  void display(const GameObj* obj1=0, const GameObj* obj2=0);
  void display(const GfxView2 &gs, const GameObj* obj1=0, const GameObj* obj2=0);
  void display_minimap(const GfxView2 &active_view);
  void draw_marker(sint4 cx, sint4 cy, sint4 r, sint4 m);

  static void add_options();

private:

  // not implemented
  GUI(const GUI &) { ERR("what"); }
  GUI &operator=(const GUI &) { ERR("what"); }
};

REGISTER_TYPEOF(66, Vector<GfxView2>::iterator);
#endif
