#ifndef GameObjFactory_H
#define GameObjFactory_H

/** @file GameObjFactory.H
    Creation and tracking of objects.

    $Id: GameObjFactory.H,v 1.17 2006/04/10 12:07:19 ddeutscher Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/GameObjFactory.H,v $
*/
// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"

class PlayerInfo;
class Game;
class ScriptObj;
class GameObjBlueprint;
class Tokenizer;

typedef ScriptObj* (*ScriptObjConstructor)();

/** Creates objects from an internal set of blueprints. */

class GameObjFactory {

public:

  struct BPConstructorData {
    ScriptObjConstructor c;
    sint4 num_actions;      ///< Components need an action count to indicate the number
                            ///< of actions available to the client.
    BPConstructorData() : num_actions(0) { }
  };

  typedef std::map<sint4,             GameObjBlueprint*> Code2BlueprintMap;
  typedef std::map<const std::string, GameObjBlueprint*> Name2BlueprintMap;
  typedef std::map<sint4,             BPConstructorData> Code2ConstructorMap;
  typedef std::map<const std::string, BPConstructorData> Name2ConstructorMap;
  typedef std::map<const std::string, sint4>             Name2CodeMap;

protected:

  Name2BlueprintMap name2blueprint;
  Code2BlueprintMap code2blueprint;
  Name2ConstructorMap name2constructor;
  Code2ConstructorMap code2constructor;
  Name2CodeMap name2code;

  sint4 code; ///< Next available blueprint code.

  std::set<std::string> blueprint_names;
  sint4 check_registration(const std::string &name);

  PlayerInfo *pi;

public:

  Game *game;

  GameObjFactory() { code = 0; pi = 0; game = 0; }
  ~GameObjFactory();
  
  void register_blueprint(const std::string &name, const GameObjBlueprint &bp);
  void register_blueprint(const std::string &name, ScriptObjConstructor constructor, sint4 num_actions);

  GameObjBlueprint* get_blueprint(const std::string &name);

  void bind(PlayerInfo &pi, Game &game);
  void lock();

  sint4 bp_code(const std::string &name);
  std::string bp_name(sint4 code);

  // environment must destroy GameObj!
  ScriptObj* create(const std::string &name, bool construct = true);

  // for client
  ScriptObj* create(sint4 code, bool construct = false);

  friend class ScriptParser;
  friend class Game;

private:

  GameObjFactory(const GameObjFactory&) { ERR("what1"); }
  void operator=(const GameObjFactory&) { ERR("what2"); }
};

REGISTER_TYPEOF(80, GameObjFactory::Name2BlueprintMap::iterator);
REGISTER_TYPEOF(82, GameObjFactory::Name2CodeMap::iterator);
REGISTER_TYPEOF(84, GameObjFactory::Code2BlueprintMap::iterator);
REGISTER_TYPEOF(86, GameObjFactory::Name2ConstructorMap::iterator);


#endif
