/** @file GameScript.C
    @see Game.H.

    $Id: GameScript.C,v 1.41 2006/04/25 03:08:38 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/GameScript.C,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "Action.H"
#include "Game.H"
#include "GameObj.H"
#include "GameTile.H"
#include "Random.H"
#include "Script.H"
#include "ScriptVector.H"

using namespace std;

#if 0 // track object creation / destruction

sint4 s_created = 0;
sint4 s_deleted = 0;

#define CREATION(n)    s_created++; cout << "created = " << s_created << " deleted = " << s_deleted\
                       << " " << s_created - s_deleted << " " << (n) << endl;
#define DESTRUCTION(n) s_deleted++; cout << "created = " << s_created << " deleted = " << s_deleted\
                       << " " << s_created - s_deleted << " " << (n) << endl;
#else

#define CREATION(n)
#define DESTRUCTION(n)

#endif

//=============================================================================

stype Game::script_GOB_Move(const SActParam &p)
{
  OBJN_CHECK(1);
  if (p.val.size() < 2) SCRIPT_ERR("not enough move paramaters");
  if (p.val.size() > 3) SCRIPT_ERR("too many move paramaters");

  SCRIPT_ASSERT(p.obj[0]);
  
  GameObj *a = dynamic_cast<GameObj*>(p.obj[0]);
  SCRIPT_ASSERT(a && "not a game object");

  sint4 x = p.val[0].to_int();
  sint4 y = p.val[1].to_int();
  sint4 speed = p.val.size() > 2 ? p.val[2].to_int() : *a->sod.max_speed;

  move_actions.add(a, x, y, speed);
  return stype();
}

//=============================================================================

stype Game::script_GOB_Fix_Position(const SActParam &p)
{
  OBJN_CHECK(1);
  GameObj *aa = dynamic_cast<GameObj*>(p.obj[0]);
  SCRIPT_ASSERT(aa && "not a game object");
  // if (aa == 0) return stype::zero;
  GameObj &a = *aa;

  update_pos(a);
  return stype();
}

//=============================================================================

stype Game::script_GOB_Can_See(const SActParam &p)
{
  OBJN_CHECK(2);
  GameObj *a = dynamic_cast<GameObj*>(p.obj[0]);
  GameObj *b = dynamic_cast<GameObj*>(p.obj[1]);

  SCRIPT_ASSERT(a != 0 && b != 0 && "not 2 game objects");
  // if (a == 0 || b == 0) return stype();

  if (can_see_obj(*a, *b)) return stype(sint4(1));
  return stype();
}

//=============================================================================

stype Game::script_GOB_Footprint(const SActParam &p)
{
  VALN_CHECK(1);
  OBJN_CHECK(1);  
  
  GameObj *a = dynamic_cast<GameObj*>(p.obj[0]);

  SCRIPT_ASSERT(a && "not a game object");
  
  //if (a == 0) return stype(1);
  
  Vector<stype> &v = *ScriptVector::stype_to_vecstype(p.val[0]);

  //  v.clear();

  Vector<sint4> obj_tiles;
  compute_footprint(*a, obj_tiles);

#if 1
  FORALL (obj_tiles, i) v.push_back(stype(sint4(*i)));

#else
  FORALL (obj_tiles, i) {
    v.push_back(tile2tx(*i));
    v.push_back(tile2ty(*i));
  }
#endif

  return stype();
}

//=============================================================================

stype Game::script_cout(const SActParam &p)
{
  VALGEN_CHECK(1);
  cout << stype::print_va_list(p.val[0].to_cstr(), p.val, 1);
  return stype();
}

//=============================================================================

stype Game::script_end_game(const SActParam &p)
{
  VALN_CHECK(1);
  cout << "Player " << p.val[0].to_int() << " wins." << endl;
  game_over = true;
  return p.val[0];
}

//=============================================================================

stype Game::script_save_game(const SActParam &)
{
  save_req = true;
  return stype();
}

//=============================================================================

stype Game::script_Tick(const SActParam &) {
  return stype(tick());
}

//=============================================================================

stype Game::script_map_width(const SActParam &) {
  return stype(tile_points * pf.get_width());
}

//=============================================================================

stype Game::script_map_height(const SActParam &) {
  return stype(tile_points * pf.get_height());
}

//=============================================================================

stype Game::script_printnum(const SActParam &p) {
  VALGEN_CHECK(1);
  cout << p.val[0].to_int() << endl;
  return stype();
}

//=============================================================================

stype Game::script_Distance(const SActParam &p)
{
  OBJN_CHECK(2);  
  GameObj *a = dynamic_cast<GameObj*>(p.obj[0]);
  GameObj *b = dynamic_cast<GameObj*>(p.obj[1]);
  SCRIPT_ASSERT(a != 0 && b != 0 && "not 2 game objects");
  // if (a == 0 || b == 0) return -1;
  return stype(static_cast<sint4>(floor(distance(*a, *b))));
}

//=============================================================================

stype Game::script_Notify_Attack(const SActParam &p)
{
  OBJN_CHECK(2);    
  GameObj *a = dynamic_cast<GameObj*>(p.obj[0]);
  GameObj *b = dynamic_cast<GameObj*>(p.obj[1]);
  SCRIPT_ASSERT(a != 0 && b != 0 && "not two game objects");
  a->notify_attack(b, p.val[0].to_int());
  return stype();
}

//=============================================================================

stype Game::script_GOB_Kill(const SActParam &p)
{
  OBJN_CHECK(1);  
  ScriptObj *s = p.obj[0];
  DESTRUCTION(s->bp_name());
  s->kill();

  if (s->creator->get_id(s) == 0) {
    ObjCont &oc = *s->creator->my_objs;
    
    // fixme: finding objects should be faster
    FORALL_B (oc, i) if (*i == s) break;
    SCRIPT_ASSERT(i != oc.end() && "object not found");
    oc.erase(i);
  }

  return stype();
}

//-----------------------------------------------------------------------------

stype Game::script_GOB_Enable(const SActParam &p)
{
  OBJN_CHECK(1);
  p.obj[0]->enable();
  return stype();
}

//-----------------------------------------------------------------------------

stype Game::script_GOB_Disable(const SActParam &p)
{ 
  OBJN_CHECK(1); 
  p.obj[0]->disable();
  return stype();
}

//-----------------------------------------------------------------------------

stype Game::script_GOB_Is_Enabled(const SActParam &p)
{
  OBJN_CHECK(1);
  return p.obj[0]->is_enabled() ? stype(1) : stype(0);
}

//-----------------------------------------------------------------------------

stype Game::script_GOB_Is_Alive(const SActParam &p)
{
  OBJN_CHECK(1);
  return stype(!p.obj[0]->is_dead());
}

//-----------------------------------------------------------------------------

stype Game::script_GOB_Type(const SActParam &p)
{
  OBJN_CHECK(1);  
  return stype(p.obj[0]->code());
}

//-----------------------------------------------------------------------------

stype Game::script_GOB_Type2Name(const SActParam &p)
{
  /// @bug Won't work if players are created differently.
  VALN_CHECK(1);
  string name = get_player_info(0).gof.bp_name(p.val[0].to_int());
  char *n = new char[name.size()+1];
  strncpy(n, name.c_str(), name.size()+1);
  return stype(n);
}

//-----------------------------------------------------------------------------

stype Game::script_GOB_Visible_By(const SActParam &p)
{
  VALN_CHECK(1);
  OBJN_CHECK(1);  
  GameObj *a = dynamic_cast<GameObj*>(p.obj[0]);
  SCRIPT_ASSERT(a && "no game object");
  return a->sod.seen_by.is_set(p.val[0].to_int()) ? stype(1) : stype(0);
}

//-----------------------------------------------------------------------------

stype Game::script_GOB_Cancel_Action(const SActParam &p)
{
  VALN_CHECK(1);
  OBJN_CHECK(1);

  string action_name = p.val[0].to_cstr(); // fixme: ScriptString?
  GameObj *a = dynamic_cast<GameObj*>(p.obj[0]);
  SCRIPT_ASSERT(a && "no game object");

  GameObj::CommandMap &m = a->last_command_map;
  FIND (m, it, action_name);
  if (it == m.end()) {
    errstr << "cannot cancel unknown action '" << action_name.c_str() << "'" << endl;
    errstr << "bp = " << a->bp_name() << endl;
    SCRIPT_ERR("unknown action");
  }

  SimpleActions::saction *q = it->second;

  if (q == 0) return stype(1);
  q->active = false;
  return stype();
}

//=============================================================================

stype Game::script_GOB_Has_Attr(const SActParam &p)
{
  VALN_CHECK(1);
  OBJN_CHECK(1);
  string attr = p.val[0].to_cstr(); // fixme: ScriptString?
  
  bool t = p.obj[0]->has_attr(attr);
  if (t) return stype(1);
  if (p.obj[0]->component(attr) != 0) return stype(1);
  return stype();
}

//=============================================================================

stype Game::script_GOB_SOD(const SActParam &p)
{
  OBJN_CHECK(1);
  GameObj *a = p.obj[0]->get_GameObj();
  if (a == 0) return stype(1);
  a->update_sod();
  return stype();
}

//=============================================================================

stype Game::script_GOB_Create(const SActParam &p)
{
  VALN_CHECK(2);
  string obj_type = p.val[0].to_cstr(); // fixme: ScriptString?

  CREATION(obj_type);

  sint4 o = p.val[1].to_int();
  ScriptObj *obj;

  if (o >= 0 && o < player_n) {

    obj = players[o]->gof.create(obj_type);
    objs[o].push_back(obj); // add the object to the game

  } else {

    obj = NPC->gof.create(obj_type);
    if (o == -1) {
      o = player_n;
      objs[o].push_back(obj); // add as npc object
    } else {
      DESTRUCTION(obj_type);
      obj->kill(); // not player owned, will be destroyed after this cycle
    }

  }
  obj->set_int("owner", o, false);
  GameObj *temp = obj->get_GameObj();
  if (temp) temp->init_sod();
  
  return stype(obj);
}

//=============================================================================

stype Game::script_GOB_CreateId(const SActParam &p)
{
  VALN_CHECK(2);
  
  sint4 obj_type = p.val[0].to_int();
  sint4 o = p.val[1].to_int();
  ScriptObj *obj;
  string name;

  if (o >= 0 && o < player_n) {

    name = players[o]->gof.bp_name(obj_type);
    obj = players[o]->gof.create(name);
    objs[o].push_back(obj); // add the object to the game

  } else {

    name = NPC->gof.bp_name(obj_type);
    obj = NPC->gof.create(name);
    if (o == -1) {
      o = player_n;
      objs[o].push_back(obj); // add as npc object
    } else {
      obj->kill(); // not player owned, will be destroyed after this cycle
    }

  }
  obj->set_int("owner", o, false);
  GameObj *temp = obj->get_GameObj();
  if (temp) temp->init_sod();
   
  return stype(obj);
}

//=============================================================================

stype Game::script_GOB_Id(const SActParam &p)
{
  VALN_CHECK(1);
  OBJN_CHECK(1);
  return stype(players[p.val[0].to_int()]->get_id(p.obj[0]));
}

//=============================================================================

// note: this code is almost identical to QueueAction::eval()

stype Game::script_eval(const SActParam &p)
{
  VALGEN_CHECK(1);
  OBJGEN_CHECK(1);  
  
  string action_name = p.val[0].to_cstr(); // fixme: ScriptString?

  SimpleActions::saction *q = new SimpleActions::saction;
  q->name = action_name;
  q->active = true;
  q->zero_tick = tick();

  for (uint4 i = 1; i < p.val.size(); ++i) q->p.val.push_back(p.val[i]);
  FORALL (p.obj, i) q->p.obj.push_back(*i);

  ScriptObj *obj = p.obj[0];
  
  ObjAction *a = obj->action(action_name);

  if (a == 0) {
    //    cout << "ACTION NOT FOUND " << action_name << endl;
    delete q;
    return stype();
  }
  q->a = a;

  q->obj_act_id = obj->action_id(action_name);

  GameObj *gob = obj->get_GameObj();
  if (gob != 0) {
    SimpleActions::saction *sact = gob->last_command_map[action_name];
    if (sact != 0) {
      if (!sact->queue_ref) { delete sact; sact = 0; }
      else                  sact->obj_ref = false;
      gob->last_command_map[action_name] = 0;
    }
  }

  q->queue_ref = true;

  if (q->obj_act_id == -1) {
    q->obj_ref = false;
    return simple_actions.spawn_now(q);
  }
  
  if (gob != 0) {
    gob->last_command_map[action_name] = q;
    q->obj_ref = true;
  } else {
    q->obj_ref = false;
  }
  simple_actions.spawn(q);
  return stype();
}
//=============================================================================
stype Game::script_GOB_Get_Attr(const SActParam &p)
{
  VALN_CHECK(1);
  OBJN_CHECK(1);  
  string attr = p.val[0].to_cstr(); // fixme: ScriptString?
  return stype(p.obj[0]->get_int(attr));
}

//=============================================================================

stype Game::script_GOB_Get_Comp(const SActParam &p)
{
  VALN_CHECK(1);
  OBJN_CHECK(1);  
  ScriptObj *a  = p.obj[0];
  string comp = p.val[0].to_cstr(); // fixme: ScriptString?

  if (comp.empty()) return stype(a);

  Vector<string> c;
  size_t i = comp.find('.', 0);
  while (i != string::npos) {
    c.push_back(comp.substr(0, i));
    comp = comp.substr(i+1, comp.size() - (i+1));
  }
  c.push_back(comp);

  FORALL (c, i) {
    a = a->component(*i);
    if (a == 0) break;
  }
  return stype(a);
}

//=============================================================================

sint4 Game::num_intersect(GameObj &a)
{
  a.update_points();
  set<GameObj*> objects;
  Vector<sint4> sectors;
  compute_ms_footprint(a, sectors);

  FORALL (sectors, it) {
    FORS (k, motion_secs.secs[*it].size()) {
      objects.insert(motion_secs.secs[*it][k]);
    }
  }
  
  sint4 count = 0;

  //  cout << "main collides = " << *a.sod.collides << endl;

  FORALL (objects, it) {
    if (*it == &a) continue;
    //    cout << "   collides = " << *(**it).sod.collides << endl;
    if (*a.sod.zcat == Object::ZCAT_UNDEF ||
	*(**it).sod.zcat == Object::ZCAT_UNDEF ||
        ((*a.sod.collides & *(**it).sod.collides) && !(*a.sod.ncollides & *(**it).sod.ncollides))) {
      real4 d = distance(a, **it);
      if (d <= 0) {
        ++count;
        //        cout << d << " " << (**it).bp_name() << endl;
      }
    }
  }
  return count;
}

//===================================================================

bool can_intersect(const GameObj &e1, const GameObj &e2)
{
  if (!(*e1.sod.collides & *e2.sod.collides)) return false;
  if (*e1.sod.ncollides & *e2.sod.ncollides) return false;
  return true;
}

//===================================================================

/** Tries to position a unit close to another unit, respecting tile
    boundaries and collision masks.
*/

stype Game::script_close_to(const SActParam &p)
{
  OBJN_CHECK(2);
  VALN_CHECK(1);  

  set<sint4> fringe_W, fringe_E;
  set<sint4> interior_W;
  set<sint4> interior_E;

  GameObj *t1 = p.obj[0]->get_GameObj();
  GameObj *t2 = p.obj[1]->get_GameObj();
  if (t1 == 0 || t2 == 0) return stype(2);

  GameObj &obj = *t1;  // the location to be near to
  GameObj &targ = *t2; // the object to place
  sint4 max_r = p.val[0].to_int();

  sint4 orig_x = *targ.sod.x;
  sint4 orig_y = *targ.sod.y;

  Vector<sint4> footprint_W;
  Vector<sint4> footprint_E;
  compute_footprint(obj, footprint_W, footprint_E);
  
  FORALL (footprint_W, i) interior_W.insert(*i);
  FORALL (footprint_E, i) interior_E.insert(*i);

  fringe_W = interior_W;
  fringe_E = interior_E;

  FORS (r, max_r+1) {
    set<sint4> nfringe_W;
    set<sint4> nfringe_E;

    if (!r) {
      nfringe_W = interior_W;
      nfringe_E = interior_E;
      interior_W.clear();
      interior_E.clear();
    }

    // expand the fringe to the whole tile if it is accessible

    FORALL (fringe_W, i) {
      sint4 tind = *i;
      const GameTile &t = pf(tind);
      if (t.get_split() == Tile::NO_SPLIT || !t.boundary_int ||
	  !can_intersect(obj, *t.boundary_int)) {
	fringe_E.insert(tind);
      }
    }
    FORALL (fringe_E, i) {
      sint4 tind = *i;
      const GameTile &t = pf(tind);
      if (t.get_split() == Tile::NO_SPLIT || !t.boundary_int ||
	  !can_intersect(obj, *t.boundary_int)) {
	fringe_W.insert(tind);
      }
    }

    // check if the fringe's neighbouring tiles are accessible
    if (r) {
      FORALL (fringe_W, i) {
        sint4 tind = *i;
        sint4 x = pf.ind2x(tind);
        sint4 y = pf.ind2y(tind);
        
        const GameTile &t = pf(tind);
        if (x > 0) {
          if (!t.boundary_W || !can_intersect(obj, *t.boundary_W))
            nfringe_E.insert(tind-1);
        }
        if (t.get_split() == Tile::TB_SPLIT) {
          if (y < pf.get_height()-1) {
            if (!t.boundary_S || !can_intersect(obj, *t.boundary_S)) {
              const GameTile &t2 = pf(tind + pf.get_width());
              if (t2.get_split() == Tile::TB_SPLIT)
                nfringe_E.insert(tind+pf.get_width());
              else
                nfringe_W.insert(tind+pf.get_width());
            }
          }
        } else {
          if (y > 0) {
            if (!t.boundary_N || !can_intersect(obj, *t.boundary_N)) {
              const GameTile &t2 = pf(tind - pf.get_width());
              if (t2.get_split() == Tile::TB_SPLIT)
                nfringe_W.insert(tind-pf.get_width());
              else
                nfringe_E.insert(tind-pf.get_width());
            }
          }
        }
      }

      FORALL (fringe_E, i) {
        sint4 tind = *i;
        sint4 x = pf.ind2x(tind);
        sint4 y = pf.ind2y(tind);
        
        const GameTile &t = pf(tind);
        if (x < pf.get_width()-1) {
          if (!t.boundary_E || !can_intersect(obj, *t.boundary_E))
            nfringe_W.insert(tind+1);
        }
        if (t.get_split() == Tile::TB_SPLIT) {
          if (y > 0) {
            if (!t.boundary_N || !can_intersect(obj, *t.boundary_N)) {
              const GameTile &t2 = pf(tind - pf.get_width());
              if (t2.get_split() == Tile::TB_SPLIT)
                nfringe_W.insert(tind - pf.get_width());
              else
                nfringe_E.insert(tind - pf.get_width());
            }
          }
        } else {
          if (y < pf.get_height()-1) {
            if (!t.boundary_S || !can_intersect(obj, *t.boundary_S)) {
              const GameTile &t2 = pf(tind + pf.get_width());
              if (t2.get_split() == Tile::TB_SPLIT)
                nfringe_E.insert(tind + pf.get_width());
              else
                nfringe_W.insert(tind + pf.get_width());
            }
          }
        }
      }
    }

    set<sint4> n_fringe_W;
    set<sint4> n_fringe_E;
    set_difference(nfringe_W.begin(), nfringe_W.end(),
                   interior_W.begin(), interior_W.end(),
                   insert_iterator< set<sint4> >(n_fringe_W, n_fringe_W.begin()) );
    set_difference(nfringe_E.begin(), nfringe_E.end(),
                   interior_E.begin(), interior_E.end(),
                   insert_iterator< set<sint4> >(n_fringe_E, n_fringe_E.begin()) );

    set<sint4> n_interior_W;
    set<sint4> n_interior_E;
    set_union(interior_W.begin(), interior_W.end(),
              n_fringe_W.begin(), n_fringe_W.end(),
              insert_iterator< set<sint4> >(n_interior_W, n_interior_W.begin()) );
    set_union(interior_E.begin(), interior_E.end(),
              n_fringe_E.begin(), n_fringe_E.end(),
              insert_iterator< set<sint4> >(n_interior_E, n_interior_E.begin()) );

    fringe_W = n_fringe_W;
    fringe_E = n_fringe_E;

    set<sint4> possible;
    set_union(nfringe_W.begin(), nfringe_W.end(),
              nfringe_E.begin(), nfringe_E.end(),
              insert_iterator< set<sint4> >(possible, possible.begin()) );

    Vector<sint4> valid;

    FORALL (possible, it) {
      sint4 tind = *it;
      sint4 tx = pf.ind2x(tind);
      sint4 ty = pf.ind2y(tind);
      *targ.sod.x = tx * tile_points + tile_points / 2;
      *targ.sod.y = ty * tile_points + tile_points / 2;
      if (num_intersect(targ) == 0) {
	valid.push_back(tind);
      }
    }
    
    if (valid.size()) {
      sint4 v = rand.rand_uint4() % valid.size();
      sint4 tind = valid[v];
      sint4 tx = pf.ind2x(tind);
      sint4 ty = pf.ind2y(tind);
      *targ.sod.x = tx * tile_points + tile_points / 2;
      *targ.sod.y = ty * tile_points + tile_points / 2;
      return stype();
    }
  }

  *targ.sod.x = orig_x;
  *targ.sod.y = orig_y;

  return stype(1);
}

//=============================================================================

stype Game::script_get_tile_type(const SActParam &p)
{
  GameTile *t = 0;
  if (p.val.size() == 1) { t = &pf(p.val[0].to_int()); }
  else if (p.val.size() == 2) { t = &pf(p.val[0].to_int(), p.val[1].to_int()); }
  SCRIPT_ASSERT(t && "illegal param size");
  sint4 v = t->get_typeW();
  if (t->is_split()) v += t->get_typeE() << 8;
  return stype(v);
}

//=============================================================================

stype Game::script_set_tile_type(const SActParam &p)
{
  Tile *t;
  sint4 tind;
  sint4 tx, ty;
  sint4 type;

  if (p.val.size() == 2) {

    sint4 v0 = p.val[0].to_int();
    t = &pf(v0);
    tx = pf.ind2x(v0);
    ty = pf.ind2y(v0);
    tind = v0;
    type = p.val[1].to_int();

  } else if (p.val.size() == 3) {

    tx = p.val[0].to_int();
    ty = p.val[1].to_int();
    t = &pf(tx, ty);
    tind = pf.xy2ind(tx,ty);
    type = p.val[2].to_int();

  } else {

    SCRIPT_ERR("illegal param size");
  }

  t->set_type(static_cast<Tile::Type>(type));

  FORS (i, player_n) {
    players[i]->explored.unset(tx, ty);
  }
  update_tile(tind);
  return stype();
}

//=============================================================================

stype Game::script_sqrt(const SActParam &p)
{
  VALN_CHECK(1);
  real4 v = static_cast<real4>(p.val[0].to_int());
  return stype(static_cast<sint4>(sqrt(v)));
}

//=============================================================================

