#ifndef GameTile_H
#define GameTile_H

// $Id: GameTile.H,v 1.7 2005/07/27 07:15:11 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "Tile.H"
#include "Vision.H"

class GameObj;

//===================================================================

/** The basic unit of terrain description.
    Boundaries are implicit between adjacent tiles whose corner heights do not match.
*/

class GameTile : public Tile {

public:

  GameTile() { clear(); }
  
  // overridden superclass functions (added vision functionality)

  void read_b(const char* &s);
  //  void read(std::istringstream &is);
  //  void read(const char* &s);
  void clear();

  sint2 min_vh;        ///< The minimum height beyond which the tile no longer fully obstructs the view.
  sint2 max_vh;        ///< The height at which the tile does not obstruct vision.

  sint2 vis_h1, vis_h2; ///< The base height of an object for both sides of the tile.

  GameObj *boundary_N;
  GameObj *boundary_S;
  GameObj *boundary_E;
  GameObj *boundary_W;
  GameObj *boundary_int;

  uint1 boundaries;    ///< Flags walls between this and neighbouring tiles for Game::update_tile(sint4 pos).
                       ///< Prevents creating duplicate (and removing non-existant) boundaries.

  sint2 sh[5];         ///< Sorted heights, lowest to highest.
  sint2 prof[5];       ///< Shadow profile (of this tile) at each height range.

  // byte-aligned TileViews
  DualView dtvW;        ///< Mask of which surrounding tiles have the West side visible.
  DualView dtvE;        ///< ... East side visible.
  DualView dknown_tv;   ///< Mask of which surrounding tiles are known.

  DualView dtvW2;       ///< If split, the West sides of tiles visible from the East part of this tile.
  DualView dtvE2;       ///< If split, the East sides of tiles visible from the East part of this tile.
  DualView dknown_tv2;  ///< If split, which surrounding tiles are known from the East part.

  sint1 v1w;   // farthest computed land view distance
               // if the client can see farther now it must recompute the tileview
  sint1 v1e;   // farthest distance computed from the east side

  static BitStrip ***edge_shadows;

  void generate_shadows();

  sint4 get_vis_h1() const { return vis_h1; }
  sint4 get_vis_h2() const { return vis_h2; }  

  /** The shadow profile of the tile when viewed from height z
      -1 iff no obstruction    \n
      12 iff full obstruction
  */
  
  inline sint4 shadow_profile(sint4 z)
  {
    //  printf("prof = %d %d %d %d %d\n", prof[0],prof[1],prof[2],prof[3],prof[4]);
    if (z > sh[2]) {
      if (z > sh[4]) return -1;
      if (z > sh[3]) return prof[4];
      return prof[3];
    }
    if (z < sh[2]) {
      if (z < sh[0]) return 12;
      if (z < sh[1]) return prof[0];
      return prof[1];
    }
    return prof[2];
  }
};

//===================================================================

#endif
