// $Id: PlayerInfo.C,v 1.17 2006/04/25 03:08:38 orts_mburo Exp $

#include "PlayerInfo.H"
#include "GameConst.H"
#include "GameObj.H"
#include "Game.H"

using namespace std;

REGISTER_TYPEOF(27, IdObjMap);

Vector<ScriptObj**> PlayerInfo::shared_objects;
Vector<string> PlayerInfo::shared_object_names;
map<string, ScriptObj*> PlayerInfo::shared_object_map;

//===================================================================

PlayerInfo::PlayerInfo(uint4 TX, uint4 TY, bool server_) :
  visible_W(TX,TY),
  visible_E(TX,TY),

  explored(TX,TY),
  p_explored(TX,TY),

  detected_W(TX,TY),
  detected_E(TX,TY),

  cidf(GameConst::ELEM_NUM_MAX),
  server(server_)
{
}

//-------------------------------------------------------------------

void PlayerInfo::save(ostream &os) const
{
  visible_W.save(os);
  visible_E.save(os);
  explored.save(os);
  p_explored.save(os);
  detected_W.save(os);
  detected_E.save(os);

  sint4 s = global_objects.size();
  OS_WRITE(s);
  FORU (i, s) {
    sint4 l = global_object_names[i].size();
    OS_WRITE(l);
    os.write(global_object_names[i].c_str(), l);

    ScriptObj **sop = global_objects[i];
    ScriptObj *so = *sop;
    ScriptObj::Ptr p(so);
    OS_WRITE(p);
  }
}

//-------------------------------------------------------------------
 
void PlayerInfo::load(istream &is)
{
  visible_W.load(is);
  visible_E.load(is);
  explored.load(is);
  p_explored.load(is);
  detected_W.load(is);
  detected_E.load(is);

  sint4 s;
  IS_READ(s);
  FORU (i, s) {
    sint4 l;
    IS_READ(l);
    string str;
    FORU (i, l) {
      char c;
      IS_READ(c);
      str += c;
    }
    
    ScriptObj **sop = new ScriptObj*;
    ScriptObj::Ptr p(0);
    IS_READ(p);
    *sop = p;

    global_objects.push_back(sop);
    global_object_names.push_back(str);
    global_object_map[str] = p;
  }
}

//-------------------------------------------------------------------

/** Returns an object's id to the available pool after a delay.
    Delay = generation + 2*GameConst::ACTION_EXPIRATION.
*/

void PlayerInfo::delete_obj(ScriptObj *obj, sint4 generation)
{
  // delete map entries
  
  FIND (obj2id, mit, obj);
  if (mit == obj2id.end()) ERR("obj not found");

  ObjId cid = (*mit).second;

  id2obj.erase(cid);
  obj2id.erase(obj);

  // schedule freeing the id 2*GameConst::ACTION_EXPIRATION generations later
  
  to_free_ids.push_front(cid);
  to_free_times.push_front(generation+2*GameConst::ACTION_EXPIRATION);

  while (!to_free_times.empty() && to_free_times.back() <= generation) {
    //cout << "DELETE " << generation << " " << to_free_ids.back() << endl; 

    // free cid
    cid = to_free_ids.back();
    if (cidf.free(cid)) ERR("cid can't be freed");

    FIND (full_id2obj, it, cid);
    if (it == full_id2obj.end()) ERR("object is already gone");
    ScriptObj *obj = it->second;
    obj->del_ref();

    full_id2obj.erase(cid);

    to_free_ids.pop_back();
    to_free_times.pop_back();
  }
}

//-------------------------------------------------------------------

/** Associates a new id with an object through the id factory.
*/

ObjId PlayerInfo::new_id(ScriptObj *obj)
{
  ObjId cid = cidf.next();
  if (!cid) ERR("no more ids");
  id2obj[cid] = obj;
  full_id2obj[cid] = obj;
  obj2id[obj] = cid;

  obj->add_ref();

  return cid;
}

//-------------------------------------------------------------------

/** The object referenced by a client id.
    @return 0 iff object id not found.
*/

ScriptObj* PlayerInfo::get_obj(ObjId cid) const
{
  if (cid == 0) return 0;

  //  IdObjMap::const_iterator mit = id2obj.find(cid);
  //  if (mit == id2obj.end()) return 0;
  FIND (full_id2obj, mit, cid);
  if (mit == full_id2obj.end()) return 0;
  return (*mit).second;
}

//-------------------------------------------------------------------

bool PlayerInfo::valid_id(ObjId cid) const
{
  FIND (id2obj, mit, cid);
  return mit != id2obj.end();
}

//-------------------------------------------------------------------

/** The client id of an object.
    @return 0 iff object not found.
*/

ObjId PlayerInfo::get_id(ScriptObj *obj) const
{
  FIND (obj2id, mit, obj);
  if (mit == obj2id.end()) return 0;
  return (*mit).second;
}

//-------------------------------------------------------------------

/** Returns the addresses of all objects in a unique order.
*/

void PlayerInfo::objs(Vector<ScriptObj::Ptr> &v)
{
  v.clear();
  FORALL (id2obj, it) v.push_back((*it).second);
}

//===================================================================

ScriptObj* PlayerInfo::shared_obj(const string &name)
{
  FIND (shared_object_map, i, name);
  if (i == shared_object_map.end()) return 0;
  return i->second;
}

//===================================================================

ScriptObj* PlayerInfo::global_obj(const string &name)
{
  FIND (global_object_map, i, name);
  if (i == global_object_map.end()) return 0;
  return i->second;
}

//-------------------------------------------------------------------

stype* PlayerInfo::global_var(const string &name)
{
  FIND (global_sym_table, i, name);
  if (i == global_sym_table.end()) return 0;
  return i->second;
}

//===================================================================

void PlayerInfo::new_global_obj(const std::string &name, ScriptObj **obj)
{
  ScriptObj *p = global_obj(name);
  if (!p) {
    global_objects.push_back(obj);
    global_object_names.push_back(name);
    global_object_map[name] = *obj;
  }
}

//-------------------------------------------------------------------

void PlayerInfo::new_shared_obj(const std::string &name, ScriptObj **obj)
{
  ScriptObj *p = shared_obj(name);
  if (!p) {
    shared_objects.push_back(obj);
    shared_object_names.push_back(name);
    shared_object_map[name] = *obj;
  }
}

//===================================================================

void PlayerInfo::new_global_var(const std::string &name, stype val)
{
  FIND (global_sym_table, i, name);
  if (i == global_sym_table.end()) {
    global_var_list.push_back(val);
    global_sym_table[name] = &global_var_list.back();
  }
}

//===================================================================
