#ifndef SaveLoad_H
#define SaveLoad_H

/** @file SaveLoad.H
    Virtual interface for saving and loading data to streams.

    $Id: SaveLoad.H,v 1.6 2006/04/03 00:41:32 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/SaveLoad.H,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"

// when using those macros make sure that v has a POD type!
#define OS_WRITE(v) os.write(reinterpret_cast<const char*>(&v), sizeof(v));
#define IS_READ(v)  is.read(reinterpret_cast<char*>(&v), sizeof(v));


// abstract interface (for polymorphic classes)

class SaveLoadAbs {
public:

  virtual void save(std::ostream &) const = 0;
  virtual void load(std::istream &) = 0;
  virtual ~SaveLoadAbs() { }
};


// plain interface and static save/load functions

class SaveLoad {

public:

  void save(std::ostream &) const { ERR("default implementation"); }
  void load(std::istream &)       { ERR("default implementation"); }
  
  template <typename T> static void save(const Vector<T> &v, std::ostream &os) {
    size_t n = v.size();
    OS_WRITE(n);
    FORALL (v, it) (*it).save(os);
  }
  
  template <typename T> static void load(Vector<T> &v, std::istream &is) {
    size_t n = 0, i;
    IS_READ(n);
    v.clear();
    FOR (i, n) { T e; e.load(is); v.push_back(e); }
  }

};

#endif
