#ifndef ScriptErr_H
#define ScriptErr_H

#include "Global.H"

struct ScriptExecException : public std::runtime_error {
  ScriptExecException(const std::string &msg) : std::runtime_error(msg) { }
};


#define SCRIPT_ERR(s) {	  \
    std::stringstream ss; \
    ss << __FILE__ << " " << __FUNCTION__ \
       << "() (line " << __LINE__ << "): " << s;	\
    errstr << " SCRIPT ERROR: " << ss.str() << std::endl; \
    throw ScriptExecException(std::string(ss.str())); \
  }

#define SCRIPT_ERR2(s1, s2) {\
    std::stringstream ss; \
    ss << __FILE__ << " " << __FUNCTION__		 \
       << "() (line " << __LINE__ << "): " << s1 << " " << s2 ; \
    errstr << " SCRIPT ERROR: " << ss.str() << std::endl; \
    throw ScriptExecException(std::string(ss.str())); \
  }

#define SCRIPT_ASSERT(t)    { if (!(t)) SCRIPT_ERR2("FAILED ASSERT: ", #t); }
#define SCRIPT_ASSERT2(t,m) { if (!(t)) SCRIPT_ERR2("FAILED ASSERT: ", #t+string(m)); }

// script function parameter checks

#define VALN_CHECK(n)   SCRIPT_ASSERT(p.val.size() == n)
#define VALGEN_CHECK(n) SCRIPT_ASSERT(p.val.size() >= n)

// also check that obj[i] != 0
#define OBJN_CHECK(n)   { SCRIPT_ASSERT(p.obj.size() == n); FORALL (p.obj, i) SCRIPT_ASSERT(*i); }
#define OBJGEN_CHECK(n) { SCRIPT_ASSERT(p.obj.size() >= n); FORALL (p.obj, i) SCRIPT_ASSERT(*i); }

#endif
