/** @file ScriptString.C
    @see ScriptString.H

    $Id: ScriptString.C,v 1.19 2006/04/25 03:08:38 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/ScriptString.C,v $
*/

// This is an ORTS file (c) Tim Furtak

#include "Global.H"
#include "ScriptString.H"
#include "GameObj.H"
#include "ByteData.H"
#include "Script.H"

using namespace std;

//===================================================================

ScriptString::ScriptString()
{
  // the stated parameter count isn't checked when calling the function
  // the function can be overloaded without running into errors
  // just ensure that the functions don't make assumptions about vector size

  add("op",        &ScriptString::op,        0, 0, 0);
  add("init",      &ScriptString::init,      0, 0, 0);
  add("clear",     &ScriptString::clear,     0, 0, 0);
  add("size",      &ScriptString::size,      0, 0, 0);
  add("append",    &ScriptString::append,    0, 1, 0);
  add("add_char",  &ScriptString::add_char,  0, 1, 0);
  add("c_str",     &ScriptString::c_str,     0, 0, 0);
  add("num_lines", &ScriptString::num_lines, 0, 0, 0);
  add("max_line",  &ScriptString::max_line,  0, 0, 0);
}

//-------------------------------------------------------------------

void ScriptString::add(const std::string &name, myfunc f, sint4 objs, sint4 p, sint4 p2)
{
  ObjActionDefinition def(name, objs, p, p2);
  ObjAction *a = new ObjAction(static_cast<ObjFunction>(f), def); // fixme: function pointer danger!
  actions.push_back(a);
  act_map[name] = a;
}

//-------------------------------------------------------------------

stype ScriptString::op(const SActParam &p)
{
  VALN_CHECK(1);
  sint4 s = data.size();
  sint4 i = p.val[0].to_int();
  if (i < 0) return stype(s); // [<0] = size
  SCRIPT_ASSERT(i < s && "index out of range");
  return stype(data[i]);
}

//-------------------------------------------------------------------

stype ScriptString::init(const SActParam &p)
{
  if (p.val.size() > 0) {
    if (!p.val[0]) {
      data.clear();
      return stype();
    }
    data = stype::print_va_list(p.val[0].to_cstr(), p.val, 1);
    return stype();
  }

  if (p.obj.size() > 0) {
    ScriptString *s = dynamic_cast<ScriptString*>(p.obj[0]);
    SCRIPT_ASSERT(s && "not a ScriptString");
    data = s->data;
    return stype();
  }

  SCRIPT_ERR("no string data");
  return stype();
}

//-------------------------------------------------------------------

stype ScriptString::clear(const SActParam &)
{
  data.clear();
  return stype();
}

//-------------------------------------------------------------------

stype ScriptString::size(const SActParam &)
{
  return stype(data.size());
}

//-------------------------------------------------------------------

/** Append a formatted printf string or another ScriptString. */

stype ScriptString::append(const SActParam &p)
{
  if (p.val.size() > 0) {
    data += stype::print_va_list(p.val[0].to_cstr(), p.val, 1);
    return stype();
  }
  if (!p.obj.empty()) {
    ScriptString *s = dynamic_cast<ScriptString*>(p.obj[0]);
    SCRIPT_ASSERT(s && "no ScriptString");
    data += s->data;
    return stype();
  }
  SCRIPT_ERR("no string data");
  return stype();
}

//-------------------------------------------------------------------

stype ScriptString::add_char(const SActParam &p)
{
  VALN_CHECK(1);
  data += static_cast<char>(p.val[0].to_int());
  return stype();
}

//-------------------------------------------------------------------

stype ScriptString::c_str(const SActParam &)
{
  return stype(data.c_str());
}

//-------------------------------------------------------------------

stype ScriptString::num_lines(const SActParam &)
{
  sint4 count = 1;
  FORALL (data, it) {
    if (*it == '\n') ++count;
  }
  return stype(count);
}

//-------------------------------------------------------------------

stype ScriptString::max_line(const SActParam &)
{
  sint4 max_count = 0;
  sint4 cur_count = 0;

  FORALL (data, it) {
    if (*it == '\n') {
      if (cur_count > max_count) max_count = cur_count;
      cur_count = 0;
    } else {
      ++cur_count;
    }
  }
  if (cur_count > max_count) max_count = cur_count;
  return stype(max_count);
}

//===================================================================

ObjAction* ScriptString::action(const std::string &name) const
{
  FIND (act_map, it, name);
  if (it != act_map.end()) return it->second;
  return 0;
}

//===================================================================

void ScriptString::gen_diff(bool first, bool allied, string &s, PlayerInfo * /*pi*/) const
{
  sint4 max_len = data.size() + 1;
  char *t = new char[max_len];
  char *u = t;

  if (!allied) {
    write_b_int4(u, 2);
    string q(t, int(u-t));
    s = q;
    delete [] t;
    return;
  }

  sint4 i = (first || data != prev_data) ? 1 : 0;
  write_b_int4(u, i);

  if (i) {
    FORALL (data, it) write_b_int4(u, *it);
    write_b_int4(u, 0);
  }

  string q(t, int(u-t));
  s = q;
  delete [] t;
}

//-------------------------------------------------------------------

bool ScriptString::apply_diff(bool /*allied*/, const char* &s)
{
  char c = *s++;
  if (c == 2) return false; // hidden
  bool first = c == 1;
  SCRIPT_ASSERT(c == 0 || c == 1)

  if (first) {
    data.clear();
    sint4 c;
    while ((c = read_b_int4(s))) {
      data += static_cast<char>(c);
    }
    return true;
  }

  return false;
}

//===================================================================

void ScriptString::copy_to_prev() {
  prev_data = data;
}
