/** @file ScriptVector.C
    @see ScriptVector.H

    $Id: ScriptVector.C,v 1.20 2006/05/12 01:27:00 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/ScriptVector.C,v $
*/

// This is an ORTS file (c) Tim Furtak

#include "Global.H"
#include "ScriptVector.H"
#include "GameObj.H"
#include "ByteData.H"
#include "Script.H"

using namespace std;

const sint4 ScriptVector::STYPE_ID = stype::next_id();

//==================================================================
 
ScriptVector::ScriptVector()
{
  add("op",        &ScriptVector::op,        0, 1, 0);
  add("op_addr",   &ScriptVector::op_addr,   0, 1, 0);
  add("size",      &ScriptVector::size,      0, 0, 0);
  add("clear",     &ScriptVector::clear,     0, 0, 0);
  add("push_back", &ScriptVector::push_back, 0, 1, 0);
  add("remove",    &ScriptVector::remove,    0, 1, 0);
  add("insert",    &ScriptVector::insert,    0, 2, 0);
  add("pop_back",  &ScriptVector::pop_back,  0, 0, 0);
  add("sort",      &ScriptVector::sort,      0, 0, 0);
  add("setify",    &ScriptVector::setify,    0, 0, 0);
  add("intersect", &ScriptVector::intersect, 1, 0, 0);
  add("subtract",  &ScriptVector::subtract,  1, 0, 0);
  add("get_vect",  &ScriptVector::get_vect,  0, 0, 0);
}

void ScriptVector::add(const std::string &name, myfunc f, sint4 objs, sint4 p, sint4 p2)
{
  ObjActionDefinition def(name, objs, p, p2);
  ObjAction *a = new ObjAction(static_cast<ObjFunction>(f), def); // fixme: function pointer danger!
  actions.push_back(a);
  act_map[name] = a;
}

stype ScriptVector::op(const SActParam &p)
{
  VALN_CHECK(1);
  sint4 i = p.val[0].to_int();
  sint4 s = data.size();
  if (i < 0) return stype(s);
  SCRIPT_ASSERT(i < s && "index out of range");
  return data[i];
}

stype ScriptVector::op_addr(const SActParam &p)
{
  VALN_CHECK(1);  
  sint4 i = p.val[0].to_int();
  sint4 s = data.size();
  SCRIPT_ASSERT(i >= 0 && i < s && "index out of range");
  return stype(&data[i]);
}

stype ScriptVector::clear(const SActParam &)
{
  data.clear();
  return stype();
}

stype ScriptVector::size(const SActParam &)
{
  return stype(data.size());
}

stype ScriptVector::push_back(const SActParam &p)
{
  VALN_CHECK(1);
  data.push_back(p.val[0]);
  return stype();
}

stype ScriptVector::resize(const SActParam &p)
{
  VALN_CHECK(1);
  data.resize(p.val[0].to_int());
  return stype();
}

/// Erases the ith element.
stype ScriptVector::remove(const SActParam &p)
{
  VALN_CHECK(1);
  const sint4 i = p.val[0].to_int();
  const sint4 s = data.size();
  SCRIPT_ASSERT(i >= 0 && i < s && "index out of range");

  data.erase(data.begin() + i);
  return stype(data.size());
}

stype ScriptVector::insert(const SActParam &p)
{
  VALN_CHECK(2);  
  data.insert(data.begin()+p.val[1].to_int(), p.val[0]);
  return stype(data.size());
}

stype ScriptVector::pop_back(const SActParam &)
{
  data.pop_back();
  return stype(data.size());
}

stype ScriptVector::sort(const SActParam &)
{
  ::sort(data.begin(), data.end());
  return stype();
}

stype ScriptVector::setify(const SActParam &)
{
  if (data.empty()) return stype();
  ::sort(data.begin(), data.end());
  Vector<stype> w;
  stype q = data[0];
  for (uint4 i = 1; i < data.size(); ++i) {
    if (q != data[i]) {
      w.push_back(data[i]);
      q = data[i];
    }
  }
  data = w;
  return stype(data.size());
}

stype ScriptVector::intersect(const SActParam &p)
{
  OBJN_CHECK(1);
  ScriptVector *q = dynamic_cast<ScriptVector*>(p.obj[0]);
  SCRIPT_ASSERT(q && "parameter not a ScriptVector");
    
  ScriptVector &gv = *q;
  Vector<stype> &v = gv.data;

  if (data.empty() || v.empty()) {
    data.clear();
    return stype();
  }

  ::sort(data.begin(), data.end());
  ::sort(v.begin(), v.end());
  Vector<stype> x(v.size() + data.size());

  Vector<stype>::iterator it = set_intersection(v.begin(), v.end(), data.begin(), data.end(), x.begin());
  x.erase(it, x.end());
  data = x;
  return stype(data.size());
}

stype ScriptVector::subtract(const SActParam &p)
{
  OBJN_CHECK(1);
  ScriptVector *q = dynamic_cast<ScriptVector*>(p.obj[0]);
  SCRIPT_ASSERT(q && "parameter not a ScriptVector");

  ScriptVector &gv = *q;
  Vector<stype> &v = gv.data;

  if (v.empty()) return stype(data.size());
  if (data.empty()) return stype();

  ::sort(v.begin(), v.end());
  ::sort(data.begin(), data.end());
  Vector<stype> x(v.size());

  Vector<stype>::iterator it = set_difference(v.begin(), v.end(),
					      data.begin(), data.end(),
					      x.begin());
  //  Vector<sint4>::iterator it = set_difference(data.begin(), data.end(), v.begin(), v.end(), x.begin());
  x.erase(it, x.end());
  data = x;
  return stype(data.size());
}

stype ScriptVector::get_vect(const SActParam &)
{
  return stype(&data, STYPE_ID);
}

//-------------------------------------------------------------------

ObjAction* ScriptVector::action(const std::string &name) const
{
  FIND (act_map, it, name);
  if (it != act_map.end()) return it->second;
  return 0;
}

//-------------------------------------------------------------------

void ScriptVector::gen_diff(bool first, bool allied, string &u, PlayerInfo * /*pi*/) const
{
  if (!allied) {
    write_b_int4(u, 2);
    return;
  }

  sint4 i = first ? 1 : 0;
  write_b_int4(u, i);

  if (first) {
    uint4 size = data.size();
    write_b_int4(u, size);
    FORU (i, size) write_b_int4(u, data[i].to_int());
  } else {
    sint4 size = data.size() - prev_data.size();
    write_b_int4(u, size);
    if (size > 0) {
      // previous indices, delta
      FORU (i, prev_data.size()) write_b_int4(u, data[i].to_int() - prev_data[i].to_int());
      // new indices, actual value
      for (uint4 i = prev_data.size(); i < data.size(); ++i) {
        write_b_int4(u, data[i].to_int());
      }
    } else {
      // same size or smaller
      FORU (i, data.size()) write_b_int4(u, data[i].to_int() - prev_data[i].to_int());
    }
  }
}

bool ScriptVector::apply_diff(bool /*allied*/, const char* &s)
{
  bool changed = false;
  char c = *s++;
  if (c == 2) return false; // hidden
  bool first = c == 1;
  SCRIPT_ASSERT(c == 0 || c == 1);

  sint4 sdelta = read_b_int4(s);
  sint4 size;
  if (first) {
    size = sdelta;
    data.resize(size);
  } else if (sdelta) {
    changed = true;
    size = data.size() + sdelta;
    data.resize(size);
  } else {
    size = data.size();
  }

  FORS (i, size) {
    SCRIPT_ASSERT(data[i].type == stype::INT && "vector element not int");
    sint4 d = read_b_int4(s);
    if (i >= size - sdelta) {
      data[i] = stype(d); // new index
    } else {
      changed |= (d != 0);
      //      data[i].val.i += d;
      data[i].val.i = data[i].val.i + d;
    }
  }

  return changed;
}

//-------------------------------------------------------------------

void ScriptVector::copy_to_prev() {
  prev_data = data;
  //  memcpy (&prev_attr_vals[0], &attr_vals[0], sizeof(AttrVal) * prev_attr_vals.size());
  //  prev_attr_vals = attr_vals;
  //  FORALL(components, it) it->second->copy_to_prev();
}
