// $Id: OFPS.C,v 1.1 2006/05/31 22:30:56 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include"OFPS.H"

#ifdef __MINGW32__ 
#include "Gettimeofday.H"
#endif

#ifdef GCC
#include <sys/time.h>

FPS::FPS(sint4 d)
{
  interval = d;
#if 1
  list = new timeval[interval];
  FOR (p, interval) gettimeofday(&list[p], 0);
#else
  list = new timeb[interval];
  FOR (p, interval) ftime(&list[p]);
#endif  
  p=0;
}

FPS::~FPS()
{
  delete [] list;
}

real8 FPS::update()
{
  sint4 p2 = (p+1) % interval;

#if 1
  gettimeofday(&list[p], 0);
  real8 diff = list[p].tv_sec-list[p2].tv_sec +
    (1.0/1000000.0)*(list[p].tv_usec-list[p2].tv_usec);

#else  
    ftime(&list[p]);
    real8 diff = list[p].time - list[p2].time +
               (1.0/1000.0)*(list[p].millitm-list[p2].millitm);
#endif
    
  p = p2;
  fps = static_cast<real8>(interval) / diff;
  dt = 1.0/fps;

  return dt;
}
 
#else // win32 FPS ---------------

FPS::FPS(sint4 d)
{
  interval = d;
  list = new DWORD[interval];
  FOR (p, interval) list[p] = GetTickCount();
  p=0;
}

FPS::~FPS()
{
  delete [] list;
}

real8 FPS::update()
{
  sint4 p2 = (p+1) % interval;

  list[p] = GetTickCount();

  if (list[p] < list[p2]) return 0.0; // counter wrapped around

  real8 diff = 0.001 * (list[p] - list[p2]);

  p = p2;
  fps = (real8) interval / diff;
  dt = 1.0/fps;

  return dt;
}

#endif
