
// $Id: OGfxModule2.C,v 1.1 2006/05/31 22:30:56 orts_mburo Exp $

// This is an ORTS file
// (c) Michael Buro
// (c) Tim Furtak
// (c) Keith Yerex
// (c) Sami Wagiaalla
// (c) Erik Benner
// licensed under the GPL

#include "Global.H"
#include "TerrainModule.H"
#include "GameStateModule.H"
#include "Options.H"
#include "GameChanges.H"
#include "GameObj.H"
#include "Game.H"
#include "PlayerInfo.H"
#include "GameTile.H"
#include "Token.H"
#include "ScriptParser.H"
#include "ScriptVector.H"

#include "OGfxGlobal.H"
#include "OGfxModule.H"
#include "OOctree.H"
#include "OGfxObject.H"
#include "OGfxTerrain.H"
#include "OWaterTile.H"
#include "OGfxAnimesh.H"
#include "OMinimap.H"
#include "OImage.H"
#include "OCWidget.H"
#include "OKeyMap.H"
#include "OScriptIdSet.H"
#include "OScriptModel.H"
#include "OScriptEmitter.H"
#include "OProfileWidget.H"

using namespace std;

#define PARSE_ERR(s) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR(s);\
}

#define PARSE_ERR2(s1, s2) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR2(s1,s2);\
}

bool GfxModule::load_interface(const string &fname_, Tokenizer *t)
{
  static sint4 depth = -1;
  ++depth;

  string fname = sanitize_path(fname_);
  fname = collapse_file_path(fname);

  sint4 i = fname.find(PATH_SEP);
  string path;
  if (i >= 0) {
    path = fname.substr(0, fname.find_last_of(PATH_SEP)+1);
  }
  i = fname.find_last_of(PATH_SEP);

  // only load each file once (files with the same name will be rejected)
  FORALL (loaded_bp_files, i) {
    if (*i == fname) {
      --depth;
      return false;
    }
  }
  loaded_bp_files.push_back(fname);
  if (depth > 0) cout << endl;

  ifstream file;
  file.open(fname.c_str(), ios::binary);
  if (!file.is_open()) ERR2("can't open file: ", fname);

  FORS (i, depth) cout << "\t";
  cout << "loading " << fname << " ... " << flush;
  splashScreen.set_current_task("Loading interface file: " + fname);
  if (!file) ERR2("can't load ", fname);

  // read file into buffer
  sint4 length;
  char *buffer;

  file.seekg(0, ios::end);
  length = file.tellg();
  file.seekg(0, ios::beg);

  buffer = new char[length+1];
  file.read(buffer, length);
  file.close();
  buffer[length] = 0;

  Tokenizer tokenizer(t);
  tokenizer.init(buffer, fname, path);
  tokenizer.set_line(0);
  string token;

  bool recursed = false;

  while (tokenizer.good()) {

    token = tokenizer.next_tag();

    if      (token == "<DEFINE>")     game->read_define_block(tokenizer);
    else if (token == "<ACTIONS>")    game->read_action_block(tokenizer);
    else if (token == "<BLUEPRINTS>") game->read_blueprint_block(tokenizer, true);
    else if (token == "<TILEDEF>")    game->read_tiledef_block(tokenizer);
    else if (token == "<BITMAPS>")    load_bitmap_list(tokenizer, path);
    else if (token == "<INCLUDE>") {
      while (tokenizer.good()) {
        token = tokenizer.next_tag();
        if (token == "</INCLUDE>") break;
        if (token != "") ERR ("</INCLUDE> expected");
        token = tokenizer.next_token();

        if (tokenizer.token_type() != Tokenizer::STRING) {
          cout << endl << "found " << token << endl;
          PARSE_ERR("filename expected");
        }

        token = token.substr(1, token.size()-2);

        recursed |= load_interface(path + token, &tokenizer);
      }
    }

    else {
      token = tokenizer.next_token();
      if (token != "")
        ERR2("expecting a tag, found:", token.c_str());
      tokenizer.putback(token);
    }
  }
  
  if (recursed) {
    cout << endl;
    FORS (i, depth) cout << "\t";
  }
  cout << "done" << flush;
  --depth;
  delete [] buffer;
  splashScreen++;
  return true;
}

