#ifndef OMD2_H
#define OMD2_H

// $Id: OMD2.H,v 1.1 2006/05/31 22:30:57 orts_mburo Exp $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "OGfxGlobal.H"
#include "OGfxAnimesh.H"

//===================================================================

struct MD2Triangle
{
  sint2 vert[3];
  sint2 tvert[3];

  void little2machine()
  {
    FORS (i, 3) { ::little2machine(vert[i]); ::little2machine(tvert[i]); }
  }
};

//-------------------------------------------------------------------

#if 0
struct MD2CommandVertex
{
  real4 tc[2];
  sint4 vert;
};

//-------------------------------------------------------------------

struct MD2Command
{
  GLenum command;
  sint4 n;
  MD2CommandVertex *list;
};
#endif

//-------------------------------------------------------------------

class MD2_Model : public Model
{
friend class GfxAnimesh;
public:

  MD2_Model();
  virtual ~MD2_Model();
  
  sint4 load(const std::string &modelfile, const std::string &texfile,
             real4 scale, real4 dy, real4 gamma);
  
  void draw_bbox(sint4 frame);
  virtual void draw(sint4 frame);
  virtual void draw(sint4 frame, sint4 team);
  virtual void draw_t(sint4 frame, sint4 team);
  virtual void draw_interp(sint4 f1, sint4 f2, real4 blend, sint4 team);

  virtual void select(sint4 frame);
  
  void draw_shadow(sint4 frame, real4 sc = 0.0);

  void free();
  
protected:
  
  virtual void init() {}

  void build_shadows();

  uint4 nvert;
  uint4 npoly;
  uint4 ntvert;
  uint4 ncommand;
  Vec3<real4> *vert;
  real4 *tvert;
  Vec3<real4> *vnorm;
  Poly *poly;
  Poly *tpoly;
  static Vec3<real4> *normal_lookup;
  sint4 *command;

  sint4 array_size;
  Vec3<real4> *vertex_array;
  Vec3<real4> *normal_array;
  real4 *texcoord_array;
  sint4 index_array_size;
  sint2 *index_array;
};

//-------------------------------------------------------------------

/** Display list MD2 model. */

class MD2_Model_DL : public MD2_Model
{
public:
  virtual void select(sint4 frame);
  virtual void draw(sint4 frame);
  virtual void draw_interp(sint4 f1, sint4 f2, real4 blend, sint4 team);

  virtual ~MD2_Model_DL() {} // fixme: delete anything?
  
protected:
  virtual void init();
  GLuint display_lists;
  GLuint select_list;
};

//-------------------------------------------------------------------

/** Vertex Buffer MD2 model. */

class MD2_Model_VB : public MD2_Model
{
public:

  virtual void draw(sint4 frame);
  virtual void draw_interp(sint4 f1, sint4 f2, real4 blend, sint4 team);
  virtual ~MD2_Model_VB() {}  // fixme: delete anything?

protected:
  virtual void init();
  GLuint vertex_buffer;
  GLuint index_buffer;
};

//===================================================================

#endif
