/** @file ScriptIdSet.C
    @see ScriptIdSet.H

    $Id: OScriptIdSet.C,v 1.1 2006/05/31 22:30:57 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/ogfx/src/OScriptIdSet.C,v $
*/

// This is an ORTS file (c) Tim Furtak

#include "Global.H"
#include "Game.H"
#include "OScriptIdSet.H"
#include "OGfxModule.H"

using namespace std;

GfxModule* ScriptIdSet::gfxm;
Vector<ScriptIdSet*> ScriptIdSet::instances;

REGISTER_TYPEOF(260, Vector<ScriptIdSet*>::iterator);

//===================================================================

ScriptIdSet::ScriptIdSet()
{
  instances.push_back(this);

  add("select",     &ScriptIdSet::select,      0, 0, 0);
  add("select_add", &ScriptIdSet::select_add,  0, 0, 0);
  add("deselect",   &ScriptIdSet::deselect,    0, 0, 0);
}

ScriptIdSet::~ScriptIdSet()
{
  FORU (i, instances.size()) {
    if (instances[i] == this) {
      instances.erase(instances.begin() + i);
      break;
    }
  }
}

//-------------------------------------------------------------------

void ScriptIdSet::add(const std::string &name, myfunc f, sint4 objs, sint4 p, sint4 p2)
{
  ObjActionDefinition def(name, objs, p, p2);
  ObjAction *a = new ObjAction(static_cast<ObjFunction>(f), def); // fixme: nonportable
  actions.push_back(a);
  act_map[name] = a;
}

//===================================================================

stype ScriptIdSet::select(const SActParam &p)
{
  gfxm->sel_objs.clear();
  gfxm->sel_ids->clear();

  return select_add(p);
}

//-------------------------------------------------------------------

stype ScriptIdSet::select_add(const SActParam&)
{
  FORALL (data, it) {
    ScriptObj *x = it->to_obj();
    GameObj *y = x->get_GameObj();
    if (y == 0) continue;
    gfxm->sel_objs.insert(gfxm->game2gfx[y]);
  }

  FORALL (data, it) {
    ScriptObj *x = it->to_obj();
    sint4 id = gfxm->game->get_cplayer_info().get_id(x); // fixme
    gfxm->sel_ids->push_back(stype(id)); // fixme: use a set
  }

  gfxm->gui_obj->set_int("sel_change", 1, false);

  return stype();
}

//-------------------------------------------------------------------

stype ScriptIdSet::deselect(const SActParam&)
{
  return stype();
}

//===================================================================

void ScriptIdSet::invalidate_id(stype id)
{
  FORALL (instances, it) {
    (*it)->invalidate(id);
  }
}

//-------------------------------------------------------------------

void ScriptIdSet::invalidate(stype id)
{
  FIND (data, it, id);
  if (it != data.end()) data.erase(it);
}

//===================================================================
