
// $Id: OShader.C,v 1.1 2006/05/31 22:30:57 orts_mburo Exp $

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "Global.H"
#include "OShader.H"

using namespace std;

//===================================================================

void Shader::clear()
{
  if (shader) {
    glDeleteShader(shader);
    shader = 0;
  }
}

void Shader::compile()
{
  glCompileShader(shader);
 
  GLint compiled = 0;
  glGetObjectParameterivARB(shader, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);
  if (compiled) {
    // OK
  } else {
    if (shader == 0) return;

    GLsizei len = 0;
    GLsizei log_len = 0;
    
    glGetObjectParameterivARB(shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, &len);
    
    char *log = new char[len];

    glGetInfoLogARB(shader, len, &log_len, log);
    cout << "COMPILE ERROR: " << log << endl;

    delete [] log;
    ERR("");
  }
}

void ShaderProgram::link()
{
  glLinkProgram(prog); 

  GLint linked = 0;
  glGetProgramiv(prog, GL_LINK_STATUS, &linked);
  
  if (linked) {
    // OK
  } else {
    if (prog == 0) return;

#if 1

    ERR2("link error! ", prog);

#else
    // fixme: not supported by some 
    GLsizei len = 0;
    GLsizei log_len = 0;
    
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &len);
    char *log = new char[len];

    glGetProgramInfoLog(prog, len, &log_len, log);
    cout << "LINK ERROR: " << log << endl;

    delete [] log;
    ERR("");
#endif    
  }
}

//-------------------------------------------------------------------

char* Shader::read_source(const string &filename)
{
  ifstream file;
  file.open(filename.c_str(), ios::binary);
  if (!file.is_open()) {
    return 0;
    ERR2("can't open file:", filename.c_str());
  }

  file.seekg(0, ios::end);
  sint4 length = file.tellg();
  file.seekg(0, ios::beg);

  char *buffer = new char[length+1];
  file.read(buffer, length);
  file.close();
  buffer[length] = 0;

  return buffer;
}

//===================================================================

bool VertexShader::load(const string &filename)
{
  if (shader) {
    glDeleteShader(shader);
    shader = 0;
  }
  const char *s = read_source(filename);
  if (s == 0) return false;
  shader = glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(shader, 1, &s, NULL);
  compile();
  return true;
}

//===================================================================

bool FragmentShader::load(const string &filename)
{
  if (shader) {
    glDeleteShader(shader);
    shader = 0;
  }
  const char *s = read_source(filename);
  if (s == 0) return false;
  shader = glCreateShader(GL_FRAGMENT_SHADER_ARB);
  glShaderSource(shader, 1, &s, NULL);
  compile();
  return true;
}

//===================================================================

ShaderProgram::ShaderProgram()
{
  fs = 0;
  vs = 0;
  prog = 0;
}

//-------------------------------------------------------------------

void ShaderProgram::fragment_shader(const string &filename)
{
  if (fs) {
    glDetachShader(prog, fs->shader);
    delete fs;
  }
  fs = 0;
  if (filename != "") {
    fs = new FragmentShader;
    if (fs->load(filename)) {
      cout << "fragment load = true" << endl;
      if (prog == 0) prog = glCreateProgram();
      glAttachShader(prog, fs->shader);
      link();
    } else {
      delete fs;
      fs = 0;
    }
  }
}

//-------------------------------------------------------------------

void ShaderProgram::vertex_shader(const string &filename)
{
  if (vs) {
    glDetachShader(prog, vs->shader);
    delete vs;
  }
  vs = 0;
  if (filename != "") {
    vs = new VertexShader;
    if (vs->load(filename)) {
      cout << "vertex load = true" << endl;
      if (prog == 0) prog = glCreateProgram();
      glAttachShader(prog, vs->shader);
      link();
    } else {
      delete vs;
      vs = 0;
    }
  }
}

//===================================================================

