// $Id: OSplashScreen.C,v 1.1 2006/05/31 22:30:57 orts_mburo Exp $

#include "OSplashScreen.H"
#include "OGfxModule.H"

using namespace std;

SplashScreen::SplashScreen(sint4 x_, sint4 y_, sint4 w_, sint4 h_, const string &version_) :
  version(version_)
{
  set_dim(x_, y_, w_, h_);
  
  status = 0; 
  background = 0;
  off = false;
}

void SplashScreen::set_dim(sint4 x_, sint4 y_, sint4 w_, sint4 h_)
{
  x = x_;
  y = y_;
  h = h_;
  w = w_;
  
  sint4 margin = 33;
  
  bar_x = x+margin + 16;
  bar_y = y+margin;
  bar_h = 28;
  bar_w = y+w-2*margin - 15;
}

void SplashScreen::set_background(const Image &image)
{
  background = image.createTexture(Image::LINEAR);
}

void SplashScreen::turn_off()
{
  off = true;
  glDeleteTextures(1, &background);
  background = 0;
}

void SplashScreen::draw()
{
  if (off) return;
  
  // real8 x1 = bar_x + bar_h/2 ;
  // real8 y1 = bar_y;
  // real8 y2 = bar_y+bar_h + 1;
  real8 x2 = bar_x+bar_w*status/max;

  //  glClearColor(1.0,1.0,1.0,1);
  glClearColor(0,0,0,1);

  glPushMatrix();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0, w, 0, h, -100, 100);    
  
  glMatrixMode(GL_MODELVIEW); 
  glLoadIdentity();
  glViewport(0, 0, w, h);

  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_DEPTH_TEST);
  
  //glEnable(GL_TEXTURE_2D); 
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  //================Draw Background====================
  if (background) {      
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);

    glBindTexture(GL_TEXTURE_2D,background);
    glEnable(GL_TEXTURE_2D);

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

    glBegin(GL_QUADS);	
    glTexCoord2f(0.f, 1.f);
    glVertex2f  (x, y);

    glTexCoord2f(0.f, 0.f);
    glVertex2f  (x, y+h);

    glTexCoord2f(1.f, 0.f);
    glVertex2f  (x+w, y+h);

    glTexCoord2f(1.f, 1.f);
    glVertex2f  (x+w, y);
    glEnd();

    glEnable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);
  }

  //================Draw Status bar box ===============
  sint4 box_x1 = bar_x-1;
  sint4 box_y1 = bar_y-2;
  sint4 box_y2 = bar_y+bar_h;
  sint4 box_x2 = bar_x+bar_w+2;

  //glColor3b(47, 53, 127); // earth blue
  glColor3f(0.09f, 0.2f, 0.56f); // space dust blue
  glBegin(GL_LINE_LOOP);
    glVertex2f( box_x2, box_y1);
    glVertex2f( box_x2, box_y2);
    glVertex2f( box_x1, box_y2);
    glVertex2f( box_x1, box_y1);
  glEnd();
 
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);
  
  GLfloat mat_ambient[] = {0.6f,0.6f,1.0f,1.0f};
  GLfloat mat_diffuse[] = {0.6f,0.6f,0.6f,1.0f};
  GLfloat mat_specular[] = {1.0f,1.0f,1.0f,1.0f};
  GLfloat mat_shiniess[] = {50.0f};

  glMaterialfv(GL_FRONT,GL_AMBIENT,mat_ambient);
  glMaterialfv(GL_FRONT,GL_DIFFUSE,mat_diffuse);
  glMaterialfv(GL_FRONT,GL_SPECULAR,mat_specular);
  glMaterialfv(GL_FRONT,GL_SHININESS,mat_shiniess);

  // Draw progress bar
  glLoadIdentity();
  glTranslated(bar_x,bar_y+bar_h/2,0);
  glRotatef(90.0, 0.0, 1.0, 0.0);
  glDepthMask(GL_FALSE);
  //glColor3b(47, 53, 127);
  gluCylinder(gluNewQuadric(),bar_h/2,bar_h/2,x2-bar_h/2-5,30,30);
  
  glDepthMask(GL_TRUE);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_LIGHTING);
  
  // Draw status & version text
  glLoadIdentity();
  print(bar_x+10, bar_y+38, task.c_str(), GLUT_BITMAP_HELVETICA_12, true);
  print(w -2 -(strlen(version.c_str())*8), h-15, version.c_str(), GLUT_BITMAP_8_BY_13, true);
  
  glEnable(GL_TEXTURE_2D);
  glPopMatrix();
  
  glFinish();
  glutSwapBuffers();
}

void SplashScreen::set_current_task(const std::string& task_)
{
  if(!off)
  {
    task = task_;
    draw();
  }
}

void SplashScreen::set_status(real8 status_)
{ 
  if (status_ > max ){
    //    cerr << "WARNING: " << "SplashScreen::set_status status > max" << endl ;
    status_ = max;
  }
  status = status_;
  draw();
}

void SplashScreen::print(uint4 x, uint4 y, const char *text, GLvoid *font, bool shadow)
{
  if(shadow) {
    glColor3f(0,0,0);
    glDepthMask(GL_FALSE);
    render_bitmap_string(x+2, y, font, text);
    glDepthMask(GL_TRUE);
  }
  glColor3f(0.9,0.9,0.9);
  render_bitmap_string(x, y+1, font, text);
}

