// $Id: OWaterTile.C,v 1.1 2006/05/31 22:30:57 orts_mburo Exp $

// This is an ORTS file
// (c) Keith Yerex

#include "Global.H"
#include "OWaterTile.H"

using namespace std;

GLuint WaterTile::tex;
real4 WaterTile::t = 0;
GLfloat WaterTile::texgen_s[] = { 0.1, 0, 0, 0 };
GLfloat WaterTile::texgen_t[] = { 0, 0, 0.1, 0 };
set<WaterTile*> WaterTile::tiles;

void WaterTile::anim(real4 dt)
{
  texgen_s[3] += dt*0.006;
  texgen_t[3] += dt*0.01;
  if (texgen_s[3] > 1) texgen_s[3] -= 1;
  if (texgen_t[3] > 1) texgen_t[3] -= 1;

  t += dt;

  FORALL (tiles, it) (*it)->update(0.0);
}

void WaterTile::beginDraw()
{
  glActiveTextureARB(GL_TEXTURE1);
  glEnable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE0);
  glEnable(GL_TEXTURE_2D);

  glEnable(GL_TEXTURE_GEN_S);
  glEnable(GL_TEXTURE_GEN_T);
  
  glTexGeni (GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  glTexGenfv(GL_S, GL_OBJECT_PLANE, texgen_s);

  glTexGeni (GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  glTexGenfv(GL_T, GL_OBJECT_PLANE, texgen_t);
  
  glEnable(GL_TEXTURE_GEN_S);
  glBindTexture(GL_TEXTURE_2D, tex);
  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glColor4f(1, 1, 1, 1);
}

void WaterTile::endDraw()
{
  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);
}

WaterTile::WaterTile(const Vec3<real4> *verts)
{
  tiles.insert(this);

  FORU (i, 4) vert[i] = orig[i] = verts[i];

  Vec3<real4> max = orig[3];
  aabb_o = orig[3];

  FORU (i, 3) {
    if (orig[i].x < aabb_o.x) aabb_o.x = orig[i].x;
    if (orig[i].y < aabb_o.y) aabb_o.y = orig[i].y;
    if (orig[i].z < aabb_o.z) aabb_o.z = orig[i].z;

    if (orig[i].x > max.x) max.x = orig[i].x;
    if (orig[i].y > max.y) max.y = orig[i].y;
    if (orig[i].z > max.z) max.z = orig[i].z;
  }

  aabb_d = max - aabb_o;

  // add some slack for wave motion
  aabb_o.y -= 0.3;
  aabb_d.y += 0.6;
}

void WaterTile::pick()
{
  glLoadName(my_id << GfxObject::SUB_ID_LOG);

  glBegin(GL_TRIANGLE_STRIP);

  glColor3fv((vert[0] / aabb_d-aabb_o).ptr());
  glVertex3fv(vert[0].ptr());

  glColor3fv((vert[1] / aabb_d-aabb_o).ptr());
  glVertex3fv(vert[1].ptr());

  glColor3fv((vert[2] / aabb_d-aabb_o).ptr());
  glVertex3fv(vert[2].ptr());

  glColor3fv((vert[3] / aabb_d-aabb_o).ptr());
  glVertex3fv(vert[3].ptr());

  glEnd();
}

void WaterTile::draw()
{
  glBegin(GL_TRIANGLE_STRIP);
  glVertex3fv(vert[0].ptr());
  glVertex3fv(vert[1].ptr());
  glVertex3fv(vert[2].ptr());
  glVertex3fv(vert[3].ptr());
  glEnd();
}

void WaterTile::update(real4)
{
  FORS (i, 4) {
    //vert[i].y = pos.y + 0.03*sin(6*t+2*(vert[i].x+10*sin(t))+2*(vert[i].z+10*sin(t+10)))-0.1;

    real4 noise = sin(6*t+2*vert[i].x + 2*vert[i].z) + 1.0;
    noise *= 0.02;

    vert[i].y = orig[i].y - noise;
  }
}

