// $Id: Net_Message.H,v 1.1 2005/01/04 19:13:17 orts_sami Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

/** a struct for quickly parsing and storing Net messages
 * initialized using a string representing a message
 * parses the string into source, type, and a Vector of strings for arguments
 */
	
#ifndef Net_Message_H
#define Net_Message_H

#include "Global.H"

/**
   Format:
    <message_id> <param1> <param2> ....
    use '\ ' for a real space

invite <game-type> <map-name> (<qualifier>)*

<game-type>        xcraft, etc.
<map-description>  lost-temple
                   random

possible qualifiers:

                   a       (anonymizing ids)
                   i       (invisible: game only announced to buddies)
                   pw=...  (password for joining this game)
                   n=...   (game name)
                   st=...  (style: island, hilly, whatnot (for random))
                   s=x*y   (size)
                   p=8     (# of players)


Details need to be fleshed out. Internally, orts.bot creates a unique
game invitation id (iid, number?). If a game request is created messages like

invitation + <iid> all parameters ...

could be sent out to update clients' lists.

abort <iid>

kick <iid> <player-alias>

-> response: kicked <iid> <player-alias>


join <iid>

  joins invitation and assigns player alias (number 1,2,...)

-> response to all members:  joined <iid> <player-alias>

   (this is also sent when creating an invitation)

leave <iid>

-> response (to all joined members):  left <iid> <player-alias>


ready   <iid>
unready <iid>

-> responses to all members
ready   <iid> <player-alias>
unready <iid> <player-alias>


when all members are ready game is assigned to worker, game id <gid>
is created, and

  game_start <gid> <worker-ip-address> <worker-port> <password-for-worker>

is sent to all players


game_end <gid> <list of winners ... points?>

indicates game is over



invitations
-> response: invitation list with details

games
-> response: list of running games with details

workers
-> response: information on all workers

tell <iid> <msg>
(send <msg> to all members of <iid> (except sender))

*/

class Net_Message {

public:
  /** standard sources and destinations where a message can be heading */
  enum{GGS = 'G', RECEIVER = 'U', CHAT = 'C'};
   
  Net_Message() {}
  Net_Message(const std::string &message);
  virtual ~Net_Message();
  
  /**
   * @return The argument at the given index, or null if the index is invalid.
   */
  std::string getArgument(sint4 index) const {
    if(index < 0 || index >= (int)arguments.size()){
      return "NULL";
    }else{
      return arguments[index];
    }
  }
  
  /**
   * @return Number of arguments in the message.
   */
  uint4 getSize() const {
    return arguments.size();
  }

  // deletes the old structure
  // parses the given message and puts it in a new structure
  void setMessage(const std::string &s);

  // returns the unparsed string of the current message
  std::string getString(){return message;}

  std::string toString() const {
    std::stringstream ss;
    ss << "\nFrom : " << from << " To : " << to << std::endl;
    for(uint4 i=0; i < arguments.size(); i++){
      ss << "  [" << i <<"] |"<< arguments[i] << "|\n" << std::endl;
    }
    ss << std::endl << std::endl;
    return ss.str();
  }

  std::string getFrom() const {return from;}
  std::string getTo() const {return to;}
  
protected:

  std::string message;  ///< The unparsed message.
  std::string from;     ///< Who sent the message.
  std::string to;       ///< _group | .chann | etc. <------ not used yet ----]
  Vector<std::string> arguments;
  
  // parses the message and stores the arguments
  sint4 parse(const std::string &s);

};

#endif
