// $Id: Net_Protocol.C,v 1.10 2006/04/03 00:41:32 orts_mburo Exp $

#include "Net_Protocol.H"
using namespace std;

/***************************************************************************************************
 * @class Message_Base
 * contains info and functions common to all messages
 ***************************************************************************************************/

/**
   convert an int to a string
*/
string Message_Base::to_string(int a){

  stringstream ss;
  ss << a;
  if(ss.bad()) return "NULL";
  return ss.str();
}

/**
   convert a string to an int
   @return false if the int is invalid, true otherwise
*/
bool Message_Base::set_int(const string &astr, int& anInt){

  istringstream is(astr);

  is >> anInt;

  if (is.fail()) {
    return false;
  }

  return true;
}


/**
  + = true
  - = false
  otherwise return false
*/
bool Message_Base::set_bool(const string& message, bool& toSet){

  if(message == "+"){
    toSet = true;
    return true;
  }

  if(message == "-"){
    toSet = false;
    return true;
  }

  return false;
}

/**
    if message == "NULL" return false
    otherwise set the string
  */
bool Message_Base::set_string(const string& message, string& toSet){

  if(message == "NULL"){
    return false;
  }

  toSet = message;
  return true;
}

/**
  if b is true return "+" otherwise "-"
*/
string Message_Base::bool_to_string(bool b){
  if(b){
    return "+";
  }else{
    return "-";
  }
}
/**
    replace all spaces with '_' so that this string is not broken up be parser
  */
void Message_Base::solidify(string& message){
  // replace all spaces in options with '_'
  // so it doesnt get parsed into seperate strings
  for(uint4 i = 0; i < message.size(); i++){
    if(message[i] == ' '){
      message[i] = '_';
    }
  }
}

/**
  replace all '_' with space
*/
void Message_Base::melt(string& message){

  // replace all '_' in options with ' '
  // so it doesnt get parsed into seperate strings
  for(uint4 i = 0; i < message.size(); i++){
    if(message[i] == '_'){
      message[i] = ' ';
    }
  }
}





/***************************************************************************************************
 * @class Invite
 * sent to the bot to create an invitation
 * <game_name> <maxPlayers> <invisible(+|-)> <anonymous (+|-)> <pass> <options>
 ***************************************************************************************************/

bool Invite::parse(const Net_Message &m){

  

  set_string(m.getArgument  (0), message);
  if(INVITE!=message) return false;
  
  /** fixme: return values should be checked but left now for debugging use*/
  set_string(m.getFrom(), sender);
  set_string(m.getArgument  (1), name);
  if(!set_int(m.getArgument (2), nplayers)) nplayers = 1; //fixme: -- just for now
  set_bool(m.getArgument    (3), invisible);
  set_bool(m.getArgument    (4), anony_ids);
  set_bool(m.getArgument    (5), passprotected);
  
  if( anony_ids ) set_string(m.getArgument  (6), alias); 
  else set_string(m.getFrom(), alias);
  
  set_string(m.getArgument  (7), password);
  cout << "PASSWORD:" << password << endl;
  set_string(m.getArgument  (8), options);

  return true;
}

/**
 * Forms a string made up of member variables including quantifier
 */
string Invite::compose( const string &name,
                        const int numPlayers,
                        const bool invisible,
                        const bool anony,
                        const bool passprotected,
                        const string &anonyname,
                        const string &pass,
                        const string &options){

  string tempOptions = options;
  solidify(tempOptions);
  
  return INVITE + " " + name + " " + to_string(numPlayers) + " " + bool_to_string(invisible)
                + " " + bool_to_string(anony) + " " + bool_to_string(passprotected) + " "  + anonyname
                + " " + pass + " " + tempOptions;
}





/************************************************************************************************
 * an invitations discriptor class used to send out and parse currently running game invitations
 * <iid> <nplayers> <owner> <game_name> <maxPlayers> <invisible(+|-)> <anonymous (+|-)> <pass> <options>
 *************************************************************************************************/

bool Invit_Disc::parse(const Net_Message &m){

  cout << "------------------------------------------" <<endl;
  cout << "NetMESSAGE:"<<m.toString()<<endl;
  cout << "------------------------------------------" <<endl;
  
  if(!set_string(m.getArgument(0), message))  return false;
  if(message != INVIT_DISC)                   return false;

  if(!set_string(m.getFrom(), sender))        return false;
  if(!set_int(m.getArgument(1), iid))         return false;
  if(!set_int(m.getArgument(2), nPlayers))    return false;
  if(!set_string(m.getArgument(3), owner))    return false;
  if(!set_string(m.getArgument(4), gameName)) return false; melt(gameName);
  if(!set_int(m.getArgument(5), maxPlayers))  return false;
  if(!set_bool(m.getArgument(6), invisible))  return false;
  if(!set_bool(m.getArgument(7), anony_ids))  return false;
  if(!set_bool(m.getArgument(8), passprotected))  return false;
  if(!set_string(m.getArgument(9), options))  return false; melt(options);
  
  return true;

}

string Invit_Disc::compose(const int iid,
                           const int nPlayers,
                           const string& owner,
                           const string& gameName,
                           const int maxPlayers,
                           const bool invisible,
                           const bool anony,
                           const bool passprotected,
                           const string& options){
  string result;

  result += to_string(iid);
  result += " ";
  
  result += to_string(nPlayers);
  result += " ";
                       
  result += owner;
  result += " ";

  result += gameName;
  result += " ";

  result += to_string(maxPlayers);
  result += " ";

  result += bool_to_string(invisible);
  result += " ";

  result += bool_to_string(anony);
  result += " ";

  result += bool_to_string(passprotected);
  result += " ";
  
  string tempOptions = options;
  solidify(tempOptions);

  result += tempOptions;

  return result;
}






/***************************************************************************************************
 * @class Abort
 * sent to the bot to abort an invitation
 * <iid>
 ***************************************************************************************************/

bool Abort::parse(const Net_Message &m){
  
  if(!set_string(m.getArgument(0), message))return false;;
  
  if(ABORT_ != message) return false;

  if(!set_string(m.getFrom(), sender))return false;;
  if(!set_int(m.getArgument(1), iid))return false;;

  return true;
}


/**
 * Forms a string made up of member variables
 */
string Abort::compose(int id){

  string result_string = ABORT_ + " " + to_string(id);
  return result_string;
}







/***************************************************************************************************
 * @class Aborted
 * sent out by the bot to all players to notify them that a game has been aborted
 * <iid>
 ***************************************************************************************************/

bool Aborted::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0),message)) return false;

  if(message != ABORTED)                    return false;

  if(!set_string(m.getFrom(), sender))              return false;
  if(!set_int(m.getArgument(1), iid))       return false;

  return true;
}


/**
 * Forms a string made up of member variables
 */
string Aborted::compose(int id){

  string result_string = ABORTED + " " + to_string(id);
  return result_string;
}





/***************************************************************************************************
 * @class Kick
 * sent by the owner of and invitation to the bot to kick a player out of an invitation
 * <iid> <pid>
 ***************************************************************************************************/
bool Kick::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;

  if(KICK!=message) return false;

  if(!set_string(m.getFrom(), sender)) return false;
  if(!set_int(m.getArgument(1), iid))  return false;
  if(!set_int(m.getArgument(2), pid))  return false;

  return true;
}

/**
 * Forms a string made up of member variables
 */
string Kick::compose(int id, int pa){

  string result_string = KICK + " " + to_string(id) + " " + to_string(pa);
  return result_string;
}






/***************************************************************************************************
 * @class Kicked
 * when a player is kick out from an invitation this is sent out by the bot to all the other players
 * in that game
 * <iid> <pid>
 ***************************************************************************************************/

bool Kicked::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;
   
  if(KICKED!=message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), iid))return false;
  if(!set_int(m.getArgument(2), pid))return false;

  return true;
}

/**
 * Forms a string made up of member variables
 */
string Kicked::compose(int id, int pa){

  string result_string = KICKED + " " + to_string(id) + " " + to_string(pa);
  return result_string;
}






/***************************************************************************************************
 * @class Join
 * sent by a player to the bot to join a specific game
 * <iid> [pass]
 ***************************************************************************************************/

bool Join::parse(const Net_Message &m){

  
  if(!set_string(m.getArgument(0), message))      return false;
  
  if(JOIN!= message) return false;
  cout << "(message)" << message << endl;
  
  if(!set_string(m.getFrom(), sender))            return false;
  cout << "(sender)" << sender << endl;
  
  if(!set_int(m.getArgument(1), iid ))            return false;
  cout << "(iid)" << iid << endl;
  
  if(!set_bool(m.getArgument(2), anony_ids))      return false;
  cout << "(anony_ids)" << anony_ids << endl;
  
  if(!set_string(m.getArgument(3), alias))        return false;
  cout << "++++(alias)" << alias << endl;
  
  if(!set_bool(m.getArgument(4), passprotected))  return false;
  cout << "(passprotected)" << passprotected << endl;
  
  if(!set_string(m.getArgument(5), pass))         return false;
  cout << "(pass)" << pass << endl;

  return true;
}

string Join::compose(int id,
                     const bool anony_ids,
                     const string &alias,
                     const bool passprotected,
                     const string& pass)
{
  string result_string;

  result_string += JOIN;
  result_string += " ";

  result_string += to_string(id);
  result_string += " ";

  result_string += bool_to_string(anony_ids);
  result_string += " ";

  result_string += alias;
  result_string += " ";

  result_string += bool_to_string(passprotected);
  result_string += " ";

  result_string += pass;

  
  return result_string;
}




/*****************************************************************************************************
 * Joined
 * sent out to aporpriet users when a player joins a game
 * <gid> <pid> <anony (+|-)> <player-alias>
 *****************************************************************************************************/

bool Joined::parse(const Net_Message &m){
  if(!set_string(m.getArgument(0), message))return false;
  
  if(JOINED!= message) return false;
  
  if(!set_string(m.getFrom(), sender))            return false;
  if(!set_int(m.getArgument(1), iid))             return false;
  if(!set_int(m.getArgument(2), pid))             return false;
  if(!set_bool(m.getArgument(3), anony))          return false;
  if(!set_string(m.getArgument(4), player_alias)) return false;

  return true;
}

string Joined::compose(int gid, int pid, bool anony, const string& alias){

  string result_string = JOINED + " " +
                          to_string(gid) + " " +
                          to_string(pid) + " " +
                          bool_to_string(anony) + " " +
                          alias;
                          
  return result_string;
}






/***************************************************************************************************
 * @class Leave
 * sent by a player to the bot to leave a particular invitation
 * <iid>
 ***************************************************************************************************/

bool Leave::parse(const Net_Message &m)
{
  if(!set_string(m.getArgument(0), message))return false;
  
  if(LEAVE!= message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), iid))return false;

  return true;
}

string Leave::compose(int id)
{
  string result_string =  LEAVE + " " + to_string(id);
  return result_string;
}





/***************************************************************************************************
 * @class Left
 * when a player leaves this is sent out to all other players in that invitation
 * <iid> <pid>
 ***************************************************************************************************/

bool Left::parse(const Net_Message &m)
{
  if(!set_string(m.getArgument(0), message))return false;
  cout << m.toString() << endl;
  if(LEFT!= message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), iid))return false;
  if(!set_int(m.getArgument(2), pid))return false;

  return true;
}

string Left::compose(int iid, int pid){

  string result_string = LEFT + " " + to_string(iid) + " " + to_string(pid);
  return result_string;
}





/***************************************************************************************************
 * @class Ready
 * sent to the bot by a player to set their ready status to true
 * <iid>
 ***************************************************************************************************/

bool Ready::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;
  
  if(READY!=message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), iid)) return false;

  return true;
}

string Ready::compose(int id){

  string result_string = READY + " " + to_string(id);
  return result_string;
}




/***************************************************************************************************
 * @class Readied
 * sent out by the bot to all players in an invitation when one of the players sets their status to
 * ready
 * <iid> <pid>
 ***************************************************************************************************/

bool Readied::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;
  
  if(READIED!= message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), iid))return false;
  if(!set_int(m.getArgument(2), pid))return false;

  return true;
}

/**
 * Forms a string made up of member variables
 */
string Readied::compose(int id, int pa){
  string result_string = READIED + " " + to_string(id) + " " + to_string(pa);
  return result_string;
}





/***************************************************************************************************
 * @class Unready
 * sent to the bot by a player to set their ready status to true
 * <iid>
 ***************************************************************************************************/

bool Unready::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;

  if(UNREADY!=message)  return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), iid))return false;

  return true;

}

string Unready::compose(int id){

  string result_string = UNREADY + " " + to_string(id);
  return result_string;
}




/***************************************************************************************************
 * @class Unreadied
 * sent out by the bot to all players in an invitation when one of the players sets their status to
 * unready
 * <iid> <pid>
 ***************************************************************************************************/

bool Unreadied::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;
  

  if(UNREADIED!= message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), iid))return false;
  if(!set_int(m.getArgument(2), pid))return false;

  return true;
}

string Unreadied::compose(int iid, int pid){

  string result_string = UNREADIED + " " + to_string(iid) + " " + to_string(pid);
  return result_string;
}




/***************************************************************************************************
 * @class Game_Start
 * set out by the bot to tell all players in an invitation to connect to the worker
 * no params
 ***************************************************************************************************/

bool Game_Start::parse(const Net_Message &m){

   if(!set_string(m.getArgument(0), message)) return false;

  if(GAME_START != message) return false;
  if(!set_string(m.getFrom(), sender)) return false;

  return true;
}

/**
 * Forms a string made up of member variables
 */
string Game_Start::compose(){
  return GAME_START;
}




/***************************************************************************************************
 * @class Game_Info
 * sent out by the bot. contains information about the worker
 * creates an options string
 * ie -shost foo -netpass bar
 * <shost> <netpass>
 ***************************************************************************************************/

bool Game_Info::parse(const Net_Message &m){

  if(!set_string(m.getFrom(), sender)) return false;

  if(!set_string(m.getArgument(0), message)) return false;

  if(GAME_INFO != message) return false;

  if(!set_string(m.getArgument(1), info)) return false;
  melt(info);

  return true;
}

string Game_Info::compose(const string& shost, const string& netpass){

  string info =  "-host " + shost/* + " -netpass " + netpass*/;
  string blah = netpass; // TODO: do something with netpass, for now shut up compiler warning
  solidify(info);
  return GAME_INFO + " " + info;
}




/***************************************************************************************************
 * @class Worker_Start
 * sent by the bot to the worker to tell it to start the game server and start listening for
 * connections
 * <numPlayers> <options> (<player1IP> <player1pass>)*...
 ***************************************************************************************************/

bool Worker_Start::parse(const Net_Message &m){

  set_string(m.getArgument(0), message);

  if(WORKER_START!=message) return false;
  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), numPlayers))return false;
  if(!set_string(m.getArgument(2), options)) return false;
  melt(options);

  for(int i = 3; i < (numPlayers*2)+3; i++){
    ipv.push_back(m.getArgument(i));
    passv.push_back(m.getArgument(++i));
  }

  return true;
}

string Worker_Start::compose(string options, int numPlayers, Vector<string> ipv, Vector<string> passv){

  solidify(options);

  stringstream ss;
  ss << WORKER_START;
  ss << " " ;
  ss << to_string(numPlayers);
  ss << " " ;
  ss << options;
  ss << " " ;

  for(int i = 0; i < numPlayers; i++){
    ss << ipv[i];
    ss << " ";
    ss << passv [i];
    ss << " ";
  }

  return ss.str();
}






/***************************************************************************************************
 * @class Worker_Ready
 * sent by the worker to the bot when the server is ready and is listening for connections
 * no params
 ***************************************************************************************************/

bool Worker_Ready::parse(const Net_Message &m) {

  if (! set_string(m.getArgument(0), message) ){   return false; }
  if (message != WORKER_READY){                    return false; }
  if (! set_string(m.getFrom(), sender) ){   return false; }
  
  return true;
}

string Worker_Ready::compose () { return WORKER_READY; }
/** END WORKER_READY */

/** GAME_END CLASS DEFINITIONS */
bool Game_End::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;
  
  if(GAME_END!=message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_int(m.getArgument(1), iid))return false;

  return true;
}


/***************************************************************************************************
 * @class Game_End
 * not used yet
 ***************************************************************************************************/

string Game_End::compose(int g_id){

  string result_string = GAME_END + " " + to_string(g_id);
  return result_string;
}



/***************************************************************************************************
 * @class Invitations
 * a list containing the current invitations. Sent out by the bot upon request
 * <Inv_Discritpor>
 * @see Inv_Discriptor
 ***************************************************************************************************/
bool Invitations::parse(const Net_Message &m){

  
  if(!set_string(m.getArgument(0), message)) return false;
  
  if(INVITATIONS != message) return false;

  if(!set_string(m.getFrom(), sender))return false;

  string temp;
  string inv;
  Invit_Disc disc;
  Net_Message mlocal;
  
  for(uint4 i = 1; i < m.getSize(); i++){
      inv = m.getFrom();
      inv += ": ";
      inv += INVIT_DISC;
      inv += " ";
      
      if(!set_string(m.getArgument(i++), temp)) return false;
      temp += " ";
      cout << "(iid)" << temp << endl;
      inv  += temp;
      
      if(!set_string(m.getArgument(i++), temp)) return false;
      temp += " ";
      cout << "(npalyers)" << temp << endl;
      inv  += temp;

      if(!set_string(m.getArgument(i++), temp)) return false;
      temp += " ";
      cout << "(owner)" << temp << endl;
      inv  += temp;

      if(!set_string(m.getArgument(i++), temp)) return false;
      temp += " ";
      cout << "(name)" << temp << endl;
      inv  += temp;
      
      if(!set_string(m.getArgument(i++), temp)) return false;
      temp += " ";
      cout << "(max)" << temp << endl;
      inv  += temp;
      
      if(!set_string(m.getArgument(i++), temp)) return false;
      temp += " ";
      cout << "(invis)" << temp << endl;
      inv  += temp;

      if(!set_string(m.getArgument(i++), temp)) return false;
      temp += " ";
      cout << "(anony)" << temp << endl;
      inv  += temp;

      
      if(!set_string(m.getArgument(i++), temp)) return false;
      temp += " ";
      cout << "(passprotected)" << temp << endl;
      inv  += temp;
      
      if(!set_string(m.getArgument(i), temp)) return false;
      cout << "(options)" << temp << endl;
      inv += temp;

      mlocal.setMessage(inv); cout << "\ninv -> " << inv << endl;
      if(!disc.parse(mlocal))                 return false;

      list=disc;

  }

  return true;
}


/** just concatinated string vector would be too inefficent */
string Invitations::compose(const string& list){
  return INVITATIONS + " " + list;
}




/***************************************************************************************************
 * @class Games
 * Used to send out a lset of games
 * not used yet
 ***************************************************************************************************/

bool Games::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;
  
  if(GAMES != message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  
  return true;
}

string Games::compose(const string &m){

  string result = m;
  return result;

}




/***************************************************************************************************
 * @class Workers
 * used to send out a list of workers
 * not used yet
 ***************************************************************************************************/

bool Workers::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;

  if(WORKERS != message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  
  return true;
}

string Workers::compose(const string &m){

  string result = m;
  return result;

}






/***************************************************************************************************
 * @class Tell
 * sent by a player in an invitation to the bot for telling something to all players in that
 * invitation
 * <message>
 ***************************************************************************************************/

bool Tell::parse(const Net_Message &m){


  if(!set_string(m.getArgument(0), message))       return false;

  if(TELL != message)                              return false;

  if(!set_string(m.getFrom(), sender))             return false;
  if(!set_int(m.getArgument(1), iid))              return false;
  if(!set_int(m.getArgument(2), pid))              return false;
  if(!set_string(m.getArgument(3), out_msg))       return false;

  melt(out_msg);
  return true;
}

string Tell::compose(int iid, int pid,  const string &outm){

  string temp = outm;
  solidify(temp);

  string result_string = TELL + " " + to_string(iid) + " " + to_string(pid) + " " + temp;
  cout << "-->"<< result_string << endl;
  return result_string;
}


/*****************************************************************************************************
 * Error                                                                                             
 * sent out to users when an error  player joins a game                                              
 * < message >                                                                                       
 *****************************************************************************************************/

bool Error::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))       return false;

  if(ERROR != message)                             return false;

  if(!set_string(m.getFrom(), sender))             return false;
  if(!set_string(m.getArgument(1), msg))           return false;

  melt(msg);
  return true;
}

string Error::compose(const string &msg){

  string temp = msg;
  solidify(temp);

  string result_string = ERROR + " " + temp;
  return result_string;
}

/***************************************************************************************************
 * @class GGS_Error
 * sent by GGS when an error occures.
 * <message>
 ***************************************************************************************************/

bool GGS_Error::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))       return false;

  if(GGS_ERROR != message)                         return false;
  if(!set_string(m.getFrom(), sender))             return false;

  if(sender != "G")                                return false;

  for(uint4 i = 1; i < m.getSize(); i++){
    msg += m.getArgument(i);
  }

  return true;
}


/***************************************************************************************************
 * @class Buddy
 * sent by a player to the bot to add/remove people from their buddy list
 * <+|-> <buddy_name>
 ***************************************************************************************************/

bool Buddy::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;

  if(BUDDY != message) return false;

  if(!set_string(m.getFrom(), sender))return false;
  if(!set_bool(m.getArgument(1), add))return false;
  if(!set_string(m.getArgument(2), buddy_name))return false;

  return true;
}

string Buddy::compose(bool add, const string &buddyName){

  string result_string = BUDDY + " " + bool_to_string(add) + " " + buddyName;
  return result_string;
}





/***************************************************************************************************
 * @class Login_Notify
 * sent out by the bot to all logged on players one someone comes online or goes offline
 * <+|-> <playerName>
 ***************************************************************************************************/

bool Login_Notify::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))            return false;

  if(LOGIN_NOTIFY!= message)               return false;

  if(!set_string(m.getFrom(), sender))                  return false;
  if(!set_bool(m.getArgument(1), online))       return false;
  if(!set_string(m.getArgument(2), player_name)) return false;

  return true;
}

string Login_Notify::compose(bool online, const string &playerName){

  string result_string = LOGIN_NOTIFY + " " + bool_to_string(online) + " " + playerName;
  return result_string;
}




/***************************************************************************************************
 * @class Added_Buddy
 * sent by the bot to a player to confirm that the buddy as been added and wether he/she is online
 * or offline
 * <player_name> <+|->
 ***************************************************************************************************/

bool Added_Buddy::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))     return false;

  if(ADDED_BUDDY != message)                     return false;

  if(!set_string(m.getFrom(), sender))           return false;
  if(!set_string(m.getArgument(1), player_name)) return false;
  if(!set_bool  (m.getArgument(2), online))      return false;
  
  return true;
}

string Added_Buddy::compose(const string &pn, bool online){

  string result_string = ADDED_BUDDY + " " + pn + " " + bool_to_string(online);
  return result_string;
}




/***************************************************************************************************
 * @class Remmed_Buddy
 * sent by the bot to a player to confirm that the buddy as been removed
 * <player_name>
 ***************************************************************************************************/
bool Remed_Buddy::parse(const Net_Message &m){

  if(!set_string(m.getArgument(0), message))return false;

  if(REMED_BUDDY!= message) return false;
  
  if(!set_string(m.getFrom(), sender))return false;
  if(!set_string(m.getArgument(1), player_name))return false;

  return true;
}




/***************************************************************************************************
 * @class Remmed_Buddy
 * sent by the bot to a player to confirm that the buddy as been removed
 * <player_name>
 ***************************************************************************************************/
bool Buddies::parse(const Net_Message &m){
  if(m.getArgument(0) != BUDDY_LIST)return false;
  if(!set_string(m.getFrom(), sender))return false;

  string name;
  bool   budOnline;
  for(uint4 i = 1; i < m.getSize(); i++){

    if(!set_string(m.getArgument(i++), name)) return false;
    if(!set_bool(m.getArgument(i), budOnline)) return false;

    buddies.push_back(name);
    online.push_back(budOnline);

  }
  return true;
}

string Buddies::compose(const Vector<string>& buddies, const Vector<uint1>& online){
  string message = BUDDY_LIST + " ";
  for(uint4 i = 0; i < buddies.size(); i++){
    message += buddies[i];
    message += " ";
    message += bool_to_string(online[i]);
    message += " ";
  }
  return message;
}
/** end CLASS Buddies */



/****************************************************************************************************
 * @class Player_List                                                                               
 * constains and parses the list of player in a game and the ready/unready status                   
 ****************************************************************************************************/
bool Player_List::parse(const Net_Message &m) {

  if(m.getArgument(0) != PLAYER_LIST)return false;
  if(!set_string(m.getFrom(), sender))return false;

  int alias;
  string name;
  bool   pReady;
  
  for(uint4 i = 1; i < m.getSize(); i++){

    if(!set_int(m.getArgument(i++), alias))   return false;
    if(!set_string(m.getArgument(i++), name)) return false;
    if(!set_bool(m.getArgument(i), pReady))   return false;

    aliases.push_back(alias);
    cout << "PLAYER_LIST: (alias)" << alias << endl;
    players.push_back(name);
    cout << "PLAYER_LIST: (name)" << name << endl;
    ready.push_back(pReady);
    cout << "PLAYER_LIST: (ready)" << pReady << endl;

  }
  return true;
}



/****************************************************************************************************
 * @class Player_List                                                                               
 * constains and parses the list of player in a game and the ready/unready status
 * (<playerName_1><+|->)*
 ****************************************************************************************************/
string Player_List::compose(const Vector<uint1>& present, const Vector<string>& players, const Vector<uint1>& ready){
  string message = PLAYER_LIST + " ";
  for(uint4 i = 0; i < players.size(); i++){

    if(!present[i]) continue;
    
    message += to_string(i);
    message += " ";
    message += players[i];
    message += " ";
    message += bool_to_string(ready[i]);
    message += " ";
    
  }
  return message;
}



/****************************************************************************************************\
 * @class Who_List                                                                                  *
 * constains and parses the list of all players currently online                                    *
\****************************************************************************************************/
bool Who_List::parse(const Net_Message &m){
  if(m.getArgument(0) != WHO_LIST)return false;
  if(!set_string(m.getFrom(), sender))return false;

  string name;
  for(uint4 i = 1; i < m.getSize(); i++){

    if(!set_string(m.getArgument(i), name)) return false;

    players.push_back(name);
    cout << "(name)" << name << endl;
  }
  return true;
}

string Who_List::compose(const string& who){
  return WHO_LIST + " " + who;
}






/**************************************************************************************************
 * Forms a string made up of member variables                                                     
 **************************************************************************************************/
string Remed_Buddy::compose(const string &pn){

  string result_string = REMED_BUDDY + " " + pn;
  return result_string;
}

/** END REMED_BUDDY DEFINITIONS */


/** GGS_Notify CLASS DEFINITIONS */
bool GGS_Notify::parse(const Net_Message &m){
  if(m.getFrom() != "G"){
    return false;
  }

  sender = m.getFrom();

  if(m.getArgument(0) == "+"){
    online = true;
  }else{
    if(m.getArgument(0) == "-"){
      online = false;
    }else{
      return false;
    }
  }

  if(m.getArgument(1) == "NULL"){
      return false;
  }

  name = m.getArgument(1);
  return true;
}




/***************************************************************************************************
 * @class GGS_Notify
 * composes a notify request and parses a GGS_Notify message
 * compose: <name> <+|>
 ***************************************************************************************************/
string GGS_Notify::compose(string names, bool notify)
{
  string result_string = "notify ";

  if(notify){
    result_string += "+";
  }else{
    result_string += "-";
  }

  result_string += " ";
  result_string += names;

  return result_string;
}




/***************************************************************************************************
 * @class Register
 * sent by a new player to the bot to create an acount
 * <level> <name> <pass>
 ***************************************************************************************************/
bool Register::parse(const Net_Message &m){

  sender = m.getFrom();

  if(m.getArgument(0) != REGISTER){
    return false;
  }

  if(!set_int(m.getArgument(1), level))return false;
  if(!set_string(m.getArgument(2), name))return false;
  if(!set_string(m.getArgument(3), pass))return false;

  return true;
}

string Register::compose(int level, string name, string pass){
  return REGISTER + " " + to_string(level) + " " + name + " " + pass;
}
/** END REGISTER CLASS*/










/***************************************************************************************************
 * @class Login
 * sent by a palyer to the bot to login
 * <level> <pass> <ip> <port>
 ***************************************************************************************************/
bool Login::parse(const Net_Message &m){

  sender = m.getFrom();

  if(!set_string(m.getArgument(0), message))return false;

  if(message != LOGIN){ return false; }

  if(!set_int(m.getArgument(1), level))return false;
  if(!set_string(m.getArgument(2), pass))return false;

  // optional arguments
  if(m.getSize() > 3 && !set_string(m.getArgument(3), ip)) return false;
  if(m.getSize() > 4 && !set_int(m.getArgument(4), port))return false;

  return true;
}

string Login::compose(int level, string pass, const string& ip, const sint4 port){
  return LOGIN + " " + to_string(level) + " " + pass + " " + ip + " " + to_string(port);
}
/** END REGISTER CLASS*/





/***************************************************************************************************
 * @class Down
 * sent by root to the bot to cleanly shut it down
 * no params
 ***************************************************************************************************/
bool Down::parse(const Net_Message &m){

  if(m.getArgument(0) != DOWN){
    return false;
  }

  sender = m.getFrom();

  return true;
}

string Down::compose(){
  return DOWN;
}
/** END DOWN*/



/**   Finger Class */


/***************************************************************************************************
 * @class Finger
 * parses a GGS finger message and composes a finger request
 * compose: <name>
 ***************************************************************************************************/
bool Finger::parse(const Net_Message &m){

  if(m.getArgument(0) != FINGER)return false;

  for(uint4 i = 1; i < m.getSize(); i++){
    string title, data;
    title = m.getArgument(i);
    i++; // skip the : after the title
    i++; // skip the : after the title
    while(m.getArgument(i+1) != "NULL" && m.getArgument(i+1) != ":"){
        data += m.getArgument(i);
        i++;
        if(m.getArgument(i+1) != ":"){
          data += " ";
        }
    }
    i--;
    list[title] = data;
  }

  return true;
}

/**
  @param title the title of the info to retrive Ex (host, pass, login ...)
  @return the info corresponding to title, return string "NULL" if not found
*/
string Finger::get(const string& title){
  FIND(i, list, title);
  if(i == list.end()){
    return "NULL";
  }else{
    return (*i).second;
  }
}

string Finger::compose(const string& name){
  return FINGER + " " + name;
}


/** end Finger */






/***************************************************************************************************
 * @class List
 * used to request a list from the bot
 * Ex: list buddies, list who, list invitations
 * <list_name>
 ***************************************************************************************************/
bool List::parse(const Net_Message &m){

  if(m.getArgument(0) != LIST){
    return false;
  }
  sender = m.getFrom();

  if(!set_string(m.getArgument(1), name)) return false;
  if(!validName(name)) return false;
  
  return true;
}


bool List::validName(const string& name_){
  if( name_ != BUDDY_LIST &&
      name_ != INVITE_LIST &&
      name_ != WHO_LIST &&
      name_ != GAME_LIST
    )return false;

  return true;
}
  
string List::compose(const string& listName){
  if(!validName(listName)){
    return "NULL";
  }
  
  return LIST + " " + listName;
}
/** END list class */





/***************************************************************************************************
 * @class Worker_Available
 * sent by a worker to the bot upon finishing a game, worker is ready to go again
 * <ip>
 ***************************************************************************************************/
bool Worker_Available::parse(const Net_Message &m) {

  if(m.getArgument(0) != WORKER_AVAILABLE) {
    return false;
  }
  
  if(!set_string(m.getArgument(1), ip)) {
    return false;
  }
  
  sender = m.getFrom();

  return true;
  
}

string Worker_Available::compose(const string& ip) {  
  return WORKER_AVAILABLE + " " + ip;  
}
