// $Id: Mem.H,v 1.1.1.1 2004/10/11 09:09:20 orts_mburo Exp $ 

// (c) Jonathan Yip, licensed under the GPL


#ifndef Mem_H
#define Mem_H

#include "Global.H"

using namespace std;


class Mem {
public:
  enum mem_types { UNKNOWN = 0,
		   VOID    = 1,
		   BOOL    = 10,
		   CHAR    = 11,
		   INT     = 40,
		   FLOAT   = 41,
		   STRING  = 42,
		   OBJECT  = 43,
		   DOUBLE  = 80
  };

  typedef sint4 mem_unit;

  //  typedef pair<uint1, uint1> mem_loc;
  //  typedef pair<mem_loc, mem_types> mem_loc_type;

  Mem() { 
    m_heap = (mem_unit *) malloc(m_init_heap_size*sizeof(mem_unit));
    m_heap_size = m_init_heap_size;
    m_stack_index = 0;
    m_heap_index = 0;
    m_frame_index = 0;
    m_last_frame_index = 0;
  }
  // fixme
  ~Mem() {}

  mem_unit* alloc_heap(sint4 elements) {


    // See if I should compact the heap
    if (m_heap_index + elements > m_heap_size) {
      GC(elements);
    }

    // See if I have enough space, technically should be (m_heap_size-1)
    uint4 index = m_heap_index;
    m_heap_index += elements;

    return &m_heap[index];
  }

  mem_unit* get_heap_loc(uint4 loc) {
    return &m_heap[loc];
  }


  // fixme: right now for functions we push at the end of stack...this includes when we exit a block and should move the stack ptr back...
  void alloc_stack(uint4 loc, bool is_obj) {
    if (loc >= m_stack.size()) {
      m_stack.resize(loc+1);
      m_is_obj.resize(loc+1);
    }
    m_stack[loc] = 0;
    m_is_obj[loc] = is_obj;
  }

  mem_unit get_stack_val(uint4 loc) {
    return m_stack[loc];
  }

  void set_stack_val(uint4 loc, mem_unit value) {
    m_stack[loc] = value;
  }

  void alloc_stack_from_frame(uint4 loc, bool is_obj) {
    if (loc+m_frame_index >= m_stack.size()) {
      m_stack.resize(loc+m_frame_index+1);
      m_is_obj.resize(loc+m_frame_index+1);
    }
    m_stack[loc+m_frame_index] = 0;
    m_is_obj[loc+m_frame_index] = is_obj;
  }

  mem_unit get_stack_val_from_frame(sint4 loc) {
    return m_stack[m_frame_index+loc];
  }
  
  void set_stack_val_from_frame(sint4 loc, mem_unit value) {
    m_stack[m_frame_index+loc] = value;
  }


  void set_next_stack_val(mem_unit value, bool is_obj) {
    m_stack.push_back(value);
    m_frame_index = m_stack.size();
    m_is_obj.push_back(is_obj);
  }

  void set_frame_index() {
    m_frame_index_stack.push(m_last_frame_index);
    m_last_frame_index = m_frame_index;
  }
  
  void reset_frame_index() {
    m_frame_index = m_frame_index_stack.top();
    m_last_frame_index = m_frame_index;
    m_frame_index_stack.pop();
  }

private:
  static const uint1 m_init_heap_size = 128;
  static const uint2 m_heap_block = 256;



  vector<mem_unit> m_stack;
  vector<bool> m_is_obj;
  uint4 m_stack_index;
  uint4 m_frame_index;

  uint4 m_last_frame_index;
  stack<uint4> m_frame_index_stack;

  mem_unit* m_heap;
  uint4 m_heap_index;
  uint4 m_heap_size;

  void GC(sint4 elements);
};



#endif
