// $Id: PathFinder.C,v 1.25 2006/04/21 00:23:15 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "PathFinder.H"
#include "Game.H"
#include "GfxModule.H"
#include "SWidget.H"
#include "ClientState.H"
#include "Options.H"

#define TIME 1

using namespace std;

const sint4 PathFinder::FIND_PATH_MSG = EventFactory::new_what();
const sint4 PathFinder::FIND_PATH_STOP = EventFactory::new_what();


PathFinder::PathFinder(ClientState &c) : state(c)
{
  widget = new SWidget;
  widget->set_pp(&pp);
  env = 0;
  width = height = 0;
}


PathFinder::~PathFinder() { 
  delete widget;
  delete env;
}


bool PathFinder::handle_event(const Event &e)
{
  static Game &game = state.gsm->get_game();
  sint4 vf = game.get_view_frame();
  sint4 ai = game.get_action_frame();
  bool behind = abs(vf-ai) > 10;

  //------------------------------
  // handle pathfinding event
  
  if (e.get_what() == PathFinder::FIND_PATH_MSG) {
    
    REM("FIND_PATH EVENT");
    const PathEvent &pe = static_cast<const PathEvent&>(e);
    taskQ.push(pe);
    return true;
  }
  
  if (e.get_what() == PathFinder::FIND_PATH_STOP) {

    REM("STOP_PATH EVENT");
    const PathStopEvent &pe = static_cast<const PathStopEvent&>(e);
    FORALL (pe.get_objs(), i) {
      pp.remove_path(*i);
    }
    return true;
  }

  //------------------------------

#if 0  
  if (e.get_who() == GfxModule::FROM) {

    if (e.get_what() == GfxModule::ADD_CWIDGETS_MSG) {
      init_widget();
      return true;
    }
  }
#endif
  
  if (e.get_who() == GameStateModule::FROM) {

    if (e.get_what() == GameStateModule::VIEW_MSG) {

      //      clock_t t1 = clock();
      
      if (!env) init_env();
      
      Vector<ScriptObj*> boundaries = state.gsm->get_changes().new_boundaries;
      //      Vector<GameObj*> boundaries;
      //      boundaries = state.gsm->get_changes().new_boundaries;      
      
      FORALL (boundaries, it) {

        GameObj *gob = dynamic_cast<GameObj*>(*it);
        if (gob == 0) continue;
        sint4 x1 = env->world2x(*gob->sod.x1);
        sint4 y1 = env->world2y(*gob->sod.y1);      
        sint4 x2 = env->world2x(*gob->sod.x2);
        sint4 y2 = env->world2y(*gob->sod.y2);
        
        bool costonly;
        Options::get("-costonly", costonly);
        if (costonly) {
          env->insert_line(x1, y1, x2, y2, Object::ON_LAND, CLOSED, 0, Environment::BOUND_COST);
        } else {
          env->insert_line(x1, y1, x2, y2, Object::ON_LAND, CLOSED, 1, 0);
        }

        update_paths(Point2i(x1, y1), Point2i(x2, y2));
      }
      
      Vector<ScriptObj*> objs;
      
      // vanished objects
      objs = state.gsm->get_changes().vanished_objs;
      FORALL (objs, i) {
        GameObj* obj = dynamic_cast<GameObj*>(*i);
        if (obj == 0) continue;
        if (!obj->has_attr("x") || !obj->has_attr("y") || !obj->has_attr("shape")) continue;
        env->remove_object(obj);
      }

      // dead objects
      objs = state.gsm->get_changes().dead_objs;
      FORALL (objs, i) {
        GameObj* obj = dynamic_cast<GameObj*>(*i);
        if (obj == 0) continue;
        if (!obj->has_attr("x") || !obj->has_attr("y") || !obj->has_attr("shape")) continue;
        env->remove_object(obj);
      }

      // new objects
      objs = state.gsm->get_changes().new_objs;
      FORALL (objs, i) {
        GameObj *obj = dynamic_cast<GameObj*>(*i);
        if (obj == 0) continue;
        if (!obj->has_attr("x") || !obj->has_attr("y") || !obj->has_attr("shape")) continue;
        env->insert_object(obj);
      }

      // changed objects
      objs = state.gsm->get_changes().changed_objs;
      FORALL (objs, i) {
        GameObj* obj = dynamic_cast<GameObj*>(*i);
        if (obj == 0) continue;
        if (obj->attr_changed("x") || obj->attr_changed("y") ||
            obj->attr_changed("zcat") || obj->attr_changed("speed")) {
          env->remove_object(obj);
          env->insert_object(obj);
        }
      }

      /*
      // retry failed paths
      Vector<UnitPath>& p = pp.get_failed();
      FORALL(p, i){
      UnitPath path = env->find_path(i->unit,i->goal );
      //path.tries = (i->tries)+1;
      path.tries = i->tries;
      if(path.get_path().size() > 0){ 
      pp.add_new_path(path);
      pp.process_paths();
      }else{
      Vector<GameObj*> ob; ob.push_back(i->unit);
      Vector<sint4> pa; char* c = "repair failed :(";
      pa.push_back((sint4)c);
      ObjAction *d = i->unit->action("say");
      if (d) d->Act(ob, pa);
      }
      }
      p.clear();
      */

      // retry failed paths
      Vector<UnitPath> &p = pp.get_failed();
      bool norepair = (p.size() == 0);
      if (p.size() > 0 && !behind) {
        Vector<UnitPath>::iterator i = p.begin();
        UnitPath path = env->find_path(i->unit, i->goal);
        cout << "retry " << i->tries << endl;
        path.tries = i->tries;
        path.spot = i->spot;
        if (path.get_path().size() > 0) {
          pp.add_new_path(path);
        } else {
          SActParam p;
          p.obj.push_back(i->unit);
          char* c = "repair failed :(";
          p.val.push_back((stype)c);
          ObjAction *d = i->unit->action("say");
          if (d) d->Act(p);
        }
        p.erase(i);
      }

      if (norepair && taskQ.size() > 0 && !behind) {
        find_path(&taskQ.front());
        taskQ.pop();
      }

      bool notasks = (taskQ.size() == 0);
      
      if (notasks && norepair && !behind ) {
        pp.process_paths();
      }
      
      //      clock_t t2 = clock();
      
      //      cout << "TIME/VIEW: " << (double)(t2-t1)*1000/(double)CLOCKS_PER_SEC << endl;
      return true;
    }
  }

  return true;
}


void PathFinder::update_paths(const Point2i &a, const Point2i &b)
{
  Vector<UnitPath>& p = pp.get_paths();
  Vector<Line> l;

  FORALL (p, i) {
    l = i->intersect(a,b);
    if (l.size() > 0) {
      UnitPath up = *i;
      p.erase(i);
      UnitPath path = env->find_path(up.unit, up.goal);
      if (path.get_path().size() > 0) { pp.add_new_path(path); }

      //cout << "[1]GOAL: (" << i->goal.x << "," << i->goal.y << ")" << endl;
    }
  }
}


void PathFinder::find_path(const PathEvent *pe)
{
  GameObj *obj;
  
  Coor3 p = pe->get_dest();
  
  real4 locX,locY;
  locX = p.x;
  locY = p.y;
  obj = (pe->get_objs())[0];

  UnitPath path = env->find_path(obj, Point2i((sint4)locX,(sint4)locY));
  if (path.get_path().size() > 0) {

    pp.add_new_path(path); 

  } else {

    char* c = "cannot find a path!";
    SActParam p;
    p.obj.push_back(obj);
    p.val.push_back((stype)c);
    ObjAction *d = obj->action("say");
    if (d) d->Act(p);
  }
}


void PathFinder::init_widget()
{
  state.gfxm->add_cwidget(*widget);
}


void PathFinder::init_env()
{
  width = state.gsm->get_game().get_map().get_width();
  height = state.gsm->get_game().get_map().get_height();
  tile_points = state.gsm->get_game().get_tile_points();
  sint4 gran;
  Options::get("-gran", gran);
  env = new Environment(state, width, height, tile_points, gran);
  
  widget->set_env(env);
  widget->set_tp(tile_points);

  map = env->get_map();
}
    
