#ifndef PathProcessor_H
#define PathProcessor_H

// $Id: PathProcessor.H,v 1.9 2005/06/28 08:42:18 orts_furtak Exp $
      
// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "Path.H"
#include "EventHandler.H"
#include "ClientModule.H"

//===================================================================

class PathStatusEvent : public Event
{
public:

  PathStatusEvent(sint4 who, sint4 what, GameObj *obj_) :
    Event(who, what), obj(obj_) {}
  
  GameObj* get_obj() const { return obj; }

protected:

  GameObj *obj;
};

//-------------------------------------------------------------------

class PathProcessor : public ClientModule
{
public:

  static const sint4 FROM;
  static const sint4 PATH_SUCCESS;
  static const sint4 PATH_FAILURE;

  Vector<UnitPath> paths;
  Vector<UnitPath> failed;
  
  int count;
  
  PathProcessor() { count = 0; }
    
  ~PathProcessor() { }

  void process_paths();

  void add_new_path(const UnitPath &new_path) {

    // delete previous path for the same object
    // fixme: SLOW!
    FORALL (paths, i){
      if (i->unit == new_path.unit) {
        paths.erase(i);
        break;
      }
    }
    paths.push_back(new_path);
  }

  void remove_dead_paths() {

    for (typeof(paths.begin()) i = paths.begin(); i != paths.end(); ) {
      if (i->unit->is_dead() || i->dead) {
        i = paths.erase(i);
      } else {
        ++i;
      }
    }
  }

  void remove_path(GameObj *obj);

  void remove_all_paths() { paths.clear(); }

  void remove_all_ground_paths(){

    for (typeof(paths.begin()) i = paths.begin(); i != paths.end();) {
      if (i->unit->get_int("zcat")==Object::ON_LAND) {
        i = paths.erase(i);
      } else {
        ++i;
      }
    }
  }

  void set_paths(Vector<UnitPath> &p) { paths = p; }

  Vector<UnitPath>& get_paths()  { return paths; }  
  Vector<UnitPath>& get_failed() { return failed; }
};

//===================================================================

#endif
