// OpenSceneGraph test file
// derived from osg/Producer tutorial files

// $Id: osgtest.C,v 1.1.1.1 2004/10/11 09:09:19 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include <Producer/Camera>
#include <Producer/CameraGroup>

#include <osg/MatrixTransform>
#include <osg/PositionAttitudeTransform>
#include <osg/Geometry>

#include <osg/Geode>
#include <osgUtil/Optimizer>
#include <osgDB/Registry>
#include <osgDB/ReadFile>
#include <osgProducer/Viewer>

#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;


#if 0

class MySceneHandler : public Producer::Camera::SceneHandler
{
public:
  
  MySceneHandler(osg::Node *scene) : Producer::Camera::SceneHandler(),
				     _angle(0.0f),
				     _initialized(false),
				     _scene(scene)
  {}

  // Producer::Camera::SceneHandler is pure virtual and
  // requires that the draw() method be implemented in 
  // derived classes.

  virtual void draw(Producer::Camera& camera)
  {
    // Lazy initialize
    if( !_initialized ) init();

    // Clear the Camera's RenderSurface ProjectionRect
    camera.clear();
                        
    // Apply the Projection parameters
    camera.applyLens();

    // Place the Camera into the world
    camera.setViewByLookat( 0.0f, 0.0f, 4.0f,   /// Position of the eye
			    0.0f, 0.0f, 0.0f,   /// Point the eye is looking at
			    0.0f, 1.0f, 0.0f ); /// Up

    glPushMatrix();
    glRotatef( _angle, 0, 1, 0 );

    // Draw the OpenGL teapot
    //glutSolidTeapot(1.0);

    camera.frame(false);

    _angle += 1;
  }

  // Initialize Graphics state

  void init()
  {
    glEnable( GL_DEPTH_TEST );
    glEnable( GL_LIGHTING );
    glEnable( GL_LIGHT0 );
    glClearColor( 0.2f, 0.2f, 0.4f, 1.0f );

    _initialized = true;
  }

private:

  float _angle;
  bool _initialized;
  osg::Node* _scene;
};

osg::AnimationPath* createAnimationPath(const osg::Vec3& center,float radius,double looptime)
{
  // set up the animation path 
  osg::AnimationPath* animationPath = new osg::AnimationPath;
  animationPath->setLoopMode(osg::AnimationPath::LOOP);
    
  int numSamples = 40;
  float yaw = 0.0f;
  float yaw_delta = 2.0f*osg::PI/((float)numSamples-1.0f);
  float roll = osg::inDegrees(30.0f);
    
  double time=0.0f;
  double time_delta = looptime/(double)numSamples;
  for (int i=0;i<numSamples;++i) {
    osg::Vec3 position(center+osg::Vec3(sinf(yaw)*radius,cosf(yaw)*radius,0.0f));
    osg::Quat rotation(osg::Quat(roll,osg::Vec3(0.0,1.0,0.0))*
		       osg::Quat(-(yaw+osg::inDegrees(90.0f)),
				 osg::Vec3(0.0,0.0,1.0)));
        
    animationPath->insert(time,osg::AnimationPath::ControlPoint(position,rotation));

    yaw += yaw_delta;
    time += time_delta;

  }
  return animationPath;    
}

#endif


osg::Node* createMovingModel(const osg::Vec3& /* center */, float radius)
{
  //float animationLength = 10.0f;
  // osg::AnimationPath* animationPath = createAnimationPath(center,radius,animationLength);

  osg::Group* model = new osg::Group;

#if 0  
  osg::Node* glider = osgDB::readNodeFile("glider.osg");
  if (glider) {
    const osg::BoundingSphere& bs = glider->getBound();

    float size = radius/bs.radius();
    osg::MatrixTransform* positioned = new osg::MatrixTransform;
    positioned->setDataVariance(osg::Object::STATIC);
    positioned->setMatrix(osg::Matrix::translate(-bs.center())*
			  osg::Matrix::scale(size,size,size)*
			  osg::Matrix::rotate(osg::inDegrees(-90.0f),0.0f,0.0f,1.0f));
    
    positioned->addChild(glider);
    
    osg::PositionAttitudeTransform* xform = new osg::PositionAttitudeTransform;    
    xform->setUpdateCallback(new osg::AnimationPathCallback(animationPath,0.0,1.0));
    xform->addChild(positioned);

    model->addChild(xform);
  }
#endif

  const int N = 10; // number of cessnas

  osg::Node* cessna_template = osgDB::readNodeFile("data/cessna.osg");
  if (!cessna_template) { cerr << "cessna.osg not found" << endl; exit(20); }

  for (int k=0; k < N; ++k) {

    osg::Node* cessna = static_cast<osg::Node*>(cessna_template->clone(osg::CopyOp::DEEP_COPY_ALL));
    const osg::BoundingSphere& bs = cessna->getBound();
    
    float size = (radius/bs.radius())*(k+1)/N;
    osg::MatrixTransform* positioned = new osg::MatrixTransform;
    positioned->setDataVariance(osg::Object::STATIC);
    osg::Vec3 t(k*0.5, 0, 0);
    t -= bs.center();
    cout << t << endl;
    positioned->setMatrix(osg::Matrix::scale(size,size,size)*
			  osg::Matrix::rotate(osg::inDegrees(180.0f),0.0f,0.0f,1.0f)*
			  osg::Matrix::translate(t));
    
    cout << positioned->addChild(cessna) << endl;

#if 0    
    osg::MatrixTransform* xform = new osg::MatrixTransform;
    xform->setUpdateCallback(new osg::AnimationPathCallback(animationPath,0.0f,2.0));
    xform->addChild(positioned);
#endif
    
    model->addChild(positioned);
  }
    
  return model;
}


osg::Node* createBase(const osg::Vec3& center,float radius)
{
  int numTilesX = 15;
  int numTilesY = 15;
    
  float width  = numTilesX*radius;
  float height = numTilesY*radius;
    
  osg::Vec3 v000(center - osg::Vec3(width*0.5f,height*0.5f,0.0f));
  osg::Vec3 dx(osg::Vec3(width/((float)numTilesX), 0.0, 0.0f));
  osg::Vec3 dy(osg::Vec3(0.0f, height/((float)numTilesY), 0.0f));
    
  // fill in vertices for grid, note numTilesX+1 * numTilesY+1...
  osg::Vec3Array* coords = new osg::Vec3Array;
  for(int iy=0; iy <= numTilesY; ++iy) {
    for(int ix=0; ix <= numTilesX; ++ix) {
      coords->push_back(v000+dx*ix+dy*iy);
    }
  }
    
  //Just two colours - black and white.
  osg::Vec4Array* colors = new osg::Vec4Array;
  colors->push_back(osg::Vec4(1.0f,1.0f,1.0f,1.0f)); // white
  colors->push_back(osg::Vec4(0.0f,0.0f,0.0f,1.0f)); // black
  int numColors=colors->size();
    
    
  int numIndicesPerRow=numTilesX+1;
  osg::UByteArray* coordIndices = new osg::UByteArray; // assumes we are using less than 256 points...
  osg::UByteArray* colorIndices = new osg::UByteArray;

  for(int iy=0; iy < numTilesY; ++iy) {
    for(int ix=0; ix < numTilesX; ++ix) {
      // four vertices per quad.
      coordIndices->push_back(ix    +(iy+1)*numIndicesPerRow);
      coordIndices->push_back(ix    +iy*numIndicesPerRow);
      coordIndices->push_back((ix+1)+iy*numIndicesPerRow);
      coordIndices->push_back((ix+1)+(iy+1)*numIndicesPerRow);
            
      // one color per quad
      colorIndices->push_back((ix+iy)%numColors);
    }
  }
    
  // set up a single normal
  osg::Vec3Array* normals = new osg::Vec3Array;
  normals->push_back(osg::Vec3(0.0f,0.0f,1.0f));
    
  osg::Geometry* geom = new osg::Geometry;
  geom->setVertexArray(coords);
  geom->setVertexIndices(coordIndices);
    
  geom->setColorArray(colors);
  geom->setColorIndices(colorIndices);
  geom->setColorBinding(osg::Geometry::BIND_PER_PRIMITIVE);
    
  geom->setNormalArray(normals);
  geom->setNormalBinding(osg::Geometry::BIND_OVERALL);
    
  geom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::QUADS,0,coordIndices->size()));
    
  osg::Geode* geode = new osg::Geode;
  geode->addDrawable(geom);
    
  return geode;
}


osg::Node* createModel()
{
  osg::Vec3 center(0.0f,0.0f,0.0f);
  float radius = 0.5f;

  osg::Group* root = new osg::Group;

  root->addChild(createMovingModel(center, radius));

  root->addChild(createBase(center-osg::Vec3(0.0f,0.0f,radius*0.5),radius));

  return root;
}


class KbdMouseCallback : public Producer::KeyboardMouseCallback
{
  bool _done;
                
public:
  float _mx, _my;
  float _px, _py;
  unsigned int _bm;
        
  KbdMouseCallback() : _done( false ), _mx(0.f), _my(0.f), _bm(0) {}
        
  virtual void mouseMotion( float mx, float my )
  {
    _mx = mx;
    _my = my;
  }
        
  virtual void passiveMouseMotion( float mx, float my )
  {
    _mx = mx;
    _my = my;
  }
        
  virtual void buttonPress( float mx, float my, unsigned int bm )
  {
    _mx = mx;
    _my = my;
    _bm |= (1<<(bm-1));
  }
        
  virtual void buttonRelease( float mx, float my, unsigned int bm )
  {
    _mx = mx;
    _my = my;
    _bm &= ~(1<<(bm-1));
  }
        
  virtual void keyPress( unsigned int ) { }

  virtual void specialKeyRelease( Producer::KeyCharacter key )
  {
    if( key == Producer::KeyChar_Escape ) shutdown();
  }
        
  virtual void shutdown() { _done = true; }
        
  bool isDone() const { return _done; }
};


static long long current_time(void)
{
   struct timeval tv;
   struct timezone tz;
   gettimeofday(&tv, &tz);
   return tv.tv_sec*1000000 + tv.tv_usec;
}

 
int main(int, char **)
{
  osg::Node* model = createModel();
  if (!model) return 1;
    
  // tilt the scene so the default eye position is looking down on the model

  osg::MatrixTransform* rootnode = new osg::MatrixTransform;
  rootnode->setMatrix(osg::Matrix::rotate(osg::inDegrees(45.0f),1.0f,0.0f,0.0f));
  rootnode->addChild(model);

  // run optimization over the scene graph
  
  osgUtil::Optimizer optimzer;
  optimzer.optimize(rootnode);
     
  osg::ref_ptr<Producer::RenderSurface> rs = new Producer::RenderSurface;

  rs->setWindowRectangle(0, 0, 800, 600);
  rs->setWindowName("osgtest");
  rs->useBorder(true);
  rs->setInputRectangle(Producer::RenderSurface::InputRectangle(0.f, 1.f, 0.f, 1.f));
        
  rs->realize();
        
  osg::ref_ptr<Producer::Camera> cam0 = new Producer::Camera;
  osg::ref_ptr<Producer::Camera> cam1 = new Producer::Camera;
        
  cam0->setRenderSurface(&*rs);
  cam1->setRenderSurface(&*rs);
        
  cam0->setProjectionRectangle (0.0f, 0.4f, 0.0f, 1.0f);
  cam1->setProjectionRectangle (0.6f, 1.0f, 0.0f, 1.0f);

  cam0->setLensOrtho(-5, 5, -5, 5, -20, 20);
  cam1->setLensOrtho(-5, 5, -5, 5, -20, 20);

  //cam0->setLensPerspective( -45.0, 45.0, 1.0, 1000.0 );
  //cam1->setLensPerspective( -45.0, 45.0, 1.0, 1000.0 );

  osg::ref_ptr<osgUtil::SceneView> sv = new osgUtil::SceneView;
  sv->setDefaults();
  sv->setSceneData(rootnode);
  
  osgProducer::OsgSceneHandler* sh0 = new osgProducer::OsgSceneHandler;
  osgProducer::OsgSceneHandler* sh1 = new osgProducer::OsgSceneHandler;
                
  sh0->setSceneView(&*sv);
  sh1->setSceneView(&*sv);
        
  cam0->setSceneHandler(sh0);
  cam1->setSceneHandler(sh1);

  osg::ref_ptr<Producer::InputArea> ia = new Producer::InputArea;
  ia->addRenderSurface(&*rs);
        
  osg::ref_ptr<Producer::KeyboardMouse> km0 = new Producer::KeyboardMouse(&*rs);
  osg::ref_ptr<KbdMouseCallback> kmcb0 = new KbdMouseCallback;

  km0->setCallback(&*kmcb0);
  // km0->setDoNormalizedMouseMotion(true);

  float angle = 45;
  int frames = 0;
  long long start = current_time();
  
  while(!kmcb0->isDone()) {

    rs->sync();
    km0->update(*kmcb0);
                
    //tb.input( kmcb0._mx, kmcb0._my, kmcb0._bm );
    //cam1->setViewByMatrix( tb.getMatrix() );

    rootnode->setMatrix(osg::Matrix::rotate(osg::inDegrees(angle),0.0f,0.0f,1.0f));
    ++angle;
    
    cam0->frame(false);
    cam1->frame(true);

    ++frames;

    if (frames >= 100) {
      cout << frames*1000000.0/double(current_time()-start+0.001) << " FPS" << endl;
      frames = 0;
      start = current_time();
    }
  }

  rs->sync();
  return 0;
}
