/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "CompAction.H"
#include "String.H"

Prop::Prop()
{
  sv_id = 0;                // invalid
  sv_owner = -1;            // not owned
  sv_radius = 0;
  sv_sight_range = 0;
  sv_min_attack_range = 0;
  sv_max_attack_range = 0;
  sv_speed = 0;
  sv_damage = 0;
  sv_hit_points = 0;
  sv_moving = 0;
  sv_pos_x = 0;   // >= 0 && < 32768
  sv_pos_y = 0;   // >= 0 && < 32768
}


void Prop::gen_diff(const Prop &prop, ostream &osdiff)
{
  // id is processed elsewhere

  osdiff << sv_owner - prop.sv_owner << " "
	 << sv_radius - prop.sv_radius << " "
	 << sv_sight_range - prop.sv_sight_range << " "
	 << sv_min_attack_range - prop.sv_min_attack_range << " "
	 << sv_max_attack_range - prop.sv_max_attack_range << " "
	 << sv_speed - prop.sv_speed << " "
	 << sv_damage - prop.sv_damage << " "
	 << sv_hit_points - prop.sv_hit_points << " "
	 << sv_moving - prop.sv_moving << " "
	 << sv_pos_x - prop.sv_pos_x << " "
	 << sv_pos_y - prop.sv_pos_y;
}

void Prop::apply_diff(istream &isdiff)
{
  // id is processed elsewhere
  
  sint4 d;
  isdiff >> d; sv_owner += d;
  isdiff >> d; sv_radius += d;
  isdiff >> d; sv_sight_range += d;
  isdiff >> d; sv_min_attack_range += d;
  isdiff >> d; sv_max_attack_range += d;
  isdiff >> d; sv_speed += d;
  isdiff >> d; sv_damage += d;
  isdiff >> d; sv_hit_points += d;
  isdiff >> d; sv_moving += d;
  isdiff >> d; sv_pos_x += d;
  isdiff >> d; sv_pos_y += d;
}

void Prop::opp_view() {}

void Prop::write_verbose(ostream &os) const
{
  os << "id=" << sv_id << endl
     << "owner=" << sv_owner << endl
     << "radius=" << sv_radius << endl
     << "sight_rang=" << sv_sight_range << endl  
     << "min_attack_range=" << sv_min_attack_range << endl
     << "max_attack_range=" << sv_max_attack_range << endl  
     << "speed=" << sv_speed << endl
     << "damage=" << sv_damage << endl
     << "hit_points=" << sv_hit_points << endl
     << "moving=" << sv_moving << endl
     << "pos_x=" << sv_pos_x << endl
     << "pos_y=" << sv_pos_y << endl;
}

void Prop::write(ostream &os) const
{
  os << sv_id << " "
     << sv_owner << " "
     << sv_radius << " "
     << sv_sight_range << " "  
     << sv_min_attack_range << " "
     << sv_max_attack_range << " "  
     << sv_speed << " "
     << sv_damage << " "
     << sv_hit_points << " "
     << sv_moving << " "
     << sv_pos_x << " "
     << sv_pos_y;
}

void Prop::write(FILE *fp) const {

  fprintf(fp, "%d %d %d %d %d %d %d %d %d %d %d %d",
	  sv_id,
	  sv_owner,
	  sv_radius,
	  sv_sight_range,  
	  sv_min_attack_range,
	  sv_max_attack_range,  
	  sv_speed,
	  sv_damage,
	  sv_hit_points,
	  sv_moving,
	  sv_pos_x,
	  sv_pos_y             
	  );
}


void Prop::read(FILE *fp)
{
  sint4 x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12;
  
  if (fscanf(fp, "%d %d %d %d %d %d %d %d %d %d %d %d",
	     &x1,&x2,&x3,&x4,&x5,&x6,&x7,&x8,&x9,&x10,&x11,&x12) != 12)
    ERR("feature list corrupt");

  sv_id               = x1;
  sv_owner            = x2;
  sv_radius           = x3;
  sv_sight_range      = x4;  
  sv_min_attack_range = x5;
  sv_max_attack_range = x6;  
  sv_speed            = x7;
  sv_damage           = x8;
  sv_hit_points       = x9;
  sv_moving           = x10;
  sv_pos_x            = x11;
  sv_pos_y            = x12;
}

// true if error

bool Prop::read_init_values(istream &is) {
  string s;

  sv_id = 0;
  sv_owner = -1;
  sv_moving = false;
  sv_pos_x = 0;
  sv_pos_y = 0;
  
  is >> s; 
  is >> sv_radius;
  if (!is || s != "radius:") { ERR("radius: expected"); }

  is >> s; 
  is >> sv_sight_range;
  if (!is || s != "sight_range:") { ERR("sight_range: expected"); }

  is >> s; 
  is >> sv_min_attack_range;
  if (!is || s != "min_attack_range:") { ERR("min_attack_range: expected"); }

  is >> s; 
  is >> sv_max_attack_range;
  if (!is || s != "max_attack_range:") { ERR("max_attack_range expected"); }

  is >> s; 
  is >> sv_speed;
  if (!is || s != "speed:") { ERR("speed: expected"); }

  is >> s; 
  is >> sv_damage;
  if (!is || s != "damage:") { ERR("damage: expected"); }
    
  is >> s; 
  is >> sv_hit_points;
  if (!is || s != "hit_points:") { ERR("hit_points: expected"); }

  return (!is);
}


void Action::write(FILE *fp) const
{
  switch (action) {
  case NO_ACTION: fputc('N', fp); return;
  case STOP:      fputc('S', fp); return;
  case MOVE:      fprintf(fp, "M%d", move_direction); return;
  case ATTACK:    fprintf(fp, "A%d", target_id);   return;
  default: ERR("unknown action");
  }

  return;
}

void Action::write(ostream &os) const
{
  switch (action) {
  case NO_ACTION: os << 'N'; return;
  case STOP:      os << 'S'; return;
  case MOVE:      os << "M" << move_direction; return;
  case ATTACK:    os << "A" << target_id; return;
  default: ERR("unknown action");
  }

  return;
}

void Action::read(FILE *fp)
{
  char c;

  if (fscanf(fp, " %c", &c) != 1) ERR("read error (action expected)");
  
  switch (c) {
  case 'N': action = NO_ACTION; return;
  case 'S': action = STOP; return;
  case 'M':
    action = MOVE;

    if (fscanf(fp, "%d", &move_direction) != 1) ERR("move direction expected");
    if (move_direction < 0 || move_direction >= DIR_N)
      ERR("move direction out of bounds");
    return;
  case 'A':
    action = ATTACK;
    if (fscanf(fp, "%d", &target_id) != 1) ERR("target id expected");
    if (target_id < 1) ERR("target id < 1");
    return;
    
  default: ERR("unknown action");
  }

  return;
}


void Action::read(istream &is)
{
  char c;

  is >> c;
  if (!is) return;

  switch (c) {
  case 'N': action = NO_ACTION; return;
  case 'S': action = STOP; return;
  case 'M':
    action = MOVE;
    is >> move_direction;
    if (!is) return;
    if (move_direction < 0 || move_direction >= DIR_N)
      ERR("move direction out of bounds");
    return;
  case 'A':
    action = ATTACK;
    is >> target_id;
    if (!is) return;
    if (target_id < 1) ERR("target id < 1");
    return;
    
  default: ERR("unknown action");
  }

  return;
}

void IdAction::write(ostream &os) const
{
  os << id << " ";
  action.write(os);
}

void ActionData::write(FILE *fp) const
{
  FORALL_CONST (vector<uint4>, vicinity_ids, i) fprintf(fp, "%d ", (*i));
  fprintf(fp, ": "); action.write(fp); fputs("", fp);
}

void ActionData::read(FILE *fp)
{
  vicinity_ids.clear();

  uint4 id;
  char c;

  FOREVER {

    if (fscanf(fp, " %c", &c) != 1) ERR("read error (':' expected) ");
    if (c == ':') break;
    ungetc(c, fp);
    if (fscanf(fp, "%d", &id) != 1) ERR("read error (id expected)");
    vicinity_ids.push_back(id);
  }

  action.read(fp);
}


void TickInfo::write(FILE *fp) const
{
  assert(owner==0 || owner == 1);
  
  uint4 n  = elements.size();
  uint4 na = act_data.size();  

  fprintf(fp, "tick: %d n: %d owner: %d actions: %d\n",
	  tick, n, owner, na);
  
  FORU (i, n) { elements[i].write(fp); fputs("\n", fp); }
  FORU (i, na) { act_data[i].write(fp); fputs("\n", fp); }
}


void TickInfo::read(FILE *fp)
{
  sint4 n, na;
  char s1[30],s2[30],s3[30],s4[30];
  
  sint4 r = fscanf(fp, "%20s %d %20s %d %20s %d %20s %d",
		   s1, &tick, s2, &n, s3, &owner, s4, &na);

  if (r != 8) {
    if (feof(fp)) return;
    ERR("tick header corrupt");
  }

  if (strcmp("tick:", s1) || tick < 0 ||
      strcmp("n:", s2) || n <= 0 ||
      strcmp("owner:", s3) || owner < 0 || owner > 1 ||
      strcmp("actions:", s4) || na < 1)
    ERR("tick header data corrupt");

  elements.clear();
  act_data.clear();

  FORS (i, n) {
    Prop f;
    f.read(fp);
    elements.push_back(f);
    if (feof(fp) || ferror(fp)) ERR("read error (elements)");
  }

  FORS (i, na) {
    ActionData ad;
    ad.read(fp);
    act_data.push_back(ad);
    if (feof(fp) || ferror(fp)) ERR("read error (actions)");    
  }

}

