/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef CompAction_H
#define CompAction_H

#include "Global.H"
#include <vector>

// update if Prop changes!
#define PROP_FIRST sv_id
#define PROP_LAST  sv_pos_y

struct Prop {

  uint4  sv_id;
  sint2  sv_owner;            // <0: no owner yet
  uint2  sv_radius;           // bounding radius
  sint2  sv_sight_range;
  sint2  sv_min_attack_range;
  sint2  sv_max_attack_range;  
  sint2  sv_speed;
  sint2  sv_damage;
  sint2  sv_hit_points;
  uint2  sv_moving;
  uint2  sv_pos_x, sv_pos_y;  // grid position

#if 0  
  sint4  sv_mobile;
  sint4  sv_mechanical;
  sint4  sv_reg_rate;
  sint4  sv_sh_max_points;
  sint4  sv_sh_reg_rate;
  sint4  sv_ground_armor;
  sint4  sv_air_armor;
  sint4  sv_sh_points;
  sint4  sv_sh_ground_armor;
  sint4  sv_sh_air_armor;
  sint4  sv_command;     // issued by client
#endif

  Prop();

  void gen_diff(const Prop &prop, ostream &osdiff);
  void apply_diff(istream &isdiff);
  void opp_view();
  uint4 checksum(uint4 id);
  void write_verbose(ostream &os = cout) const;
  void write(ostream &os = cout) const;
  void write(FILE *fp) const;
  void read(FILE *fp);
  void read(const char *);
  bool read_init_values(istream &is);
};


// Ant actions

struct Action {

  enum { DIR_N = 8 };  // must be divisible by 4!
  
  enum { NO_ACTION=0, STOP, MOVE, ATTACK };
  
  sint4 action;         // either MOVE or ATTACK
  
  uint4 target_id;      // ATTACK parameter
  sint4 move_direction; // MOVE parameter

  Action() { action = NO_ACTION; }

  void write(FILE *fp) const;
  void write(ostream &oss) const;
  void read(FILE *fp);
  void read(istream &is);
};



struct ActionData {

  sint4 owner;
  vector<uint4> vicinity_ids;
  Action action;
  
  void write(FILE *fp) const;
  void read(FILE *fp);
};


struct IdAction {

  uint4 id;
  Action action;
  
  void write(ostream &os) const;
  void read(istream &os);
};


struct TickInfo {
  sint4 tick;
  sint4 owner; // 0|1
  vector<Prop> elements;
  vector<ActionData> act_data;

  void write(FILE *fp) const;
  void read(FILE *fp);
};


class CompAction
{
public:

  virtual void write_params(ostream &os) const = 0;
  virtual void read_params(istream &is) = 0;

  virtual Action action(const vector<Prop> &vicinity) = 0;
  virtual ~CompAction() {}
};

typedef CompAction *(*CompActCtor)(void);

#endif
