
// $Id: bsimtest_main.C,v 1.8 2006/06/16 05:34:44 orts_mburo Exp $

#include "GfxGlobal.H"
#include "Options.H"
#include "GfxModule.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "SoundMixer.H"
#include "LoopFunctor.H"
#include "Game.H"
#include "GameChanges.H"
#include "SquadCombatAI.H"
#include "SquadCombatAIWidget.H"
#include "GatherAI.H"
//#include "ClusterSquads.H"
//#include "ClusterSquadsWidget.H"
#include "BuildTasksAI.h"
//#include "HeuristicMapWidget.h"
//#include "MapEvalHandler.h"
#include "ClientState.H"
#include "TerrainModule.H"
//#include "SimpleTerrain.H"
//#include "SimpleTerrainWidget.H"
#include "ST_ForceField.H"

//#include "PRATerrain.H"

using namespace std;

typedef SimpleTerrain::ST_Terrain Terrain;
//typedef SimpleTerrain::ST_ForceField Terrain;

//===================================================================

const char *ORTSG_VERSION = "GfxClient version 0.91 (2005-Apr-10)";

#define USE_EFENCE 0

extern int EF_PROTECT_FREE, EF_FILL, EF_PROTECT_BELOW, EF_ALLOW_MALLOC_0;

//===================================================================

void add_options()
{
  Options o("ortsg");
  o.put("-soundtest", "test sound module");
  o.put("-Moverlay",false,"Embed the evaluation map over the minimap");
  o.put("-Mx",sint4(10),"x position of the evaluation map");
  o.put("-My",sint4(250),"y position of the evaluation map");
  o.put("-Mscale",real8(3.0),"amount we should scale the evaluation map");
  o.put("-Malpha",real8(0.5),"amount of alpha transparency the evaluation map has");
  o.put("-Mwidth",real8(40.0),"the desired width of each abstract square(in terms of tile points)");
  o.put("-Mheight",real8(40.0),"the desired height of each abstract square(in terms of tile points)");

  o.put("-Eimmobile",sint4(3),"the range of influence of immobile objects(in terms of tiles)");
  o.put("-Erange",real8(1.5),"amount that the range of influence of objects is scaled by");
  o.put("-Edecay",real8(0.5),"how much the influence decays per grid square");
}

//===================================================================

static bool just_drew = false;

struct GfxEventHandler : public EventHandler
{
  ClientState &state;

  GfxEventHandler(ClientState &state_) : state(state_) { }

  bool handle_event(const Event& e)
  {
    if (e.get_who() == GameStateModule::FROM) {

      //cout << "Event: " << e.who() << " " << e.what() << endl;

      if (e.get_what() == GameStateModule::STOP_MSG) exit(0);

      if (e.get_what() == GameStateModule::READ_ERROR_MSG) exit(10);

      if (e.get_what() == GameStateModule::VIEW_MSG) {

#if 0
        if (state.gui) {
          state.gui->event();
          state.gui->display();
          if (state.gui->quit) exit(0);
        }
#endif

        // update LowAI objects

        //        const sint4 me = state.gsm->get_game().get_client_player();

        // add new objects

        //        FORALL (state.gsm->get_changes().new_objs, i) {
        //          GameObj *gob = (*i)->get_GameObj();
        //        }

        // remove dead objects

        //        FORALL (state.gsm->get_changes().dead_objs, i) {
        //          GameObj *gob = (*i)->get_GameObj();
        //        }

        // remove vanished gfx objects

        //        FORALL (state.gsm->get_changes().vanished_objs, i) {
        //          GameObj *gob = (*i)->get_GameObj();
        //        }

        // compute actions

        // turret_ai.act(*state.gsm);


        state.gfxm->process_changes();

        ////////////////
        Vector<string> temp;
        Game &game = state.gsm->get_game();
        FORS (i, game.get_player_num()) temp.push_back(""); // dummy list of commands
        game.exec_actions(temp); // evaluate client actions and send actions to the server
        ////////////////

        state.gsm->send_actions();

        sint4 vf = game.get_view_frame();
        sint4 ai = game.get_action_frame();
        sint4 sk = game.get_skipped_actions();
        sint4 md = game.get_merged_actions();

        if (vf != ai || sk || md > 1) {

          cout << "frame " << vf;
          if (ai < 0) cout << " [no action]";
          else if (vf != ai) cout << " [behind by " << (vf - ai) << " frame(s)]";
          if (sk) cout << " [skipped " << sk << "]";
          if (md > 1) cout << " [merged " << md << "]";
          cout << endl;
        }

        // don't draw if we are far behind
        if (abs(vf-ai) < 10) {
          state.gfxm->draw();
        }
        just_drew = true;

        return true;
      }

    }

    return false;
  }
};

//===================================================================

class ortsgLooper : public LoopFunctor
{
private:
  ClientState &state;
  bool refresh;

public:

  ortsgLooper(ClientState &s) : state(s)
  {
    ::Options::get("-refresh", refresh);
  }

  void loop() const {

    if (state.gfxm->get_quit()) exit(0);

    if (refresh) {

      if (!just_drew) // allow some waiting time after a message
        state.gfxm->draw();
      else
        just_drew = false;
    }

    // server events
    if (!state.gsm->recv_view()) SDL_Delay(1);
  }
};

//===================================================================

int main(int argc, char **argv)
{
  char mydir[81];
  getcwd(mydir, 80);

  glutInit(&argc, argv);

  chdir(mydir);

  try {

#if USE_EFENCE
    // aggressive efence checks
    EF_PROTECT_FREE = 1;
    EF_FILL = 0xee;
    EF_PROTECT_BELOW = 1;
    EF_ALLOW_MALLOC_0 = 1;
#endif

    signals_shut_down(true);
    atexit(SDL_Quit);

    // populate opt with options from various modules

    add_options();
    GameStateModule::Options::add();
    GfxModule::Options::add();
    Terrain::add_options();
    SquadCombatAI::add_options();

    // process command line arguments
    if (Options::process(argc, argv, cerr, ORTSG_VERSION)) exit(20);

    bool st = false;
    Options::get("-soundtest", st);

    // frano stuff
    /*
    bool overlay;
    sint4 mapx,mapy;
    real8 scale,alpha,width,height;
    
    Options::get("-Moverlay",overlay);
    Options::get("-Mx",mapx);
    Options::get("-My",mapy);
    Options::get("-Mscale",scale);
    Options::get("-Malpha",alpha);
    Options::get("-Mwidth",width);
    Options::get("-Mheight",height);

    HeuristicMapWidget widget(Point2i(mapx,mapy),scale,width,height,alpha,overlay);  

    widget.addAttribute("hp",750.0);
    widget.addAttribute("max_speed",10.0);
    
    widget.addAttribute("armor.AC",10.0);
    widget.addAttribute("armor.slashing",10.0);
    widget.addAttribute("armor.piercing",10.0);
    widget.addAttribute("armor.explosive",10.0);
    widget.addAttribute("armor.blunt",10.0);
    
    widget.addAttribute("weapon.min_damage",10.0);
    widget.addAttribute("weapon.max_damage",10.0);
    widget.addAttribute("weapon.bonus",20.0);
    widget.addAttribute("weapon.max_air_range",200.0);
    widget.addAttribute("weapon.max_ground_range",200.0);
    
    //widget.initializeMap();
    sint4 immobileRange;
    real8 rangeMod, decay;
  
    Options::get("-Eimmobile",immobileRange);
    Options::get("-Erange",rangeMod);
    Options::get("-Edecay",decay);
    */




    if (st) {
      string path;
      Options::get("-p", path);
      sound_test(path);
      exit(0);
    }

    // change seed value to time(0) if 0
    sint4 seed;
    Options::get("-seed", seed);
    if (!seed) seed = time(0);
    Options::set("-seed", seed);
    cout << "seed=" << seed << endl;

    srand48(seed); // used by STL, fixme: avoid (global)

    SDLinit::video_init();
    SDLinit::network_init();

    // populate option objects with command line values
    // and create modules

    GameStateModule::Options gsmo;
    GameStateModule gsm(gsmo);

    GfxModule::Options gfxmo;
    GfxModule gfxm;

    // gives event handler access to the state
    ClientState state(gsm, gfxm);

    // ortsg event handler
    GfxEventHandler gev(state);

  //=====================================
  // create pathfinding module and implementation
  Terrain timp;
  TerrainModule tm(gsm, timp, 30);

  // hook up terrain widget to gfx module
    //SimpleTerrain::SimpleTerrainWidget st_widget(timp, &gfxm);
    //gfxm.add_cwidget(st_widget);
  //=====================================

    //=====================================
    // create pathfinding module and implementation
    //    Terrain timp;
    //TerrainModule tm(gsm, timp);

    // hook up terrain widget to gfx module
    //SimpleTerrain::SimpleTerrainWidget st_widget(timp, &gfxm);
    //gfxm.add_cwidget(st_widget);
    //=====================================

  BSimul::BuildTasksAI builder_ai(gsm, tm);
    builder_ai.add_handler(&tm);
    gsm.add_handler(&builder_ai);
    tm.add_handler(&builder_ai);

    //=====================================
    //   GatherAI gather_ai(gsm);
    //   gather_ai.add_handler(&tm);
    //   gfxm.add_handler(&gather_ai);
    //   gsm.add_handler(&gather_ai);
    //   tm.add_handler(&gather_ai);
    //=====================================
  //   SquadCombatAI shooter_ai(gsm);
    //     shooter_ai.add_handler(&tm);
    //     gfxm.add_handler(&shooter_ai);
    //     gsm.add_handler(&shooter_ai);
    //     tm.add_handler(&shooter_ai);
    //     //    shooter_ai.enable();
    //     //-------------------------------------
    //     SquadCombatAIWidget *sw = new SquadCombatAIWidget(shooter_ai, &gfxm);
    //    sw->attach(&shooter_ai);
    
    SquadCombatAI shooter_ai(gsm);
    gfxm.add_handler(&shooter_ai);
    builder_ai.add_handler(&shooter_ai);
    shooter_ai.add_handler(&builder_ai);

    shooter_ai.enable_motion();
    shooter_ai.enable_attack();
    //-------------------------------------
    //ShooterAIWidget *sw = new ShooterAIWidget;
    //sw->attach(&shooter_ai);

    //=====================================

    //MapEvalHandler handler(state, &widget, immobileRange, rangeMod, decay);

    // register pathfinder as gfx and gsm event handler
    gfxm.add_handler(&tm);
    //gsm.add_handler(&tm); done in TerrainModule

    // gsm events also handled in gev
    gsm.add_handler(&gev);
    /*
    frano's stuff
    gfxm.add_handler(&handler);
    gsm.add_handler(&handler);
    tm.add_handler(&handler);
    */
    // connect game state module to server
    gsm.connect();
    gfxm.init(gsm, gfxmo);


    //  gfxm.add_cwidget(*sw);
    //   gfxm.add_cwidget(*csw);

    // gfx loop (should become a thread)
    ortsgLooper l(state);

    cout << "start loop" << endl;

    gfxm.start_loop(&l);
  }
  catch (ExitException& ee) {
    errstr << std::endl << *(ee.get_reason()) << std::endl;
#ifdef GCC
    abort();
#else
    ABORT;
#endif
    exit(20);
  }
  return 0;
}

//===================================================================

