// $Id: AutoGatherAI.C,v 1.15 2006/06/15 22:03:23 ddeutscher Exp $

#include "Global.H"
#include "Options.H"
#include "AutoGatherAI.H"

using namespace std;

typedef TerrainBase::Loc Loc;
typedef TerrainBase::Path Path;
typedef TerrainBase::Goal Goal;
typedef TerrainBase::Task Task;

#define TEST_SEC 0
//#define TEST(x)
#define TEST(x) x
#define TEST2(x)
//#define TEST2(x) x

/**************************************************************************/
// Constants

const sint4 AutoGatherAI::FROM = EventFactory::new_who();

const string AutoGatherAI::workerBPname("worker");
const string AutoGatherAI::mineralBPname("mineral");
const string AutoGatherAI::controlCenterBPname("controlCenter");

const real4 AutoGatherAI::MAX_WORKERS_PER_CLUSTER_FACTOR = 0.5;

/**************************************************************************/
#if 0
static real8 path_length(Path &path)
{
  if( path.locs.size() < 2 )
    return 0;
  real8 total = 0;
  for( Vector<Loc>::iterator i = ++path.locs.begin() ; i != path.locs.end() ; i++ )
    total += i->distance(*(i-1));
  return total;
}

// Assuming path is in forward order, with first waypoint at locs[0]
static real8 path_length_from(Path &path, const Loc& start_loc)
{
  if( path.locs.size() < 1 ) 
    return 0;
  return start_loc.distance(path.locs[0]) + path_length(path);
}
#endif

/**************************************************************************/
static Loc get_center_loc(const Object *obj)
{
  Loc l;
  obj->get_center(l.x, l.y);
  return l;
}

/**************************************************************************/
#ifndef USE_DEBUG_GFX
AutoGatherAI::AutoGatherAI(GameStateModule &gsm_, TerrainModule &tm_, Game1_SimpleTerrain::ST_Terrain &st_)
  : map_width(-1), map_height(-1),
    gsm(gsm_),
    tm(tm_),
    st(st_),
    handlerOfPathFinderEvents(this), 
    handlerOfViewMsgEvents(this)
{
  tm.add_handler(&handlerOfPathFinderEvents);     // terrain -> ai
  gsm.add_handler(&handlerOfViewMsgEvents);       // gsm -> ai
  add_handler(&tm);                               // ai -> terrain

  noClusterPlanning = true;
  bool useClusterPlanning;
  ::Options::get("-useclusters", useClusterPlanning);
  if(useClusterPlanning)
    noClusterPlanning = false;

  dummy_against_segfault.statically_blocked = 1;
  dummy_against_segfault.dont_use_until = 1000000;
  dummy_against_segfault.static_distance_from_cc = 1000000;
}
#else
AutoGatherAI::AutoGatherAI(GameStateModule &gsm_, TerrainModule &tm_, Game1_SimpleTerrain::ST_Terrain &st_, GfxModule *gfxm_)
  : gsm(gsm_),
    tm(tm_),
    st(st_),
    handlerOfPathFinderEvents(this), 
    handlerOfViewMsgEvents(this)
{
  tm.add_handler(&handlerOfPathFinderEvents);     // terrain -> ai
  gsm.add_handler(&handlerOfViewMsgEvents);       // gsm -> ai
  add_handler(&tm);                               // ai -> terrain

  noClusterPlanning = true;
  bool useClusterPlanning;
  ::Options::get("-useclusters", useClusterPlanning);
  if(useClusterPlanning)
    noClusterPlanning = false;

  if( gfxm_ )
    dbg_draw.reset(new DbgDraw(this, *gfxm_));
}
#endif
/**************************************************************************/

AutoGatherAI::~AutoGatherAI()
{
  gsm.remove_handler(&handlerOfViewMsgEvents);
  tm.remove_handler(&handlerOfPathFinderEvents);
}

/**************************************************************************/

void AutoGatherAI::add_options(void)
{
  ::Options o("AutoGatherAI");
  o.put("-useclusters", "Use the cluster-based planning scheme");
}

/**************************************************************************/

void AutoGatherAI::removeWMC(const Vector<ScriptObj*>& objects)
{
  FORALL_DYN(objects, it) {
    GameObj* gob = dynamic_cast<GameObj*>(*it);
    if( !gob )
      continue;

    // Remove a worker
    // ---------------
    if( gob->bp_name() == workerBPname ) {

      // find the worker wrapper
      Workers::iterator workerIt = workers.find(Worker(gob, gsm));
      assert(workerIt != workers.end());
      const Worker& worker = *workerIt;

      // is the worker associated with a mineral patch ?
      const GobPtrWrapper* mineral = getMineral(worker);
      if( mineral != NULL ) {
        // remove the worker from minerals2workers
        pair<M2W::iterator, M2W::iterator> workersOfTheMineral = minerals2workers.equal_range(mineral);
        for( M2W::iterator m2wIt = workersOfTheMineral.first ; m2wIt != workersOfTheMineral.second ; m2wIt++ ) {
          if( *(m2wIt->second) == worker ) {
            minerals2workers.erase(m2wIt);
            break;
          }
        }
      }

      // remove the worker itself
      workers2state.erase(&worker);
      workers.erase(workerIt);
    }

    // Remove a mineral
    // ----------------
    else if( gob->bp_name() == mineralBPname ) {

      // find the mineral wrapper
      GameObjects::iterator mineralIt = minerals.find(GobPtrWrapper(gob, gsm));
      assert(mineralIt != minerals.end());
      const GobPtrWrapper& mineral = *mineralIt;

      // find all workers associated with this mineral patch
      pair<M2W::iterator, M2W::iterator> workersOfTheMineral = minerals2workers.equal_range(&mineral);
      // remove these associations from the workers2state map
      for( M2W::iterator m2wIt = workersOfTheMineral.first ; m2wIt != workersOfTheMineral.second ; m2wIt++ ) {
        workers2state[m2wIt->second].mineral = NULL;
      }
      // remove these associations from the minerals2workers map
      minerals2workers.erase(workersOfTheMineral.first, workersOfTheMineral.second);

      // remove the mineral itself
      minerals2state.erase(&mineral);
      minerals.erase(mineralIt);
    }
    else if( gob->bp_name() == controlCenterBPname ) {
      if( controlCenters.size()>=1 && *controlCenters.begin() == gob )
        controlCenters.clear();
    }
  }
}

/**************************************************************************/

void AutoGatherAI::insertWMC(const Vector<ScriptObj*>& objects)
{
  const sint4 me = gsm.get_game().get_client_player();

  FORALL_DYN(objects, it) {
    GameObj* gob = dynamic_cast<GameObj*>(*it);
    if( !gob )
      continue;

    if( gob->bp_name() == workerBPname && *(gob->sod.owner) == me ) {
      pair<Workers::iterator,bool> result = workers.insert(Worker(gob, gsm));
      const Worker& worker = *result.first;
      workers2state.insert(W2S::value_type(&worker,WorkerState()));
    }
    else if( gob->bp_name() == mineralBPname ) {
      pair<GameObjects::iterator,bool> result = minerals.insert(GobPtrWrapper(gob, gsm));
      const GobPtrWrapper& mineral= *result.first;
      minerals2state.insert(M2S::value_type(&mineral, MineralState()));
    }
    else if( gob->bp_name() == controlCenterBPname && *(gob->sod.owner) == me ) {
      controlCenters.clear();
      controlCenters.insert(GobPtrWrapper(gob, gsm));
    }
  }
}

void AutoGatherAI::updateWorkers(const Vector<ScriptObj*>& objects)
{
  const sint4 me = gsm.get_game().get_client_player();

  FORALL_DYN(objects, it) {
    GameObj* gob = dynamic_cast<GameObj*>(*it);
    if( !gob )
      continue;

    if( gob->bp_name() == workerBPname && *(gob->sod.owner) == me ) {

      // find the worker wrapper and its current internal state
      Workers::iterator workerIt = workers.find(Worker(gob, gsm));
      assert(workerIt != workers.end());
      const Worker& worker = *workerIt;
      WorkerState &state = getState(worker);

      Loc cur_pos = get_center_loc(gob);
      sint4 now = gsm.get_game().get_view_frame();

      // if it moved, update last_pos
      if( cur_pos != state.last_pos ) {
        state.last_pos = cur_pos;
        state.in_this_pos_since = now;
      }
      // else, verify it is not stuck for too long
      else if( now-state.in_this_pos_since > MAX_TICKS_TO_FORCED_IDLE ) {
        state.job = WorkerState::IDLE;
        state.in_this_pos_since = now; // prevent resetting to IDLE every tick
      }
    }
  }
}

/**************************************************************************/

void AutoGatherAI::applyViewMessage(const GameChanges &changes)
{
  // Vanished workers, minerals and control center
  removeWMC(changes.vanished_objs);

  // Dead workers, minerals and control center
  removeWMC(changes.dead_objs);

  // New workers, minerals and control center
  insertWMC(changes.new_objs);

  // Track changes
  updateWorkers(changes.changed_objs);

#if TEST_SEC
  FORALL(changes.changed_objs, it) {
    GameObj* gob = dynamic_cast<GameObj*>(*it);
    // print 'changed speed' events for workers:
    if( gob && gob->bp_name() == workerBPname && gob->attr_changed("speed") )
      cout  << dynamic_cast<Object*>(gob) << " changed speed; new speed = " << *gob->sod.speed << " location=(" << *gob->sod.x << "," << *gob->sod.y << ")" << endl;
    // print location of moving workers:
    else if( gob && gob->bp_name() == workerBPname && *gob->sod.speed > 0 )
      cout  << dynamic_cast<Object*>(gob) << " location=(" << *gob->sod.x << "," << *gob->sod.y << ")" << endl;

    // print location of all workers (with some changed attribute):
    if( gob && gob->bp_name() == workerBPname )
      cout  << dynamic_cast<Object*>(gob) << " location=(" << *gob->sod.x << "," << *gob->sod.y << ")" << endl;
  }
#endif
}

/**************************************************************************/

void AutoGatherAI::associateWorkerAndMineral(Worker& worker, const GobPtrWrapper* mineral)
{
  // is this worker already associated with a different mineral ?
  const GobPtrWrapper* otherMineral = getMineral(worker);
  if( otherMineral != NULL ) {
    // remove this association
    pair<M2W::iterator, M2W::iterator> workersOfTheMineral = minerals2workers.equal_range(otherMineral);
    for( M2W::iterator m2wIt = workersOfTheMineral.first ; m2wIt != workersOfTheMineral.second ; m2wIt++ ) {
      if( *(m2wIt->second) == worker ) {
        minerals2workers.erase(m2wIt);
        break;
      }
    }
    getState(worker).mineral = NULL;
  }

  // associate with the new mineral
  if( mineral != NULL ) {
    minerals2workers.insert(M2W::value_type(mineral, &worker));
    getState(worker).mineral = mineral;

    TEST(cout << static_cast<const Object*>(worker) << " associated with mineral at (" << mineral->getPosition().x << "," << mineral->getPosition().y << ")" << endl);
  }
}

/**************************************************************************/

inline bool AutoGatherAI::isNearTheCC(const Worker& worker) const
{
  if( controlCenters.size() < 1 )
    return false;

  const GobPtrWrapper& controlCenter = *(controlCenters.begin());
  bool nearCC = (sint4)(Game::distance(*worker,*controlCenter)) <= WORKER_CC_DISTANCE_THRESHOLD;
  TEST(cout << static_cast<const Object*>(worker) << " NEAR_CC? " << nearCC << " (dist=" << ((sint4)(Game::distance(*worker,*controlCenter))) << ")" << endl);
  return nearCC;
}

/**************************************************************************/

inline bool AutoGatherAI::isNearTheMineral(const Worker& worker) const
{
  // find the allotted mineral patch
  const GobPtrWrapper* mineral = getMineral(worker);
  if( mineral == NULL )
    return false;

  bool nearMineral = (sint4)(Game::distance(*worker,**mineral)) <= WORKER_MINERAL_DISTANCE_THRESHOLD;
  TEST(cout << static_cast<const Object*>(worker) << " NEAR_MINERAL? " << nearMineral << " for mineral at ("; );
  TEST(cout << mineral->getPosition().x << "," << mineral->getPosition().y << ") (dist=" << (sint4)(Game::distance(*worker,**mineral)) << ")" << endl; );
  return nearMineral;
}

/**************************************************************************/
void AutoGatherAI::moveTo(Worker& worker, const Loc &location, const Path *cached_path)
{
  TEST( cout << static_cast<const Object*>(worker) << " moving to location (" << location.x << "," << location.y << ")" << endl; );
  if( !cached_path ) {
    call_handlers( TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG,
                  TerrainBase::Obj2LocTask(worker, location)) );
  } else {
    st.execute_task(TerrainBase::Obj2LocTask(worker, location), *cached_path);
  }
  // update the worker's state
  getState(worker).job = WorkerState::MOVING;
}

/**************************************************************************/
void AutoGatherAI::moveTo(Worker& worker, const GobPtrWrapper& target)
{
  ScalarPoint targetPosition = target->getPosition();

/* The to-location version:
  TerrainBase::Task t(TerrainBase::Task::LOCATION, TerrainBase::Task::TOUCH);
  t.goal_loc = TerrainBase::Loc(targetPosition.x, targetPosition.y); */

  call_handlers( TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG,
                 TerrainBase::Obj2ObjTask(worker, target)) );

  // update the worker's state
  getState(worker).job = WorkerState::MOVING;
}

/**************************************************************************/

void AutoGatherAI::moveToCC(Worker& worker)
{
  if( controlCenters.size() >= 1 ) {
    const GobPtrWrapper& controlCenter = *(controlCenters.begin());
    moveTo(worker, controlCenter);
    TEST(cout << static_cast<const Object*>(worker) << " MOVE_CC" << endl);
  }
}

/**************************************************************************/

void AutoGatherAI::moveToMineral(Worker& worker)
{
  // find the allotted mineral patch
  const GobPtrWrapper* mineral = getMineral(worker);
  if( mineral == NULL )
    return;

  moveTo(worker,*mineral);
  TEST(cout << static_cast<const Object*>(worker) << " MOVE_TO_MINEARL at (" << mineral->getPosition().x << "," << mineral->getPosition().y << ")" << endl);
}

/**************************************************************************/

void AutoGatherAI::startMining(Worker& worker)
{
  // find the allotted mineral patch
  const GobPtrWrapper* mineral = getMineral(worker);
  if( mineral == NULL )
    return;

  worker->startMining(*mineral);
  // update the worker's state
  getState(worker).job = WorkerState::MINING;
}

/**************************************************************************/

void AutoGatherAI::returnMinerals(Worker& worker)
{
  const GobPtrWrapper& controlCenter = *controlCenters.begin();
  worker->returnMinerals(controlCenter);
  // update the worker's state - returning resources takes a single game tick,
  // and since there is no event for "finished returning the resources", 
  // we already mark the worker as idle.
  getState(worker).job = WorkerState::RETURNING_MINERALS;
}

/**************************************************************************/

bool AutoGatherAI::isIdle(const Worker& worker) const
{
  // The end of mining/returningMinerals is not signaled automatically 
  // by an external event, so check these jobs explicitly.

  // Mining might take > 1 tick and the GameObj has an internal status to check
  if( getState(worker).job == WorkerState::MINING )
    return !worker->isMining() && !worker->is_pending_action();

  // Returning minerals to the CC takes 1 tick, no internal status. The moment this
  // action is executed, it is finished.
  else if( getState(worker).job == WorkerState::RETURNING_MINERALS )
    return !worker->is_pending_action();

  // MOVING/IDLE differentiation is handled correctly by events.
  else
    return getState(worker).job == WorkerState::IDLE;
}

/**************************************************************************/

void AutoGatherAI::act()
{
  int i=0, max_workers=20;

  // walk the workers list, find idle workers
  FORALL(workers, it) {

    Worker& worker          = const_cast<Worker&>(*it);
    WorkerState &state      = getState(worker);

    TEST2(GameObj* gob = (GameObj*)worker;
         cout  << dynamic_cast<Object*>(gob) << " is at (" << *gob->sod.x << "," << *gob->sod.y << ")" << endl;); 

    if( isIdle(worker) ) {

      // If using cluster planning for this worker
      if( state.clusterId != -1 ) {

// ----------------------------------------------------------------------------------------
// POLICY: use cached paths for CC->cluster and cluster->CC, 
//         and freely plan the other segments
// ----------------------------------------------------------------------------------------

        sint4 &waypoint         = state.clusterPathWaypoint;
        sint4 waypoint_count    = (sint4)mineralClusters[state.clusterId].cachedPaths.size();

        // Does it have minerals ?
        if( worker->getMinerals() > 0 ) {
          // If it's near the cont rol center, return the minerals
          if( isNearTheCC(worker) ) {
            returnMinerals(worker);
          }

          // Else, move to the CC
          else {
            // Do this by moving to the last 2 waypoints, which are the 
            // 'launching' point near the mineral cluster and the CC itself.
            // In case of plan/move failures (collisions, occupied waypoint etc.),
            // fall back to free planning staright to the CC.

            // Do we need to freely plan to the CC
            const GobPtrWrapper& controlCenter = *(controlCenters.begin());
            const Loc &launch_point = mineralClusters[state.clusterId].cachedPaths[waypoint_count-2].locs.back();
            if( state.failedToArrive || 
              worker.distance_to(controlCenter) < worker.distance_to(launch_point.x, launch_point.y))
            {
              waypoint = waypoint_count-1;
              moveToCC(worker);
            }
            else {
              // move to the 'launching' point
              if( waypoint <= waypoint_count-2 ) {
                waypoint = waypoint_count-2;
                moveTo(worker, launch_point); // not using the cached path
              } 
              // or move to the CC itself
              else {
                const Path &cached_path = mineralClusters[state.clusterId].cachedPaths[waypoint];
                const Loc &waypoint_loc = cached_path.locs.back();
                moveTo(worker, waypoint_loc, &cached_path);
              }
            }
          }
        }
        // No minerals:
        else {
          // verify that there is an allocated mineral
          allocateMineralPatchFor(worker);
          if( state.mineral == NULL )
            continue;
            
          // If the worker is near the mineral, start mining
          if( isNearTheMineral(worker) ) {
            startMining(worker);
          }

          // Else, move to the mineral
          else {              
            // Go to the first waypoint (close to the cluster) using a cached path,
            // and plan freely from there to the specific mineral
            // In case of plan/move failures (collisions, occupied waypoint etc.),
            // fall back to free planning staright to the mineral.

            // Need to go to the mineral
            if( state.failedToArrive || waypoint > 0 ) {
              waypoint = 1;
              moveToMineral(worker);
            }

            // Else, use the first cached path to go from the CC to the cluster:
            else {
              const Path &cached_path = mineralClusters[state.clusterId].cachedPaths[waypoint];
              const Loc &waypoint_loc = cached_path.locs.back();
              // When leaving the CC towards the first waypoint, send the workers 
              // of each cluster with a time spacing between them to minimize 
              // collisions.
              sint4 &ldt = mineralClusters[state.clusterId].last_departure_from_CC;
              sint4 cur_frame = gsm.get_game().get_view_frame();
              if( ldt == -1 || cur_frame-ldt >= SAME_CLUSTER_DELAY_TICKS ) {
                ldt = cur_frame;
                // In the initial move plan an independent path for each worker
                if( state.initial_move ) {
                  moveTo(worker, waypoint_loc);
                }
                // else, use the cached path
                else {
                  moveTo(worker, waypoint_loc, &cached_path);
                }
              }
            }
          }
        }

      



#if 0 
// ----------------------------------------------------------------------------------------
// OLD TESTING: just walk through the cached paths, no mining at all
// ----------------------------------------------------------------------------------------
        sint4 waypoint = state.clusterPathWaypoint;
        const Path &cached_path = mineralClusters[state.clusterId].cachedPaths[waypoint];
        Loc goal_loc = cached_path.locs.back();

        // If moving according to a cached path failed, try getting there using a
        // freshly planned path
        if( state.failedToArrive ) {
          moveTo(worker, goal_loc);
        }
        else {
          // If leaving the CC towards the first waypoint, send the workers of each 
          // cluster with a time spacing between them to minimize collisions.
          sint4 &ldt = mineralClusters[state.clusterId].last_departure_from_CC;
          sint4 cur_frame = gsm.get_game().get_view_frame();
          if( waypoint != 0 || ldt == -1 || cur_frame-ldt >= SAME_CLUSTER_DELAY_TICKS ) {
            // In the initial move plan an independent path for each worker
            if( state.initial_move ) {
              moveTo(worker, goal_loc);
            }
            // else, use the next leg of the cached paths          
            else {
              moveTo(worker, goal_loc, &cached_path);
            }

            if( waypoint == 0 ) {
              ldt = cur_frame;
            }
          }
        }
#endif
      }

      // else, using single-mineral planning
      else {

        // Does it have minerals ?
        if( worker->getMinerals() > 0 ) {
          // If it's near the control center, return the minerals
          if( isNearTheCC(worker) ) {
            returnMinerals(worker);

            // Else, move to the CC
          } else {
            moveToCC(worker);
          }
        }

        // No minerals:
        else {

          // verify that there's an allotted mineral patch for this worker
          allocateMineralPatchFor(worker);
          if( getState(worker).mineral == NULL )
            continue;

          // If it's near the minerals, start mining
          if( isNearTheMineral(worker) ) {
            startMining(worker);
          }
          // Else, move to the minerals
          else {
            moveToMineral(worker);
          }
        }
      }
    }
    if( ++i >= max_workers )
      break; // handle only the first X workers*/
  }
}

/**************************************************************************/

bool AutoGatherAI::HandlerOfPathFinderEvents::handle_event(const Event &event)
{
  assert( event.get_who() == TerrainModule::FROM );

  if( event.get_what() == TerrainModule::PATH_SUCCESS ||
      event.get_what() == TerrainModule::PATH_FAILURE ) {

      // is this a notification regarding one of our workers?
      const TerrainModule::PathStatusEvent* pse = 
        dynamic_cast<const TerrainModule::PathStatusEvent*>(&event);
      if( pse ) {
        GameObj* gob = pse->get_obj();
        // find the worker wrapper, if it's indeed a worker
        Workers::iterator workerIt = parent->workers.find(Worker(gob, parent->gsm));
        if( workerIt != parent->workers.end() ) {

          const Worker& worker = *workerIt;
          WorkerState &state = parent->getState(worker);

	  TEST(
	       cout << static_cast<const Object*>(worker) << "got PF notification: ";
	       if( event.get_what() == TerrainModule::PATH_SUCCESS )
		 cout << "arrived" << endl;
	       else if ( pse->get_details() == TerrainModule::PATH_PLAN_FAILURE )
		 cout << "planning failed" << endl;
	       else if ( pse->get_details() == TerrainModule::PATH_MOVE_FAILURE )
		 cout << "movement failed" << endl;
	       else
		 cout << "unknown pf notification!" << endl;
	       );

          // This worker is now idle
          state.job = AutoGatherAI::WorkerState::IDLE;

          // if this worker uses cluster planning, update waypoints etc.
          if( state.clusterId >= 0 ) {

            // if the worker arrived
            if( pse->get_what() == TerrainModule::PATH_SUCCESS ) {
              // next move is no longer an initial or a failed move
              state.initial_move = false;
              state.failedToArrive = false;
              // increment the next waypoint to go to
              sint4 waypoint_count = (sint4)parent->mineralClusters[state.clusterId].cachedPaths.size();
              TEST(cout << "wp_count=" << waypoint_count << " waypoint was: " << state.clusterPathWaypoint; );
              state.clusterPathWaypoint = (state.clusterPathWaypoint+1) % waypoint_count;
              TEST(cout << " and is now: " << state.clusterPathWaypoint << endl; );
            }
          
            // if it failed to reach the target (probably due to collisions), remember it
            if( pse->get_what() == TerrainModule::PATH_FAILURE ) {
              state.failedToArrive = true;
            }
          }
          // else, this worker is using a single-mineral planning
          else {
            // If a path cannot be planned to this mineral, switch to a different mineral
            if( pse->get_what() == TerrainModule::PATH_FAILURE &&
                pse->get_details() == TerrainModule::PATH_PLAN_FAILURE &&
                worker.getMinerals() == 0 )
            {
              TEST(cout << "Reallocating worker because of a failed planning" << endl; );

              // Mark that there was a difficulty reaching this mineral
              sint4 now = parent->gsm.get_game().get_view_frame();
              if( state.mineral != NULL ) {
                //parent->getMineralState(*state.mineral).failed_moves++;
                parent->getMineralState(*state.mineral).dont_use_until = now + FAILED_MINERAL_SLEEP_TICKS;
              }

              // switch this worker to a different mineral, since this one is blocked
              parent->allocateMineralPatchFor(const_cast<Worker&>(worker), true);
            }
          }
        }
      }
      return true;
    }
    return false;
}

/**************************************************************************/

bool AutoGatherAI::HandlerOfViewMsgEvents::handle_event(const Event& e)
{
  assert(e.get_who() == GameStateModule::FROM);

  if( e.get_what() == GameStateModule::VIEW_MSG ) {
    parent->map_width   = parent->gsm.get_game().get_map().get_width() * 
                          parent->gsm.get_game().get_tile_points();
    parent->map_height  = parent->gsm.get_game().get_map().get_height() * 
                          parent->gsm.get_game().get_tile_points();
    parent->applyViewMessage(parent->gsm.get_changes());

    Game &game = parent->gsm.get_game();
    sint4 vf = game.get_view_frame();
    sint4 ai = game.get_action_frame();
    bool behind = abs(vf-ai) > 10 || ai < 0;

    if( !behind ) {
      parent->plan();
      parent->act();
    }
    return true;
  }
  return false;
}

/**************************************************************************/
/**************************************************************************/

#ifdef USE_DEBUG_GFX

AutoGatherAI::DbgDraw::DbgDraw(const AutoGatherAI *parent_, GfxModule &gfxm_)
  : DrawOnTerrain(gfxm_), parent(parent_), enabled(true)
{
  gfxm->add_handler(this);
}

AutoGatherAI::DbgDraw::~DbgDraw()
{
  gfxm->remove_handler(this);
}

bool AutoGatherAI::DbgDraw::handle_event(const Event &e)
{
  if (e.get_who() == GfxModule::FROM && e.get_what() == GfxModule::DRAW_MSG) {
    draw();
    return true;
  }
  return false;
}

void AutoGatherAI::DbgDraw::draw()
{
  Vec3<real4> white(1,1,1);
  Vec3<real4> yellow(215.0/255,227.0/255,0);

  if( !is_enabled() ) return;

  sint4 gw = gfxm->get_gsm().get_game().get_map().get_width();
  sint4 gh = gfxm->get_gsm().get_game().get_map().get_height();
  
  real4 dx = (real4)gw / (real4)(parent->map_width);
  real4 dy = (real4)gh / (real4)(parent->map_height);

  begin_lines();
  draw_line(0,0,10,10,yellow,5);
  FORALL(parent->mineralClusters, c) {
    FORALL(c->cachedPaths, p) {
      typeof(p->locs.begin()) wp = p->locs.begin();
      draw_circle(wp->x * dx, wp->y * dy, 3, yellow);
      for( typeof(wp) prev = wp++ ; wp != p->locs.end() ; wp++, prev++ ) {
        draw_line(prev->x * dx, prev->y * dy, wp->x * dx, wp->y * dy, white, 5.0);
        //draw_circle(wp->x * dx, wp->y * dy, 0.3, yellow);
      }
    }
  }
  end_lines();
  glEnable(GL_TEXTURE_2D);
}
#endif
