#ifndef Game1_Accessories_H
#define Game1_Accessories_H

// $ Id $

#include "Global.H"
#include "Coord.H"

/** Perform a basic clustering of points (designed with mineral locations in mind).
    Implemented using K-means, with K=the number of clusters chosen automatically
    using a maximum-average-silhouette criterion from a range of values supplied
    by the caller.

    @param data     Input locations
    @param min_k    Minimal value of K to test
    @param max_k    Maximal value of K to test
    @param result   Cluster IDs are stored in this vector, one per input point
*/
extern void simple_clustering(const Vector<ScalarPoint> &data, 
                              uint4 min_k, uint4 max_k, 
                              Vector<sint4> &result);


/** Given a set S of points (designed for minerals' locations) and
    a reference location P (designed for a control center's location),
    compute the subsequence of their convex hull between the leftmost and
    rightmost points in S (relative to P) on the far side of S (again 
    relative to P).
    This sequence can be used to create a path from P that goes around S
    and returns back to P, hence the name 'roundabout'.

    @param point      The set S of points.
    @param reference  The reference point P.
    @param result     The computed subsequence of the convex hull is stored in this vector.
*/
extern void compute_roundabout(const Vector<ScalarPoint> &points, 
                               const ScalarPoint &reference,
                               Vector<ScalarPoint> &result);

/** Enlarge (blowup) a given polygon by moving each vertex outward by a given distance.

    @param polygon    Input polygon.
    @param distance   The distance between each original vertex and its moved counterpart.
    @param result     Output is stored in this vector.
*/
extern void blowup_polygon(const Vector<ScalarPoint> &polygon, 
                           real8 distance, 
                           Vector<ScalarPoint> &result);


#endif
