#ifndef ShooterAI_H
#define ShooterAI_H

// $Id: Game2_AI.H,v 1.1 2006/05/09 15:44:30 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, Tim Furtak, licensed under the GPL

#include "Global.H"
#include "EventHandler.H"
#include "ClientModule.H"
#include "PlayerInfo.H"

class GameStateModule;
class GameChanges;
class GameObj;

//===================================================================

class ShooterAI : public EventHandler, public ClientModule
{
protected:
  //------------------------------

  class ShooterData {
  public:
    Vector<GameObj*> in_range;
  };

  //------------------------------

  GameStateModule &gsm;

  typedef std::map<GameObj*, ShooterData> ShooterDataMap;

  ShooterDataMap friendly_data; // which enemies friendlies can hit
  ShooterDataMap enemy_data;    // which friendlies can hit a target

  std::set<GameObj*> friendlies;
  std::set<GameObj*> enemies;

  void add_friendly(GameObj*);
  void recompute_targets();
  GameObj* select_target(const Vector<GameObj*> &targets);

  bool enabled;

public:

  ShooterAI(GameStateModule &gsm);
  ~ShooterAI();

  void act(GameStateModule &gsm);
  static const std::string bp_name;

  bool handle_event(const Event &e);

  void update(const GameChanges &changes);

  bool is_enabled() const { return enabled; }
  void enable()  { enabled = true; }
  void disable() { enabled = false; }
};

//===================================================================

REGISTER_TYPEOF(400, ShooterDataMap::iterator);

#endif
