// $Id: tortsg_main.C,v 1.4 2006/06/01 20:24:17 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "GfxGlobal.H"
#include "Options.H"
#include "GfxModule.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "SoundMixer.H"
#include "LoopFunctor.H"
#include "Game.H"
#include "GameChanges.H"

#include "TurretAI.H"
#include "GatherAI.H"

#include "ClientState.H"
#include "TerrainModule.H"
#include "SimpleTerrain.H"

using namespace std;


const char *ORTSG_VERSION = "GfxClient version 0.91 (2005-Apr-10)";

#define USE_EFENCE 0

extern int EF_PROTECT_FREE, EF_FILL, EF_PROTECT_BELOW, EF_ALLOW_MALLOC_0;


void add_options()
{
  Options o("ortsg");
  o.put("-soundtest", "test sound module");
}

static bool just_drew = false;

struct GfxEventHandler : public EventHandler
{
  ClientState &state;
  TurretAI turret_ai;
  
  GfxEventHandler(ClientState &state_) : state(state_) { }

  bool handle_event(const Event& e)
  {
    if (e.get_who() == GameStateModule::FROM) {

      //cout << "Event: " << e.who() << " " << e.what() << endl;
      
      if (e.get_what() == GameStateModule::STOP_MSG) exit(0);

      if (e.get_what() == GameStateModule::READ_ERROR_MSG) exit(10);      
      
      if (e.get_what() == GameStateModule::VIEW_MSG) {

#if 0
        if (state.gui) {
          state.gui->event();
          state.gui->display();
          if (state.gui->quit) exit(0);
        }
#endif

        // update LowAI objects

        const sint4 me = state.gsm->get_game().get_client_player();
        
        // add new objects

        FORALL (state.gsm->get_changes().new_objs, i) {
          GameObj *gob = (*i)->get_GameObj();
          if (gob == 0 || *gob->sod.owner != me) continue;
          if (gob->bp_name() == TurretAI::bp_name) turret_ai.add(gob);
        }

        // remove dead objects
        
        FORALL (state.gsm->get_changes().dead_objs, i) {
          GameObj *gob = (*i)->get_GameObj();
          if (gob == 0 || *gob->sod.owner != me) continue;
          if (gob->bp_name() == TurretAI::bp_name) turret_ai.del(gob);
        }
        
        // remove vanished gfx objects

        FORALL (state.gsm->get_changes().vanished_objs, i) {
          GameObj *gob = (*i)->get_GameObj();
          if (gob == 0 || *gob->sod.owner != me) continue;
          if (gob->bp_name() == TurretAI::bp_name) turret_ai.del(gob);
        }

        // compute actions
        
        // turret_ai.act(*state.gsm);
        
        
        state.gfxm->process_changes();

        ////////////////
        Vector<string> temp;
        Game &game = state.gsm->get_game();
        FORS (i, game.get_player_num()) temp.push_back(""); // dummy list of commands
        game.exec_actions(temp); // evaluate client actions and send actions to the server
        ////////////////

        state.gsm->send_actions();

        sint4 vf = game.get_view_frame();
        sint4 ai = game.get_action_frame();
        sint4 sk = game.get_skipped_actions();
        sint4 md = game.get_merged_actions();
        
        if (vf != ai || sk || md > 1) {
          
          cout << "frame " << vf;
          if (ai < 0) cout << " [no action]";
          else if (vf != ai) cout << " [behind by " << (vf - ai) << " frame(s)]";
          if (sk) cout << " [skipped " << sk << "]";
          if (md > 1) cout << " [merged " << md << "]";
          cout << endl;
        }
        
        // don't draw if we are far behind
        if (abs(vf-ai) < 10) {
          state.gfxm->draw();
        }
        just_drew = true;

        return true;
      }
      
    }

    return false;
  }
};


class ortsgLooper : public LoopFunctor
{
private:
  ClientState &state;
  bool refresh;
  
public:

  ortsgLooper(ClientState &s) : state(s)
  {
    ::Options::get("-refresh", refresh);
  }
  
  void loop() const {

    if (state.gfxm->get_quit()) exit(0);
    
    if (refresh) {

      if (!just_drew) // allow some waiting time after a message
        state.gfxm->draw();
      else
        just_drew = false;
    }

    // server events
    if (!state.gsm->recv_view()) SDL_Delay(1);
  }
};

typedef SimpleTerrain::ST_Terrain Terrain;

int main(int argc, char **argv)
{
 char mydir[81];
 getcwd(mydir, 80);
 glutInit(&argc, argv);
 chdir(mydir);

 try {
#if USE_EFENCE
  // aggressive efence checks
  EF_PROTECT_FREE = 1;
  EF_FILL = 0xee;
  EF_PROTECT_BELOW = 1;
  EF_ALLOW_MALLOC_0 = 1;
#endif

  signals_shut_down(true);
  atexit(SDL_Quit);

  // populate opt with options from various modules
  
  add_options();
  GameStateModule::Options::add();
  GfxModule::Options::add();
  Terrain::add_options();
  
  // process command line arguments
  if (Options::process(argc, argv, cerr, ORTSG_VERSION)) exit(20);

  bool st = false;
  Options::get("-soundtest", st);

  if (st) {
    string path;
    Options::get("-p", path);
    sound_test(path);
    exit(0);
  }

  // change seed value to time(0) if 0
  sint4 seed;
  Options::get("-seed", seed);
  if (!seed) seed = time(0);
  Options::set("-seed", seed);
  cout << "seed=" << seed << endl;

  srand48(seed); // used by STL, fixme: avoid (global)

  SDLinit::video_init();
  SDLinit::network_init();

  // populate option objects with command line values
  // and create modules
  
  GameStateModule::Options gsmo;
  GameStateModule gsm(gsmo);
  
  GfxModule::Options gfxmo;
  GfxModule gfxm;

  // gives event handler access to the state
  ClientState state(gsm, gfxm);

  // ortsg event handler
  GfxEventHandler gev(state);

  // create pathfinding module and implementation
  //boost::shared_ptr<TerrainBase> timp(new Terrain);
  Terrain timp;
  TerrainModule tm(gsm, timp);

  // hook up terrain widget to gfx module
  // gfxm.add_cwidget(*timp->create_widget());
  
  GatherAI gather_ai(gsm);
  gather_ai.add_handler(&tm);
  gfxm.add_handler(&gather_ai);
  gsm.add_handler(&gather_ai);
  tm.add_handler(&gather_ai);

  // register pathfinder as gfx and gsm event handler
  gfxm.add_handler(&tm);
  //gsm.add_handler(&tm); done in TerrainModule
  
  // gsm events also handled in gev
  gsm.add_handler(&gev);

  // connect game state module to server
  if (!gsm.connect()) ERR("connection problems");
  
  gfxm.init(gsm, gfxmo);

  // gfx loop (should become a thread)
  ortsgLooper l(state);

  cout << "start loop" << endl;
  
  gfxm.start_loop(&l);
    
 } 
 catch (ExitException& ee)
 {
 	errstr << std::endl << *(ee.get_reason()) << std::endl;
#ifdef GCC
    abort();
#else
 	ABORT;
#endif
 	exit(20);
 }
 return 0;
}
