// $Id: SegmentMerge.C,v 1.1 2006/05/12 04:18:01 demyen Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "SegmentMerge.H"

using namespace std;

void SegmentMerge::clear_delta() {
  FORS (i, 4) {
    del[i].erase(del[i].begin(), del[i].end());
    add[i].erase(add[i].begin(), add[i].end());
  }
}


void SegmentMerge::new_segment(const DCDT::Segment &seg)
{
  SortSegment s(seg);
  s.s.normalize();
  s.label = -1; // new
    
  Kind k = H;
    
  if (s.s.y1 == s.s.y2) {  // horizontal
    s.q  = s.s.y1;
    s.p1 = s.s.x1;
    s.p2 = s.s.x2;
    k = H;
  } else if (s.s.x1 == s.s.x2) { // vertical
    s.q  = s.s.x1;
    s.p1 = s.s.y1;
    s.p2 = s.s.y2;
    k = V;
  } else if (s.s.x1 - s.s.y1 == s.s.x2 - s.s.y2) {  //  / (diagonal minus)
    s.q  = s.s.x1 - s.s.y1;
    s.p1 = s.s.x1 + s.s.y1;
    s.p2 = s.s.x2 + s.s.y2;
    k = DM;
  } else if (s.s.x1 + s.s.y1 == s.s.x2 + s.s.y2) {  //  \ (diagonal plus)
    s.q  = s.s.x1 + s.s.y1;
    s.p1 = s.s.x1 - s.s.y1;
    s.p2 = s.s.x2 - s.s.y2;
    k = DP;
  } else ERR("illegal slope");

  SSet &alls = all[k];
  ISet &adds = add[k];
  typeof(del[0]) &dels = del[k];        

  SSet::iterator it = alls.lower_bound(s);

  bool rm = it != alls.end() && it->q == s.q && it->p1 == s.p2;
  bool lm = false;

  if (it != alls.begin()) {
    SSet::iterator it2 = it;
    --it2;
    lm = it2->q == s.q && it2->p2 == s.p1;
  }
    
  if (lm && rm) {

    // fill hole
    //                   it->p2
    // o-----o----o------o
    //   ^      s   ^
    //   it2        it
    //   |          |
    //  adjust p2  erase
    //  & set new
      
    SSet::iterator it2 = it;
    --it2;

    it2->p2 = it->p2; // adjust element
    it2->s.x2 = it->s.x2;
    it2->s.y2 = it->s.y2;    
    
    // predecessor ...
      
    if (it2->label >= 0) {

      // existing segment => schedule for delete
      dels.push_back(it2->label); 

      it2->label = -1;  // set new
      adds.insert(&*it2); // new segment to be added

    } // else: already in adds

    // successor ...

    if (it->label >= 0) {

      // existing segment => schedule for delete

      dels.push_back(it->label); 

    } else {

      // new segment => remove from add set

      if (adds.erase(&*it) != 1) ERR("not found");	
    }

    // delete

    alls.erase(it);

  } else if (lm) {

    // merge with predecessor
    //
    // o-----o----o
    //   ^      s  
    //   it2       
    //   |         
    //  adjust p2
    //  & set new

    SSet::iterator it2 = it;
    --it2;

    // cout << &*it2 << " " << it2->p2 << endl;
      
    it2->p2 = s.p2; // adjust element
    it2->s.x2 = s.s.x2;
    it2->s.y2 = s.s.y2;
    
    // cout << it2->p2 << endl;      
      
    if (it2->label >= 0) {

      // existing segment => schedule for delete
      dels.push_back(it2->label); 

      it2->label = -1;  // set new
      adds.insert(&*it2); // new segment to be added
	  
    } else {
      
      // else: already in

      assert(adds.find(&*it2) != adds.end());
    }


  } else if (rm) {

    // merge with successor
    //     s.p2 <- it->p2
    // o----o-----o
    //   s    ^      
    //        it       
    //
    // adjust s.p2, insert s before it
    // erase *it

    s.p2 = it->p2;
    s.s.x2 = it->s.x2;
    s.s.y2 = it->s.y2;
    
    if (it->label >= 0) {

      // existing segment => schedule for delete
      dels.push_back(it->label); 
	  
    } else {

      if (adds.erase(&*it) != 1) ERR("not found");

    }

    alls.erase(it);
    adds.insert(&*alls.insert(s).first);

  } else {

    // no merge

    adds.insert(&*alls.insert(s).first);
  }
}

