#ifndef ClusterSquad_H
#define ClusterSquad_H

/** @file ClusterSquad.H
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"

class GameObj;

//===================================================================

/** A multi-dimensional cluster finder. */

class ClusterSquad
{
protected:
  //------------------------------

  class ObjData {
  public:
    sint4 cluster_id;
    real4 inertia; // acts as a hysteresis threshold
  };

  //------------------------------

  class Point {
  public:
    sint4 x, y;
    Point(sint4 x_, sint4 y_) : x(x_), y(y_) {}
    Point() : x(0), y(0) {}
  };

  //------------------------------

  std::map<const GameObj*, ObjData> obj2dat;

  static const uint4 MAX_K;
  Vector<Point> kmeans;

  Vector< Vector<Point> > prev_means;

  Vector<sint4> decay;
  Vector<Point> kdata;
  real4 tension;

  sint4 max_x, max_y; // size of the board

  // Average and standard deviation for using k-means to classify uniformly distributed data.
  struct UniformData {
    real4 err;
    real4 std;
    UniformData() { err = std = 0.0; }
  };
  Vector<UniformData> baseline; // Entry i corresponds to using i means.

  void compute_baseline();

  //------------------------------

  real4 distance(const Point &a, const Point &b) const;
  real4 distance(const GameObj *a, const GameObj *b) const;
  void compute_mean(const Vector<Point> &p, Point &mean) const;
  void update_data();
  void update_means();

  sint4 classify(const Point &p) const;
  sint4 classify(const GameObj *g) const;
  sint4 classify(const GameObj *g, const ObjData &d);

  real4 compute_means(const Vector<Point> &data, sint4 k, Vector<Point> &means) const;

  void assign_hard(const Vector<Point> &p, const Vector<Point> &means,
                   Vector< Vector<sint4> > &bin, real4 &err) const;

  //  void update_cluster();

  sint4 closest(const Point &p, const Vector<Point> &means, real4 &err) const;
  void match_assignment(const Vector<Point> &orig, Vector<Point> &pts) const;


public:
  ClusterSquad(sint4 w, sint4 h);
  ~ClusterSquad();

  void add(const GameObj *obj);
  void del(const GameObj *obj);

  void update();

  friend class ClusterSquadsWidget;
};

//===================================================================

#endif
