/** @file ClusterSquads.C
    @see ClusterSquads.H

    $Id: ClusterSquads.C,v 1.3 2006/05/23 18:28:24 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvssoftware/orts3loc/libs/ai/squad/src/ClusterSquads.C,v $
*/

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "Global.H"
#include "ClusterSquad.H"
#include "ClusterSquads.H"
#include "Game.H"
#include "GameStateModule.H"
#include "GameChanges.H"

using namespace std;

//===================================================================

ClusterSquads::ClusterSquads(GameStateModule &gsm_) : gsm(gsm_)
{
  initialized = false;
  enabled = false;
}

//===================================================================

ClusterSquads::~ClusterSquads()
{
}

//===================================================================

void ClusterSquads::init()
{
  num_player = 1; //gsm.get_game().get_player_num()+1;
  sint4 w = gsm.get_game().get_map().get_width();
  sint4 h = gsm.get_game().get_map().get_height();

  FORU (i, num_player) {
    squads.push_back(new ClusterSquad(w, h));
  }

  initialized = true;
}

//===================================================================

bool ClusterSquads::handle_event(const Event &e)
{
  if (e.get_who() == GameStateModule::FROM) {

    if (e.get_what() == GameStateModule::VIEW_MSG) {

      if (!initialized) init();

      // add new objects
      FORALL (gsm.get_changes().new_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (gob->get_shape() != Object::CIRCLE) continue;
        sint4 o = *gob->sod.owner;
        if (o >= num_player || o < 0) continue;

        squads[o]->add(gob);
      }
      
      // remove dead objects
      FORALL (gsm.get_changes().dead_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (gob->get_shape() != Object::CIRCLE) continue;
        sint4 o = *gob->sod.owner;
        if (o >= num_player || o < 0) continue;

        squads[o]->del(gob);
      }

      // remove vanished objects
      FORALL (gsm.get_changes().vanished_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (gob->get_shape() != Object::CIRCLE) continue;
        sint4 o = *gob->sod.owner;
        if (o >= num_player || o < 0) continue;

        squads[o]->del(gob);
      }
      
      update();

      return true;
    }
  }

  return false;
}

//===================================================================

void ClusterSquads::update()
{
  if (!enabled) return;

  FORALL (squads, it) {
    (*it)->update();
  }
}

//===================================================================
