#ifndef SUICIDEATTACKAI_H
#define SUICIDEATTACKAI_H

// $Id: SuicideAttackAI.H,v 1.4 2006/05/24 15:05:32 sailer Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL
// Adapted from GatherAI class

#include "Global.H"
#include "LowAI.H"
#include "EventHandler.H"
#include "ClientModule.H"

class GameStateModule;

//===================================================================

/**
 * Class to hold the data needed for each object controlled by SuicideAttackAI
 */
class SuicideData {
public:
  GameObj *base;
  enum State { MOVE_TO, ATTACK};
  State state;
};

//-------------------------------------------------------------------

/**
 * This AI module causes all units to attack the nearest (geometric distance)
 * base.
 */
class SuicideAttackAI : public LowAI<SuicideData>, public EventHandler, public ClientModule
{
protected:
  std::set<GameObj*> bases; ///< Enemy Bases

  GameStateModule &gsm;

  Vector<GameObj*> need_act; ///< GameObjects needing an action in the next tick
  /**
   * Removes the given GameObj from the control of this class.
   */
  void stop(GameObj *obj);


  /**
   * Finds the nearest base to travel to, based on geometric distance.
   */
  GameObj* find_target(GameObj *targeter, GameObj *baseToAvoid = 0);
public:

  static const sint4 FROM;

  SuicideAttackAI(GameStateModule &gsm);
  ~SuicideAttackAI();

  /**
   * Causes the units in the need_act vector to act according to
   * their current state.
   */
  void act(GameStateModule &gsm);

  //static const std::string bp_name; ///< ?

  void add(GameObj*) { ERR("?"); }

  /**
   * Addes an enemy base to the list of enemy bases, and updates targets
   */
  void add_base(GameObj *gob);

  void del(GameObj*) { ERR("?"); }

  /**
   * Deletes a base from the list of enemy bases, and updates targets
   */
  void del_base(GameObj *gob);

  /**
   * Handles events sent to SuicideAttackAI from other modules
   */
  bool handle_event(const Event &e);
};

//===================================================================

typedef std::map<GameObj*, SuicideData> SuicideDataMap;

#endif
