// $Id: AutoGatherAI_G3.C,v 1.3 2006/06/15 22:22:20 dtom Exp $

#include "Global.H"
#include "AutoGatherAI_G3.H"
//#include "Game1_GlobalState.H"
#include "Game3_accessories.H"

using namespace std;
typedef TerrainBase::Loc Loc;

#define TEST_SEC 0
#define TEST(x)
//#define TEST(x) x

/**************************************************************************/
// Constants

const sint4 AutoGatherAI::FROM = EventFactory::new_who();

const string AutoGatherAI::workerBPname("worker");
const string AutoGatherAI::mineralBPname("mineral");
const string AutoGatherAI::controlCenterBPname("controlCenter");

const uint4 AutoGatherAI::MINERAL_SLEEP_TIME_MS = 10*1000;
const sint4 AutoGatherAI::WORKER_MINERAL_DISTANCE_THRESHOLD = 2;
const sint4 AutoGatherAI::WORKER_CC_DISTANCE_THRESHOLD = 3;
const uint4 AutoGatherAI::MAX_WORKERS_PER_MINERAL = 1;

const sint4 AutoGatherAI::MIN_NUM_CLUSTERS = 2;
const sint4 AutoGatherAI::MAX_NUM_CLUSTERS = 10;

/**************************************************************************/

AutoGatherAI::AutoGatherAI(GameStateModule &gsm_, TerrainModule &tm_)
  : gsm(gsm_),
    tm(tm_),
    handlerOfPathFinderEvents(this), 
    handlerOfViewMsgEvents(this)
{
  tm.add_handler(&handlerOfPathFinderEvents);     // terrain -> ai
  gsm.add_handler(&handlerOfViewMsgEvents);       // gsm -> ai
  add_handler(&tm);                               // ai -> terrain
  minsMined=0;
}

/**************************************************************************/

AutoGatherAI::~AutoGatherAI()
{
  gsm.remove_handler(&handlerOfViewMsgEvents);
  tm.remove_handler(&handlerOfPathFinderEvents);
}

/**************************************************************************/

void AutoGatherAI::removeWMC(const Vector<ScriptObj*>& objects)
{
  FORALL_DYN(objects, it) {
    GameObj* gob = dynamic_cast<GameObj*>(*it);
    if( !gob )
      continue;

    // Remove a worker
    // ---------------
    if( gob->bp_name() == workerBPname ) {

      // find the worker wrapper
      Workers::iterator workerIt = workers.find(Worker(gob, gsm));
      assert(workerIt != workers.end());
      const Worker& worker = *workerIt;

      // is the worker associated with a mineral patch ?
      const GobPtrWrapper* mineral = getMineral(worker);
      if( mineral != NULL ) {
        // remove the worker from minerals2workers
        pair<M2W::iterator, M2W::iterator> workersOfTheMineral = minerals2workers.equal_range(mineral);
        for( M2W::iterator m2wIt = workersOfTheMineral.first ; m2wIt != workersOfTheMineral.second ; m2wIt++ ) {
          if( *(m2wIt->second) == worker ) {
            minerals2workers.erase(m2wIt);
            break;
          }
        }
      }

      // remove the worker itself
      workers2state.erase(workers2state.find(&worker));
      workers.erase(workerIt);
    }

    // Remove a mineral
    // ----------------
    else if( gob->bp_name() == mineralBPname ) {

      // find the mineral wrapper
      GameObjects::iterator mineralIt = minerals.find(GobPtrWrapper(gob, gsm));
      assert(mineralIt != minerals.end());
      const GobPtrWrapper& mineral = *mineralIt;

      // find all workers associated with this mineral patch
      pair<M2W::iterator, M2W::iterator> workersOfTheMineral = minerals2workers.equal_range(&mineral);
      // remove these associations from the workers2state map
      for( M2W::iterator m2wIt = workersOfTheMineral.first ; m2wIt != workersOfTheMineral.second ; m2wIt++ ) {
        workers2state[m2wIt->second].mineral = NULL;
      }
      // remove these associations from the minerals2workers map
      minerals2workers.erase(workersOfTheMineral.first, workersOfTheMineral.second);

      // remove the mineral itself
      minerals.erase(mineralIt);
    }
    else if( gob->bp_name() == controlCenterBPname ) {
      if( controlCenters.size()>=1 && *controlCenters.begin() == gob )
        controlCenters.clear();
    }
  }
}

/**************************************************************************/

void AutoGatherAI::insertWMC(const Vector<ScriptObj*>& objects)
{
  const sint4 me = gsm.get_game().get_client_player();

  FORALL_DYN(objects, it) {
    GameObj* gob = dynamic_cast<GameObj*>(*it);
    if( !gob )
      continue;

    if( gob->bp_name() == workerBPname && *(gob->sod.owner) == me ) {
      pair<Workers::iterator,bool> result =
        workers.insert(Worker(gob, gsm));
      const Worker& worker = *result.first;
      //workers2state.insert(W2S::value_type(&worker, WorkerState()));
      workers2state.insert(W2S::value_type(&worker,WorkerState()));
    }
    else if( gob->bp_name() == mineralBPname ) {
      minerals.insert(GobPtrWrapper(gob, gsm));
    }
    else if( gob->bp_name() == controlCenterBPname && *(gob->sod.owner) == me ) {
      controlCenters.clear();
      controlCenters.insert(GobPtrWrapper(gob, gsm));
    }
  }
}

/**************************************************************************/
void AutoGatherAI::removeW(Vector<GameObj*>& objects)
{
  FORALL_DYN(objects, it) {
    GameObj* gob = (*it);
    if( !gob )
      continue;

    // Remove a worker
    // ---------------
    if( gob->bp_name() == workerBPname ) {
      // find the worker wrapper
      Workers::iterator workerIt = workers.find(Worker(gob, gsm));
      assert(workerIt != workers.end());
      const Worker& worker = *workerIt;

      // is the worker associated with a mineral patch ?
      const GobPtrWrapper* mineral = getMineral(worker);
      if( mineral != NULL ) {
        // remove the worker from minerals2workers
        pair<M2W::iterator, M2W::iterator> workersOfTheMineral = minerals2workers.equal_range(mineral);
        for( M2W::iterator m2wIt = workersOfTheMineral.first ; m2wIt != workersOfTheMineral.second ; m2wIt++ ) {
          if( *(m2wIt->second) == worker ) {
            minerals2workers.erase(m2wIt);
            break;
          }
        }
      }

      // remove the worker itself
      workers2state.erase(workers2state.find(&worker));
      workers.erase(workerIt);
    }

  }
}

/**************************************************************************/

void AutoGatherAI::insertW(Vector<GameObj*>& objects)
{
  const sint4 me = gsm.get_game().get_client_player();

  FORALL_DYN(objects, it) {
    GameObj* gob = (*it);
    if( !gob )
      continue;

    if( gob->bp_name() == workerBPname && *(gob->sod.owner) == me ) {
      fprintf(stderr, "should be returning worker now\n");
      pair<Workers::iterator,bool> result =
        workers.insert(Worker(gob, gsm));
      const Worker& worker = *result.first;
      //workers2state.insert(W2S::value_type(&worker, WorkerState()));
      workers2state.insert(W2S::value_type(&worker,WorkerState()));
    }
    }
}
/*************************************************************************/
void AutoGatherAI::applyViewMessage(const GameChanges &changes)
{
  // Vanished workers, minerals and control center
  removeWMC(changes.vanished_objs);

  // Dead workers, minerals and control center
  removeWMC(changes.dead_objs);

  // New workers, minerals and control center
  insertWMC(changes.new_objs);

#if TEST_SEC
  FORALL(changes.changed_objs, it) {
    GameObj* gob = dynamic_cast<GameObj*>(*it);
    // print 'changed speed' events for workers:
    if( gob && gob->bp_name() == workerBPname && gob->attr_changed("speed") )
      cout  << dynamic_cast<Object*>(gob) << " changed speed; new speed = " << *gob->sod.speed << " location=(" << *gob->sod.x << "," << *gob->sod.y << ")" << endl;
    // print location of moving workers:
    else if( gob && gob->bp_name() == workerBPname && *gob->sod.speed > 0 )
      cout  << dynamic_cast<Object*>(gob) << " location=(" << *gob->sod.x << "," << *gob->sod.y << ")" << endl;

    // print location of all workers (with some changed attribute):
    if( gob && gob->bp_name() == workerBPname )
      cout  << dynamic_cast<Object*>(gob) << " location=(" << *gob->sod.x << "," << *gob->sod.y << ")" << endl;
  }
//  if(GlobalState::gsm->get_game().get_view_frame() >= 106)
//    cout << "!" << endl;
#endif
}

/**************************************************************************/

bool AutoGatherAI::isMineralFreeFor(const GobPtrWrapper* mineral, const Worker& worker) const
{
  return minerals2workers.count(mineral) - ((getMineral(worker)==mineral)? 1:0) < MAX_WORKERS_PER_MINERAL;
}

/**************************************************************************/

void AutoGatherAI::associateWorkerAndMineral(Worker& worker, const GobPtrWrapper& mineral)
{
  // is this worker already associated with a different mineral ?
  const GobPtrWrapper* otherMineral = getMineral(worker);
  if( otherMineral != NULL ) {
    // remove this association
    pair<M2W::iterator, M2W::iterator> workersOfTheMineral = minerals2workers.equal_range(otherMineral);
    for( M2W::iterator m2wIt = workersOfTheMineral.first ; m2wIt != workersOfTheMineral.second ; m2wIt++ ) {
      if( *(m2wIt->second) == worker ) {
        minerals2workers.erase(m2wIt);
        break;
      }
    }
    // and for the sake of clarity:
    getState(worker).mineral = NULL;
  }

  // associate with the new mineral
  minerals2workers.insert(M2W::value_type(&mineral, &worker));
  getState(worker).mineral = &mineral;

  TEST(cout << static_cast<const Object*>(worker) << " associated with mineral at (" << mineral->getPosition().x << "," << mineral->getPosition().y << ")" << endl);
}

/**************************************************************************/

void AutoGatherAI::allocateMineralPatchFor(Worker& worker, bool forcedChange)
{
  /*    // pick a random mineral patch
  if( minerals.size() > 0 ) {
  vector<GameObjects::value_type> rm(minerals.begin(),minerals.end());
  random_shuffle(rm.begin(), rm.end());
  associateWorkerAndMineral(worker, rm[0]);
  }
  return;
  */

  // Is there a good currently allocated mineral ?
  const GobPtrWrapper* current_mineral = getMineral(worker);
  if( !forcedChange && current_mineral != NULL && isMineralFreeFor(current_mineral, worker) ) {
      return;
  }

  // find the closest mineral patch which is not allocated to too many workers

  // sort according to distance
  vector<const GameObjects::value_type*> sortable_minerals;
  FORALL(minerals, mIt) {
    const GobPtrWrapper& mineral = *mIt;
    sortable_minerals.push_back(&mineral);
  }
  sort(sortable_minerals.begin(), sortable_minerals.end(), 
       CompareMineralsByDistanceToWorker(worker));

  // find the first free mineral, taking into account the 'forcedChange' instruction
  FORALL_B(sortable_minerals, freeMineralIt) {
    if( (!forcedChange || current_mineral != *freeMineralIt) && 
        isMineralFreeFor(*freeMineralIt, worker) )
      break;
  }

  // if there is such a mineral, use it
  if( freeMineralIt != sortable_minerals.end() ) {
    const GobPtrWrapper* mineral = *freeMineralIt;
    associateWorkerAndMineral(worker, *mineral);
  }
}

/**************************************************************************/

inline bool AutoGatherAI::isNearTheCC(const Worker& worker) const
{
  if( controlCenters.size() < 1 )
    return false;

  const GobPtrWrapper& controlCenter = *(controlCenters.begin());
  bool nearCC = (sint4)(Game::distance(*worker,*controlCenter)) <= WORKER_CC_DISTANCE_THRESHOLD;
  TEST(cout << static_cast<const Object*>(worker) << " NEAR_CC? " << nearCC << " (dist=" << ((sint4)(Game::distance(*worker,*controlCenter))) << ")" << endl);
  return nearCC;
}

/**************************************************************************/

inline bool AutoGatherAI::isNearTheMineral(const Worker& worker) const
{
  // find the allotted mineral patch
  const GobPtrWrapper* mineral = getMineral(worker);
  if( mineral == NULL )
    return false;

  bool nearMineral = (sint4)(Game::distance(*worker,**mineral)) <= WORKER_MINERAL_DISTANCE_THRESHOLD;
  TEST(cout << static_cast<const Object*>(worker) << " NEAR_MINERAL? " << nearMineral << " for mineral at ("; );
  TEST(cout << mineral->getPosition().x << "," << mineral->getPosition().y << ") (dist=" << (sint4)(Game::distance(*worker,**mineral)) << ")" << endl; );
  return nearMineral;
}

/**************************************************************************/
void AutoGatherAI::moveTo(Worker& worker, const Loc &location)
{
  TEST( cout << static_cast<const Object*>(worker) << " moving to location (" << location.x << "," << location.y << ")" << endl; );
  call_handlers( TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG,
                 TerrainBase::Obj2LocTask(worker, location)) );
  // update the worker's state
  getState(worker).job = WorkerState::MOVING;
}

/**************************************************************************/
void AutoGatherAI::moveTo(Worker& worker, const GobPtrWrapper& target)
{
  ScalarPoint targetPosition = target->getPosition();

/* The to-location version:
  TerrainBase::Task t(TerrainBase::Task::LOCATION, TerrainBase::Task::TOUCH);
  t.goal_loc = TerrainBase::Loc(targetPosition.x, targetPosition.y); */

  call_handlers( TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG,
                 TerrainBase::Obj2ObjTask(worker, target)) );

  // update the worker's state
  getState(worker).job = WorkerState::MOVING;
}

/**************************************************************************/

void AutoGatherAI::moveToCC(Worker& worker)
{
  if( controlCenters.size() >= 1 ) {
    const GobPtrWrapper& controlCenter = *(controlCenters.begin());
    moveTo(worker, controlCenter);
    TEST(cout << static_cast<const Object*>(worker) << " MOVE_CC" << endl);
  }
}

/**************************************************************************/

void AutoGatherAI::moveToMineral(Worker& worker)
{
  // find the allotted mineral patch
  const GobPtrWrapper* mineral = getMineral(worker);
  if( mineral == NULL )
    return;

  moveTo(worker,*mineral);
  TEST(cout << static_cast<const Object*>(worker) << " MOVE_TO_MINEARL at (" << mineral->getPosition().x << "," << mineral->getPosition().y << ")" << endl);
}

/**************************************************************************/

void AutoGatherAI::startMining(Worker& worker)
{
  // find the allotted mineral patch
  const GobPtrWrapper* mineral = getMineral(worker);
  if( mineral == NULL )
    return;

  worker->startMining(*mineral);
  // update the worker's state
  getState(worker).job = WorkerState::MINING;
}

/**************************************************************************/

void AutoGatherAI::returnMinerals(Worker& worker)
{
  const GobPtrWrapper& controlCenter = *controlCenters.begin();
  worker->returnMinerals(controlCenter);
  // update the worker's state - returning resources takes a single game tick,
  // and since there is no event for "finished returning the resources", 
  // we already mark the worker as idle.
  getState(worker).job = WorkerState::RETURNINGMINERALS;
}

/**************************************************************************/

bool AutoGatherAI::isIdle(const Worker& worker)
{
  // The end of mining/returningMinerals is not signaled automatically 
  // by an external event, so check these jobs explicitly.

  // Mining might take > 1 tick and the GameObj has an internal status to check
  if( getState(worker).job == WorkerState::MINING )
    return !worker->isMining() && !worker->is_pending_action();

  // Returning minerals to the CC takes 1 tick, no internal status. The moment this
  // action is executed, it is finished.
  else if( getState(worker).job == WorkerState::RETURNINGMINERALS ){
    if(!worker->is_pending_action()) minsMined+=10;
    return !worker->is_pending_action();
  }
  

  // MOVING/IDLE differentiation is handled correctly by events.
  else
    return getState(worker).job == WorkerState::IDLE;
}

/**************************************************************************/

void AutoGatherAI::act()
{
  int i=0, max_workers=20;

  // get a random permutation of the workers list
  /*    vector<GameObjects::value_type> random_workers(workers.begin(),workers.end());
  random_shuffle(random_workers.begin(), random_workers.end());
  */

  //TEST(cout << "ACT ACT ACT ACT ACT ACT ACT ACT" << endl;);

  // walk the workers list, find idle workers
  FORALL(workers, it) {

    Worker& worker = const_cast<Worker&>(*it);

    TEST(GameObj* gob = (GameObj*)worker;
         cout  << dynamic_cast<Object*>(gob) << " is at (" << *gob->sod.x << "," << *gob->sod.y << ")" << endl;
    ); 

    if( isIdle(worker) ) {

      // Does it have minerals ?
      if( worker->getMinerals() > 0 ) {
        // If it's near the control center, return the minerals
        if( isNearTheCC(worker) ) {
          returnMinerals(worker);

          // Else, move to the CC
        } else {
          moveToCC(worker);
        }

        // No minerals:
      } else {
/*        // move to the next point in the path, or back to the CC
        sint4 &waypoint = getState(worker).clusterPathWaypoint;
        if( waypoint <= (sint4)(mineralClusters[0].pathAroundCluster.size()) )
          moveTo(worker, mineralClusters[0].pathAroundCluster[waypoint++]);
        else
          moveToCC(worker);
*/
        // verify that there's an allotted mineral patch for this worker
        allocateMineralPatchFor(worker);

        // If it's near the minerals, start mining
        if( isNearTheMineral(worker) ) {
          startMining(worker);

          // Else, move to the minerals
        } else {
          moveToMineral(worker);
        }
      }
    }

    if( ++i >= max_workers )
      break; // handle only the first X workers*/
  }
}

/**************************************************************************/

bool AutoGatherAI::HandlerOfPathFinderEvents::handle_event(const Event &event)
{
  assert( event.get_who() == TerrainModule::FROM );

  if( event.get_what() == TerrainModule::PATH_SUCCESS ||
      event.get_what() == TerrainModule::PATH_FAILURE ) {

      // is this a notification regarding one of our workers?
      const TerrainModule::PathStatusEvent* pse = 
        dynamic_cast<const TerrainModule::PathStatusEvent*>(&event);
      if( pse ) {
        GameObj* gob = pse->get_obj();
        // find the worker wrapper, if it's indeed a worker
        Workers::iterator workerIt = parent->workers.find(Worker(gob, parent->gsm));
        if( workerIt != parent->workers.end() ) {

          const Worker& worker = *workerIt;

          TEST(
            cout << static_cast<const Object*>(worker) << "got PF notification: ";
            if( event.get_what() == TerrainModule::PATH_SUCCESS )
              cout << "arrived" << endl;
            else
              cout << "failed" << endl;
          );

          // This worker is now idle
          parent->getState(worker).job = AutoGatherAI::WorkerState::IDLE;

          // If it failed to arrive, switch to a different mineral
          if( pse->get_what() == TerrainModule::PATH_FAILURE &&
              pse->get_details() == TerrainModule::PATH_PLAN_FAILURE ) {

              // Mark this mineral as sleeping, since it might be blocked
              // (alternatively, the worker might be blocked, so sleep for a short time)
              // switch this worker to a different mineral, since this one is blocked
              parent->allocateMineralPatchFor(const_cast<Worker&>(worker), true);
          }
        }
      }
      return true;
    }
    return false;
}

/**************************************************************************/

bool AutoGatherAI::HandlerOfViewMsgEvents::handle_event(const Event& e)
{
  assert(e.get_who() == GameStateModule::FROM);
  
  if( e.get_what() == GameStateModule::VIEW_MSG ) {
    parent->applyViewMessage(parent->gsm.get_changes());
    Game &game = parent->gsm.get_game();
    sint4 vf = game.get_view_frame();
    sint4 ai = game.get_action_frame();
    bool behind = abs(vf-ai) > 10 || ai < 0;
    if( !behind ) {
      parent->plan();
      parent->act();
    }
    return true;
  }
  return false;
}

/**************************************************************************/
void AutoGatherAI::computeMineralClusters()
{

  // List all mineral locations
  Vector<ScalarPoint> mineralLocations;
  mineralLocations.reserve(minerals.size());
  FORALL(minerals, mIt) {
    mineralLocations.push_back(mIt->getPosition());
  }

  // Cluster
  Vector<sint4> clusterOf;
  sint4 clusterCount;
  if(mineralLocations.size()>10){
    simple_clustering(mineralLocations, MIN_NUM_CLUSTERS, MAX_NUM_CLUSTERS, clusterOf);
    clusterCount = *max_element(clusterOf.begin(), clusterOf.end()) + 1;
  }
  // Store the results
  else
    clusterCount = 1;
  mineralClusters.clear(); mineralClusters.resize(clusterCount);
  int i=0; 
  if(clusterCount>1)
    FORALL(minerals, mIt) {
      const GobPtrWrapper &mineral = *mIt;
      mineralClusters[clusterOf[i++]].mineralsInCluster.push_back(&mineral);
    }
  else
    FORALL(minerals, mIt) {
      const GobPtrWrapper &mineral = *mIt;
      mineralClusters[0].mineralsInCluster.push_back(&mineral);
    }
  TEST(
    cout << "mineral2cluster = [" << endl;
    FORU(i, mineralLocations.size()) 
      cout << mineralLocations[i].x << "," << mineralLocations[i].y << "," << clusterOf[i] << ";" << endl;
    cout << "];" << endl;
  );
}


/**************************************************************************/
void AutoGatherAI::computePathsAroundClusters()
{
  // for each cluster, compute a roundabout path and blow it up a little
  FORU(i, mineralClusters.size())       // FORU io FORALL to avoid REGISTER_TYPEOF
  {
    Vector<ScalarPoint> clusterPoints;
    clusterPoints.reserve(mineralClusters[i].mineralsInCluster.size());
    FORALL(mineralClusters[i].mineralsInCluster, mIt) {
      clusterPoints.push_back((*mIt)->getPosition());
    }
    Vector<ScalarPoint> roundabout;
    const GobPtrWrapper& controlCenter = *(controlCenters.begin());
    compute_roundabout(clusterPoints, controlCenter->getPosition(), roundabout);

    // close the polygon using the control center's position
    roundabout.push_back(controlCenter->getPosition());

    // enlarge it a little bit
    Vector<ScalarPoint> blownup;
    blowup_polygon(roundabout, WORKER_MINERAL_DISTANCE_THRESHOLD, blownup);

    TEST( cout << "blownup = [" << endl;
          FORALL(blownup, j)  cout << j->x << "," << j->y << ";" << endl;
          cout << "];" << endl; );
    
    // copy to mineralClusters, remove the CC vertex of the path
    mineralClusters[i].pathAroundCluster.clear();
    FORALL(blownup, vertexIt) {
      mineralClusters[i].pathAroundCluster.push_back(Loc(vertexIt->x, vertexIt->y));
    }
    mineralClusters[i].pathAroundCluster.pop_back();
    mineralClusters[i].usePathAroundCluster = true;
  }
}

/**************************************************************************/
void AutoGatherAI::plan()
{
  static bool firstTime = true;
  if( !firstTime )
    return;
  firstTime = false;
  computeMineralClusters();
  computePathsAroundClusters();
}
