
#include "BuildTasksAI.h"

#include "BuildTasksState.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameObj.H"
#include "GfxModule.H"
#include "TerrainModule.H"
#include "TerrainClassifier.H"


using namespace std;

namespace BSimul{

enum tileType{ TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

const sint4 BuildTasksAI::FROM = EventFactory::new_who();

  BuildTasksAI::BuildTasksAI(GameStateModule &gsm_, TerrainModule &tm) : gsm(gsm_){
    simulationState = new BuildTasksState(this, gsm, tm);
    simulationState->setCurrent(600);
    currentTask = NULL;

  }
  BuildTasksAI::~BuildTasksAI(){simulationState->printMap();}


void BuildTasksAI::stop(GameObj *obj){
  FIND (obj2dat, it, obj);
  if (it != obj2dat.end()) obj2dat.erase(it);
}

//===================================================================

void BuildTasksAI::act(GameStateModule &gsm){
  //if(currentTask == NULL) return;
  if(!(currentTasks.size())) return;
  bool isDone;
  Vector<Task*>::iterator current;
  for(current=currentTasks.begin(); current!=currentTasks.end();){
    isDone=(*current)->update();
    if(isDone){
      currentTask=(*current);
      current=currentTasks.erase(current);
      delete currentTask;
      currentTask=NULL;
      isDone=false;
    }
    else
      current++;
  }
}

//===================================================================

/**
 * Returns true if event was handled.
 */
bool BuildTasksAI::handle_event(const Event &e)
{
  if(simulationState->getConCens().size())
  if (e.get_what() == TerrainModule::PATH_SUCCESS) {

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "SUCCESS: not a PathStatusEvent");

    FIND(simulationState->getFriendlyUnits(),it,p->get_obj());
    if (it == simulationState->getFriendlyUnits().end()) return false;

    // if we were just moving... then we are done!
    if(it->second.state == ExtraObjectData::BUILD){
      it->second.isDone = true;
      //it->second.state = ExtraObjectData::IDLE;
    }
    if(it->second.state == ExtraObjectData::BUILD){
    }
    if(it->second.state == ExtraObjectData::SCOUT){
      int x,y;
      (it->first)->get_center(x,y);
      if((x!=it->second.location.x)||(y!=it->second.location.y))
	simulationState->removeBuildableTile((it->second.location.x-8)/16,(it->second.location.y-8)/16);
      it->second.isDone = true;
    }
  }

  if (e.get_what() == TerrainModule::PATH_FAILURE) {

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "FAILURE: not a PathStatusEvent");
    FIND (simulationState->getFriendlyUnits(), it, p->get_obj());
    if (it != simulationState->getFriendlyUnits().end()){
      if(it->second.state == ExtraObjectData::BUILD){
      }
      if(p->get_details() == TerrainModule::PATH_PLAN_FAILURE){
	if(it->second.state==ExtraObjectData::SCOUT){
	  simulationState->removeBuildableTile((it->second.location.x-8)/16,(it->second.location.y-8)/16);
	  it->second.isDone = true;
	}
      }
      else if(p->get_details() == TerrainModule::PATH_MOVE_FAILURE){
	//it->second.isDone = true;
	if(it->second.state==ExtraObjectData::SCOUT){
	  int x,y;
	  simulationState->removeBuildableTile((it->second.location.x-8)/16,(it->second.location.y-8)/16);
	  (it->first)->get_center(x,y);
	  it->second.isDone = true;
	}
      }

    }
  }


  if (e.get_who() == GameStateModule::FROM) {

    if (e.get_what() == GameStateModule::VIEW_MSG) {
      const Game &game = gsm.get_game();
      const sint4 me = gsm.get_game().get_client_player();
      const sint4 numplayers = game.get_player_num();
      PlayerInfo &pi = game.get_cplayer_info();
      ScriptObj *player = pi.global_obj("player");
      simulationState->addMins(simulationState->autogatherer->getMinsMined());
//----------------------------TERRAIN------------------------------//
      //from sampleai
      const Map<GameTile> &map = game.get_map();
      FORS (y, map.get_height()) {
	FORS (x, map.get_width()) {
	  const GameTile &tile = map(x,y);
	  
	  sint4 hNW, hNE, hSE, hSW;
	  Tile::Split split;

	  // get corner heights and split type
	  tile.get_topo(hNW, hNE, hSE, hSW, split);

	  // tile types: see enum on top
	  // split types: see orts3/libs/kernel/src/Tile.H
	  // if split == Tile::NO_SPLIT => typeW is the tile type
	  
	  if((!split)&&(tile.get_typeW()==BSimul::TILE_GROUND))
	    simulationState->addBuildableTile(x, y);
	  if(tile.get_max_h() != 0)
	    simulationState->removeBuildableTile(x,y);
	}

      }
//------------------------------------------------------------------//

      //check on new buildings
      if(simulationState->incompleteBuildings()){
	FORALL(simulationState->getNewBuildings(), it){
	  if((*it)->bp_name() == "worker"){
	    if((*it)->get_int("is_mobile")){
	      simulationState->removeNewBuilding(*it);
	      simulationState->getFriendlyUnits()[(*it)].isDone=true;
	      ObjectVector freeBuilder;
	      freeBuilder.push_back((*it));
	      simulationState->autogatherer->insertW(freeBuilder);
	      simulationState->toggle();
	    }
	  }
	  else if((*it)->get_int("amount_built")==(*it)->get_int("build_time")){
	    if((*it)->bp_name() == "barracks")
	      simulationState->addBarracks(*it);
	    else if((*it)->bp_name() == "controlCenter")
	      simulationState->addConCen(*it);
	    else if((*it)->bp_name() == "factory")
	      simulationState->addFactory(*it);
	    simulationState->removeNewBuilding(*it);
	  }
	}
      }
      
      if((!(simulationState->getHarassers().size()))&&(simulationState->getFighters().size()>=5)&&(simulationState->getEnemyBases().size())){
	ObjectVector harassers;
	ObjectVector fighters=simulationState->getFighters();
	FORS(i, 5){
	  harassers.push_back(fighters[i]);
	}
	simulationState->setHarassers(harassers);
	currentTasks.push_back(createHarass(me,simulationState,harassers, 0.0));
      }
      

      if(simulationState->getFighters().size()>40);


      //check on enemy attackers
      if(simulationState->enemyNearBases()){
	FORALL(simulationState->getFriendlyUnits(), it)
	  if(it->second.state==ExtraObjectData::IDLE){
	    GameObj* offender=simulationState->enemyNearBases();
	    TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::TOUCH,
				   TerrainBase::Task::INDIVIDUAL);
	    task.goal.obj = offender;
	    task.objs.insert(it->first);
	    simulationState->simAI->doPath(task);
	    ExtraObjectData newData;
	    newData.target = offender;
	    newData.state = ExtraObjectData::DEFEND;
	    simulationState->getFriendlyUnits()[it->first] = newData;
	  }
      }

      if(simulationState->needWorker())
	if((simulationState->getConCens().size())&&(!simulationState->trainingWorker())){
	  currentTasks.push_back(createTrainWorker(me,simulationState,simulationState->getConCens(), 0.0));
	  simulationState->workerCompleted();
	  simulationState->uptrainWorker();
	}
      if(simulationState->needMarine())
	if((simulationState->getBarracks().size())){
	   currentTasks.push_back(createTrainMarine(me,simulationState,simulationState->getBarracks(), 0.0));
	   simulationState->marineCompleted();
	   simulationState->uptrainMarine();
	}
      if(simulationState->needTank())
	if((simulationState->getFactories().size())){
	  currentTasks.push_back(createTrainTank(me,simulationState,simulationState->getFactories(), 0.0));
	  simulationState->tankCompleted();
	  simulationState->uptrainTank();
	}

      sint4 x, y, l, t;
      FORALL (gsm.get_changes().new_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (*gob->sod.owner == me){

          if(isControllable(gob)){
            simulationState->addFriendlyUnit(gob);
	    simulationState->getFriendlyUnits()[gob].state=ExtraObjectData::IDLE;
	    if(gob->bp_name() == "worker"){
	      simulationState->addWorker(gob);
	      simulationState->workerDone();
	    }
	    else{
	      if(gob->bp_name() == "marine")
		simulationState->marineDone();
	      else if(gob->bp_name() == "tank")
		simulationState->tankDone();
	      simulationState->addFighter(gob);
	    }
	    gob->get_center(x,y);
	    simulationState->removeBuildableTile(x/16,y/16);
          }else if(gob->bp_name() == "barracks"){   //redundant, especially if james' map works
            simulationState->addFriendlyBase(gob);
	    simulationState->addNewBuilding(gob);
	    gob->get_p1(l,t);
	    simulationState->removeDimensions(l, t,
	    			      gob->get_int("width"), gob->get_int("height"));
          }else if(gob->bp_name() == "controlCenter"){
            simulationState->addFriendlyBase(gob);
	    simulationState->addNewBuilding(gob);
	    gob->get_p1(l,t);
	    simulationState->removeDimensions(l, t,
	    			      gob->get_int("width"), gob->get_int("height"));
          }else if(gob->bp_name() == "factory"){
            simulationState->addFriendlyBase(gob);
	    simulationState->addNewBuilding(gob);
	    gob->get_p1(l,t);
	    simulationState->removeDimensions(l, t,
	    			      gob->get_int("width"), gob->get_int("height"));
          }
	}else if(*gob->sod.owner == numplayers){
	  if(gob->bp_name() == "mineral"){
	    simulationState->addMinerals(gob);
	    gob->get_center(x,y);
	    simulationState->removeBuildableTile(x/16,y/16);
	  }
	}else if(*gob->sod.owner != numplayers){
          if(isControllable(gob)){
	    if(gob->bp_name() == "worker"){
	      simulationState->addEnemyWorker(gob);
	    }
	    else{
	      if((gob->bp_name() == "marine")||(gob->bp_name() == "tank"))
		simulationState->addEnemyFighter(gob);
	    }
          }else if (gob->bp_name() == "controlCenter"){
            simulationState->addEnemyBase(gob);	 
	    gob->get_p1(l,t);
	    simulationState->removeDimensions(l, t,
					      gob->get_int("width"), gob->get_int("height"));
          }
	}
      }

      

      if((simulationState->getWorkers().size()+simulationState->needWorker()+simulationState->trainingWorker())<(simulationState->getMinerals().size())){
	simulationState->queueWorkers(1);
      }
      
      if((!simulationState->needMarine())&&(simulationState->spendMins(50)))
	simulationState->queueMarines(1);
      
      

      
      if((simulationState->spendMins(400))&& (!(simulationState->queued()))
	                                           && ((simulationState->getWorkers()).size())){
	simulationState->toggle();
	ObjectVector builder;
	builder.push_back(simulationState->getWorkers()[0]);
	currentTasks.push_back(createBuildandTrain(me,simulationState,builder, 0.0));
      }

      // remove dead objects
      FORALL (gsm.get_changes().dead_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (*gob->sod.owner == me){
          if(gob->bp_name() == "controlCenter"){
            simulationState->removeFriendlyBase(gob);
	    simulationState->removeConCen(gob);
	  }else if(gob->bp_name() == "barracks"){
            simulationState->removeFriendlyBase(gob);
	    simulationState->removeBarracks(gob);
	  }else if(gob->bp_name() == "factory"){
            simulationState->removeFriendlyBase(gob);
	    simulationState->removeFactory(gob);
	  }else{
	    gob->get_center(x,y);
            simulationState->removeFriendlyUnit(gob);
	    if(gob->bp_name()=="worker")
	      simulationState->removeWorker(gob);
	    else
	      simulationState->removeFighter(gob);
          }
	  FORALL(currentTasks, it) 
	    (*it)->removeObject(gob);
        }else{
          if(gob->bp_name() == "controlCenter"){	    
	    simulationState->removeEnemyBase(gob);
          }else if(gob->bp_name() == "mineral"){	    
	    simulationState->removeMinerals(gob);
          }else if(gob->bp_name() == "worker"){
	    simulationState->removeEnemyWorker(gob);
          }else if((gob->bp_name() == "marine")||(gob->bp_name() == "tank")){
	    simulationState->removeEnemyFighter(gob);
          }

        }
        informObjectsOfDeath(gob);

      }

      //remove vanished objects
      FORALL (gsm.get_changes().vanished_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;
        if (*gob->sod.owner == me){
          if(gob->bp_name() == "controlCenter"){
            simulationState->removeFriendlyBase(gob);
	    simulationState->removeConCen(gob);
	  }else if(gob->bp_name() == "barracks"){
            simulationState->removeFriendlyBase(gob);
	    simulationState->removeBarracks(gob);
	  }else if(gob->bp_name() == "factory"){
            simulationState->removeFriendlyBase(gob);
	    simulationState->removeFactory(gob);
	  }else{
	    gob->get_center(x,y);
            simulationState->removeFriendlyUnit(gob);
	    if(gob->bp_name()=="worker")
	      simulationState->removeWorker(gob);
	    else
	      simulationState->removeFighter(gob);
          }
	  FORALL(currentTasks, it) 
	      (*it)->removeObject(gob);
        }else{
          if(gob->bp_name() == "controlCenter"){	    
	    simulationState->removeEnemyBase(gob);
          }else if(gob->bp_name() == "mineral"){	    
	    simulationState->removeMinerals(gob);
          }else if(gob->bp_name() == "worker"){
	    simulationState->removeEnemyWorker(gob);
          }else if((gob->bp_name() == "marine")||(gob->bp_name() == "tank")){
	    simulationState->removeEnemyFighter(gob);
          }
        }
        informObjectsOfDeath(gob);
      }

    }
  }
  //  cout << "Event was not handled!" << endl; 

  act(gsm);
  return true; 
}

void BuildTasksAI::informObjectsOfDeath(GameObj* obj){
  FORALL(simulationState->getFriendlyUnits(),it){
    // if the objects we had a target to was killed... this unit is done its task
    if((it->second.state == ExtraObjectData::HARASS)||(it->second.state == ExtraObjectData::ATTACK)){
      if(it->second.target != NULL){
        if(it->second.target == obj){
          it->second.isDone = true;
	  it->second.target=NULL;
        }
      }
    }else if(it->second.state == ExtraObjectData::DEFEND){
      if(it->second.target != NULL){
        if(it->second.target == obj){
	  it->second.state=ExtraObjectData::IDLE;
	  it->second.target=NULL;
        }
      }
    }
  }

  // TODO: maybe add updates for other objects later.  not needed right now, since we only control friendly units
}


  bool BuildTasksAI::isControllable(GameObj* obj){
  // first, check if we even have a weapon...and can move...
    if ((obj->get_max_speed() > 0) && (obj->component("weapon") != NULL)){
            // if we can, make sure it can be used....this will likely always pass
      if (obj->component("weapon")->has_attr("max_damage")){
              // make sure we can actually do dmg
        if(obj->component("weapon")->get_int("max_damage") > 0){ // if we can attack
          return true;
        }
      }
    }
    return false;
  }
  sint4 BuildTasksAI::doPath(TerrainBase::Task task){
    std::cout << "Send Pathfinding Message\n";
    return call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, task));
  }

}
