#ifndef BUILDTASKS_AI_H
#define BUILDTASKS_AI_H


#include "Global.H"
#include "LowAI.H"
#include "EventHandler.H"
#include "ClientModule.H"
#include "TerrainBase.H"

#include "TerrainModule.H"
#include "BuildTasks.h"

class GameStateModule;
namespace BSimul{
//===================================================================

class Task;
class BuildTasksState;
/**
 * Struct for extra data associated with each object
 */
class ExtraObjectData {
public:
  ExtraObjectData(){
    target = NULL;
    isDone = false;
  }
  GameObj *target;
  TerrainBase::Loc location;
  enum State {MOVE, ATTACK, IDLE, SCOUT, POSITIONING, FLEE, BUILD, HARASS, DEFEND};
  State state;
  bool isDone;
};

typedef std::map<GameObj*, ExtraObjectData> ObjectDataMap;


//-------------------------------------------------------------------

/**
 * This AI module causes all units to attack the nearest (geometric distance)
 * base.
 */
class BuildTasksAI : public LowAI<ExtraObjectData>, public EventHandler, public ClientModule{
protected:
  //std::set<GameObj*> bases; ///< Enemy Bases

  GameStateModule &gsm;
  BuildTasksState* simulationState;
  
  Vector<Task*> currentTasks;
  Task* currentTask;

  //Vector<GameObj*> need_act; ///< GameObjects needing an action in the next tick
  /**
   * Removes the given GameObj from the control of this class.
   */
  void stop(GameObj *obj);


  /**
   * Finds the nearest base to travel to, based on geometric distance.
   */
  //GameObj* find_target(GameObj *targeter, GameObj *baseToAvoid = 0);
public:

  static const sint4 FROM;

  BuildTasksAI(GameStateModule &gsm, TerrainModule &tm);
  ~BuildTasksAI();

  /**
   * Causes the units in the need_act vector to act according to
   * their current state.
   */
  void act(GameStateModule &gsm);

  //static const std::string bp_name; ///< ?

  void add(GameObj*) { ERR("?"); }

  /**
   * Addes an enemy base to the list of enemy bases, and updates targets
   */
  //void add_base(GameObj *gob);

  void del(GameObj*) { ERR("?"); }

  /**
   * Deletes a base from the list of enemy bases, and updates targets
   */
  //void del_base(GameObj *gob);

  /**
   * Handles events sent to SuicideAttackAI from other modules
   */
  bool handle_event(const Event &e);

  void informObjectsOfDeath(GameObj* obj);
  bool isControllable(GameObj* obj);

  sint4 doPath(TerrainBase::Task task);

  GameStateModule& get_gsm(){ return gsm;}
  

};

}

#endif

