
#include "TerrainClassifier.H"

//#define PRINT(x) x;
#define PRINT(x)

using namespace std;

namespace mapTerrain{

  /*---------------------------------------------------------------*/
  /*                         CONSTRUCTOR                           */
  /*---------------------------------------------------------------*/
  //Constuctor, DO NOT CALL WITH A NULL GAME!!
  TerrainMap::TerrainMap(Game *game){
    //Create a copy of the current game map.
    if(!game){
      ERR("Game cannot be NULL!");
      return;
    }
      
    game_map = &(game->get_map()); 
    
    PRINT(printf("game = %p .\n" , game);
	  printf("game_map = %p .\n" , game_map););
      
    //Set our interal variables for readability
    width = game_map->get_width();
    height = game_map->get_height();
    num_players = game->get_player_num();
    tile_points = game->get_tile_points();
    rand.srand(0);

    FORS(i, num_players + 1)
      all_objs.push_back(game->get_objs(i));
    
    //Create a 2D array of the given sizes.
    map = new Tile*[width];
    FORS(i, width)
      map[i] = new Tile[height];

    init_blocked_tiles();
    compute_free_areas();

  }

  /*---------------------------------------------------------------*/
  /*                         DESTRUCTOR                            */
  /*---------------------------------------------------------------*/
  //Destuctor
  TerrainMap::~TerrainMap(){
    //Free all of the used memory
    FORS (i, width) 
      delete [] map[i];
    delete [] map;

    height = 0; width = 0; tile_points = 0;
    num_players = 0;
    openList.clear();
    all_objs.clear();
    buildings.clear();
    mineral_patches.clear();

  }
 
  /*---------------------------------------------------------------*/
  /*                      INITIALIZATION                           */
  /*---------------------------------------------------------------*/
  //Initialize the map to mirror the current game map
  void TerrainMap::init_blocked_tiles(){

    FORS(i, width) {
      FORS(j, height) {
	//Step through all the tiles and reproduce a simplified version here
	GameTile *temp = &game_map->get_tile(i,j);
	sint4 height = temp->get_max_h();
	map[i][j].x = i;
	map[i][j].y = j;

	//Tile type is cliff or water
	if(height != 0 || temp->get_typeW() == 1 || temp->get_typeW() == 3 ){
	  map[i][j].status = Tile::OBSTACLE;
	  blocked_tiles.push_back(map[i][j]);
	}
	//UNKNOWN Tile type -> behind FOW
	else if(temp->get_typeW() == 0) 
	  map[i][j].status = Tile::FOG;
	//Tile type is free
	else if(temp->get_typeW() == 2) 
	  map[i][j].status = Tile::FREE;
	else
	  ERR("Unknown tile type!!");
      }
    }


    FORU (i, width) {
      map[i][0].status = Tile::OBSTACLE;
      map[i][height-1].status = Tile::OBSTACLE;
      blocked_tiles.push_back(map[i][0]);
      blocked_tiles.push_back(map[i][height-1]);
    }

    FORU (i, height) {
      map[0][i].status = Tile::OBSTACLE;
      map[width-1][i].status = Tile::OBSTACLE;
      blocked_tiles.push_back(map[0][i]);
      blocked_tiles.push_back(map[width-1][i]);
    }

    
    //Get all of the objects from all of the players
    FORALL(all_objs, i){
      ObjCont player_objs = (*i);
      FORALL(player_objs, j){
	GameObj* obj = (*j)->get_GameObj();
	if(!obj)
	  continue;

	if(obj->get_shape() == Object::RECTANGLE)
	  buildings.push_back(*j); //Insert the buildings
	else if(obj->get_shape() == Object::CIRCLE && obj->bp_name() == "mineral" )
	  mineral_patches.push_back(*j); //Insert the mineral patches
	else if(obj->get_shape() == Object::CIRCLE && obj->bp_name() != "sheep" )
	  units.push_back(*j); //Insert the units
      }
    }
    /**Step through all the buildings, first convert the 
       coordinates to tiles and then for all such tiles
       mark as an obstacle and add to the initial open list.
    */
    FORALL(buildings, i){
      sint4 topLeftx, topLefty;
      sint4 bottomRightx, bottomRighty;
      GameObj* obj = (*i)->get_GameObj();
      if(!obj)
	continue;

      obj->get_p1(topLeftx, topLefty);
      obj->get_p2(bottomRightx, bottomRighty);

      sint4 tilei = topLeftx / tile_points;
      sint4 tilej = topLefty / tile_points;

      sint4 endi = bottomRightx / tile_points;
      sint4 endj = bottomRighty / tile_points;

      tilei = max(tilei, 0);
      tilej = max(tilej, 0);
      endi = min(endi, width-1);
      endj = min(endj, height-1);

      for(int j = tilei; j <= endi; j++){
	for(int k = tilej; k <= endj; k++){
	  map[j][k].status = Tile::OBSTACLE;
	  blocked_tiles.push_back(map[j][k]);
	}
      }
    }

    FORALL(mineral_patches, i){
      sint4 topLeftx, topLefty;
      sint4 bottomRightx, bottomRighty;

      GameObj* obj = (*i)->get_GameObj();
      if(!obj)
	continue;

      sint4 x,y; 
      obj->get_center(x,y);
      sint4 rad = obj->get_radius();
      
      topLeftx = (x - rad) / tile_points;
      topLefty = (y - rad) / tile_points;

      bottomRightx = (x + rad) / tile_points;
      bottomRighty = (y + rad) / tile_points;

      topLeftx = max(topLeftx, 0);
      topLefty = max(topLefty, 0);
      bottomRightx = min(bottomRightx, width-1);
      bottomRighty = min(bottomRighty, height-1);

      for(int j = topLeftx; j <= bottomRightx; j++){
	for(int k = topLefty; k <= bottomRighty; k++){
	  map[j][k].status = Tile::OBSTACLE;
	  blocked_tiles.push_back(map[j][k]);
	}
      }
    }

    FORALL(units, i){
      sint4 topLeftx, topLefty;
      sint4 bottomRightx, bottomRighty;

      GameObj* obj = (*i)->get_GameObj();
      if(!obj)
	continue;

      sint4 x,y; 
      obj->get_center(x,y);
      sint4 rad = obj->get_radius();

      topLeftx = (x - rad) / tile_points;
      topLefty = (y - rad) / tile_points;

      bottomRightx = (x + rad) / tile_points;
      bottomRighty = (y + rad) / tile_points;

  
      topLeftx = max(topLeftx, 0);
      topLefty = max(topLefty, 0);
      bottomRightx = min(bottomRightx, width-1);
      bottomRighty = min(bottomRighty, height-1);

      for(int j = topLeftx; j <= bottomRightx; j++){
	for(int k = topLefty; k <= bottomRighty; k++){
	  map[j][k].status = Tile::UNIT;
	}
      }
    }
  }


  //Compute the distances to all obstacles
  void TerrainMap::compute_free_areas(){  
   
    /**First go through all blocked tiles and mark all the neighbors as 1 away from an obstacle,
       pushing onto the open list all such tiles.
    */ 
    FORALL(blocked_tiles, i){
      Tile current = *i;
      sint4 startx = current.x;
      sint4 starty = current.y;

      //Left side
      checkNode(startx - 1, starty, 1);
      //Upper left corner
      checkNode(startx - 1, starty - 1, 1);
      //Top side
      checkNode(startx, starty - 1, 1);
      //Upper right corner
      checkNode(startx + 1, starty - 1, 1);
      //Right side
      checkNode(startx + 1, starty, 1);
      //Bottom right corner
      checkNode(startx + 1, starty + 1, 1);
      //Bottom side
      checkNode(startx, starty + 1, 1);
      //Bottom left corner
      checkNode(startx - 1, starty + 1, 1);
	
    }

    /**Next we will go through the open list assigning distances as we go and pushing 
       all unmarked tiles on to the list.
    */
    openList = tempList;
    tempList.clear();
    sint4 distance = 2;
    FOREVER{
      //Mark the bounds of the current version of the list
      //std::deque<Tile*>::iterator start = openList.begin();
      //std::deque<Tile*>::iterator end = openList.end();
      
      //Check to see if we are done
      if(openList.empty())
	break;

      //Go through the old version of the list adding tiles only past the end iterator
      //      for(std::deque<Tile*>::iterator j = start; j != end; j++){
      FORALL(openList, j){

	PRINT(std::cout << "Open List = {" ;
	      FORALL(openList, i)
	      std::cout << "(" << (*i)->x << "," << (*i)->y << ") " ;
	      std::cout << "}" << std::endl;);

	Tile *temp = *j;
	sint4 startx = temp->x;
	sint4 starty = temp->y;
	
	//Left side
	checkNode(startx - 1, starty, distance);
	//Upper left corner
	checkNode(startx - 1, starty - 1, distance);
	//Top side
	checkNode(startx, starty - 1, distance);
	//Upper right corner
	checkNode(startx + 1, starty - 1, distance);
	//Right side
	checkNode(startx + 1, starty, distance);
	//Bottom right corner
	checkNode(startx + 1, starty + 1, distance);
	//Bottom side
	checkNode(startx, starty + 1, distance);
	//Bottom left corner
	checkNode(startx - 1, starty +1, distance);
	
      }
      //Clear all expanded nodes.
      openList = tempList;
      tempList.clear();
      //openList.erase(start, end);
      distance++;
    }

  
  }      

  /*---------------------------------------------------------------*/
  /*                   Public Member Functions                     */
  /*---------------------------------------------------------------*/
  
  //Print the current map
  void TerrainMap::print_map(){
    //Print the map in terms of status.
    FORS(j, height) {
      FORS(i, width) {
	std::cout << map[i][j].status << " ";
      }
      std::cout << std::endl;
    }
    std::cout << std::endl << std::endl;

    //Print the map in terms of distances to objects.
    FORS(j, height) {
      FORS(i, width) {
	std::cout << map[i][j].freeValue << " ";
      }
      std::cout << std::endl;
    }
    std::cout << std::endl << std::endl;
  }
 
  bool TerrainMap::find_free_loc(TerrainBase::Loc point, TerrainBase::Loc &location, sint4 bound, sint4 min){

    init_blocked_tiles();
    compute_free_areas(); 

    sint4 i = point.x / tile_points;
    sint4 j = point.y / tile_points;
    bound /= tile_points;

#if 1
    const sint4 range = 6;

    sint4 x0 = i - range;
    sint4 x1 = i + range;
    sint4 y0 = j - range;
    sint4 y1 = j + range;

    if (x0 < 0) x0 = 0;
    if (y0 < 0) y0 = 0;
    if (x1 >= width) x1 = width-1;
    if (y1 >= height) y1 = height-1;

    sint4 current_value = map[i][j].freeValue;

    for (sint4 y = y0; y <= y1; ++y) {
      for (sint4 x = x0; x <= x1; ++x) {
        if (map[x][y].freeValue > current_value) {
          i = x;
          j = y;
          
          location.x = x * tile_points;
          location.y = y * tile_points;

          current_value = map[x][y].freeValue;

        }
      }
    }

    return true;
    return current_value >= min;

#else
    sint4 distance = 0;
    
    while(distance <= bound){
      if(move_one_tile(i, j)){
	distance++;
	continue;
      }
      else
	break;
    }
    PRINT(  std::cout << "found tile i =" << i << " and j = " << j << std::endl;
	    std::cout << "Distance = " << distance << " and bound = " << bound << std::endl; );
    
    location = TerrainBase::Loc(i*tile_points , j*tile_points);
    if(map[i][j].freeValue >= min){
      return true;
    }
    else{
      return false;
    }
#endif
  }

  bool TerrainMap::is_free_of_units(TerrainBase::Loc point, sint4 bound, sint4 max_units){
    init_blocked_tiles();
    compute_free_areas();

    sint4 no_of_units = 0;
    sint4 starti, startj;
    sint4 endi, endj;

    starti = (point.x - bound) / tile_points;
    startj = (point.y - bound) / tile_points;

    endi = (point.x + bound) / tile_points;
    endj = (point.y + bound) / tile_points;
     
    starti = max(starti, 0);
    startj = max(startj, 0);
    endi = min(endi, width-1);
    endj = min(endj, height-1);

    for(int i = starti; i <= endi; i++){
      for(int j = startj; j <= endj; j++){
	if(map[i][j].status == Tile::UNIT)
	  no_of_units++;
	if(no_of_units > max_units)
	  return false;
      }
    }
    return true; 
  }

  bool TerrainMap::get_random_unexplored_point(TerrainBase::Loc point, TerrainBase::Loc &location,sint4 time_limit , sint4 bound ){
    init_blocked_tiles();
    compute_free_areas();

    sint4 starti, startj;
    sint4 endi, endj;
    sint4 boundy = bound;
    sint4 stop_time;

    if(time_limit == 0){
      stop_time = max_sint4;
    }
    else
      stop_time = SDL_GetTicks() + time_limit;

    if(bound == 0){
      bound = width;
      boundy = height;
    }
    

    starti = (point.x - bound) / tile_points;
    startj = (point.y - boundy) / tile_points;

    endi = (point.x + bound) / tile_points;
    endj = (point.y + boundy) / tile_points;
    
    starti = max(starti, 0);
    startj = max(startj, 0);
    endi = min(endi, width-1);
    endj = min(endj, height-1);

    while( (sint4)SDL_GetTicks() < stop_time ){
      sint4 randi = ((sint4)rand.rand_uint4() % (endi - starti) ) + starti;
      sint4 randj = ((sint4)rand.rand_uint4() % (endj - startj) ) + startj;

      randi = max(randi, 0);
      randi = min(randi, width-1);
      randj = max(randj, 0);
      randj = min(randj, height-1);

      if(map[randi][randj].status == Tile::FOG){
	location = TerrainBase::Loc(randi*tile_points , randj*tile_points);
	return true;
      }
    }
    location = TerrainBase::Loc(point.x , point.y);
    return false;


  }



  /*---------------------------------------------------------------*/
  /*                      PRIVATE MEMBERS                          */
  /*---------------------------------------------------------------*/
  void TerrainMap::checkNode(sint4 i, sint4 j, sint4 distance){
    
    //Check to ensure the given tile is in the map boundary.
    if(i<0 || i>=width || j<0 || j>=height){
      PRINT(std::cout << "Invalid index ingoring tile:" << std::endl;
	    std::cout << "i = " << i << "j = " << j << std::endl;)
	return;
    }
    
    PRINT( std::cout << "i = " << i << "j = " << j << std::endl;)

      //Check the current value of the tile skip it if it has been marked before or if it is not free.
      Tile *temp = &map[i][j];	
    if(!temp->freeValue && temp->status == Tile::FREE){
      temp->freeValue = distance;
      tempList.push_back(temp); 
      PRINT(cout << "Adding new tile (" << i << "," << j << ") to the open list." << endl;);
    }
  }


  bool TerrainMap::move_one_tile(sint4 &i, sint4 &j){

    sint4 d[8][2] = {{-1,0},{-1,-1},{0,-1},{1,-1},
                     {1,0},{1,1},{0,1},{-1,1}};

    sint4 current_value = map[i][j].freeValue;

    FORU (i, 8) {
      sint4 nx = i+d[i][0];
      sint4 ny = i+d[i][1];

      if (nx >= 0 && nx < width && ny >= 0 && ny < height && map[nx][ny].freeValue > current_value) {
        i = nx;
        j = ny;
        return true;
      }
    }

    FORU (i, 8) {
      sint4 nx = i+d[i][0];
      sint4 ny = i+d[i][1];

      if (nx >= 0 && nx < width && ny >= 0 && ny < height && map[nx][ny].freeValue == current_value) {
        i = nx;
        j = ny;
        return true;
      }
    }

    return false;
  }


}//End of namespace mapTerrain
