// $Id: PRAEngine.C,v 1.11 2006/05/18 15:27:59 nicholas Exp $
// This is an ORTS file 

#include "PRAEngine.H"


PR_Terrain::PRAEngine::PRAEngine(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, int gran): 
  map(tiles_x_ * gran +1, tiles_x_ * gran +1),
  aMap(&map),
  res_provider((tiles_x_ * gran+1) * (tiles_y_ * gran  +1))
{
  subtile_points = tile_points_ / gran;
  disp = subtile_points / 2;
// sint4 width, height;
//   width = tiles_x_ * gran + 1;
//   height = tiles_y_ * gran + 1;
//   map(width, height);
//   aMap(&map);
//   res_provider(width * height);

}

PR_Terrain::PRAEngine::~PRAEngine(){
}

//-----------------------------------------------------------------------------
void PR_Terrain::PRAEngine::insert_object(const Object *obj)
{
  ir_obj(obj, true, false);
}

//-----------------------------------------------------------------------------

// void PR_Terrain::PRAEngine::ir_bucket(const Demo_SimpleTerrain::Bucket &bucket, Object::ZCat zcat, bool insert){
//   sint4 x1, y1;
//   const TerrainBase::Loc &c = bucket.get_center();
//   const Vector<TerrainBase::Loc> &points = bucket.get_points();
  
//   FORALL (points, i) {
//     x1 = c.x + i->x;
//     y1 = c.y + i->y;
//     res_provider->setNodeStatus(aMap->getNodeFromMap(x1, y1), insert);
//   }
// }

void PR_Terrain::PRAEngine::ir_rectangle(TerrainBase::Loc &topleft, TerrainBase::Loc &bottomright, Object::ZCat /*zcat*/, bool insert){

  sint4 tlTileX = world2x(topleft.x);
  sint4 tlTileY = world2y(topleft.y);
  sint4 brTileX = world2x(bottomright.x);
  sint4 brTileY = world2y(bottomright.y);
  

 //code for taking the edges of a rectangle out of the map:
//   FORT (i, brTileX-tlTileX+1) remove_edges(tlTileX+i, tlTileY); 
//   FORT (i, brTileY-tlTileY+1) remove_edges(tlTileX, tlTileY+i);
//   FORT (i, brTileX-tlTileX+1) remove_edges(tlTileX+i, brTileY); 
//   FORT (i, brTileY-tlTileY+1) remove_edges(brTileX, tlTileY+i);

//   remove_cross_edge(tlTileX, tlTileY, HOG::kTopLeft);
//   remove_cross_edge(tlTileX, brTileY, HOG::kBottomLeft);
//   remove_cross_edge(brTileX, tlTileY, HOG::kTopRight);
//   remove_cross_edge(brTileX, brTileY, HOG::kBottomRight);
  
  //code for setting the reservation provider to take the nodes out of the map
  FORT(i, brTileX - tlTileX + 1){
    FORT(j, brTileY - tlTileY + 1){
      res_provider.setNodeStatus(aMap.getNodeFromMap(tlTileX + i, tlTileY + j), insert);
    }
  }
}


void PR_Terrain::PRAEngine::update_object(const Object *obj)
{
  //remove previous
  ir_obj(obj, false, true);
  //insert new
  ir_obj(obj, true, false);
}

void PR_Terrain::PRAEngine::remove_object(const Object *obj)
{
  ir_obj(obj, false, false);
}

void PR_Terrain::PRAEngine::ir_obj(const Object *obj, bool insert, bool use_prev_attributes){
 {

    if (obj->get_shape() == Object::RECTANGLE) {
      TerrainBase::Loc topleft, bottomright;
      Object::ZCat zcat;

      if( !use_prev_attributes ) {
        obj->get_p1(topleft.x, topleft.y);
        obj->get_p2(bottomright.x, bottomright.y);
        zcat = obj->get_zcat();
      } else {
        // get previous attribute values
        sint4 dummy1, dummy2;
        obj->get_p1(dummy1, dummy2, &(topleft.x), &(topleft.y));
        obj->get_p2(dummy1, dummy2, &(bottomright.x), &(bottomright.y));
        obj->get_zcat(&zcat);
      }
      ir_rectangle(topleft, bottomright, zcat, insert);
      
    } else if (obj->get_shape() == Object::CIRCLE) {

      sint4 r = obj->get_radius();
      TerrainBase::Loc c;
      Object::ZCat zcat;

      if( !use_prev_attributes ) {
        obj->get_center(c.x, c.y);
        zcat = obj->get_zcat();
      } else {
        // get previous attribute values
        sint4 dummy1, dummy2;
        obj->get_center(dummy1, dummy2, &(c.x), &(c.y));
        obj->get_zcat(&zcat);
      }

      int newr = r / subtile_points;
      if (r % subtile_points >= subtile_points / 2)
	newr++;
      if (newr == 0){
	newr++;
      }
     
      assert(newr <= 1 && "Tile Size smaller than at least one unit");
      
      sint4 x, y;
      x = world2x(c.x);
      y = world2y(c.y);
      res_provider.setNodeStatus(aMap.getNodeFromMap(x, y), insert);
      
//       Demo_SimpleTerrain::Bucket b = Demo_SimpleTerrain::BucketFactory::get_bucket(newr);
//       Demo_SimpleTerrain::Bucket b2 = Demo_SimpleTerrain::BucketFactory::get_bucket(newr-1);
//       TerrainBase::Loc cworld(c.x / subtile_points, c.y / subtile_points;     
//   b.set_center(cworld);
//        ir_bucket(b, zcat, insert);

//       b2.set_center(cworld);
//       ir_bucket(b2, zcat, insert);

      // fixme: heading is not accessible through Object
      /* We wanted to add some cost to the position where the object will be
      in the next tick to prefer not to go there; However, heading is not
      currently accessible with the new TerrainBase interface:
      // one second distance
      Loc head = cworld;
      Loc tail = cworld;
      get_head_and_tail(head, tail, heading);

      b.set_center(head);
      insert_bucket(b, zcat, TEMP, 0, id, cost*CROSS_COST);

      b.set_center(tail);
      //  insert_bucket(b, zcat, TEMP, 0, id, cost*CROSS_COST);
      */
    } else ERR("unhandled shape");
  }
} 

void PR_Terrain::PRAEngine::insert_boundary(const TerrainBase::Segment& aSeg)
{

  std::cout << "PR_Terrain::PRAEngine::insert_boundary \n";
  sint4 x1Coord = world2x(aSeg.l1.x);
  sint4 y1Coord = world2y(aSeg.l1.y);
  sint4 x2Coord = world2x(aSeg.l2.x);
  sint4 y2Coord = world2y(aSeg.l2.y);

  sint4 temp;

  if (x1Coord > x2Coord){
    temp = x1Coord;
    x1Coord = x2Coord;
    x2Coord = temp;

    temp = y1Coord;
    y1Coord = y2Coord;
    y2Coord = temp;
  }

  sint4 dx = aSeg.l2.x - aSeg.l1.x;
  sint4 dy = aSeg.l2.y - aSeg.l1.y;

  printf("aSeg.l1.x =%d, aSeg.l1.y = %d\n",aSeg.l1.x , aSeg.l1.y );
  printf("aSeg.l2.x =%d, aSeg.l2.y = %d\n",aSeg.l2.x , aSeg.l2.y );

  printf("x1Coord = %d, y1Coord = %d, x2Coord = %d, y2Coord = %d.\n", x1Coord, y1Coord, x2Coord, y2Coord);
  printf("dx = %d, dy = %d\n", dx ,dy);

  /*Add a horizontal line. */    
  if(dx == 0){
   
    if (y1Coord < y2Coord) {
      FORT (i, y2Coord-y1Coord+1) remove_edges(x1Coord, y1Coord+i); 

      remove_cross_edge(x1Coord, y1Coord, HOG::kTopLeft);
      remove_cross_edge(x1Coord, y1Coord, HOG::kBottomLeft);
      remove_cross_edge(x2Coord, y2Coord, HOG::kTopRight);
      remove_cross_edge(x2Coord, y2Coord, HOG::kBottomRight);
    }
    else{
      FORT (i, y1Coord-y2Coord+1) remove_edges(x1Coord, y2Coord+i);

      remove_cross_edge(x1Coord, y1Coord, HOG::kBottomLeft);
      remove_cross_edge(x1Coord, y1Coord, HOG::kBottomRight);
      remove_cross_edge(x2Coord, y2Coord, HOG::kTopLeft);
      remove_cross_edge(x2Coord, y2Coord, HOG::kTopRight);
    }
  }
  /*Add a vertical line. */  
  else if(dy == 0){
    
    remove_cross_edge(x1Coord, y1Coord, HOG::kTopLeft);
    remove_cross_edge(x1Coord, y1Coord, HOG::kBottomLeft);
    remove_cross_edge(x2Coord, y2Coord, HOG::kBottomRight);
    remove_cross_edge(x2Coord, y2Coord, HOG::kTopRight);

    FORT (i, x2Coord-x1Coord+1) remove_edges(x1Coord+i, y1Coord); 
  }
  
  /*Add a diagonal line.*/
  else if(dx == dy){
    remove_cross_edge(x1Coord, y1Coord, HOG::kTopLeft);
       
    int i = 0; // Go right one tile. 
  
    remove_edges(x1Coord, y1Coord);

    while(!(x1Coord + i == x2Coord && y1Coord + i == y2Coord)){
      i++;

      remove_edges(x1Coord + i, y1Coord + i);
      remove_cross_edge(x1Coord + i, y1Coord + i, HOG::kBottomRight);
    }
  }
  else if(dx == -dy){
  
    remove_cross_edge(x1Coord, y1Coord, HOG::kBottomLeft);
  
    int i = 0; // Go right one tile. 
    
    remove_edges(x1Coord, y1Coord);

    while(!(x1Coord + i == x2Coord && y1Coord - i == y2Coord)){
      i++;

      remove_edges(x1Coord + i, y1Coord - i);
      remove_cross_edge(x1Coord + i, y1Coord - i, HOG::kTopRight);
    } 
  }  
}


void PR_Terrain::PRAEngine::repair_abstraction(){
  aMap.repairAbstraction();
}

bool PR_Terrain::PRAEngine::find_path(const Object *obj, const TerrainBase::Loc &goal, TerrainBase::Path &output){
  sint4  goalX, goalY;

  goalX = world2x(goal.x);
  goalY = world2y(goal.y);

  HOG::node *start, *end;
  
  start = get_center_node(obj);
  end = aMap.getNodeFromMap(goalX, goalY);
  
  return find_path(start, end, output);

}

bool PR_Terrain::PRAEngine::find_path(const Object *obj, const Object *goal, TerrainBase::Path& output){
  
  //find the nodes on the center of the object and the goal
  
  HOG::node *start, *end;  
  start = get_center_node(obj);
  end = get_center_node(goal);
 
  //temporarily remove the goal unit
  remove_object(goal);

  //find the path
  bool path_found = find_path(start, end, output);
  
  //add the goal unit back in
  insert_object(goal);

  return path_found;
  
}

bool PR_Terrain::PRAEngine::find_path(HOG::node *start, HOG::node *end, TerrainBase::Path& output){
  HOG::path *hog_path;
  
  hog_path = prastar.getPath(&aMap, start, end, &res_provider);
     
  if (!hog_path){
    std::cout << "NO PATH FOUND!!\n";
    return false; // path not found
  }
  
  std::cout << "Path found! \n";
   
  HOG::node *n;
  sint4 locX, locY;

  Vector<TerrainBase::Loc> temp_output;

  n = hog_path->n;
  aMap.getTileFromNode(n, locX, locY);
  
  locX = x2world(locX);
  locY = y2world(locY);
  temp_output.push_back(TerrainBase::Loc(locX, locY));
  //std::cout << "First loc x = " << locX << " loc y = " << locY << ".\n";
  
  while (hog_path->next){
    hog_path = hog_path->next;
    n = hog_path->n;
    aMap.getTileFromNode(n, locX, locY);
    
    locX = x2world(locX);
    locY = y2world(locY);
    temp_output.push_back( TerrainBase::Loc(locX, locY));
    // std::cout << "Next loc x = " << locX << " loc y = " << locY << ".\n";
  }
 
  //To be more compatible with code from SimpleTerrain, reverse the 
  //order of the locations.
  
  FORALL_REV(temp_output, rev_it){
    output.locs.push_back(*rev_it);
  }
  
  smooth_path(output.locs);
  
  return true;
}

HOG::node* PR_Terrain::PRAEngine::get_center_node(const Object *obj){
  sint4 x1, y1;
  if (obj->get_shape() == Object::CIRCLE){
    obj->get_center(x1, y1);    
    x1 = world2x(x1);
    y1 = world2y(y1);
  }
  else if (obj->get_shape() == Object::RECTANGLE){
    sint4 x2, y2;
    obj->get_p1(x1, y1);
    obj->get_p2(x2, y2);
    x1 = (x1 + x2) / 2; //integer division, so might not be exactly at the center
    y1 = (y1 + y2) / 2;
    
    x1 = world2x(x1);
    y1 = world2x(y1);
  }
  else {
    ERR("Unsupported Shape - cannot find central node");
  }
  
  return aMap.getNodeFromMap(x1, y1);
}

void PR_Terrain::PRAEngine::remove_edges(sint4 x, sint4 y){
    
  sint4 currentNode = map.getNodeNum(x, y);
  HOG::graph *g = aMap.getAbstractGraph(0); 
  HOG::edge *e;

  //Left going edge
  e = g->findEdge(currentNode, map.getNodeNum(x-1, y)); 
  if(e){
    aMap.removeEdge(e, 0); 
    std::cout << "Removing left going edge.\n";
  }
	
  //Right going edge
  e = g->findEdge(currentNode, map.getNodeNum(x+1, y)); 
  if(e){
    aMap.removeEdge(e, 0);
    std::cout << "Removing right going edge.\n";
  }
  //Up going edge
  e = g->findEdge(currentNode, map.getNodeNum(x, y-1)); 
  if(e){
    aMap.removeEdge(e, 0);
    std::cout << "Removing up going edge.\n";
  }  
  //Down going edge
  e = g->findEdge(currentNode, map.getNodeNum(x, y+1));
  if(e){
    aMap.removeEdge(e, 0);
    std::cout << "Removing left going edge.\n";
  }	
  //Top left
  e = g->findEdge(currentNode, map.getNodeNum(x-1, y-1)); 
  if(e){
    aMap.removeEdge(e, 0);
    std::cout << "Removing Top left edge.\n";
  }
  //Top Right
  e = g->findEdge(currentNode, map.getNodeNum(x+1, y-1)); 
  if(e){
    aMap.removeEdge(e, 0);
    std::cout << "Removing top right edge.\n";
  } 
  //Bottom Left
  e = g->findEdge(currentNode, map.getNodeNum(x-1, y+1)); 
  if(e){
    aMap.removeEdge(e, 0);
    std::cout << "Removing bottom left edge.\n";
  }
  //Bottom Right
  e = g->findEdge(currentNode, map.getNodeNum(x+1, y+1));
  if(e){
    aMap.removeEdge(e, 0);
    std::cout << "Removing bottom right edge.\n";
  }
		    
}

/**
 * Removes the edge between the nodes that passes through the corner
 * of the given node (described by its x and y coordinates).
 *
 * eg: for top-left corner, remove this edge:
 *      /_
 *     /|_|
 */
void PR_Terrain::PRAEngine::remove_cross_edge(sint4 x, sint4 y, HOG::tCorner corner){

  sint4 node1, node2;
  HOG::graph *g = aMap.getAbstractGraph(0); 
  HOG::edge *e;
  switch(corner){
  case HOG::kTopLeft :
    node1 =  map.getNodeNum(x, y-1);
    node2 = map.getNodeNum(x-1, y);
    break;

  case HOG::kTopRight:
    node1 =  map.getNodeNum(x, y-1);
    node2 = map.getNodeNum(x+1, y);
    break;

  case HOG::kBottomLeft:
    node1 =  map.getNodeNum(x, y+1);
    node2 = map.getNodeNum(x-1, y);
    break;

  case HOG::kBottomRight:
    node1 =  map.getNodeNum(x, y+1);
    node2 = map.getNodeNum(x+1, y);
    break;

  default:
    ERR("Unkown corner - cannot remove_cross_edge");
  }
  e = g->findEdge(node1, node2); 
  
  if(e != 0)
    aMap.removeEdge(e, 0);
}
    
  /**
  * go throught the waypoints in the given vector and remove
  * all redundent points that fall on a straight line
  *
  * Taken from ST_PFEngine.
  * */
  void PR_Terrain::PRAEngine::smooth_path(Vector<TerrainBase::Loc> &path)
  {
    uint4 i1 = 0;
    uint4 i2 = 1;
    for (uint4 i = 2; i < path.size(); i++) {

      while (i < path.size() && on_line(path[i1], path[i2], path[i])) {
        path.erase(path.begin()+i2);
      }

      i1 = i-1;
      i2 = i;
    }
  }


/*
 *
 */
bool PR_Terrain::PRAEngine::on_line(const TerrainBase::Loc &a, const TerrainBase::Loc &b, const TerrainBase::Loc &c)
{
  if (a.x == b.x && b.x == c.x) { return true; }
  if (a.y == b.y && b.y == c.y) { return true; }
  double tx = (double)(b.x - a.x) / ((double)(c.x - a.x));
  double ty = (double)(b.y - a.y) / ((double)(c.y - a.y));
  return tx >=0 && tx <= 1 && tx == ty;
}  
