#ifndef SI_PLAYER_H
#define SI_PLAYER_H

// $Id: SI_Player.h,v 1.6 2006/06/06 11:33:45 ddeutscher Exp $

#include "Global.H"
#include "Vector.H"

#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <set>
#include <map>
#include <cmath>

#include "SI_Base.h"
#include "SI_Group.h"
#include "SI_SimulationAI.H"

namespace Simul{

  class MapObject;
  class Group;
  class Base;
  class SimulationState;

  struct Task;

  typedef Vector<Group*> GroupVector;
  typedef Vector<Base*> BaseVector;
  typedef std::map<Task*,Task*> TaskCopiesMap;


  typedef std::map<GameObj*, ExtraObjectData> ObjectDataMap;

  class Player
  {
  public:
    Player(uint4 playerID);
    ~Player();
    Player(const Player& p);

    Player* clone();

    uint4 getID() const {return playerID;}

    void addGroup(Group* newGroup);
    void addBase(Base* newBase);

    //void drawObjects();
    //void drawObjectsInSelectionMode();

    void executeOrders(real8 timeElapsed);
    void updateOrders(SimulationState& state);


    void addOrtsBase(GameObj* obj){
      ortsBases[obj] = ExtraObjectData();
      // we can add a base right away, since there is no grouping of bases, they are what they are
      addBase(new Base(playerID,obj->get_int("hp")));
    }

    void addOrtsUnit(GameObj* obj){
      ortsUnits[obj] = ExtraObjectData();
    }

    void removeOrtsBase(GameObj* obj){ ortsBases.erase(obj);}
    void removeOrtsUnit(GameObj* obj){ ortsUnits.erase(obj);}

    ObjectDataMap& getOrtsUnits(){return ortsUnits;}
    ObjectDataMap& getOrtsBases(){return ortsBases;}

    GroupVector* getGroups() {return &groups;}
    BaseVector* getBases() {return &bases;}


    real8 getVictoryValue(){return (real8)bases.size();}

    void setStrategy(Task* strategy);
    Task* getStrategy() {return currentStrategy;}

  private:

    Player &operator=(const Player &);

    GroupVector groups;   // owns objects pointed to
    BaseVector bases;     // owns objects pointed to

    Task* currentStrategy;   // owner
    uint4 playerID;

    ObjectDataMap ortsUnits;
    ObjectDataMap ortsBases;

  };
}

REGISTER_TYPEOF(2841, Simul::ObjectDataMap::iterator);
REGISTER_TYPEOF(2842, Simul::ObjectDataMap::const_iterator);
REGISTER_TYPEOF(2843, Simul::BaseVector::iterator);
REGISTER_TYPEOF(2844, Simul::BaseVector::const_iterator);
REGISTER_TYPEOF(2845, Simul::GroupVector::iterator);
REGISTER_TYPEOF(2846, Simul::GroupVector::const_iterator);

#endif

