
#include "SI_SimulationAI.H"

#include "SI_SimulationState.h"
#include "SI_Player.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
//#include "GfxModule.H"
#include "TerrainModule.H"
#include "SquadCombatAI.H"

#include "Options.H"

using namespace std;

namespace Simul{

  const sint4 SimulationAI::FROM = EventFactory::new_who();

  void SimulationAI::add_options()
  {
    ::Options o("SimulationAI");
    o.put("-hunter", "Use Hunter, with targets being closest enemy unit to any unit in group");
    o.put("-hunter2","Use Hunter, with target being closest to ALL our units in group");
    o.put("-spread", "every unit continuously attacks nearest base to it");
    o.put("-halfdef", "Use task that send half units to mass and attack, while the other half defends bases spread out evenly");
    o.put("-joinspread","Join into small groups, which then attack nearest bases");
    o.put("-groupsize",sint4(50),"Specify how large groups are(used by hunter currently(default is use all 50 units)");
    o.put("-smartjoinsize",sint4(5),"specify how large local groups are formed before doing a larger join");
    o.put("-smartjoin", "Specify whether to form local groups when joining, or whether to just go and join");
    o.put("-moveforattack","When told to attack a unit, instead of using an attack task, use a move to location task(means shooterAI is not called, and there is less replanning(used by hunter1/2 right now)");
    o.put("-findspot","finds a smarter spot to do joins than just the average location of the units");
  }

  SimulationAI::SimulationAI(GameStateModule &gsm_) : gsm(gsm_){
    simulationState = new SimulationState(this);
    // crashes here for some reason, need to put it later
    //simulationState->setOurID(gsm.get_game().get_client_player());
  }
  SimulationAI::~SimulationAI(){}

//===================================================================

void SimulationAI::act(GameStateModule &/*gsm*/){
  fprintf(stderr,"u...");
  simulationState->updateStrategies();

}

//===================================================================

/**
 * Returns true if event was handled.
 */
bool SimulationAI::handle_event(const Event &e)
{
  // stuff for tim
  // for now, do something dumb, we don't have time to put in smarter things
  FORALL(simulationState->getPlayers(),it){
    FORALL(it->second->getOrtsUnits(),it2){
      if(it2->second.state == ExtraObjectData::ATTACK){ // if we have an attack order
        assert(it2->second.target != NULL);

        real8 distance = gsm.get_game().distance(*it2->first, *it2->second.target);

        if (distance < 75.0) { // FIXME: this is a magic number, should be something better -frano

          Vector<GameObj*> foo;
          foo.push_back(it2->first);

          doCombat(foo, it2->second.target);
        }

      }
    }
  }


  bool needStrategyUpdate = false;

  if (e.get_what() == TerrainModule::PATH_SUCCESS) {

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "SUCCESS: not a PathStatusEvent");

    FORALL(simulationState->getPlayers(),it){

      FIND((it->second)->getOrtsUnits(),it2,p->get_obj());
      if (it2 == (it->second)->getOrtsUnits().end()) continue;

      // if we were just moving... then we are done!
      //if(it->second.state == ExtraObjectData::MOVE){
	it2->second.isDone = true;
	needStrategyUpdate = true;
    }
  }

  if (e.get_what() == TerrainModule::PATH_FAILURE) {

    const TerrainModule::PathStatusEvent *p =
      dynamic_cast<const TerrainModule::PathStatusEvent*>(&e);
    assert(p && "FAILURE: not a PathStatusEvent");

    //if (p->get_details() == TerrainModule::PATH_PLAN_FAILURE){
      std::cout << "Path failure! This is bad, and shouldn't happen often" << endl;

    FORALL(simulationState->getPlayers(),it){

	FIND((it->second)->getOrtsUnits(), it2, p->get_obj());
	if (it2 == (it->second)->getOrtsUnits().end()) continue;

	// TODO: right now, just say we are done if path fails.. not sure what to do here

      /*if(it2->second.hasReplanned){
        it2->second.isDone = true;
        needStrategyUpdate=true;
        continue;
      }

        if(it2->second.target == NULL){
          if(it2->second.location.x != -1){
            TerrainBase::Task task(TerrainBase::Goal::LOCATION, TerrainBase::Goal::TOUCH,
                                   TerrainBase::Task::INDIVIDUAL);



            task.goal.loc = it2->second.location;
            task.objs.insert(p->get_obj());
            doPath(task);
            it2->second.hasReplanned = true;

          }else{
            ERR("somethign is wrong, we should have either a target, or location");
          }
        }else{
          TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::TOUCH,
                                 TerrainBase::Task::INDIVIDUAL);

          task.goal.obj = it2->second.target;
          task.objs.insert(p->get_obj());
          doPath(task);
          it2->second.hasReplanned = true;
        }*/
        it2->second.isDone = true;
        needStrategyUpdate=true;

    }
  }

  if (e.get_what() == SquadCombatAI::TARGET_DESTROYED) {

    const SquadCombatEvent *s =
      dynamic_cast<const SquadCombatEvent*>(&e);
    assert(s && "FAILURE: not a SquadCombatEvent");

    FORALL (s->units, u) {
      Player *player = simulationState->getPlayer(*(*u)->sod.owner);
      FIND (player->getOrtsUnits(), it, *u);
      if (it == player->getOrtsUnits().end()) continue;

      it->second.isDone = true;
      needStrategyUpdate = true;
    }

    stopCombat(s->units);
  }


  if (e.get_who() == GameStateModule::FROM) {

    if (e.get_what() == GameStateModule::VIEW_MSG) {

      const sint4 me = gsm.get_game().get_client_player();

      // this is a dumb place to put it, but it crashes if it is put into the constuctor
      simulationState->setOurID(gsm.get_game().get_client_player());

      // add new objects
      // NOTE: we assume this is only called once right now, since we see the whole map
      // this WILL NOT work for fog of war!

      bool addedUnit = false;

      FORALL (gsm.get_changes().new_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;

        Player* player = simulationState->getPlayer(*gob->sod.owner);
        if(player == NULL){ // player does not exist yet.. create one
          player = new Player(*gob->sod.owner);
          simulationState->addPlayer(player);
        }

        if(isControllable(gob)){
          player->addOrtsUnit(gob);
          addedUnit = true;

        }else if(gob->bp_name() == "controlCenter"){
          player->addOrtsBase(gob);
        }
      }
      if(addedUnit){

        Player* ourPlayer = simulationState->getPlayer(simulationState->getOurID());
        if(ourPlayer == NULL) ERR("umm, weirdness, our player should exist always");

        if(ourPlayer->getStrategy() != NULL){
          ERR("GOT NEW OBJECTS IN MULTIPLE FRAMES! CAN'T HANDLE THIS RIGHT NOW!");
        }

        ObjectVector ourControllableObjects;

        FORALL(ourPlayer->getOrtsUnits(),it){
          ourControllableObjects.push_back(it->first);
        }

        if(!ourControllableObjects.empty()){
          sint4 gSize,smartSize;
          bool useSmart = false;
          bool useMove = false;
	  bool useFindSpot = false;
          ::Options::get("-groupsize", gSize);
          ::Options::get("-smartjoinsize", smartSize);
          ::Options::get("-smartjoin", useSmart);
          ::Options::get("-moveforattack", useMove);
          ::Options::get("-findspot", useFindSpot);
		

          simulationState->groupSize = (uint4)gSize;
          simulationState->smartJoinSize = (uint4)smartSize;
          simulationState->smartJoin = useSmart;
          simulationState->useMoveForAttack = useMove;
          simulationState->findspot = useFindSpot;

	  bool mode;
	  ::Options::get("-hunter", mode);
	  if (mode)
	    ourPlayer->setStrategy(createHunter(me, simulationState, ourControllableObjects, 0.0));
	  else{
  	   ::Options::get("-hunter2", mode);
  	    if (mode)
    	     ourPlayer->setStrategy(createHunter2(me, simulationState, ourControllableObjects, 0.0));
  	    else{
	      ::Options::get("-spread", mode);
	      if (mode)
		ourPlayer->setStrategy(createSpreadAttack(me, simulationState, ourControllableObjects, 0.0));
	      else{
		::Options::get("-halfdef", mode);
		if (mode)
		  ourPlayer->setStrategy(createHalfUnitsBaseDefence_MassAttack(me, simulationState, ourControllableObjects, 0.0));
		else{
		  ::Options::get("-joinspread", mode);
		  if (mode)
		    ourPlayer->setStrategy(createJoinSpreadAttack(me, simulationState, ourControllableObjects, 0.0));
		  else{
		      ourPlayer->setStrategy(createMassAttack(me, simulationState, ourControllableObjects, 0.0));
		  }
		}
	      }
	    }
	  }
        }

        needStrategyUpdate = true;
      }

      // remove dead objects
      FORALL (gsm.get_changes().dead_objs, i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob == 0) continue;

        Player* player = simulationState->getPlayer(*gob->sod.owner);
        if(player == NULL) ERR("problem, we have an object that does not belong to a player we know about");

	if(gob->bp_name() == "controlCenter"){
	  player->removeOrtsBase(gob);
	}else{
	  player->removeOrtsUnit(gob);

	  // if it's our unit, we have to make sure our current task also loses it
	  if(player->getStrategy() != NULL){
	   player->getStrategy()->removeObject(gob);
  	   needStrategyUpdate=true;
	  }
        }

        needStrategyUpdate = informObjectsOfDeath(gob);

      }
    }
  }

  // TODO for now, leave this check out.  it just results in slightly more work.  if i leave the check in
  // it seems to fail sometimes... not sure why.
  //if(needStrategyUpdate){
    act(gsm);
  //}

  return false;
}

bool SimulationAI::informObjectsOfDeath(GameObj* obj){
  bool needStrategyUpdate=false;
  //Player* player = simulationState->getPlayer(*obj->sod.owner);
  //if(player == NULL) ERR("bad stuff happening, our player should always exist!");

  FORALL(simulationState->getPlayers(), it){
    FORALL((it->second)->getOrtsUnits(),it2){
      // if the objects we had a target to was killed... this unit is done its task
      //if(it->second.state == ExtraObjectData::ATTACK || it->second.state == ExtraObjectData::DEFEND){
	if(it2->second.target != NULL){
	  if(it2->second.target == obj){
	    it2->second.isDone = true;
	    needStrategyUpdate = true;
	  }
	}
      /*
	else{
	  ERR("ERROR: we have an attack order, but no target!!!");
	}
      */
      //}
    }
  }
  return needStrategyUpdate;
  // TODO: maybe add updates for other objects later.  not needed right now, since we only control friendly units
}

  bool SimulationAI::isControllable(GameObj* obj){
  // first, check if we even have a weapon...and can move...
    if ((obj->get_max_speed() > 0) && (obj->component("weapon") != NULL)){
            // if we can, make sure it can be used....this will likely always pass
      if (obj->component("weapon")->has_attr("max_damage")){
              // make sure we can actually do dmg
        if(obj->component("weapon")->get_int("max_damage") > 0){ // if we can attack
          return true;
        }
      }
    }
    return false;
  }

  sint4 SimulationAI::doPath(TerrainBase::Task& task){
    std::cout << "Send Pathfinding Message from GAME 2\n";
    return call_handlers(TerrainModule::PathEvent(FROM, TerrainModule::FIND_PATH_MSG, task));
  }

  sint4 SimulationAI::doCombat(Vector<GameObj*> units, GameObj *target) {
        std::cout << "attack target" << endl;
    return call_handlers(SquadCombatEvent(FROM, SquadCombatAI::ATTACK_TARGET, units, target));
  }

  sint4 SimulationAI::stopCombat(Vector<GameObj*> units) {
    std::cout << "stop attack" << endl;
    return call_handlers(SquadCombatEvent(FROM, SquadCombatAI::STOP_ATTACK, units));
  }

}
