#ifndef ST_SimpleMap_H
#define ST_SimpleMap_H

// $Id: ST_SimpleMap.H,v 1.9 2006/06/06 17:29:00 orts_mburo Exp $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla
// licensed under the GPL

#include "TerrainBase.H"

namespace SimpleTerrain
{
  // ------------------------------------------------------------------------------
  //enum CellFlag { OPEN, TEMP, CLOSED }; // open, temporarily closed, open

  // ------------------------------------------------------------------------------
  struct PACKED_STRUCT SizeCell
  {
    //TerrainBase::ObjId obj_id;   // id of the object occupying this cell
    //uint1 flags;
    sint1 max_size; // <=0 iff the cell is totally blocked
    uint4 cost;

    SizeCell() {
      //flags = OPEN;
      max_size = 0;
      cost = 0;
    }

    void write(std::ostream &os = std::cout) const {
      os << sint4(max_size);
      //if (flags) os << '*'; else os << ' ';
    }
  };

  // ------------------------------------------------------------------------------
  struct PACKED_STRUCT AStarCell
  {
    uint4 parent;       // path predecessor
    uint4 open, closed; // set membership: in set <=> (value == magic)
    uint4 g;
  
    AStarCell() { parent = 0; open = closed = g = 0; }
  };

  // ------------------------------------------------------------------------------
  struct HeapElement
  {
    uint4 cell; // index
    uint4 f, g; // can't use cell->g because otherwise heap gets corrupted
    // when cell->g is changed and used for heap element comparison

    HeapElement() { cell = 0; f = g = 0; }
  };

  // ------------------------------------------------------------------------------
  struct HeapElementCmp
  {
    bool operator()(HeapElement &h1, HeapElement &h2) const {
      sint4 d = h1.f - h2.f;
      if (d != 0) return d > 0;
      return h1.g < h2.g;
    }

  };

  // ------------------------------------------------------------------------------
  // mailbox representation with boundary (width->width+2, height->height+2)
  template <typename T> struct SimpleMap {

  public:

    SimpleMap(sint4 w_, sint4 h_) : w(w_), h(h_)
    { 
      assert(w >= 0 && h >= 0);
    
      d[0] =  1; d[1] = 1+w;  d[2] =  w; d[3] = w-1;
      d[4] = -1; d[5] = -1-w; d[6] = -w; d[7] = 1-w;
    
      n = w * h;
      cells.reserve(n);
      T c = T();
      FORT (i, n) cells.push_back(c);
      magic = 0;
    }

    virtual ~SimpleMap() {};
  
    sint4 get_d(sint4 i) const { return d[i]; }
    T &operator()(sint4 i) { return cells[i]; }
    const T &operator()(sint4 i) const { return cells[i]; }
    T &at(sint4 i) { return operator()(i); };
    const T &at(sint4 i) const { return operator()(i); };

    // important: add 1 to both when using original coordinates  
    T &operator()(sint4 x, sint4 y) { return cells[index(x, y)]; }
    const T &operator()(sint4 x, sint4 y) const { return cells[index(x, y)]; }  

    sint4 get_n() const { return n; }
    sint4 get_w() const { return w; }
    sint4 get_h() const { return h; }  
  
    void write(std::ostream &os = std::cout) const;

    inline sint4 i2x(sint4 index) const { return index % w; }
    inline sint4 i2y(sint4 index) const { return index / w; }

    inline sint4 index(sint4 x, sint4 y) const {
      assert(x >= 0 && x < w);
      assert(y >= 0 && y < h);    
      return y*w + x;
    }

    uint4 get_magic() const { return magic; }
    void  set_magic(uint4 m) { magic = m; }

    // interface for iterating over all cells
    typedef typename Vector<T>::iterator iterator;
    typedef typename Vector<T>::const_iterator const_iterator;
    iterator begin(void) { return cells.begin(); };
    const_iterator begin(void) const { return cells.begin(); };
    iterator end(void) { return cells.end(); };
    const_iterator end(void) const { return cells.end(); };

  protected:
  
    uint4 magic; // used for set membership test
    sint4 w, h, n;
    Vector<T> cells;
    sint4 d[8];
  };

} // End of namespace SimpleTerrain

REGISTER_TYPEOF(2701,Vector<SimpleTerrain::SizeCell>::iterator);
REGISTER_TYPEOF(2702,Vector<SimpleTerrain::SizeCell>::const_iterator);

#endif
